Overview of the MoSeL artifact
==============================

This is the artifact for the paper "MoSeL: A General, Extensible Modal
Framework for Interactive Proofs in Separation Logic".


Getting started
---------------

The artifact comes in two forms: As a `mosel.zip` file to build yourself, and as
a preconfigured VM `mosel.ova` exported from VirtualBox.

### Booting the VM

The VM was exported and tested with VirtualBox 5.2.10; any newer version should
also work.  You can download VirtualBox for Linux, Solaris, Mac and Windows at
<https://www.virtualbox.org/>.  The VM needs 8GB of RAM and 4 CPU cores in its
default settings, but you can change the VM settings to 4GB of RAM and 2 CPU
cores to save resources (and slow down compilation), just make sure to allocate
2GB of RAM per CPU core.  When booted, the appliance should log in automatically
as the "aec" user.  In case that does not work, the password for this account
(and for "root" as well) is "aec".

After booting, if you are not using a US keyboard layout, you probably want to
configure the correct keyboard layout by double-clicking the corresponding icon
on the Desktop, and then choosing the "Layout" tab.  Please make sure to remove
all but your desired layout from the list; Xfce will otherwise not always pick
the right layout.

This appliance comes with our artifact already fully set up and compiled in the
`mosel` directory on the Desktop.  You can skip the next section and jump
directly to "Explore the artifact".

### Set up the artifact (already done in the VM)

These steps can be used to reproduce the artifact setup available in the VM,
using the contents of the `mosel.zip` file.

1. **Installing build-dependencies**

    We have tested everything with Coq 8.8.0.  We recommend to use opam to
    install the build-dependencies.  After installing opam with your OS package
    manager, do the following:

    ```
    opam repo add coq-released https://coq.inria.fr/opam/released
    opam repo add iris-dev https://gitlab.mpi-sws.org/FP/opam-dev.git
    opam install coq.8.8.0
    (cd iris && make build-dep)
    opam install coq-tlc.20180316
    ```

2. **Building everything**

    Type `./build.sh` in the root of the artifact to build everything. This
    script has been tested on Linux, and it should also work on Mac.  The build
    process can take an hour.

3. **Prepare for interactive use**

    To explore the artifact interactively with emacs+ProofGeneral or CoqIDE, you
    have two choices:
    - The easier choice is to run `make install` in the `iris`
      subdirectory. This will install Iris system-wide.
    - If you want to avoid system-wide modifications, you can set the
      environment variable `COQPATH` to the absolute path for the `coqlibdir`
      subdirectory that has been created during build.  This environment
      variable instructs Coq to load Iris from the artifact directory.

### Explore the artifact

Apart from MoSeL itself (which is part of the Iris development, as can be found
in the `iris/theories/bi` and `iris/theories/proofmode` directories), the
artifact contains a number of directories corresponding to the various
instantiations of MoSeL:

- `iris` : Iris as an instantiation of MoSeL. In particular, see the file
  `iris/theories/base_logic/upred.v` (Section 3.1 of the paper).
- `fairis` : Fairis as an instantiation of MoSeL (Section 3.2 of the paper)
- `ora` : the ordered RA model as an instantiation of MoSeL (Section 3.3 of the
  paper)
- `igps` : iGPS as an instantiation of MoSeL (Section 4.1 of the paper). See
  also `iris/theories/bi/monpred.v` for the generic definitions of monotone BI
  predicates in MoSeL (also described in Section 4.1 of the paper).
- `cfml` :  CFML as an instantiation of MoSeL (Section 4.2 of the paper)
- `chl` : CHL as an instantiation of MoSeL (Section 4.3 of the paper)

Moreover, we included a port of the Coq development accompanying the
[POPL'18 RustBelt paper](https://plv.mpi-sws.org/rustbelt/popl18/) in the
`lambda-rust` directory. This development used the original Iris Proof Mode
before, but has been ported to MoSeL.  We have also included the version of this
Coq development that works with the "old" Iris proofmode in the
`lambda-rust-orig`, and a diff file showing that less than 200 of the more than
15000 (non-whitespace, non-comment) lines changed.  The diff was created using
`git diff --no-index --stat --patch lambda-rust-orig/ lambda-rust/` with both
directories being clean.

You can verify that everything builds by running `./clean.sh && ./build.sh` --
however, this can easily take an hour.  If you only want to selectively rebuild
one of the projects, you can run `make clean && make -j4` inside that project
folder.

To interactively explore a particular file, double-click it.  That will open
emacs, with ProofGeneral set up.  Alternatively, you can right-click and select
"Open With" - "coqide" to open the file in CoqIDE (see also "Prepare for
interactive use" above).

Step-by-step instructions to evaluate the artifact
--------------------------------------------------

Below we describe some possible steps to evaluate the artifact.

1. **The BI++ interface**

    The core of MoSeL is its interface for BIs with a persistence modality
    (called BI++, see Section 2.4 and 5.1 of the paper). The Coq version of this
    interface can be found in `iris/theories/bi/interface.v`.

    In order to provide additional tactics for logics that have a later modality
    (like Iris), MoSeL also has a step-indexed BI++ interface called `sbi`. This
    is not discussed in the paper for reasons of space.

2. **The MoSeL tactics**

    The file `iris/ProofMode.md` contains a detailed list of all the MoSeL
    tactics. The Coq implementations of these tactics can be found in the
    directory `iris/theories/proofmode`, and is briefly described in Section 5.2
    of the paper.

3. **Examples from the paper**

    The easiest way to get started is to try out the examples from the paper,
    These can be found in `iris/tests/mosel_paper.v`. It is recommended to try
    variations of these examples to explore the various tactics of MoSeL.

    More advanced examples can be found in the files `iris/tests/proofmode.v`
    and `iris/tests/proofmode_monpred.v`.

4. **Examples of program verification in Iris using MoSeL**

    Some simple examples of program verification using Iris in MoSeL can be
    found in the `iris/theories/heap_lang/lib` directory. Although these
    examples are not novel, it is worth noting that their proofs are mostly the
    same as in Iris using the original Iris Proof Mode (IPM), i.e. they show
    that MoSeL is backwards compatible.

5. **Verification in a Linear Variant of Iris (Fairis)**

    The file `fairis/theories/tests/heap_lang.v` shows proofs of some simple
    refinements in Fairis using MoSeL -- as with the examples from Iris, the
    proofs are almost exactly the same as in the original Fairis paper (we took
    the opportunity to clean up some use of autogenerated hypothesis names).  To
    keep compatibility with the original Fairis notation, the affine modality is
    written as `⧆` in these scripts. To see an example of how the non-affine
    nature of the logic arises, see line 175, where to prove the goal at that
    point we must introduce an affine modality. The tactic `iAlways` succeeds in
    doing so because the context contains only intuitionistic assertions,

6. **Instances of the BI++ interface**

    In order to see how MoSeL can be instantiated, it is recommend to take a
    look at the following files:
    - `iris/theories/base_logic/upred.v` : The Iris model (Section 3.1 of the
      paper).
    - `iris/theories/bi/monpred.v` : The monotone predicates model, as used by
      the iGPS instance (Section 4.1 of the paper).
    - `ora/theories/logic/oupred.v` : The ordered RA model (Section 3.3 of the
      paper).

    Other instantiations can be found in the `fairis`, `cfml` and `chl`
    directories.

7. **Using MoSeL on non-Iris developments**

    As described in Section 4 of the paper, MoSeL can be used with separation
    logics that are quite different from Iris. Several examples are included in
    the artifact:
    - iGPS (Section 4.1): iGPS uses the monotone predicate model mentioned
      above, so there is no new BI instance to be found here.  The files in
      `igps/theories/examples` are several case studies implemented inside the
      iGPS logic, using MoSeL to avoid reasoning directly about views.
    - CFML (Section 4.2): The files `cfml/ExampleListProofMode.v` and
      `cfml/ExampleROProofMode.v` show how MoSeL can be used to prove programs
      with the CFML separation logic. The former uses the standard separation
      logic while the second proves specifications using the "read-only"
      version of CFML.

        The instantiation of MoSeL for CFML is done in the files
      `cfml/SepTactics.v` (in the ProofMode submodule, for the generic part of
      the instantiation), `cfml/LambdaSep.v` and `cfml/LambdaSepRO.v` (in the
      ProofMode submodule, for the part specific to the standard and
      read-only versions of CFML, respectively).
    - CHL (Section 4.3): The file `chl/theories/PredBi.v` shows how the BI++
      interface can be instantiated with CHL's assertion logic and contains
      various tactics. The tactics whose names are prefixed by `iCancel`
      reproduce much of the functionality of the CHL tactic `cancel` (which is
      defined in `chl/theories/SepAuto.v`) (we have several variants because the
      exact behavior of `cancel` is hard to specify, and we found it useful to be
      able to control how aggressively existentials are instantiated/eliminated).

        At the bottom of the file, the typeclass instances whose names are prefixed
      by `crash_xform_` extend MoSeL tactics with special support for the
      `crash_xform` predicate (e.g. by automatically distributing it over
      separating conjunction when using `iDestruct`, as described in the paper).

        The file `chl/theories/CacheIPM.v` is a modification of FSCQ's "buffer
      cache" verification using a mixture of MoSeL and pre-existing FSCQ
      automation tactics (compare with the original in `chl/theories/Cache.v`).
      The idea is not to entirely re-write the original proofs, but to show how
      MoSeL is compatible with the previous automation style, while making it
      easy to develop new automation from its primitives and carry out structured
      proofs in situations when automation cannot entirely discharge goals.
