From iris.base_logic.lib Require Import own invariants.
From iris.algebra Require Import agree.
From stdpp Require Import gmap.
From iris.proofmode Require Import tactics.

From igps Require Import viewpred proofmode.
From igps.base Require Import ghosts.

Definition escrowN : namespace := nroot .@ "escrow".

Section Escrows.

  Context `{!foundationG Σ}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).

  Definition exchange (P Q: vProp Σ) V : vProp Σ :=
    tc_opaque (monPred_in V ∗ ⎡inv escrowN (P V ∨ Q V)⎤)%I.

  (* TODO: 3rd argument was = instead of ≡ - why? *)
  Global Instance exchange_proper :
    Proper ((≡) ==> (≡) ==> (≡) ==> (≡)) exchange.
  Proof. solve_proper. Qed.

  Global Instance exchange_persistent P Q β :
    Persistent (exchange P Q β).
  Proof. rewrite /exchange /=. apply _. Qed.

  Global Instance exchange_ne n: Proper (dist_later n ==> dist_later n ==> (≡) ==> dist n) exchange.
  Proof. solve_contractive. Qed.

  Definition escrow P Q : vProp Σ :=
    (∃ β, exchange (<subj> P) Q β)%I.

  Global Instance escrow_proper : Proper ((≡) ==> (≡) ==> (≡)) escrow.
  Proof. intros ?? H ?? H2. apply: bi.exist_proper => ?. by rewrite H H2. Qed.

End Escrows.

(* "↭"	U21AD	 	# LEFT RIGHT WAVE ARROW *)
Notation "[ex P ↭ Q ]" := (exchange P Q)
  (at level 99, format "[ex  P  ↭  Q ]"): bi_scope.

Notation "[es P ⇝ Q ]" := (escrow P Q)
  (at level 99, format "[es  P  ⇝  Q ]"): bi_scope.

Section proof.

  Context `{!foundationG Σ}.

  Set Default Proof Using "Type".

  Lemma exchange_in P Q V:
    [ex P ↭ Q] V ⊢ monPred_in V.
  Proof. rewrite /exchange /=. iIntros "[$ _]". Qed.

  Lemma exchange_comm P Q V :
    [ex P ↭ Q] V ⊣⊢ [ex Q ↭ P] V.
  Proof. rewrite /exchange /=. by rewrite (comm bi_or). Qed.

  Lemma exchange_alloc P Q: (▷ (P ∨ Q) ={↑escrowN}=∗ ∃ V, [ex P ↭ Q] V)%I.
  Proof.
    iStartProof (uPred _). iIntros (V) "P".
    iExists V. rewrite /exchange /=. iSplitR; [done|].
    iApply (inv_alloc escrowN (↑escrowN) (P _ ∨ Q _)%I with "P").
  Qed.

  Lemma exchange_alloc_l P Q: (▷ P ={↑escrowN}=∗ ∃ V, [ex P ↭ Q] V)%I.
  Proof. iIntros. iApply exchange_alloc. eauto. Qed.

  Lemma exchange_alloc_r P Q: (▷ Q ={↑escrowN}=∗ ∃ V, [ex P ↭ Q] V)%I.
  Proof. iIntros. iApply exchange_alloc. eauto. Qed.

  Lemma exchange_elim P0 Q0 V (b : bool):
    let P := if b then P0 else Q0 in
    let Q := if b then Q0 else P0 in
    (P ∗ P → False) ⊢ ([ex P0 ↭ Q0] V ∗ ▷ ⎡P V⎤ ={↑escrowN}=∗ ▷ ⎡Q V⎤).
  Proof.
    intros P Q. iStartProof (uPred _). iIntros (V') "ND".
    rewrite /exchange /=. iIntros (V0 ?) "(#[% Inv] & P) /=".
    iAssert (inv escrowN (monPred_at P V ∨ monPred_at Q V))%I with "[Inv]" as "#Inv'".
    { destruct b; subst P Q; auto. by rewrite bi.or_comm. }
    iClear "Inv". iInv escrowN as "Inv" "HClose".
    iAssert (▷ (Q V ∗ (P V ∨ Q V)))%I with "[P Inv ND]" as "[Q Inv]".
      { iNext.
        iDestruct "Inv" as "[Inv|$]".
        - iExFalso. iApply ("ND" $! V0 with "[%//] [$P $Inv]").
        - by iLeft. }
    by iMod ("HClose" with "Inv").
  Qed.

  Lemma exchange_elim_l P Q V:
    (P ∗ P → False) ⊢ ([ex P ↭ Q] V ∗ ▷ ⎡P V⎤ ={↑escrowN}=∗ ▷ ⎡Q V⎤).
  Proof. exact: (exchange_elim _ _ _ true). Qed.

  Lemma exchange_exchange P0 Q0 V (b : bool):
    let P := if b then P0 else Q0 in
    let Q := if b then Q0 else P0 in
    (P ∗ P → False) ⊢ ([ex P0 ↭ Q0] V ∗ ▷ ⎡P V⎤ ={↑escrowN}=∗ ▷ Q).
  Proof.
    intros.
    iIntros "ND (#Ex & P)".
    iAssert ([ex P ↭ Q] V)%I with "[Ex]" as "Ex'".
    { destruct b; auto. by rewrite exchange_comm. }
    iDestruct (exchange_in with "Ex") as "aSeen".
    iDestruct (exchange_elim_l with "ND [$Ex' $P]") as ">Q".
    iModIntro. iNext. iApply monPred_in_elim; auto.
  Qed.

  Lemma exchange_left P Q V:
    (P ∗ P → False) ⊢ ([ex P ↭ Q] V ∗ ▷ ⎡P V⎤ ={↑escrowN}=∗ ▷ Q).
  Proof. exact: (exchange_exchange _ _ _ true). Qed.

  Lemma exchange_elim_r P Q V:
    (Q ∗ Q → False) ⊢ ([ex P ↭ Q] V ∗ ▷ ⎡Q V⎤ ={↑escrowN}=∗ ▷ ⎡P V⎤).
  Proof. exact: (exchange_elim _ _ _ false). Qed.

  Lemma exchange_right P Q V:
    (Q ∗ Q → False) ⊢ ([ex P ↭ Q] V ∗ ▷ ⎡Q V⎤ ={↑escrowN}=∗ ▷ P).
  Proof. exact: (exchange_exchange _ _ _ false). Qed.

  Lemma escrow_alloc P Q:
    (▷ Q ={↑escrowN}=∗ [es P ⇝ Q])%I.
  Proof. by apply exchange_alloc_r. Qed.

  Lemma escrow_apply_obj P Q:
    (P ∗ P → False) ⊢ ([es P ⇝ Q] ∗ ▷ <subj> P ={↑escrowN}=∗ ▷ Q).
  Proof.
    iStartProof (uPred _). iIntros (V0) "ND".
    rewrite /escrow /exchange /=. iIntros (V ?) "(Esc & oP)".
    iDestruct "Esc" as (V') "[% ?]".
    iInv escrowN as "Inv" "HClose".
    iAssert (▷ (Q V' ∗ ((<subj> P)%I V' ∨ Q V')))%I with "[ND oP Inv]" as "[Q Inv]".
      { iNext. iDestruct "Inv" as "[Inv|$]".
        - iExFalso.
          iDestruct "oP" as (V1) "P1".
          iDestruct "Inv" as (V2) "P2".
          iApply ("ND" $! (V1 ⊔ V2 ⊔ V0) with "[%] [$P1 $P2]"). solve_jsl.
        - iLeft. iDestruct "oP" as (?) "oP". by iExists _.  }
    iMod ("HClose" with "Inv"). auto.
  Qed.

  Lemma escrow_apply P Q:
    (P ∗ P → False) ⊢ ([es P ⇝ Q] ∗ ▷ P ={↑escrowN}=∗ ▷ Q).
  Proof.
    iIntros "ND (Esc & oP)".
    iApply (escrow_apply_obj with "ND [-$Esc]"). by iNext.
  Qed.
End proof.
