From iris.base_logic Require Import upred derived.
From iris.proofmode Require Import tactics.
From igps Require Import derived_vProp.
From igps Require Export weakestpre notation wp_tactics_vProp lifting_vProp.
Import uPred lang.base.

Definition for_loop i (cond incr body : expr) := ((rec: "loop" "i" :=
  if: cond "i" then body "i";; "loop" (incr "i") else #()) i)%E.

Notation "'for:' ( i := v ; cond ; incr ) body" :=
  (for_loop v (λ: i%bind, cond%E) (λ: i%bind, incr%E) (λ: i%bind, body%E))
  (at level 102) : expr_scope.

(* For now, force the induction variable to be a number. *)
Lemma wp_for `{ghosts.foundationG Σ} E (v : Z) (f : Z -> Z) (c : Z -> bool)
  (cond incr body : expr) Φ
  `{_ : Closed nil cond} `{_ : Closed nil incr} `{_ : Closed nil body}:
  □(∀i : Z, {{{ ⌜c i = true⌝ ∗ Φ i }}} body #i @ E {{{ v, RET v; Φ (f i) }}}) -∗
  □(∀i : Z, {{{True}}} incr #i @ E {{{ v, RET v; ⌜v = #(f i)⌝ }}}) -∗
  □(∀i : Z, {{{True}}} cond #i @ E {{{ v, RET v; ⌜v = #(lit_of_bool (c i))⌝ }}}) -∗
   {{{ Φ v }}} for_loop #v cond incr body @ E {{{ RET #(); ∃v, ⌜c v = false⌝ ∗ Φ v }}}.
Proof.
  intros; iIntros "#Hbody #Hincr #Hcond !#" (?) "Hv Hpost".
  unfold for_loop.
  iLöb as "I" forall (v).
  wp_rec.
  wp_bind (cond _).
  iApply "Hcond"; first done.
  iNext; iIntros (?) "%"; subst.
  destruct (c v) eqn: Hc.
  - wp_if_true.
    wp_bind (body _).
    iApply ("Hbody" with "[$Hv]"); first done.
    iNext; iIntros (?) "Hf".
    wp_seq.
    wp_bind (incr _).
    iApply "Hincr"; first done.
    iNext; iIntros (?) "%"; subst.
    iApply ("I" with "Hf Hpost").
  - wp_if_false.
    iApply "Hpost"; iExists v; auto.
Qed.

Lemma wp_for_simple `{ghosts.foundationG Σ} E (v : Z) (i : string) N (body : expr) Φ
  `{_ : Closed nil body}:
  □(∀i : Z, {{{ ⌜i < N⌝ ∗ Φ i }}} body #i @ E {{{ v, RET v; Φ (i + 1) }}}) -∗
   {{{ Φ v }}} for_loop #v (λ: i, i < #N) (λ: i, i + #1) body @ E {{{ RET #(); ∃v, ⌜v >= N⌝ ∗ Φ v }}}.
Proof.
  iIntros "#Hbody !#" (?) "Hpre Hpost". iApply (wp_for _ _ (fun i => i + 1) (fun i => i <? N) with "[#] [#] [#] Hpre").
  - unfold Closed, is_closed; simpl.
    unfold bool_decide, decide_rel.
    destruct (elem_of_list_dec _); auto.
    contradiction n; constructor.
  - unfold Closed, is_closed; simpl.
    unfold bool_decide, decide_rel.
    destruct (elem_of_list_dec _); auto.
    contradiction n; constructor.
  - iIntros "!#" (??) "!# [% HP] H".
    apply Z.ltb_lt in H1.
    iApply ("Hbody" with "[$HP] H"); auto.
  - iIntros "!#" (??) "!# _ H".
    assert (Closed (<> :b: i :b: nil) (i + #1)).
    { unfold Closed, is_closed; simpl.
      unfold bool_decide, decide_rel.
      destruct (elem_of_list_dec _); auto.
      set_solver.
    }
    wp_let.
    rewrite decide_True; [|done]. (* FIXME: all of the above! *)
    wp_op.
    iApply "H"; auto.
  - iIntros "!#" (j?) "!# _ H".
    assert (Closed (<> :b: i :b: nil) (i < #N)).
    { unfold Closed, is_closed; simpl.
      unfold bool_decide, decide_rel.
      destruct (elem_of_list_dec _); auto.
      set_solver.
    }
    wp_let.
    rewrite decide_True; [|done]. (* FIXME: all of the above! *)
    pose proof (Zlt_cases j N).
    wp_op; destruct (j <? N); iApply "H"; auto; iPureIntro.
    + by rewrite bool_decide_true.
    + rewrite bool_decide_false; [done|omega].
  - iNext; iIntros "H"; iDestruct "H" as (?) "[%?]".
    iApply "Hpost".
    apply Z.ltb_nlt in H2; eauto.
Qed.

Instance for_closed l i cond incr body {Hi : Closed l i} {Hc : Closed l cond}
  {Hn : Closed l incr} {Hb : Closed l body} : Closed l (for_loop i cond incr body).
Proof.
  unfold for_loop, Closed; simpl.
  rewrite !andb_true_r.
  assert (l ⊆ "loop" :: "i" :: l) as Hl.
  { repeat intro.
    do 2 constructor; done. }
  pose proof (is_closed_weaken _ _ _ Hc Hl).
  pose proof (is_closed_weaken _ _ _ Hn Hl).
  pose proof (is_closed_weaken _ _ _ Hb Hl).
  repeat (destruct (is_closed _); try done).
Qed.

Lemma subst_for s e i cond incr body (Hs : s ≠ "loop" /\ s ≠ "i") :
  subst s e (for_loop i cond incr body) =
  for_loop (subst s e i) (subst s e cond) (subst s e incr) (subst s e body).
Proof.
  unfold for_loop, subst; fold subst.
  destruct Hs, (decide _); last by contradiction n; split; congruence.
  destruct (decide _); first done.
  destruct (decide _); last by contradiction n0.
  destruct (decide _); done.
Qed.

Opaque for_loop.
