From iris.algebra Require Import frac_auth.
From iris.program_logic Require Export weakestpre.
From iris.proofmode Require Import tactics.

From igps Require Export malloc repeat notation escrows.
From igps Require Import persistor viewpred proofmode weakestpre protocols logically_atomic
  invariants fork.
From igps.gps Require Import shared plain recursive.
From igps.examples Require Import abs_hashtable snapshot hist_protocol repeat for_loop list_lemmas.
Import uPred lang.base.

Lemma NoDup_remove: forall {A} (l l' : list A) a, NoDup (l ++ a :: l') ->
  NoDup (l ++ l') /\ a ∉ (l ++ l')%list.
Proof.
  intros.
  apply NoDup_app in H as (? & Ha & Hl'); inversion Hl'; subst.
  rewrite NoDup_app; repeat split; auto.
  - intros ? Hin ?; contradiction (Ha _ Hin); constructor; done.
  - rewrite elem_of_app; intros [X|]; last done.
    contradiction (Ha _ X); constructor.
Qed.

Lemma filter_find_count : forall {A} (P : A -> Prop) {H : ∀ x, Decision (P x)}
  l li (Hunique : NoDup li) (Hli : Forall (fun i => exists x, l !! i = Some x /\ P x) li)
  (Hrest : forall i x, i ∉ li -> l !! i = Some x -> ~P x),
  length (filter P l) = length li.
Proof.
  induction l using rev_ind; simpl; intros.
  - destruct li; auto.
    inversion Hli; subst.
    destruct H2 as (? & ? & ?); done.
  - rewrite filter_app; unfold filter at 2; simpl.
    destruct (decide (length l ∈ li)).
    + assert (P x).
      { eapply Forall_forall in Hli; last done.
        destruct Hli as (? & X & ?); subst.
        rewrite list_lookup_middle in X; last done; inversion X; done. }
      destruct (decide _); last done.
      destruct (elem_of_list_split _ _ e) as (li1 & li2 & ?); subst.
      apply NoDup_remove in Hunique as [].
      rewrite !app_length (IHl (li1 ++ li2)%list); auto.
      rewrite app_length; simpl; omega.
      * rewrite Forall_forall; intros j Hin.
        eapply Forall_forall in Hli;
          last by rewrite elem_of_app; apply elem_of_app in Hin as [|]; eauto; repeat constructor.
        destruct Hli as (? & Hi & ?).
        pose proof (lookup_lt_Some _ _ _ Hi) as Hlt.
        rewrite lookup_app_l in Hi; eauto.
        rewrite app_length in Hlt; simpl in Hlt.
        destruct (decide (j = length l)); [subst; done | omega].
      * intros ?? Hout Hi.
        pose proof (lookup_lt_Some _ _ _ Hi) as Hlt.
        apply (Hrest i).
        { rewrite elem_of_app; intro Hin; contradiction Hout.
          rewrite elem_of_app; destruct Hin as [|X]; last inversion X; auto; omega. }
        { rewrite lookup_app_l; done. }
    + pose proof (Hrest _ _ n (list_lookup_middle _ _ _ _ eq_refl)).
      destruct (decide _); first done.
      rewrite app_length Nat.add_0_r; apply IHl; auto.
      * eapply Forall_impl; first done.
        intros j (? & Hi & ?).
        pose proof (lookup_lt_Some _ _ _ Hi) as Hlt.
        destruct (decide (j = length l));
          first by rewrite list_lookup_middle in Hi; last done; inversion Hi; subst.
        rewrite lookup_app_l in Hi; eauto.
        rewrite app_length in Hlt; simpl in Hlt; omega.
      * intros ?? Hout Hi.
        pose proof (lookup_lt_Some _ _ _ Hi) as Hlt.
        apply (Hrest i); auto.
        rewrite lookup_app_l; done.
Qed.

Section hashtable.

Definition const := 654435761.

Instance hf1 : hash_fun := { size := Z.to_nat 16384; hash i := Z.to_nat ((i * const) mod 16384) }.
Proof.
  - hnf.
    by apply (Z2Nat.inj_le 1 16384); compute.
  - intro.
    lapply (Z_mod_lt (i * const) 16384); last by compute.
    by intros []; split; [apply (Z2Nat.inj_le 0) | apply (Z2Nat.inj_lt _ 16384)].
Defined.

Definition integer_hash : val := λ: "i", "i" * #654435761.

Context (m_entries : loc) (results : loc).

Local Notation key := 0.
Local Notation value := 1.

(* There's no default location, so make sure i is in-bounds. *)
Local Notation undef := 1%positive.

Section code.

Definition get_item : val := λ: "key",
  (rec: "f" "idx" :=
     let: "idx" := "idx" `mod` #size in
     let: "i0" := [ #m_entries + "idx"]_na in
     let: "i" := (loc) "i0" in
     let: "probed_key" := ["i" + #key]_at in
     if: "probed_key" = "key" then ["i" + #value]_at
     else if: "probed_key" = #0 then #0
     else "f" ("idx" + #1)) (integer_hash "key").

Definition set_key : val := λ: "key",
  (rec: "f" "idx" :=
     let: "idx" := "idx" `mod` #size in
     let: "i0" := [ #m_entries + "idx"]_na in
     let: "i" := (loc) "i0" in
     let: "probed_key" := ["i" + #key]_at in
     (if: "probed_key" ≠ "key" then
        if: "probed_key" ≠ #0 then "f" ("idx" + #1)
        else let: "result" := base.CAS ("i" + #key) #0 "key" in
          if: ~"result" then let: "probed_key" := ["i" + #key]_at in
            if: "probed_key" ≠ "key" then "f" ("idx" + #1) else "i"
          else "i"
      else "i")) (integer_hash "key").

Definition set_item : val := λ: "key" "value",
  let: "i" := set_key "key" in
  ["i" + #value]_at <- "value".

Definition add_item : val := λ: "key" "value",
  let: "i" := set_key "key" in
  base.CAS ("i" + #value) #0 "value".

Definition init_table : val := λ: <>,
  for: ("i" := #0; "i" < #size; "i" + #1)
    let: "p" := malloc #2 in
    ["p" + #key]_na <- #0;;
    ["p" + #value]_na <- #0;;
    [ #m_entries + "i"]_na <- (int) "p".

(* Coq has trouble proving that for loops are closed, even with the instance in for_loop, but
   this can easily be proved by reduction. *)
Program Definition test : val := @RecV <> "t" (λ: "l",
  (for: ("i" := #0; "i" < #3; "i" + #1)
    let: "r" := add_item ("i" + #1) #1 in
    if: "r" then [ #results + "t"]_na <- ([ #results + "t"]_na + #1) else #());;
  ["l"]_at <- #1) _.
Next Obligation.
Proof.
  apply I.
Qed.

Program Definition main : val := @RecV <> <>
  (init_table #();;
  let: "t" := malloc #3 in
  (for: ("i" := #0; "i" < #3; "i" + #1)
     let: "ti" := "t" + "i" in
     ["ti"]_na <- #0;;
     [ #results + "i"]_na <- #0;;
     Fork (test "i" "ti"));;
  let: "total" := Alloc in
  ["total"]_na <- #0;;
  (for: ("i" := #0; "i" < #3; "i" + #1)
     repeat ["t" + "i"]_at;;
     ["total"]_na <- (["total"]_na + [ #results + "i"]_na));;
  ["total"]_na) _.
Next Obligation.
Proof.
  apply I.
Qed.

End code.

Definition value_of (m : gmap nat Z) v := exists n, m !! n = Some v /\
  forall n', (n < n')%nat -> m !! n' = None.

Definition zero_ord a b := a = 0 \/ a = b.

Canonical Structure zeroProtocol : protocolT := ProtocolT _ zero_ord.

Instance zeroPrt_facts : protocol_facts zeroProtocol.
Proof. esplit; try apply _.
  - intro; simpl; unfold zero_ord; auto.
  - intros ???[|][|]; subst; simpl; unfold zero_ord; auto.
Qed.

Instance zero_unit : Unit ZC := 0.

Instance zero_join : Join zero_ord :=
  { join a b := if decide (a = 0) then Some b else if decide (b = 0) then Some a
      else if decide (a = b) then Some a else None }.
Proof.
  - constructor; first constructor; repeat intro; unfold zero_ord in *; auto.
    + destruct H, H0; subst; auto.
    + destruct H, H0; subst; auto.
  - constructor; auto.
  - intros.
    destruct (decide (a = 0)), (decide (b = 0)); last (destruct (decide (a = b))); inversion H;
      subst; eauto; destruct (decide (c = 0)); subst; eauto; try done.
    destruct (decide _); last destruct (decide _); inversion H0; subst; do 2 eexists; eauto;
      destruct (decide _); auto; rewrite decide_left; auto.
  - repeat intro.
    destruct (decide _), (decide _); subst; auto.
    destruct (decide _), (decide _); subst; auto; done.
  - intros.
    unfold zero_ord.
    destruct (decide _); subst; auto.
    destruct (decide _); last destruct (decide _); inversion H; auto.
  - unfold zero_ord; intros; destruct (decide _); auto.
    destruct H; [contradiction | subst].
    destruct (decide _); auto.
    rewrite decide_left; auto.
Qed.

Section Interpretation.
  Context `{foundationG Σ} `{histG : inG Σ (snapR (hist_join ZC))}
    {ZG : inG Σ (snapR zero_join)} `{inG Σ (exclR unitC)}.

  Definition vP (γ : gname) : interpC Σ histProtocol := (fun b _ s v =>
    ⌜value_of s v /\ (v = 0 -> forall j v', s !! j = Some v' -> v' = 0)%Z⌝ ∗
    ⎡own γ (if b then Master (1/2) s else Snapshot s)⎤)%I.

  Definition kP (γ : gname) : interpC Σ zeroProtocol := (fun b _ s v =>
    ⌜(s = v)%Z⌝ ∗ ⎡own γ (if b then Master (1/2) v else Snapshot v)⎤)%I.

  (* adapted from message_passing *)
  Definition XE (γ: gname) (P : vProp Σ) := [es ⎡own γ (Excl ())⎤ ⇝ P ]%I.

  Definition onceP (γ : gname) (P : vProp Σ) : interpC Σ boolProtocol := (fun b _ s v =>
    ⌜(if s then v = 1 else v = 0)%Z⌝ ∗ if s then XE γ P else True)%I.

End Interpretation.

Definition table_incl (T1 T2 : list (Z * gmap nat Z)) := forall i,
  zero_ord (fst (nth i T1 (0, ∅))) (fst (nth i T2 (0, ∅))) /\
  map_subseteq (snd (nth i T1 (0, ∅))) (snd (nth i T2 (0, ∅))).

Instance table_incl_refl : Reflexive table_incl.
Proof.
  done.
Qed.

Instance table_incl_trans : Transitive table_incl.
Proof.
  intros ??? Hincl1 Hincl2 ?.
  destruct (Hincl1 i), (Hincl2 i).
  split; etransitivity; eauto.
Qed.

Lemma table_incl_upd : forall T (i : nat) k0 v0 k v, T !! i = Some (k0, v0) ->
  zero_ord k0 k -> v0 ⊆ v -> table_incl T (<[i := (k, v)]>T).
Proof.
  repeat intro.
  rewrite nth_insert; last (apply lookup_lt_is_Some_1; eauto).
  destruct (decide _); last done.
  subst; rewrite nth_lookup H; auto.
Qed.

Section proof.
  Context `{fG : foundationG Σ}
          `{persG : persistorG Σ}
          {histG : inG Σ (snapR (hist_join ZC))}
          {ZG : inG Σ (snapR zero_join)}
          {hashG : inG Σ (exclR (Z -c> optionC (gmapC nat ZC)))}
          `{gpGk: !gpsG Σ zeroProtocol _}
          `{gpGv: !gpsG Σ histProtocol _}
          `{gpGb: !gpsG Σ boolProtocol _}
          `{agreeGk : !gps_agreeG Σ zeroProtocol}
          `{agreeGv : !gps_agreeG Σ histProtocol}
          `{agreeGb : !gps_agreeG Σ boolProtocol}
          {cG : cancelable_invariants.cinvG Σ}.

  Set Default Proof Using "Type".

  Lemma integer_hash_spec i: {{{ True }}}
    integer_hash #i {{{ j, RET #j; ⌜j = (i * const)%Z⌝ }}}.
  Proof.
    intros; iIntros "_ Hpost".
    wp_lam.
    wp_op.
    by iApply "Hpost".
  Qed.

  Notation Tnth i T := (nth i T (0, ∅)).

  Definition wf_table (T : list (Z * gmap nat Z)) := forall k i, k ≠ 0 ->
    fst (Tnth i T) = k -> hlookup (map fst T) k = Some i.

  Definition hashtable_entry x lgk lgv i : vProp Σ :=
    let '(k, lv) := x in
    (⌜exists v, value_of lv v /\ (k = 0 -> v = 0) /\
     (v = 0 -> forall j v', lv !! j = Some v' -> v' = 0)⌝ ∗
    ⎡own (nth i lgk undef) (Master (1/2) k)⎤ ∗ ⎡own (nth i lgv undef) (Master (1/2) lv)⎤)%I.

  Definition hashtable H (g : gname) lgk lgv : vProp Σ :=
    (∃ T : _, ⌜length T = size /\ wf_table T /\
       forall k lv, H k = Some lv <-> (k, lv) ∈ T /\ exists v, value_of lv v /\ v ≠ 0⌝ ∗
    ⎡@own _ _ hashG g (Excl H)⎤ ∗ [∗ list] i↦x ∈ T, hashtable_entry x lgk lgv i)%I.

  Definition hashtable_entry_A x lgk lgv entries i : vProp Σ :=
    let p := nth i entries undef in
    let '(k, lv) := x in (⌜exists v, value_of lv v /\ (k = 0 -> v = 0) /\
     (v = 0 -> forall j v', lv !! j = Some v' -> v' = 0)⌝ ∗
     [PP (ZplusPos key p) in k | kP (nth i lgk undef)] ∗
     [PP (ZplusPos value p) in lv | vP (nth i lgv undef)])%I.

  Instance hashtable_entry_persistent x lgk lgv entries i:
    Persistent (hashtable_entry_A x lgk lgv entries i).
  Proof.
    destruct x; apply _.
  Qed.

  Definition hashtable_A T lgk lgv entries :=
    ([∗ list] i↦x ∈ T, hashtable_entry_A x lgk lgv entries i)%I.

  Opaque integer_hash const size.

  Definition log0 : gmap nat Z := <[O:=0]>∅.

  Lemma log0_lookup : forall i x, log0 !! i = Some x -> i = O /\ x = 0.
  Proof.
    destruct i; intro.
    - rewrite lookup_insert; split; congruence.
    - rewrite lookup_insert_ne; done.
  Qed.

  Lemma init_table_spec :
    {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ [∗ list] i↦x ∈ replicate size 0, ZplusPos i m_entries ↦ ? }}}
    init_table #()
    {{{ v, RET v; ∃ entries g lgk lgv, ⌜length entries = size⌝ ∗
      ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦ Z.pos x) ∗
      hashtable_A (replicate size (0, log0)) lgk lgv entries ∗
      hashtable (fun _ => None) g lgk lgv }}}.
  Proof.
    intros; iIntros "(#Hctx & #Hctx' & Hentries) Hpost".
    unfold init_table.
    iMod (@own_alloc _ _ hashG (Excl (fun _ => None))) as (g) "?".
    { hnf; auto. }
    wp_seq.
    iApply (wp_for_simple with "[] [Hentries]").
    - instantiate (1 := (fun i : Z => ∃ entries lgk lgv,
        ⌜0 <= i <= size /\ length entries = Z.to_nat i /\ length lgk = Z.to_nat i /\
         length lgv = Z.to_nat i⌝ ∗
        ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦ Z.pos x) ∗
        ([∗ list] j↦x ∈ replicate (size - Z.to_nat i) 0, (ZplusPos (j + i) m_entries) ↦ ?) ∗
        ([∗ list] i↦x ∈ replicate (Z.to_nat i) (0, log0), hashtable_entry_A x lgk lgv entries i) ∗
        ([∗ list] i↦x ∈ replicate (Z.to_nat i) (0, log0), hashtable_entry x lgk lgv i))%I).
      iIntros "!#" (i ?) "!# (% & H) Hpost".
      iDestruct "H" as (entries lgk lgv) "((% & % & % & %) & Hentries & Hrest & HA & HH)".
      wp_let.
      wp_bind (malloc _).
      iApply wp_mask_mono; first auto.
      iApply (malloc_spec with "Hctx"); [omega|done|].
      iNext; iIntros (p) "(Hk & Hv & _)".
      wp_let.
      iMod (own_alloc (Master 1 0 ⋅ Snapshot 0)) as (gk) "Hgk".
      { apply init_valid. }
      iMod (own_alloc (Master 1 log0 ⋅ Snapshot log0)) as (gv) "Hgv".
      { apply init_valid. }
      rewrite -half_join -!master_op !own_op.
      iDestruct "Hgk" as "[[Hgk1 Hgk2] Hgk3]"; iDestruct "Hgv" as "[[Hgv1 Hgv2] Hgv3]".
      wp_bind ([_]_na <- _)%E.
      wp_op.
      iApply (GPS_PP_Init (p:=tt) (kP gk) 0 with "[$Hctx $Hctx' $Hk $Hgk2 $Hgk3]");
        [done|done|done|].
      iNext; iIntros "Hgk".
      wp_seq.
      wp_bind ([_]_na <- _)%E.
      wp_op.
      assert (value_of log0 0 /\ ∀ j v', log0 !! j = Some v' → v' = 0) as [].
      { split.
        - exists O.
          split; first apply lookup_insert.
          intros; rewrite lookup_insert_ne; auto; omega.
        - intros ???%log0_lookup; tauto. }
      iApply (GPS_PP_Init (p:=tt) (vP gv) log0 with "[$Hctx $Hctx' $Hv $Hgv2 $Hgv3]");
        [done|done|done|].
      iNext; iIntros "Hgv".
      wp_seq.
      wp_op.
      wp_op.
      destruct (size - Z.to_nat i)%nat eqn: Hi.
      { apply Z2Nat.inj_lt in H; try omega.
        rewrite Nat2Z.id in H; omega. }
      simpl replicate; rewrite big_sepL_cons.
      iDestruct "Hrest" as "[Hi Hrest]".
      rewrite Z.add_0_l.
      iApply (na_write with "[$Hctx $Hi]"); first done.
      iNext; iIntros "Hi".
      iApply "Hpost".
      iExists (entries ++ p :: nil)%list, (lgk ++ gk :: nil)%list, (lgv ++ gv :: nil)%list.
      rewrite Z2Nat.inj_add; try omega.
      rewrite replicate_plus !big_sepL_app; simpl.
      rewrite replicate_length !app_nth2; try omega.
      rewrite !Nat.add_0_r.
      replace (length entries) with (Z.to_nat i); replace (length lgk) with (Z.to_nat i);
        replace (length lgv) with (Z.to_nat i); rewrite Nat.sub_diag; simpl.
      rewrite Z2Nat.id; last omega.
      iFrame.
      change (Z.to_nat 1) with 1%nat.
      iSplit.
      { iPureIntro; rewrite !app_length; simpl; omega. }
      iSplitL "Hrest"; [|iSplitL "HA"].
      + replace (size - (Z.to_nat i + 1))%nat with n by omega.
        iApply (big_sepL_proper with "Hrest"); intros; simpl.
        rewrite Zpos_P_of_succ_nat.
        unfold Z.succ; rewrite <- Z.add_assoc, (Z.add_comm i 1); done.
      + iSplit; last eauto.
        iApply (big_sepL_proper with "HA"); intros ? (?, ?) ?%lookup_replicate_1; simpl.
        unfold hashtable_entry_A.
        rewrite !app_nth1; auto; omega.
      + iSplit; last eauto.
        iApply (big_sepL_proper with "HH"); intros ? (?, ?) ?%lookup_replicate_1; simpl.
        unfold hashtable_entry.
        rewrite !app_nth1; auto; omega.
    - iExists nil, nil, nil; simpl; repeat iSplit; try done.
      rewrite Nat.sub_0_r; iApply (big_sepL_proper with "Hentries").
      simpl; intros; rewrite Z.add_0_r; done.
    - iNext; iIntros "H"; iDestruct "H" as (i) "(% & H)".
      iDestruct "H" as (entries lgk lgv) "((% & % & % & %) & ? & ? & ? & ?)".
      assert (i = size) by omega; subst.
      rewrite Nat2Z.id in H1 |- *.
      iApply "Hpost".
      iExists entries, g, lgk, lgv; iFrame.
      iSplit; first done.
      iExists (replicate size (0, log0)); iFrame; iPureIntro.
      split; first (rewrite replicate_length; reflexivity).
      split.
      + intros ?? Hk Hi; absurd (k = 0); first done.
        destruct (decide (i < size)%nat).
        * by rewrite nth_lookup lookup_replicate_2 in Hi.
        * rewrite nth_overflow in Hi; first done.
          rewrite replicate_length; omega.
      + split; first done.
        intros [X%elem_of_replicate_inv (? & (? & Hx & ?) & ?)]; inversion X; subst.
        apply log0_lookup in Hx as []; contradiction.
  Qed.

  Lemma size_pos' : (size > 0)%Z.
  Proof.
    apply Z.lt_gt, (Nat2Z.inj_lt 0), size_pos.
  Qed.

  Lemma big_sepL_insert : forall {A} (l : list A) Φ i x,
    ((([∗ list] k↦x ∈ l, Φ k x) : vProp Σ) -∗ Φ i x -∗ [∗ list] k↦x ∈ <[i:=x]>l, Φ k x)%I.
  Proof.
    induction l; intros; iIntros "H ?"; auto.
    iDestruct "H" as "[? Hlist]".
    destruct i; simpl; iFrame.
    by iApply (IHl with "Hlist").
  Qed.

  (* When we're not making changes, we can use wand instead of this, but this is useful for
     modifying an element in a list of resources. *)
  Lemma big_sepL_remove : forall {A} (l : list A) Φ i x, l !! i = Some x ->
    (([∗ list] k↦x ∈ l, Φ k x) : vProp Σ) -∗
    Φ i x ∗ [∗ list] k↦x ∈ l, if decide (k = i) then emp else Φ k x.
  Proof.
    induction l; destruct i; simpl in *; try done; intros; iIntros "[? Hrest]".
    - inversion H; subst; iFrame.
    - iFrame.
      apply (IHl (fun i x => Φ (S i) x)) in H.
      iPoseProof (H with "Hrest") as "[? Hrest]"; iFrame.
      iApply big_sepL_proper; last done.
      intros; simpl.
      destruct (decide _), (decide _); auto; omega.
  Qed.

  Lemma big_sepL_replace : forall {A} (l : list A) Φ i x,
    ([∗ list] k↦x ∈ l, if decide (k = i) then emp else Φ k x) -∗ Φ i x -∗
    ([∗ list] k↦x ∈ <[i:=x]>l, Φ k x) : vProp Σ.
  Proof.
    induction l; intros; iIntros "H ?"; simpl; auto.
    iDestruct "H" as "[? Hlist]".
    destruct i; simpl; iFrame.
    iApply (IHl (fun k x => Φ (S k) x) with "[Hlist]"); last done.
    iApply big_sepL_proper; last done.
    intros; simpl.
    destruct (decide _), (decide _); auto; omega.
  Qed.

  Definition base_mask : coPset := ↑physN ∪ ↑fracN ∪ ↑persist_locN.

  Lemma phys_base : ↑physN ⊆ base_mask.
  Proof.
    etransitivity; apply union_subseteq_l.
  Qed.

  Lemma frac_base {A} {_ : EqDecision A} {_ : Countable A} (i : A) :
    ↑namespaces.ndot fracN i ⊆ base_mask.
  Proof.
    etransitivity; last apply union_subseteq_l.
    etransitivity; last apply union_subseteq_r.
    apply namespaces.nclose_subseteq.
  Qed.

  Lemma persist_base {A} {_ : EqDecision A} {_ : Countable A} (i : A) :
    ↑namespaces.ndot persist_locN i ⊆ base_mask.
  Proof.
    etransitivity; last apply union_subseteq_r.
    apply namespaces.nclose_subseteq.
  Qed.

  Hint Resolve phys_base frac_base persist_base.

  Opaque hash.

  Lemma get_item_spec k q (entries : list loc) g lgk lgv T0 (Hk : k ≠ 0)
    (Hlen : length entries = size) (HlenT0 : length T0 = size):
    ⎡PSCtx⎤ ∗ ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
        hashtable_A T0 lgk lgv entries -∗
    atomic_wp (fun H => hashtable H g lgk lgv)%I
      (fun H val => ∃ lv : _, ∃ v : Z, ⌜val = LitV v⌝ ∗
        (* private *)
        ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
        (∃ T : _, ⌜table_incl T0 T /\ ∃ i, hlookup (map fst T) k = Some i /\ snd (Tnth i T) = lv⌝ ∗
          hashtable_A T lgk lgv entries) ∗
        (* public *)
        ⌜value_of lv v /\ (v ≠ 0 -> exists lv', H k = Some lv' /\ lv ⊆ lv')⌝ ∗
          hashtable H g lgk lgv)%I
      base_mask ⊤ (get_item #k).
  Proof.
    unfold atomic_wp, atomic_shift.
    iIntros "(#kI & Hentries & #HA)" (Φ) "H".
    iDestruct "H" as (P) "(HP & #Hshift)".
    match goal with |-context[(▷ P ={_,_}=∗ ?R)%I] => set (Q := R) end.

    unfold get_item.
    wp_let.
    wp_bind (integer_hash #k)%E.
    iApply integer_hash_spec; first done.
    iNext. iIntros (??); subst.

    (* loop invariant *)
    iAssert (∃i : nat, ∃T, ⌜(Z.to_nat (k * const mod size) = (i + hash k) mod size)%nat /\
      (i < size)%nat /\ length T = size /\ table_incl T0 T /\
      Forall (fun z => z ≠ 0 /\ z ≠ k) (sublist 0 i (rebase (map fst T) (hash k)))⌝ ∗
      hashtable_A T lgk lgv entries)%I as "Hpre".
    { iExists O, T0; iSplit; last by simpl.
      iPureIntro; rewrite sublist_nil; repeat split; try done.
      - simpl.
        change size with (Z.to_nat 16384); rewrite Z2Nat.id; last by compute.
        rewrite Nat.mod_small; first done.
        apply Nat2Z.inj_lt; rewrite !Z2Nat.id; try omega; apply Z_mod_lt; by compute.
      - apply size_pos. }
    set (i1 := (k * const)%Z); clearbody i1.
    iLöb as "IH" forall (i1).
    iDestruct "HA" as "_".

    wp_rec.
    wp_op. wp_lam.
    wp_op.
    wp_bind ([_]_na)%E.
    iDestruct "Hpre" as (i T) "(% & #HA)".
    destruct H as (Hi1 & Hi & HlenT & Hincl & Hseen).
    assert (Z.to_nat (i1 `mod` size) < size)%nat.
    { apply Nat2Z.inj_lt.
      rewrite Z2Nat.id; apply Z_mod_lt, size_pos'. }
    destruct (lookup_lt_is_Some_2 entries (Z.to_nat (i1 `mod` size))) as [p Hp];
      first by rewrite Hlen.
    iPoseProof (big_sepL_lookup_acc with "Hentries") as "[Hentry Hentries]"; first apply Hp.
    rewrite (Z2Nat.id (i1 `mod` _)); last apply Z_mod_lt, size_pos'.
    iApply (na_read_frac with "[kI $Hentry]"); auto.
    iNext; iIntros (?) "[% Hentry]"; subst.
    iPoseProof ("Hentries" with "Hentry") as "Hentries".
    wp_lam.
    wp_op. wp_lam.

    wp_op.
    wp_bind ([_]_at)%E.
    iMod ("Hshift" with "HP") as (HT) "(HH & Hrest)".
    destruct (lookup_lt_is_Some_2 T (Z.to_nat (i1 `mod` size))) as [(k0, lv0) Hpi];
      first by rewrite HlenT.
    iPoseProof (big_sepL_lookup_acc with "HA") as "[#HentryA HA']"; first apply Hpi.
    iPoseProof "HentryA" as "#[Hdata [Hk Hv]]".
    rewrite (nth_lookup_Some _ _ _ _ Hp).
    set (gki := nth _ lgk _).
    set (gvi := nth _ lgv _).
    assert (length (map fst T) = size) by (rewrite map_length; auto).
    iApply (GPS_PP_Read (p:=tt)(s := k0) (kP _) emp%I (fun s k' => (⌜s = k'⌝ ∗
      |={base_mask,⊤}=> if decide (k' = 0) then Φ (LitV 0)
        else ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗ ⎡own gki (Snapshot k')⎤ ∗ ▷P))%I
      with "kI [-]"); [auto | auto | ..].
    { iSplitR "".
      - iNext; iIntros (s' Hs').
        iLeft; iIntros (k') "[[% ?] ?]"; subst.
        iModIntro.
        iSplit; first done.
        destruct (decide _).
        + iPoseProof (and_elim_r with "Hrest") as "HQ".
          iApply ("HQ" $! (LitV 0) with "[Hentries HA HH]").
          iExists lv0, 0; iSplit; first auto.
          iDestruct "Hdata" as %(v & ? & ? & ?).
          assert (k0 = 0) by (destruct Hs'; subst; auto).
          assert (v = 0) by auto; subst.
          iFrame; iSplit; last by iPureIntro.
          iExists T; iSplit; last by iApply "HA".
          iPureIntro.
          split; first done.
          eexists; erewrite (nth_lookup_Some _ _ _ _ Hpi); split; auto.
          eapply (hlookup_succeeds _ i); try done.
          erewrite nth_lookup_Some; eauto.
          rewrite map_lookup Hpi; auto.
        + iPoseProof (and_elim_l with "Hrest") as "HP".
          iFrame.
          iApply ("HP" with "HH").
      - iSplit; [|iSplit; auto].
        iNext; iApply objective_objectively.
        iIntros (? ?) "[? #?]"; iModIntro; iFrame "#". }
    iNext. iIntros (s' k') "(% & #Hk' & % & Hpost)"; subst.
    iMod "Hpost"; iModIntro.
    iDestruct "HentryA" as "_"; iDestruct "HA'" as "_".

    wp_lam.

    assert (0 ≤ hash k < length (map fst T))%nat.
    { rewrite map_length HlenT; apply hash_range. }
    assert (length (rebase (map fst T) (hash k)) = size).
    { rewrite length_rebase; last done.
      by rewrite map_length. }
    wp_op.
    destruct (decide (k' = k)).
    - rewrite bool_decide_true; last done.
      wp_if_true.
      wp_op.
      destruct (decide _); subst; first contradiction.
      iDestruct "Hpost" as "(Hentries & Hsnap & HP)".
      iMod ("Hshift" with "HP") as (HT1) "(HH & Hrest)".
      iApply (GPS_PP_Read' (p:=tt)(s := lv0) (vP gvi) bi_emp
        (fun s v => |={base_mask,⊤}=> Φ (LitV v))%I with "[-]"); [auto | auto | ..].
      { iSplit; first auto.
        iSplitR "".
        + iNext; iIntros (s' Hs').
          iLeft; iIntros (v') "[[[% %] Hs'] [Hv' _]]".
          iPoseProof (and_elim_r with "Hrest") as "HQ".
          iApply ("HQ" $! (LitV v') with "[-]").
          iExists s', v'; iSplit; first auto.
          iSplitL "Hentries"; first done.
          subst; iSplitL "Hv' HA".
          * iExists (<[Z.to_nat (i1 mod size) := (k, s')]>T); iSplit.
            { iPureIntro.
              split.
              { etransitivity; eauto.
                eapply table_incl_upd; eauto. }
              exists (Z.to_nat (i1 mod size)); rewrite nth_insert; last omega.
              rewrite decide_left; split; last done.
              eapply (hlookup_succeeds _ i); eauto.
              { rewrite map_length insert_length; auto. }
              { rewrite map_insert Hi1.
                replace size with (length (map fst T)).
                rewrite rebase_insert; try omega.
                rewrite sublist_insert_l; auto; omega. }
              { rewrite map_insert nth_insert.
                rewrite decide_left; auto.
                rewrite map_length HlenT; auto. } }
            iApply big_sepL_insert; first done.
            iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
            iPureIntro; do 2 eexists; eauto; done.
          * iSplit; last done.
            iDestruct "HH" as (HTa) "[% [? HH]]".
            destruct H6 as (? & ? & HHTa).
            destruct (lookup_lt_is_Some_2 HTa (Z.to_nat (i1 mod size))) as [(k1, lv1) Hpi'].
            { replace (length HTa) with size; omega. }
            iPoseProof (big_sepL_lookup_acc with "HH") as "[Hentry HH]"; first apply Hpi'.
            iDestruct "Hentry" as "[% [Hk1 Hlv1]]".
            iPoseProof (own_snap_ord _ _ _ k1 with "Hk1 Hsnap") as "[% | %]"; [done | subst k1].
            iPoseProof (own_snap_ord _ _ _ lv1 with "Hlv1 Hs'") as "%".
            iPureIntro.
            split; first done; intro Hv'; do 2 eexists; eauto.
            rewrite HHTa.
            split; first by eapply elem_of_list_lookup_2.
            destruct H8 as (? & ? & ? & Hzero); do 2 eexists; eauto; intro; contradiction Hv'.
            destruct H4 as (? & ? & ?).
            eapply Hzero; auto.
            eapply lookup_weaken; eauto.
        + iSplit; last auto.
          iNext; iApply objective_objectively.
          iIntros (? ?) "[? #?]"; iModIntro; iFrame "#". }
      iNext. iIntros (s' v) "[% ?]"; auto.
    - rewrite bool_decide_false; last done.
      wp_if_false.
      wp_op.
      destruct (decide _); subst.
      + rewrite bool_decide_true; last done.
        by wp_if_true.
      + rewrite bool_decide_false; last done.
        wp_if_false.
        wp_op.
        iDestruct "Hpost" as "(Hentries & Hsnap & HP)".
        iApply ("IH" with "Hentries HP [Hsnap]").
        iExists (i + 1)%nat, (<[Z.to_nat (i1 mod size) := (k', lv0)]>T); iSplit(*; last iSplit*).
        * iPureIntro; split.
          -- assert ((0 ≤ i1 `mod` Z.of_nat size)%Z) by (apply Z_mod_lt, size_pos').
             rewrite <- (Z2Nat.id (_ + 1)); last omega.
             rewrite <- mod_Zmod; last omega.
             rewrite Z2Nat.inj_add; try omega.
             rewrite Hi1 Nat2Z.id Nat.add_mod_idemp_l; last omega.
             rewrite <- plus_assoc, (plus_comm (hash k)), plus_assoc; done.
          -- split.
             { admit. (* we don't run out of space *) }
             split; first by rewrite insert_length.
             split; first by (etransitivity; eauto; eapply table_incl_upd; eauto).
             rewrite map_insert Hi1.
             replace size with (length (map fst T)).
             rewrite rebase_insert; try omega.
             rewrite sublist_insert_m; try omega.
             rewrite Nat.sub_0_r sublist_split; last omega.
             rewrite (sublist_len_1 _ _ 0); last omega.
             rewrite insert_app_r_alt; rewrite length_sublist; try omega.
             rewrite Nat.sub_0_r Nat.sub_diag; simpl.
             rewrite Forall_app; split; first done; repeat constructor; auto.
        * iApply big_sepL_insert; first done.
          iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
          iDestruct "Hdata" as %(? & ? & ? & ?).
          iPureIntro; do 2 eexists; eauto; done.
  Admitted.

  Definition map_upd {A} (H : Z -> option A) k v k' := if decide (k' = k) then Some v else H k'.

  (* up *)
  Lemma hlookup_range : forall m k i, hlookup m k = Some i -> (i < size)%nat.
  Proof.
    unfold hlookup; intros.
    pose proof (index_of'_spec k (rebase m (hash k))).
    destruct (index_of' _ _); inversion H.
    apply Nat.mod_upper_bound.
    pose proof size_pos; omega.
  Qed.

  Lemma wf_table_upd : forall T k v i (Hwf : wf_table T) (HT : length T = size)
    (Hi : hlookup (map fst T) k = Some i) (Hk : k ≠ 0), wf_table (<[i:=(k, v)]>T).
  Proof.
    intros; intros ?? Hj ?.
    rewrite map_insert.
    pose proof (hlookup_range _ _ _ Hi).
    rewrite nth_insert in H; last omega.
    destruct (decide _); simpl in *; subst.
    - by rewrite lookup_insert_same; rewrite ?map_length; try omega.
    - specialize (Hwf _ _ Hj eq_refl).
      rewrite lookup_insert_diff; simpl; rewrite ?map_length; try congruence.
      split; congruence.
  Qed.

  Definition past_snaps lgk k (T : list (Z * gmap nat Z)) i : vProp Σ :=
    ([∗ list] i ∈ seq 0 i, ⎡own (nth ((i + hash k) mod size) lgk undef)
        (Snapshot (nth ((i + hash k) mod size) (map fst T) 0))⎤)%I.

  Lemma mod_smallish : forall (x y : nat), (y ≠ 0 -> x < 2 * y -> x mod y = x \/ (x mod y) + y = x)%nat.
  Proof.
    intros.
    destruct (decide (x < y)%nat); first by left; apply Nat.mod_small; omega.
    right.
    replace x with ((x - y) + 1 * y)%nat at 1; last omega.
    rewrite Nat.mod_add; last done.
    rewrite Nat.mod_small; omega.
  Qed.

  Lemma mod_plus_inv : forall a b c d, (c ≠ 0 -> (a + b) mod c = (d + b) mod c -> a mod c = d mod c)%nat.
  Proof.
    intros ???? Hc Heq.
    rewrite Nat.add_mod in Heq; last done.
    rewrite (Nat.add_mod d) in Heq; last done.
    pose proof (Nat.mod_upper_bound a c Hc).
    pose proof (Nat.mod_upper_bound b c Hc).
    pose proof (Nat.mod_upper_bound d c Hc).
    destruct (mod_smallish (a mod c + b mod c) c), (mod_smallish (d mod c + b mod c) c); omega.
  Qed.

  Lemma entries_lookup : forall k i i1 T0 lgk lgv T (Hk : k ≠ 0) (Hi : (i < size)%nat)
    (Hi1 : ((i + hash k) mod size = Z.to_nat (i1 mod size))%nat)
    (HT : length T = size) (Hkeys: length T0 = size)
    (Hfail : Forall (fun z => z ≠ 0 /\ z ≠ k) (sublist 0 i (rebase (map fst T0) (hash k))))
    (Hhit : fst (Tnth (Z.to_nat (i1 mod size)) T) = k \/ fst (Tnth (Z.to_nat (i1 mod size)) T) = 0),
    ([∗ list] k0↦x ∈ T, if decide (k0 = Z.to_nat (i1 `mod` size)) then emp
      else hashtable_entry x lgk lgv k0) -∗ past_snaps lgk k T0 i -∗ 
    ⌜hlookup (map fst T) k = Some (Z.to_nat (i1 mod size))⌝.
  Proof.
    intros; iIntros "HT Hsnaps".
    iAssert (⌜Forall (fun x => x ≠ 0 /\ x ≠ k) (sublist 0 i (rebase (map fst T) (hash k)))⌝)%I
      as %Hfail'.
    - intros.
      rewrite Forall_forall.
      iIntros (k') "%".
      destruct (elem_of_list_lookup_1 _ _ a) as (j & Hjth).
      pose proof (lookup_lt_Some _ _ _ Hjth) as Hj.
      apply nth_lookup_Some with (d := 0) in Hjth.
      pose proof (hash_range k).
      rewrite length_sublist in Hj; try (rewrite length_rebase map_length; omega).
      rewrite nth_sublist in Hjth; last omega.
      rewrite nth_rebase in Hjth; try (rewrite map_length; omega).
      unfold past_snaps.
      lapply (Nat.mod_upper_bound (j + hash k) size); last omega; intro.
      destruct (lookup_lt_is_Some_2 T ((j + hash k) mod size)) as [(k1, v1) Hp]; first omega.
      iPoseProof (big_sepL_lookup_acc with "HT") as "[HTi _]"; first apply Hp.
      iPoseProof (big_sepL_lookup_acc with "Hsnaps") as "[Hsnap _]"; first by apply (lookup_seq _ _ j); omega.
      destruct (decide _).
      { rewrite <- Hi1 in e.
        apply mod_plus_inv in e; last omega.
        rewrite !Nat.mod_small in e; omega. }
      unfold hashtable_entry.
      iDestruct "HTi" as "(_ & Hk & _)".
      iPoseProof (own_snap_ord _ _ _ k1 with "Hk Hsnap") as "%".
      iPureIntro.
      apply Forall_nth with (i0 := j)(d := 0) in Hfail.
      rewrite nth_sublist in Hfail; last done.
      rewrite Nat.add_0_r (nth_rebase _ j) in Hfail; try (rewrite map_length; omega).
      rewrite map_length in Hfail; replace (length T0) with size in Hfail.
      match goal with H : zero_ord _ _ |- _ => destruct H; subst; first by destruct Hfail end.
      rewrite Nat.add_0_r nth_lookup map_lookup map_length.
      replace (length T) with size; rewrite Hp; simpl; done.
      { rewrite length_sublist; auto; try omega.
        rewrite length_rebase map_length; omega. }
    - iPureIntro; eapply hlookup_succeeds; eauto.
      { by rewrite map_length. }
      { by rewrite <- map_nth in Hhit. }
  Qed.

  Lemma insert_in_old : forall {A} (l : list A) i x y, x ∈ l -> l !! i ≠ Some x -> x ∈ <[i:=y]>l.
  Proof.
    induction l; auto; intros.
    destruct i; inversion H; subst; constructor; eauto; done.
  Qed.

  Lemma in_insert : forall {A} (l : list A) i x y, x ∈ <[i:=y]>l -> x = y \/ x ∈ l.
  Proof.
    induction l; auto; intros.
    destruct i; subst; simpl in *.
    - inversion H; subst; auto; by right; constructor.
    - inversion H; subst; first by right; constructor.
      destruct (IHl _ _ _ H2); auto; by right; constructor.
  Qed.

  Lemma past_snaps_add : forall lgk k T i i1 k' lv'
    (Hi1 : Z.to_nat (i1 `mod` size) = ((i + hash k) `mod` size)%nat),
    (i < size)%nat -> (Z.to_nat (i1 `mod` size) < length T)%nat ->
    past_snaps lgk k T i -∗
    ⎡own (nth (Z.to_nat (i1 `mod` size)) lgk undef) (Snapshot k')⎤ -∗
    past_snaps lgk k (<[Z.to_nat (i1 mod size):=(k', lv')]> T) (i + 1).
  Proof.
    intros; iIntros "? ?".
    unfold past_snaps.
    rewrite Nat.add_1_r seq_S big_sepL_app map_insert; iSplit.
    - iApply big_sepL_proper; last done.
      intros; simpl.
      rewrite nth_insert; last by rewrite map_length.
      destruct (decide _); auto.
      rewrite Hi1 in e; apply mod_plus_inv in e; last done.
      match goal with H : seq _ _ !! _ = Some _ |- _ => apply lookup_seq_inv in H as []; subst end.
      rewrite !Nat.mod_small in e; omega.
    - simpl.
      rewrite nth_insert; last by rewrite map_length.
      rewrite Hi1 decide_left -Hi1; auto.
  Qed.

  Lemma set_key_spec (k : Z) q entries g lgk lgv T0 (Hk : k ≠ 0)
        (Hlen : length entries = size) (HlenT0 : length T0 = size):
    ⎡PSCtx⎤ ∗ ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
    hashtable_A T0 lgk lgv entries -∗
    atomic_wp (fun H => hashtable H g lgk lgv)%I
      (fun H v => ∃i : nat, ∃p : loc, ⌜v = LitV p /\ entries !! i = Some p⌝ ∗
      (* private *) ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
        (∃T, ⌜length T = size /\ table_incl T0 T /\ hlookup (map fst T) k = Some i /\
              fst (Tnth i T) = k⌝ ∗ hashtable_A T lgk lgv entries) ∗
            ⎡own (nth i lgk undef) (Snapshot k)⎤ ∗ (* public *) hashtable H g lgk lgv)%I
      base_mask ⊤ (set_key #k).
  Proof.
    unfold atomic_wp, atomic_shift.
    iIntros "(#kI & Hentries & #HA)" (Φ) "H".
    iDestruct "H" as (P) "(HP & #Hshift)".
    match goal with |-context[(▷ P ={_,_}=∗ ?R)%I] => set (Q := R) end.

    unfold set_key.
    wp_let.
    wp_bind (integer_hash #k)%E.
    iApply integer_hash_spec; first done.
    iNext; iIntros (??); subst.

    (* loop invariant *)
    iAssert (∃i : nat, ∃T, ⌜(Z.to_nat (k * const mod size) = (i + hash k) mod size)%nat /\
      (i < size)%nat /\ length T = size /\ table_incl T0 T /\
      Forall (fun z => z ≠ 0 /\ z ≠ k) (sublist 0 i (rebase (map fst T) (hash k)))⌝ ∗
      hashtable_A T lgk lgv entries ∗ past_snaps lgk k T i)%I as "Hpre".
    { unfold past_snaps; iExists O, T0; iSplit; last by iSplit; simpl.
      iPureIntro; rewrite sublist_nil; repeat split; try done.
      - simpl.
        change size with (Z.to_nat 16384); rewrite Z2Nat.id; last by compute.
        rewrite Nat.mod_small; first done.
        apply Nat2Z.inj_lt; rewrite !Z2Nat.id; try omega; apply Z_mod_lt; by compute.
      - apply size_pos. }
    set (i1 := (k * const)%Z); clearbody i1.
    iLöb as "IH" forall (i1).
    iDestruct "HA" as "_".

    wp_rec.
    wp_op. wp_lam.
    wp_op.
    wp_bind ([_]_na)%E.
    iDestruct "Hpre" as (i T) "(% & #HA & #Hsnaps)".
    destruct H as (Hi1 & Hi & HlenT & Hincl & Hseen).
    assert (Z.to_nat (i1 `mod` size) < size)%nat.
    { apply Nat2Z.inj_lt.
      rewrite Z2Nat.id; apply Z_mod_lt, size_pos'. }
    destruct (lookup_lt_is_Some_2 entries (Z.to_nat (i1 `mod` size))) as [p Hp];
      first by rewrite Hlen.
    iPoseProof (big_sepL_lookup_acc with "Hentries") as "[Hentry Hentries]"; first apply Hp.
    rewrite (Z2Nat.id (i1 `mod` _)); last apply Z_mod_lt, size_pos'.
    iApply (na_read_frac with "[$kI $Hentry]"); [done | done |].
    iNext; iIntros (?) "[% Hentry]"; subst.
    iPoseProof ("Hentries" with "Hentry") as "Hentries".
    wp_lam.
    wp_op. wp_lam.

    wp_op.
    wp_bind ([_]_at)%E.
    iMod ("Hshift" with "HP") as (HT) "(HH & Hrest)".
    destruct (lookup_lt_is_Some_2 T (Z.to_nat (i1 `mod` size))) as [(k0, lv0) Hpi];
      first by rewrite HlenT.
    iPoseProof (big_sepL_lookup_acc with "HA") as "[#HentryA HA']"; first apply Hpi.
    iPoseProof "HentryA" as "#[Hdata [Hk Hv]]".
    rewrite (nth_lookup_Some _ _ _ _ Hp).
    set (gki := nth _ lgk _).
    set (gvi := nth _ lgv _).
    assert (length (map fst T) = size) by (rewrite map_length; auto).
    assert (0 ≤ hash k < length (map fst T))%nat.
    { rewrite map_length HlenT; apply hash_range. }
    assert (length (rebase (map fst T) (hash k)) = size).
    { rewrite length_rebase; last done.
      by rewrite map_length. }
    iApply (GPS_PP_Read' (p:=tt)(s := k0) (kP gki) bi_emp (fun s k' => (⌜s = k'⌝ ∗
      |={base_mask,⊤}=> if decide (k' = k) then Φ (LitV p)
        else ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗ ⎡own gki (Snapshot k')⎤ ∗
        [PP ZplusPos 0 p in k' | kP gki] ∗ ▷P))%I
      with "[$kI Hentries HA HH Hrest]"); [done | done | ..].
    { iSplitR "".
      - iNext; iIntros (s' Hs').
        iLeft; iIntros (k') "((% & ?) & #Hk' & _)"; subst.
        iModIntro.
        iSplit; first done.
        destruct (decide _).
        + iApply ("Hrest" with "[-]").
          subst; iExists (Z.to_nat (i1 mod size)), p; iSplit; first done; fold gki; iFrame.
          iExists (<[Z.to_nat (i1 mod size) := (k, lv0)]>T); iSplit.
          { iPureIntro.
            split; first by rewrite insert_length.
            split.
            { etransitivity; eauto.
              eapply table_incl_upd; eauto. }
            rewrite nth_insert; last omega.
            rewrite decide_left; split; last done.
            eapply (hlookup_succeeds _ i); eauto.
            { rewrite map_length insert_length; auto. }
            { rewrite map_insert Hi1.
              replace size with (length (map fst T)).
              rewrite rebase_insert; try omega.
              rewrite sublist_insert_l; auto; omega. }
            { rewrite map_insert nth_insert.
              rewrite decide_left; auto.
              rewrite map_length HlenT; auto. } }
          iApply big_sepL_insert; first done.
          iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
          iDestruct "Hdata" as %(? & ? & ? & ?).
          iPureIntro; do 2 eexists; eauto; done.
        + iPoseProof (and_elim_l with "Hrest") as "HP".
          iFrame.
          by iMod ("HP" with "HH"); iFrame.
      - iSplit; [|iSplit; auto].
        iNext; iApply objective_objectively.
        iIntros (? ?) "[? #?]"; iModIntro; iFrame "#". }
    iNext. iIntros (s' k') "(% & % & Hpost)"; subst.
    iMod "Hpost"; iModIntro.
    iDestruct "HentryA" as "_"; iDestruct "HA'" as "_".

    wp_lam.
    wp_op.
    intros. destruct (decide _).
    { rewrite bool_decide_true; last done.
      by wp_op; wp_if_false. }
    rewrite bool_decide_false; last done.
    wp_op; wp_if_true.
    iDestruct "Hpost" as "(Hentries & Hsnap & #Hk' & HP)".
    wp_op.
    destruct (decide (k' = 0)).
    - rewrite bool_decide_true; last done.
      wp_op; wp_if_false.
      subst; assert (k0 = 0) by (destruct H3; done); subst.
      wp_op.
      wp_bind (CAS _ _ _).
      iDestruct "Hk'" as "_".
      iMod ("Hshift" with "HP") as (HT1) "(HH & Hrest)".
      match goal with |-context[bi_and ?Pa ?Pb] =>
        iApply (GPS_PP_CAS' (p:=tt)(s := 0) (kP gki) (⎡own gki (Snapshot 0)⎤ ∗
          ([∗ list] i0↦x ∈ entries, ZplusPos i0 m_entries ↦{q} Z.pos x) ∗
          hashtable HT1 g lgk lgv ∗ bi_and Pa Pb)%I (fun s => |={base_mask,⊤}=> Φ (LitV p))%I
         (fun s k' => ⌜s = k' /\ k' ≠ 0⌝ ∗ ([∗ list] i0↦x ∈ entries, (ZplusPos i0 m_entries ↦{q} Z.pos x)) ∗
          ⎡own gki (Snapshot k')⎤ ∗ [PP ZplusPos 0 p in k' @ () | kP gki ] ∗ |={base_mask,⊤}=> ▷P)%I with "[-]");
          [done | done | ..] end.
      { iSplit; first auto.
        iSplitL "Hsnaps"; [|iSplitL ""; [|iSplitL ""]].
        + iNext; iIntros (s') "(% & (% & Hhalf1) & Hsnap & Hentries & HH & Hrest)"; subst.
          iDestruct "HH" as (HTa) "(% & ? & HH)".
          destruct H5 as (? & ? & HHTa).
          destruct (lookup_lt_is_Some_2 HTa (Z.to_nat (i1 mod size))) as [(k1, lv1) Hpi'].
          { replace (length HTa) with size; omega. }
          iPoseProof (big_sepL_remove _ _ _ _ Hpi' with "HH") as "[Hentry HH]".
          iDestruct "Hentry" as "[% [Hhalf2 Hlv1]]".
          iPoseProof (own_master_eq _ _ _ _ 0 with "Hhalf1 Hhalf2") as "%"; subst.
          iPoseProof (own_update_3 with "Hhalf1 Hhalf2 Hsnap") as "Hupd".
          { rewrite master_op half_join.
            apply @snap_update with (a' := k).
            unfold zero_ord; auto. }
          rewrite <- half_join, <- master_op, half_join.
          iMod "Hupd" as "[[Hhalf1 Hhalf2] #Hsnap]".
          iModIntro.
          iExists k; iSplit.
          { iPureIntro; hnf; auto. }
          unfold kP; iFrame.
          repeat (iSplit; first auto).
          iIntros "#Hk'".
          iModIntro; iApply "Hrest".
          iExists (Z.to_nat (i1 mod size)), p; iSplit; first done; iFrame.
          iSplit.
          * iExists (<[Z.to_nat (i1 mod size) := (k, lv0)]>T); iSplit.
            { iPureIntro.
              split; first by rewrite insert_length.
              split.
              { etransitivity; eauto.
                eapply table_incl_upd; unfold zero_ord; eauto. }
              rewrite nth_insert; last omega.
              rewrite decide_left; split; last done.
              eapply (hlookup_succeeds _ i); eauto.
              { rewrite map_length insert_length; auto. }
              { rewrite map_insert Hi1.
                replace size with (length (map fst T)).
                rewrite rebase_insert; try omega.
                rewrite sublist_insert_l; auto; omega. }
              { rewrite map_insert nth_insert.
                rewrite decide_left; auto.
                rewrite map_length HlenT; auto. } }
            iApply big_sepL_insert; first done.
            iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
            iDestruct "Hdata" as %(? & ? & ? & ?).
            iPureIntro; do 2 eexists; eauto; done.
          * iSplit; first done.
            destruct H7 as (v & ? & ? & ?); assert (v = 0) by auto; subst.
            iExists (<[Z.to_nat (i1 mod size) := (k, lv1)]>HTa); iSplit.
            { iPoseProof (entries_lookup with "HH Hsnaps") as "%"; auto.
              { rewrite (nth_lookup_Some _ _ _ _ Hpi'); auto. }
              iPureIntro.
              split; first by rewrite insert_length.
              split; first by apply wf_table_upd.
              intros; rewrite HHTa.
              split; intros (Hin & ?); split; auto.
              * apply insert_in_old; first done.
                rewrite Hpi'; intro X; inversion X; subst.
                match goal with H : exists v, _ /\ _ |- _ => destruct H as (? & (? & ? & ?) & Hnz) end.
                contradiction Hnz; eauto.
              * apply in_insert in Hin as [X|]; auto.
                inversion X; subst.
                match goal with H : exists v, _ /\ _ |- _ => destruct H as (? & (? & ? & ?) & Hnz) end.
                contradiction Hnz; eauto. }
            iApply (big_sepL_replace with "HH").
            iSplit; last iFrame.
            iPureIntro; exists 0; done.
        + iNext; unfold kP.
          iIntros (s' v0) "(% & % & Hsnap' & ? & Hsnap & ? & HH & HP)".
          iMod "Hsnap'" as "[% ?]"; subst.
          iModIntro; iSplit; [done | iFrame].
          iApply "HP"; iFrame; iSplit; done.
        + iNext; iApply objective_objectively.
          iIntros (? ?) "[? #?]"; iModIntro; iFrame "#".
        + iFrame; done. }
      iNext. iIntros (? b k') "(% & Hpost)".

      destruct b.
      + iMod "Hpost"; iModIntro.
        wp_let.
        wp_op.
        wp_if_false.
        done.
      + iDestruct "Hpost" as "([% %] & Hentries & Hsnap & #Hk' & Hpost)"; subst.
        iMod "Hpost"; iModIntro.
        wp_let.
        wp_op.
        wp_if_true.
        wp_op.
        wp_bind ([_]_at)%E.
        iMod ("Hshift" with "Hpost") as (HT2) "(HH & Hrest)".
        iApply (GPS_PP_Read' (p:=tt)(s := k') (kP gki) emp%I (fun s k'' => (⌜s = k''⌝ ∗
           ⎡own gki (Snapshot k'')⎤ ∗ |={base_mask,⊤}=> if decide (k'' = k) then Φ (LitV p)
           else ([∗ list] i0↦x ∈ entries, (ZplusPos i0 m_entries ↦{q} Z.pos x)) ∗ ▷P))%I
          with "[-]"); [done | done | ..].
        { iSplit; first auto.
          iSplitR "".
          - iNext; iIntros (s' [|]); [contradiction | subst].
            iLeft; iIntros (k1) "[[% ?] [? ?]]"; subst.
            iModIntro.
            iSplit; first done.
            iFrame.
            destruct (decide _).
            + subst; iApply ("Hrest" with "[-]").
              iExists (Z.to_nat (i1 mod size)), p; iSplit; first done; fold gki; iFrame.
              iExists (<[Z.to_nat (i1 mod size) := (k, lv0)]>T); iSplit.
              { iPureIntro.
                split; first by rewrite insert_length.
                split.
                { etransitivity; eauto.
                  eapply table_incl_upd; eauto. }
                rewrite nth_insert; last omega.
                rewrite decide_left; split; last done.
                eapply (hlookup_succeeds _ i); eauto.
                { rewrite map_length insert_length; auto. }
                { rewrite map_insert Hi1.
                  replace size with (length (map fst T)).
                  rewrite rebase_insert; try omega.
                  rewrite sublist_insert_l; auto; omega. }
                { rewrite map_insert nth_insert.
                  rewrite decide_left; auto.
                  rewrite map_length HlenT; auto. } }
              iApply big_sepL_insert; first done.
              iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
              iDestruct "Hdata" as %(? & ? & ? & ?).
              iPureIntro; do 2 eexists; eauto; done.
            + iPoseProof (and_elim_l with "Hrest") as "HP".
              by iMod ("HP" with "HH"); iFrame.
          - iSplit; [|iSplit; auto].
            iNext; iApply objective_objectively.
            iIntros (? ?) "[? #?]"; iModIntro; iFrame "#". }
        iNext. iIntros (? k'') "(% & % & Hsnap & Hpost)"; subst.
        destruct H5; try contradiction; subst k''.
        iMod "Hpost" as "Hpost"; iModIntro.
        wp_lam.
        wp_op. destruct (decide (k' = k)).
        * rewrite bool_decide_true; last done.
          by wp_op; wp_if_false.
        * rewrite bool_decide_false; last done.
          wp_op; wp_if_true.
          wp_op.
          iDestruct "Hpost" as "[Hentries HP]".
          iApply ("IH" with "Hentries HP [Hsnap]").
          iExists (i + 1)%nat, (<[Z.to_nat (i1 mod size) := (k', lv0)]>T); iSplit; last iSplit.
          -- iPureIntro; split.
             ++ assert ((0 ≤ i1 `mod` Z.of_nat size)%Z) by (apply Z_mod_lt, size_pos').
                rewrite <- (Z2Nat.id (_ + 1)); last omega.
                rewrite <- mod_Zmod; last omega.
                rewrite Z2Nat.inj_add; try omega.
                rewrite Hi1 Nat2Z.id Nat.add_mod_idemp_l; last omega.
                rewrite <- plus_assoc, (plus_comm (hash k)), plus_assoc; done.
             ++ split.
                { admit. (* we don't run out of space *) }
                split; first by rewrite insert_length.
                split; first by (etransitivity; eauto; eapply table_incl_upd; eauto).
                rewrite map_insert Hi1.
                replace size with (length (map fst T)).
                rewrite rebase_insert; try omega.
                rewrite sublist_insert_m; try omega.
                rewrite Nat.sub_0_r sublist_split; last omega.
                rewrite (sublist_len_1 _ _ 0); last omega.
                rewrite insert_app_r_alt; rewrite length_sublist; try omega.
                rewrite Nat.sub_0_r Nat.sub_diag; simpl.
                rewrite Forall_app; split; first done; repeat constructor; auto.
        -- iApply big_sepL_insert; first done.
           iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
           iDestruct "Hdata" as %(? & ? & ? & ?).
           iPureIntro; do 2 eexists; eauto; done.
        -- iApply (past_snaps_add with "Hsnaps Hsnap"); auto; omega.
    - rewrite bool_decide_false; last done.
      wp_op; wp_if_true.
      wp_op.
      iApply ("IH" with "Hentries HP [Hsnap]").
      iExists (i + 1)%nat, (<[Z.to_nat (i1 mod size) := (k', lv0)]>T); iSplit; last iSplit.
      + iPureIntro; split.
         * assert ((0 ≤ i1 `mod` Z.of_nat size)%Z) by (apply Z_mod_lt, size_pos').
           rewrite <- (Z2Nat.id (_ + 1)); last omega.
           rewrite <- mod_Zmod; last omega.
           rewrite Z2Nat.inj_add; try omega.
           rewrite Hi1 Nat2Z.id Nat.add_mod_idemp_l; last omega.
           rewrite <- plus_assoc, (plus_comm (hash k)), plus_assoc; done.
         * split.
           { admit. (* we don't run out of space *) }
           split; first by rewrite insert_length.
           split; first by (etransitivity; eauto; eapply table_incl_upd; eauto).
           rewrite map_insert Hi1.
           replace size with (length (map fst T)).
           rewrite rebase_insert; try omega.
           rewrite sublist_insert_m; try omega.
           rewrite Nat.sub_0_r sublist_split; last omega.
           rewrite (sublist_len_1 _ _ 0); last omega.
           rewrite insert_app_r_alt; rewrite length_sublist; try omega.
           rewrite Nat.sub_0_r Nat.sub_diag; simpl.
           rewrite Forall_app; split; first done; repeat constructor; auto.
      + iApply big_sepL_insert; first done.
        iSplit; last by rewrite (nth_lookup_Some _ _ _ _ Hp); auto.
        iDestruct "Hdata" as %(? & ? & ? & ?).
        iPureIntro; do 2 eexists; eauto; done.
      + iApply (past_snaps_add with "Hsnaps Hsnap"); auto; omega.
  Admitted.

  Opaque set_key.

  Lemma wf_table_inj : forall T k v1 v2 (Hwf : wf_table T) (Hk : k ≠ 0)
    (Hin1 : (k, v1) ∈ T) (Hin2 : (k, v2) ∈ T), v1 = v2.
  Proof.
    intros.
    apply elem_of_list_lookup in Hin1 as (i1 & Hi1);
      apply elem_of_list_lookup in Hin2 as (i2 & Hi2).
    generalize (Hwf _ i1 Hk) (Hwf _ i2 Hk).
    rewrite !nth_lookup Hi1 Hi2; simpl; intros H1 H2.
    specialize (H1 eq_refl); specialize (H2 eq_refl); congruence.
  Qed.

  Lemma excl_update : forall {A : ofeT} (a a' : A), Excl a ~~> Excl a'.
  Proof.
    repeat intro.
    destruct mz; done.
  Qed.

  Lemma add_item_spec (k v : Z) q entries g lgk lgv T0 (Hk : k ≠ 0) (Hv : v ≠ 0)
        (Hlen : length entries = size) (HlenT0 : length T0 = size):
    ⎡PSCtx⎤ ∗ ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
     hashtable_A T0 lgk lgv entries -∗
    atomic_wp (fun H => hashtable H g lgk lgv)%I
      (fun H r => ∃(b : bool) lv,
      (* private *) ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
        (∃T, ⌜table_incl T0 T /\ exists i, hlookup (map fst T) k = Some i /\ Tnth i T = (k, lv)⌝ ∗
          hashtable_A T lgk lgv entries) ∗
        (* public *) 
        ⌜r = LitV (lit_of_bool b) /\ b = (match H k with Some _ => false | None => true end) /\
         (b = true -> value_of lv v)⌝ ∗ hashtable (if b then map_upd H k lv else H) g lgk lgv)%I
      base_mask ⊤ (add_item #k #v).
  Proof.
    unfold atomic_wp, atomic_shift.
    iIntros "(#kI & Hentries & HA)" (Φ) "H".
    iDestruct "H" as (P) "(HP & #Hshift)".
    match goal with |-context[(▷ P ={_,_}=∗ ?R)%I] => set (Q := R) end.

    unfold add_item.
    wp_lam.
    wp_lam.
    wp_bind (set_key _).
    iApply (set_key_spec with "[$Hentries $HA]"); [done | done | done | auto |].

    (* Instead of directly applying set_key_spec, we probably want a rule for using LA triples. *)
    unfold atomic_shift.
    iExists P; iSplit; first done.
    iIntros "!# HP".
    iMod ("Hshift" with "HP") as (x) "(? & Hrest)"; iModIntro.
    iExists x.
    iSplitR "Hrest"; first by iFrame; auto.
    iSplit; first by iApply and_elim_l.
    iIntros (?) "Hfound".
    iDestruct "Hfound" as (i p) "([% %] & Hentries & Hfound & Hsnap & HH)"; subst.
    iDestruct "Hfound" as (T) "[(% & % & % & %) #HA]".
    iPoseProof (and_elim_l with "Hrest") as "HP".
    iMod ("HP" with "HH") as "HP".
    iModIntro.

    wp_let.
    wp_op.
    iMod ("Hshift" with "HP") as (HT1) "(HH & Hrest)".
    match goal with H : _ !! _ = _ |- _ => pose proof (lookup_lt_Some _ _ _ H) end.
    destruct (nth_lookup_or_length T i (0, ∅)); last omega.
    iPoseProof (big_sepL_lookup_acc with "HA") as "[#HentryA HA']"; first done.
    destruct (Tnth i T) as (?, lv) eqn: HTi; simpl in *; subst.
    iDestruct "HentryA" as "#[Hdata [Hk Hv]]".
    rewrite (nth_lookup_Some _ _ _ _ H0).
    iApply wp_fupd.
    match goal with |-context[bi_and ?Pa ?Pb] =>
      iApply (GPS_PP_CAS' (p:=tt)(s := lv) (vP (nth i lgv undef))
        (⎡own (nth i lgk 1%positive) (Snapshot k)⎤ ∗
         ([∗ list] i0↦x ∈ entries, ZplusPos i0 m_entries ↦{q} Z.pos x) ∗
         hashtable HT1 g lgk lgv ∗ bi_and Pa Pb)%I
        (fun s => |={base_mask,⊤}=> Φ (LitV 1))%I (fun s v' => |={base_mask,⊤}=> Φ (LitV 0))%I
        with "[-]"); [done | done | ..] end.
    { iSplit; first auto.
      iSplitL ""; [|iSplitL ""; [|iSplitL ""]].
      - iNext; iIntros (s') "(% & ([% %] & Hhalf1) & Hsnap & Hentries & HH & Hrest)".
        match goal with H : value_of s' 0 |- _ => destruct H as (j & ? & Hlast) end.
        set (lv' := <[(j+1)%nat := v]>s').
        assert (map_subseteq s' lv').
        { simpl in *; apply insert_subseteq, Hlast; omega. }
        iDestruct "HH" as (HTa) "(% & Hexcl & HH)".
        destruct H8 as (? & Hwf & HHTa).
        destruct (lookup_lt_is_Some_2 HTa i) as [(k1, lv1) Hpi'].
        { replace (length HTa) with size; omega. }
        iPoseProof (big_sepL_remove _ _ _ _ Hpi' with "HH") as "[Hentry HH]".
        iDestruct "Hentry" as "[% [Hk1 Hhalf2]]".
        iPoseProof (own_snap_ord _ _ _ k1 with "Hk1 Hsnap") as "[% | %]"; [done | subst k1].
        iPoseProof (own_master_eq _ _ _ _ s' with "Hhalf1 Hhalf2") as "%"; subst.
        iPoseProof (own_update_2 with "Hhalf1 Hhalf2") as "Hupd".
        { rewrite master_op half_join.
          etransitivity; first apply make_snap.
          apply @snap_update with (a' := lv'); auto. }
        rewrite own_op -half_join -master_op half_join own_op.
        iMod "Hupd" as "[[Hhalf1 Hhalf2] #Hsnapv]".
        iMod (@own_update _ _ hashG with "Hexcl") as "Hexcl".
        { apply (@excl_update (Z -c> optionC (gmapC nat ZC))). }
        iModIntro; iExists lv'; iSplit; first done.
        assert (value_of lv' v).
        { do 2 eexists; first by simpl in *; apply lookup_insert.
          intros; destruct (decide (n' = j + 1)%nat); first omega.
          subst lv'; rewrite lookup_insert_ne; auto.
          apply Hlast; omega. }
        unfold vP; iFrame.
        repeat (iSplit; first auto).
        iIntros "#Hv'"; iModIntro.
        iApply "Hrest".
        iExists true, lv'; iFrame.
        iSplit.
        * iExists (<[i:=(k, lv')]>T); iSplit.
          { iPureIntro.
             (*split; first by rewrite insert_length.*)
             split.
             { etransitivity; eauto.
               eapply table_incl_upd; unfold zero_ord; eauto.
               etransitivity; eauto. }
             exists i; rewrite nth_insert; last omega.
             rewrite decide_left; split; auto.
             rewrite map_insert lookup_insert_same; auto; rewrite map_length; auto; omega. }
         iApply big_sepL_insert; first done.
         iSplit; last by rewrite (nth_lookup_Some _ _ _ _ H0); auto.
         iPureIntro; do 2 eexists; eauto; done.
       * iSplit.
         { iPureIntro; destruct (HT1 k) eqn: HHT1; auto.
           apply HHTa in HHT1; destruct HHT1 as (Hi & ? & (? & ? & ?) & Hnz); contradiction Hnz.
           assert (g0 = lv1).
           { eapply wf_table_inj with (k := k); eauto.
             apply elem_of_list_lookup; eauto. }
           subst; eauto. }
         iExists (<[i := (k, lv')]>HTa); iSplit.
         { iPureIntro; split; first by rewrite insert_length.
           assert (wf_table (<[i:=(k, lv')]>HTa)).
           { apply wf_table_upd; auto.
             apply Hwf; auto.
             rewrite nth_lookup Hpi'; auto. }
           split; auto.
           intros; unfold map_upd.
           destruct (decide _).
           + split; intros X.
             * inversion X; subst; split; eauto.
               apply insert_in; omega.
             * subst; destruct X as [X _].
               eapply f_equal, wf_table_inj with (k := k); eauto.
               apply insert_in; omega.
           + rewrite HHTa.
             split; intros (Hin & ?); split; auto.
             * apply insert_in_old; congruence.
             * apply in_insert in Hin as [X|]; congruence. }
         iApply (big_sepL_replace with "HH").
         iSplit; last iFrame.
         iPureIntro; do 2 eexists; eauto; done.
     - iNext; iIntros (s' v') "(% & % & Hlv & ? & Hsnap & Hentries & HH & Hrest)".
       iMod "Hlv" as "[[% %] Hsnapv]".
       iAssert (⌜exists x, HT1 k = Some x⌝)%I as %[? Hx].
       { iDestruct "HH" as (HTa) "(% & ? & HH)".
         destruct H8 as (? & ? & HHTa).
         destruct (lookup_lt_is_Some_2 HTa i) as [(k1, lv1) Hpi'].
         { replace (length HTa) with size; omega. }
         iPoseProof (big_sepL_remove _ _ _ _ Hpi' with "HH") as "[Hentry HH]".
         iDestruct "Hentry" as "[% [Hk1 Hhalf2]]".
         destruct H10 as (? & ? & ? & Hzero).
         iPoseProof (own_snap_ord _ _ _ k1 with "Hk1 Hsnap") as "[% | %]"; [done | subst k1].
         iPoseProof (own_snap_ord _ _ _ lv1 with "Hhalf2 Hsnapv") as "%".
         iPureIntro; eexists; rewrite HHTa.
         split; first by eapply elem_of_list_lookup_2.
         do 2 eexists; eauto.
         intro; assert (v' = 0); last done.
         match goal with H : value_of _ v' |- _ => destruct H as (? & ? & ?) end.
         eapply Hzero; eauto.
         eapply lookup_weaken; eauto. }
       iModIntro; iApply ("Hrest" with "[-]").
       iExists false, s'; iFrame.
       iSplit.
       + iExists (<[i:=(k, s')]>T); iSplit.
         { iPureIntro; split.
           { etransitivity; eauto.
             eapply table_incl_upd; unfold zero_ord; eauto. }
           exists i; split.
           { rewrite map_insert lookup_insert_same; auto; rewrite map_length; auto; omega. }
           rewrite nth_insert; last omega.
           rewrite decide_left; done. }
         iApply big_sepL_insert; first done.
         iSplit; last by rewrite (nth_lookup_Some _ _ _ _ H0); auto.
         iPureIntro; do 2 eexists; eauto; done.
       + rewrite Hx; auto.
     - iNext; iApply objective_objectively.
       iIntros (? ?) "[? #?]"; iModIntro; iFrame "#".
     - iFrame; done. }
    iNext. iIntros (? b ?) "[% Hpost]".
    destruct b; done.
  Qed.

  (* Definitions for linearizability *)
  Inductive hashtable_hist_el :=
  (*| HSet (k : Z) (v : Z)*) | HGet (k : Z) (v : Z) | HAdd (k : Z) (v : Z) (r : bool).

  Notation hist := (list (nat * hashtable_hist_el)).

  Definition HT_upd (H : Z -> option (gmap nat Z)) k v H' := exists lv, H' = map_upd H k lv /\
    value_of lv v /\ match H k with Some lv0 => lv0 ⊆ lv | None => True end.

  Fixpoint apply_hist H h H' :=
    match h with
    | nil => H' = H
  (*  | HSet k v :: h' => exists H1, HT_upd H k v H1 /\ apply_hist H1 h' H'*)
    | HGet k v :: h' => (v ≠ 0 -> exists lv, H k = Some lv /\ exists j, lv !! j = Some v) /\
        apply_hist H h' H'
    | HAdd k v r :: h' => if r then H k = None /\ exists H1, HT_upd H k v H1 /\ apply_hist H1 h' H'
                          else H k ≠ None /\ apply_hist H h' H'
    end.

  Instance el_equiv : Equiv hashtable_hist_el := eq.

  Context {linG : inG Σ (frac_authR (gmapR nat (exclR (discreteC hashtable_hist_el))))}.

  Definition hist_list {A} (m : gmap nat A) (l : list A) :=
    forall t e, m !! t = Some e <-> l !! t = Some e.

  Notation ghost_hist g e := (⎡@own _ _ linG g e⎤)%I.

  Definition hashtable_inv gh g lgk lgv H := (hashtable H g lgk lgv ∗
    ∃ hr, ⌜apply_hist (fun _ => None) hr H⌝ ∗ ∃ mr, ⌜hist_list mr hr⌝ ∗
      ghost_hist gh (●! (Excl <$> mr)))%I.

  Definition newer {A} (m : gmap nat A) t := forall i, is_Some (m !! i) -> (i < t)%nat.

  Inductive add_events {A} h : list A -> gmap nat A -> Prop :=
  | add_events_nil : add_events h nil h
  | add_events_snoc : forall le h' t e (Hh' : add_events h le h') (Ht : newer h' t),
      add_events h (le ++ e :: nil) (<[t:=e]>h').
  Hint Resolve add_events_nil.

  Lemma hist_list_newer : forall {A} h (l : list A) (Hhist : hist_list h l), newer h (length l).
  Proof.
    unfold hist_list, newer; intros.
    destruct H as (? & H).
    apply Hhist in H.
    eapply lookup_lt_Some; eauto.
  Qed.

  Lemma incl_newer : forall {A : ofeT} (h h' : gmap nat A) t (Hincl : Excl <$> h ≼ Excl <$> h')
    (Ht : newer h' t), newer h t.
  Proof.
    unfold newer; intros.
    apply Ht.
    assert (Excl <$> h !! i ≼ Excl <$> h' !! i) as Heq.
    { rewrite <- !lookup_fmap.
      apply lookup_included; auto. }
    destruct (h' !! i); eauto.
    destruct H as [? H]; rewrite H in Heq.
    apply option_included in Heq as [|(? & ? & ? & ? & ?)]; done.
  Qed.

  Lemma apply_hist_app : forall h1 h2 H H', apply_hist H (h1 ++ h2) H' <->
    exists H1, apply_hist H h1 H1 /\ apply_hist H1 h2 H'.
  Proof.
    induction h1; auto; simpl; intros.
    { split; [|intros (? & ? & ?)]; subst; eauto. }
    destruct a.
    - split.
      + intros (? & H1); apply IHh1 in H1.
        destruct H1 as (? & ? & ?); eauto.
      + intros (? & (? & ?) & ?); rewrite IHh1; eauto.
    - split.
      + destruct r.
        * intros (? & ? & ? & H1); apply IHh1 in H1.
          destruct H1 as (? & ? & ?); eauto 6.
        * intros (? & H1); apply IHh1 in H1.
          destruct H1 as (? & ? & ?); eauto.
      + destruct r.
        * intros (? & (? & ? & ? & ?) & ?); split; auto.
          eexists; rewrite IHh1; eauto.
        * intros (? & (? & ?) & ?); rewrite IHh1; eauto.
  Qed.

  Lemma hist_list_snoc : forall {A} h l (e : A), hist_list h l ->
    hist_list (<[length l := e]>h) (l ++ e :: nil).
  Proof.
    unfold hist_list; intros.
    destruct (decide (t = length l)).
    - subst; rewrite lookup_insert list_lookup_middle; done.
    - rewrite lookup_insert_ne; last done.
      rewrite H.
      split; intro Ht.
      + rewrite lookup_app_l; auto.
        eapply lookup_lt_Some; eauto.
      + pose proof (lookup_lt_Some _ _ _ Ht) as Hlen; rewrite app_length in Hlen; simpl in Hlen.
        rewrite lookup_app_l in Ht; auto; omega.
  Qed.

  Lemma add_fails : forall k v b l H H' (HH : apply_hist H l H') (Hadd : HAdd k v b ∈ l)
    (Hk : H k ≠ None), b = false.
  Proof.
    induction l; simpl; intros.
    { apply not_elem_of_nil in Hadd; done. }
    destruct a; inversion Hadd; subst.
    - destruct HH; eauto.
    - destruct r; auto.
      destruct HH; done.
    - destruct r.
      + destruct HH as (? & ? & (? & ? & ?) & ?); eapply IHl; eauto.
        subst; unfold map_upd; destruct (decide _); auto.
      + destruct HH; eauto.
  Qed.

  Lemma only_one_add_succeeds : forall k v1 v2 l i1 i2 H0 H (HH : apply_hist H0 l H)
    (Hin1 : l !! i1 = Some (HAdd k v1 true)) (Hin2 : l !! i2 = Some (HAdd k v2 true)),
    i2 = i1 /\ v2 = v1.
  Proof.
    induction l; simpl; intros; first done.
    assert (i2 = i1); [|subst; rewrite Hin2 in Hin1; inversion Hin1; auto].
    destruct i1, i2; auto; simpl in *.
    - inversion Hin1; subst.
      destruct HH as (Hk & ? & (? & ? & ?) & HH).
      eapply add_fails in HH; [|eapply elem_of_list_lookup_2; eauto|]; first done.
      subst; unfold map_upd; rewrite decide_left; done.
    - inversion Hin2; subst.
      destruct HH as (Hk & ? & (? & ? & ?) & HH).
      eapply add_fails in HH; [|eapply elem_of_list_lookup_2; eauto|]; first done.
      subst; unfold map_upd; rewrite decide_left; done.
    - apply f_equal.
      destruct a.
      + destruct HH; eapply IHl; eauto.
      + destruct r.
        * destruct HH as (? & ? & ? & ?); eapply IHl; eauto.
        * destruct HH; eapply IHl; eauto.
  Qed.

  Lemma one_add_succeeds : forall k v b l H0 H (HH : apply_hist H0 l H) (Hk : H0 k = None)
    (Hin : HAdd k v b ∈ l), exists v', HAdd k v' true ∈ l.
  Proof.
    induction l; simpl; intros; first by inversion Hin.
    destruct a; inversion Hin; subst.
    - destruct HH as [? HH].
      destruct (IHl H0 H); auto.
      eexists; constructor; eauto.
    - destruct r; first eauto.
      destruct HH; done.
    - destruct r.
      + destruct (decide (k0 = k)); first by subst; eexists; constructor.
        destruct HH as (_ & ? & (v' & ? & ?) & ?); subst.
        destruct (IHl (map_upd H0 k0 v') H); auto.
        { unfold map_upd.
          destruct (decide _); done. }
        eexists; constructor; eauto.
      + destruct HH as [? HH].
        destruct (IHl H0 H); auto.
        eexists; constructor; eauto.
  Qed.

  Lemma add_events_incl : forall {A} (h : gmap nat A) le h', add_events h le h' -> h ⊆ h'.
  Proof.
    induction 1; first done.
    etransitivity; first done.
    apply insert_subseteq.
    destruct (h' !! t) eqn: Ha; last done.
    lapply (Ht t); [omega | eauto].
  Qed.

  Lemma add_events_newer : forall {A} (h : gmap nat A) le h' t, add_events h le h' ->
    newer h' t -> newer h t.
  Proof.
    repeat intro.
    apply H0.
    eapply lookup_weaken_is_Some; first done.
    eapply add_events_incl; eauto.
  Qed.

  Lemma add_events_in {A} (h : gmap nat A) l h' e: add_events h l h' -> e ∈ l ->
    exists t, newer h t /\ h' !! t = Some e.
  Proof.
    induction 1; first by inversion 1.
    rewrite elem_of_app; intros [X|X].
    - destruct (IHadd_events X) as (? & ? & ?); do 2 eexists; eauto.
      rewrite lookup_insert_ne; first done.
      lapply (Ht x); last eauto; omega.
    - inversion X; subst; last by (contradiction (not_elem_of_nil e)).
      exists t; split; last apply lookup_insert.
      eapply add_events_newer; eauto.
  Qed.

  Hint Resolve not_elem_of_nil.
  Hint Constructors elem_of_list.

  Lemma Permutation_filter {A} P {H : forall x, Decision (P x)} (l l' : list A):
    Permutation l l' -> Permutation (filter P l) (filter P l').
  Proof.
    induction 1; auto; unfold filter; simpl.
    - destruct (decide _); auto.
    - unfold filter; simpl.
      destruct (decide _), (decide _); auto.
      constructor.
    - etransitivity; eauto.
  Qed.

  Lemma hist_list_snoc_inv : forall {A} (l : list A) h e, hist_list h (l ++ e :: nil) ->
    exists h', hist_list h' l /\ h = <[length l := e]>h'.
  Proof.
    intros; exists (h ∖ <[length l := e]>∅).
    split; first (unfold hist_list; split; intro Hl).
    - apply lookup_difference_Some in Hl as [Hl Ht].
      apply H, lookup_app_Some in Hl as [|[? Hl]]; auto.
      destruct (t - length l)%nat eqn: Hminus; inversion Hl.
      assert (t = length l) by omega; subst; rewrite lookup_insert in Ht; done.
    - rewrite lookup_difference_Some.
      pose proof (lookup_lt_Some _ _ _ Hl); rewrite lookup_insert_ne; last omega; split; auto.
      rewrite H lookup_app_l; done.
    - eapply map_eq; intro.
      destruct (decide (i = length l)).
      + subst; rewrite lookup_insert.
        apply H, list_lookup_middle; done.
      + rewrite lookup_insert_ne; last done.
        destruct (h !! i) eqn: Hi.
        * symmetry; rewrite lookup_difference_Some lookup_insert_ne; done.
        * symmetry; rewrite lookup_difference_None; auto.
  Qed.

  Lemma hist_list_perm : forall {A} (l : list A) h, hist_list h l ->
    Permutation (map snd (map_to_list h)) l.
  Proof.
    induction l using rev_ind; unfold hist_list; intros.
    - rewrite (map_empty h); auto.
      intros; destruct (h !! i) eqn: Hi; auto.
      apply H in Hi; done.
    - apply hist_list_snoc_inv in H as (h' & ? & ?); subst.
      rewrite map_to_list_insert; simpl.
      etransitivity; last apply Permutation_cons_append.
      constructor; apply IHl; done.
      { destruct (h' !! length l) eqn: Hl; last done.
        apply H in Hl.
        pose proof (lookup_lt_Some _ _ _ Hl); omega. }
  Qed.

  Lemma add_events_inv {A} (h : gmap nat A) le h' t e: add_events h le h' ->
    h' !! t = Some e -> h !! t = Some e \/ e ∈ le.
  Proof.
    induction 1; auto; intro Hi.
    rewrite elem_of_app; destruct (decide (t0 = t)).
    - subst; rewrite lookup_insert in Hi; inversion Hi; auto.
    - rewrite lookup_insert_ne in Hi; last done; tauto.
  Qed.

  Lemma add_events_perm {A} (le : list A) h : add_events ∅ le h ->
    Permutation (map snd (map_to_list h)) le.
  Proof.
    induction 1; auto.
    rewrite map_to_list_insert; simpl.
    etransitivity; last apply Permutation_cons_append; constructor; done.
    { specialize (Ht t).
      destruct (h' !! t); auto; lapply Ht; [omega | eauto]. }
  Qed.

  Lemma excl_map_op (m m1 : gmap nat hashtable_hist_el) m2:
    @Excl (discreteC hashtable_hist_el) <$> m ≡ (Excl <$> m1) ⋅ m2 ->
    exists m2', m2 = Excl <$> m2' /\ m1 ##ₘ m2' /\ m = m1 ∪ m2'.
  Proof.
    intro.
    assert (∀ i, m2 !! i = match m1 !! i with Some _ => None | _ => Excl <$> (m !! i) end) as Hm2.
    { intro; specialize (H i).
      apply option_leibniz in H; rewrite lookup_merge !lookup_fmap in H.
      destruct (m1 !! i) eqn: Hm1, (m2 !! i) eqn: Hm2, (m !! i) eqn: Hm;
        rewrite Hm1 Hm2 Hm in H; inversion H; done. }
    set (m2' := (fun x => match x with Excl a => a | _ => HGet 0 0 end) <$> m2); exists m2'.
    assert (m1 ##ₘ m2'); last repeat split; auto.
    - intro i.
      rewrite lookup_fmap Hm2.
      destruct (m1 !! i); try done.
      destruct (m !! i); done.
    - apply map_eq; intro i.
      rewrite -map_fmap_compose lookup_fmap; unfold compose.
      rewrite Hm2; destruct (m1 !! i); auto; destruct (m !! i); auto.
    - apply map_eq; intro i; specialize (H i).
      apply option_leibniz in H; rewrite lookup_merge !lookup_fmap Hm2 in H.
      destruct (m !! i) eqn: Hm; rewrite Hm in H; symmetry.
      + rewrite lookup_union_Some; last done.
        subst m2'; rewrite lookup_fmap Hm2 Hm.
        destruct (m1 !! i) eqn: Hm1; auto.
        rewrite Hm1 in H; inversion H; auto.
      + rewrite lookup_union_None.
        subst m2'; rewrite lookup_fmap Hm2 Hm.
        destruct (m1 !! i) eqn: Hm1; auto.
        rewrite Hm1 in H; inversion H.
  Qed.

  Lemma excl_map_inj {A} (m1 m2 : gmap nat A): Excl <$> m1 = Excl <$> m2 -> m1 = m2.
  Proof.
    intros.
    apply map_eq; intro.
    assert (Excl <$> (m1 !! i) = Excl <$> (m2 !! i)) as Heq.
    { rewrite -!lookup_fmap H; done. }
    destruct (m1 !! i), (m2 !! i); inversion Heq; done.
  Qed.

  Lemma map_to_list_union {K A M} `{FinMapToList K A M} `{EqDecision K} `{Countable K}
    (m1 m2 : gmap K A): m1 ##ₘ m2 ->
    Permutation (map_to_list (m1 ∪ m2)) (map_to_list m1 ++ map_to_list m2).
  Proof.
    intro Hdisj.
    rewrite -(map_of_to_list m1) -foldr_insert_union.
    pose proof (NoDup_fst_map_to_list m1) as HNoDup.
    remember (map_to_list m1) as l1.
    assert (Permutation (map_to_list m1) l1) as Hl1 by (rewrite Heql1; done).
    clear Heql1; revert dependent m1; induction l1; auto; intros; simpl.
    destruct a as (k, v).
    pose proof (map_to_list_insert_inv _ _ _ _ Hl1); subst; simpl.
    inversion HNoDup as [|??? HNoDup']; subst.
    apply map_disjoint_insert_l in Hdisj as [].
    pose proof (not_elem_of_map_of_list_1 _ _ H1).
    rewrite !map_to_list_insert; auto.
    constructor; apply (IHl1 HNoDup' (map_of_list l1)); try done.
    { rewrite map_to_of_list; done. }
    { rewrite foldr_insert_union lookup_union_None; auto. }
  Qed.

  Lemma add_three a1 a2 a3 ha b1 b2 b3 hb c1 c2 c3 hc m l HT
    (Hha : add_events ∅ (HAdd 1 1 a1 :: HAdd 2 1 a2 :: HAdd 3 1 a3 :: nil) ha)
    (Hhb : add_events ∅ (HAdd 1 1 b1 :: HAdd 2 1 b2 :: HAdd 3 1 b3 :: nil) hb)
    (Hhc : add_events ∅ (HAdd 1 1 c1 :: HAdd 2 1 c2 :: HAdd 3 1 c3 :: nil) hc)
    (Hm : @Excl (discreteC hashtable_hist_el) <$> m ≡
      ∅ ⋅ ((Excl <$> ha) ⋅ ((Excl <$> hb) ⋅ (Excl <$> hc)))) (Hl : hist_list m l)
    (HHT : apply_hist (fun _ => None) l HT):
    length (filter (fun b => b = true) (a1 :: a2 :: a3 :: b1 :: b2 :: b3 :: c1 :: c2 :: c3 :: nil))
      = 3%nat.
  Proof.
    rewrite -(fmap_empty Excl) in Hm.
    apply excl_map_op in Hm as (m1 & Hm1 & _ & ?); subst.
    rewrite left_id in Hl.
    lapply (excl_map_op m1 ha ((@Excl (discreteC hashtable_hist_el) <$> hb) ⋅ (Excl <$> hc)));
      last by rewrite Hm1.
    clear Hm1; intros (m2 & Hm2 & ? & ?); subst.
    lapply (excl_map_op m2 hb (Excl <$> hc)); last by rewrite Hm2.
    clear Hm2; intros (m3 & ?%excl_map_inj & ? & ?); subst m3; subst.
    rewrite (Permutation_filter _ _
      (map (fun e => match e with HAdd _ _ b => b | _ => false end) l)).
    assert (forall t e, ha !! t = Some e -> l !! t = Some e) as Hin.
    { intros; apply Hl.
      rewrite lookup_union_Some; auto. }
    destruct (one_add_succeeds 1 1 a1 _ _ _ HHT) as [v1 Hv1]; auto.
    { destruct (add_events_in _ _ _ (HAdd 1 1 a1) Hha) as (? & ? & ?); first constructor.
      eapply elem_of_list_lookup_2, Hin; eauto. }
    destruct (one_add_succeeds 2 1 a2 _ _ _ HHT) as [v2 Hv2]; auto.
    { destruct (add_events_in _ _ _ (HAdd 2 1 a2) Hha) as (? & ? & ?); first repeat constructor.
      eapply elem_of_list_lookup_2, Hin; eauto. }
    destruct (one_add_succeeds 3 1 a3 _ _ _ HHT) as [v3 Hv3]; auto.
    { destruct (add_events_in _ _ _ (HAdd 3 1 a3) Hha) as (? & ? & ?); first repeat constructor.
      eapply elem_of_list_lookup_2, Hin; eauto. }
    apply elem_of_list_lookup_1 in Hv1 as [i1 Hi1].
    apply elem_of_list_lookup_1 in Hv2 as [i2 Hi2].
    apply elem_of_list_lookup_1 in Hv3 as [i3 Hi3].
    rewrite (filter_find_count _ _ (i1 :: i2 :: i3 :: nil)); auto.
    - repeat constructor; auto.
      + rewrite !elem_of_cons; intros [|[|]]; subst; last by eapply not_elem_of_nil.
        * rewrite Hi1 in Hi2; done.
        * rewrite Hi1 in Hi3; done.
      + rewrite elem_of_cons; intros [|]; subst; last by eapply not_elem_of_nil.
        rewrite Hi2 in Hi3; done.
    - repeat constructor; rewrite map_lookup; [rewrite Hi1 | rewrite Hi2 | rewrite Hi3]; eauto.
    - intros ???; rewrite map_lookup; intros Hhit ?.
      destruct (l !! i) as [[|]|] eqn: Hi; rewrite Hi in Hhit; inversion Hhit; subst; try done.
      assert (k = 1 \/ k = 2 \/ k = 3) as Hk.
      { apply Hl in Hi.
        apply lookup_union_Some in Hi; auto.
        destruct Hi as [Hi | Hi]; last (apply lookup_union_Some in Hi; auto;
          destruct Hi as [Hi | Hi]).
        * destruct (add_events_inv _ _ _ _ _ Hha Hi) as [X|X]; first by rewrite lookup_empty in X.
          inversion X as [|??? X']; auto; subst.
          inversion X' as [|??? X'']; auto; subst.
          apply elem_of_list_singleton in X''; inversion X''; auto.
        * destruct (add_events_inv _ _ _ _ _ Hhb Hi) as [X|X]; first by rewrite lookup_empty in X.
          inversion X as [|??? X']; auto; subst.
          inversion X' as [|??? X'']; auto; subst.
          apply elem_of_list_singleton in X''; inversion X''; auto.
        * destruct (add_events_inv _ _ _ _ _ Hhc Hi) as [X|X]; first by rewrite lookup_empty in X.
          inversion X as [|??? X']; auto; subst.
          inversion X' as [|??? X'']; auto; subst.
          apply elem_of_list_singleton in X''; inversion X''; auto. }
      destruct Hk as [|[|]]; subst; eapply only_one_add_succeeds in Hi as [];
        try apply Hi1; try apply Hi2; try apply Hi3; try eassumption; subst; contradiction H1;
        repeat constructor.
    - apply hist_list_perm in Hl.
      etransitivity; last by apply Permutation_map.
      apply add_events_perm in Hha; apply add_events_perm in Hhb; apply add_events_perm in Hhc.
      rewrite !map_to_list_union; auto.
      rewrite !map_app.
      etransitivity; last apply Permutation_app; last apply Permutation_app;
        try (apply Permutation_map; symmetry; eassumption); done.
  Qed.

  Instance hashtable_inv_objective : Objective (hashtable_inv gh g lgk lgv H).
  Proof.
    intros; unfold hashtable_inv.
    apply sep_objective, _.
    unfold hashtable.
    apply exists_objective; intro.
    apply sep_objective; first apply _.
    apply sep_objective; first apply _.
    apply big_sepL_objective; intros ? (?, ?).
    apply _.
  Qed.

  Instance hashtable_inv_timeless : Timeless (hashtable_inv gh g lgk lgv H).
  Proof.
    intros; unfold hashtable_inv.
    apply sep_timeless, _.
    unfold hashtable.
    apply exist_timeless; intro.
    apply sep_timeless; first apply _.
    apply sep_timeless.
    { apply embed_timeless, own_timeless, @Excl_discrete.
      repeat intro.
      apply option_ofe_discrete; auto.
      apply _. }
    apply big_sepL_timeless; intros ? (?, ?).
    apply _.
  Qed.

  Opaque add_item.

  Definition test_inv gq gh t results := (∃ b1 b2 b3 h,
    ⌜add_events ∅ (HAdd 1 1 b1 :: HAdd 2 1 b2 :: HAdd 3 1 b3 :: nil) h⌝ ∗
    ghost_hist gh (◯!{gq} (Excl <$> h)) ∗
    ZplusPos t results ↦ length (filter (fun b => b = true) (b1 :: b2 :: b3 :: nil)))%I.

  Context `{mG : inG Σ (exclR unitC)}.

  Lemma subst_for s e i cond incr body (Hs : s ≠ "loop" /\ s ≠ "i") :
    subst s e (for_loop i cond incr body) =
    for_loop (subst s e i) (subst s e cond) (subst s e incr) (subst s e body).
  Proof.
    unfold for_loop, subst; fold subst.
    destruct Hs, (decide _); last by contradiction n; split; congruence.
    destruct (decide _); first done.
    destruct (decide _); last by contradiction n0.
    destruct (decide _); done.
  Qed.

  Opaque for_loop.

  Ltac wp_rec' := lazymatch goal with |-context[Rec ?f ?x ?e] => let H := fresh "Hclosed" in
    assert (Closed (f :b: x :b: nil) e) as H by apply I; wp_rec; clear H;
    unfold subst; fold subst; try (rewrite !subst_for; try done); unfold subst; fold subst; simpl end.

  Lemma test_spec t l N entries gt gh g lgk lgv q gq T0 (Hentries : length entries = size)
    (HlenT0 : length T0 = size) (HN : base_mask ## ↑N):
    {{{ ⎡PSCtx⎤ ∗ ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
        hashtable_A T0 lgk lgv entries ∗ inv N (∃H, hashtable_inv gh g lgk lgv H) ∗
        ghost_hist gh (◯!{gq} ∅) ∗ ZplusPos t results ↦ 0 ∗
        [PP l in false | onceP gt (test_inv gq gh t results)] }}}
    test #t #l
    {{{ v, RET v; [PP l in true | onceP gt (test_inv gq gh t results)] }}}.
  Proof.
    intros; iIntros "(#Hctx & Hentries & #HA & #Hinv & Hhist & Ht & #Hl) Hpost".
    unfold test.
    unfold of_val at 1.
    wp_rec'. rewrite is_closed_nil_subst; last apply I.
    wp_rec'. rewrite is_closed_nil_subst; last apply I.
    wp_bind (for_loop _ _ _ _).
    iApply (wp_for_simple with "[] [Hentries Hhist Ht]").
    - instantiate (1 := (fun i => ∃ ls h,
        ⌜0 <= i <= 3 /\ length ls = Z.to_nat i /\
         add_events ∅ (imap (fun j b => HAdd (j + 1) 1 b) ls) h⌝ ∗
        ([∗ list] i↦x ∈ entries, (ZplusPos i m_entries) ↦{q} Z.pos x) ∗
        ghost_hist gh (◯!{gq} (Excl <$> h)) ∗
        ZplusPos t results ↦ length (filter (fun b => b = true) ls))%I).
      iIntros "!#" (i ?) "!# [% H] Hpost".
      iDestruct "H" as (ls h) "((% & % & %) & Hentries & Hhist & Ht)".
      wp_rec'.
      wp_op.
      iApply (wp_wand with "[-Hpost]"); last iApply "Hpost".
      wp_bind (add_item _ _)%E.
      (* We need the full postcondition of add_item in the atomic shift, so we can't use LA_Inv. *)
      iApply (LA_Mono _ _ _ _ _ (⊤ ∖ ↑N) with "[Hentries]");
        [| by iApply (add_item_spec with "[$]"); first omega |].
      { by apply namespaces.ndisj_subseteq_difference. }
      unfold atomic_shift.
      iExists _; iSplit; first by iCombine "Hhist Ht" as "H"; iApply "H".
      iIntros "!# [>Hhist Ht]".
      iMod (inv_open with "Hinv") as "[>HH Hclose]"; first auto.
      iDestruct (monPred_objectively_elim with "HH") as (HT) "[HH H]".
      iModIntro; iExists HT; iFrame "HH".
      iSplit.
      { iIntros "HH".
        iMod ("Hclose" with "[HH H]"); last by iFrame.
        iNext; iApply objective_objectively.
        iExists HT; iFrame. }
      iDestruct "H" as (hr) "[% H]"; iDestruct "H" as (mr) "[% Hg]".
      iIntros (v) "H"; iDestruct "H" as (b lv) "(Hentries & H & % & HH)"; iDestruct "H" as (T) "[% HA']".
      iAssert ⌜newer h (length hr)⌝%I as %Hnew.
      { iPoseProof (own_valid_2 with "Hg Hhist") as "%".
        iPureIntro.
        eapply (incl_newer(A := discreteC hashtable_hist_el)).
        * eapply @frac_auth_included_total; eauto; apply _.
        * apply hist_list_newer; auto. }
      iPoseProof (own_update_2 with "Hg Hhist") as "Hupd".
      { apply frac_auth_update.
        apply alloc_local_update with (i0 := length hr)(x := Excl (HAdd (i + 1) 1 b)); last done.
        rewrite lookup_fmap.
        destruct (mr !! length hr) eqn: Hr; auto.
        apply H4, lookup_lt_Some in Hr; omega. }
      iMod "Hupd" as "[Hg Hhist]".
      rewrite <- !fmap_insert.
      destruct H5 as (? & Hb & Hadd); subst v.
      iMod ("Hclose" with "[HH Hg]").
      { iNext.
        iApply objective_objectively.
        iExists (if b then map_upd HT (i + 1) lv else HT); iFrame.
        iExists (hr ++ HAdd (i + 1) 1 b :: nil)%list; iSplit.
        { iPureIntro; rewrite apply_hist_app.
          do 2 eexists; eauto; simpl.
          subst; destruct (HT (i + 1)) eqn: Hi; split; auto.
          do 2 eexists; eauto.
          do 2 eexists; first eauto.
          rewrite Hi; auto. }
        iExists (<[length hr:=HAdd (i + 1) 1 b]> mr); iSplit; last done.
        iPureIntro.
        apply hist_list_snoc; auto. }

      iModIntro.
      wp_let.
      destruct b.
      + wp_if_true.
        wp_op.
        wp_op.
        wp_bind ([_]_na)%E.
        iApply (na_read with "[$Hctx $Ht]"); first auto.
        iNext; iIntros (?) "[% Ht]"; simpl.
        wp_op.
        iApply (na_write with "[$Hctx $Ht]"); first auto.
        iNext; iIntros "Ht".
        iExists (ls ++ true :: nil)%list, (<[length hr:=HAdd (i + 1) 1 true]> h).
        rewrite filter_app !app_length imap_app; simpl.
        rewrite !Nat.add_0_r Nat2Z.inj_add.
        subst; iSplit; last (iFrame "#"; iFrame).
        iPureIntro.
        split; first omega.
        rewrite Z2Nat.inj_add; try omega.
        change (Z.to_nat 1) with 1%nat.
        split; first (simpl; omega).
        rewrite H1 Z2Nat.id; last omega.
        constructor; auto.
      + wp_if_false.
        iExists (ls ++ false :: nil)%list, (<[length hr:=HAdd (i + 1) 1 false]> h).
        rewrite filter_app !app_length imap_app; simpl.
        rewrite !Nat.add_0_r.
        iSplit; last (iFrame "#"; iFrame).
        iPureIntro.
        split; first omega.
        rewrite Z2Nat.inj_add; try omega.
        change (Z.to_nat 1) with 1%nat.
        split; first (simpl; omega).
        rewrite H1 Z2Nat.id; last omega.
        constructor; auto.
    - simpl.
      iExists nil, ∅; iFrame; simpl.
      rewrite fmap_empty; iFrame "#"; iFrame.
      iPureIntro; split; auto; omega.
    - iNext; iIntros "H"; iDestruct "H" as (i) "[% H]".
      iDestruct "H" as (ls h) "((% & % & %) & Hentries & Hhist & Ht)".
      wp_seq.
      iApply (GPS_PP_Write (p:=tt) (onceP _ _) true emp%I (fun s => emp)%I with "[$Hctx Hhist Ht]");
        [auto | auto | ..].
      { intro; hnf.
        rewrite ssrbool.implybT; done. }
      { iSplit; last auto.
        iNext; iIntros "_".
        iPoseProof (escrow_alloc (⎡own gt (Excl ())⎤%I) (test_inv gq gh t results) with "[Hhist Ht]")
          as "HX".
        { iNext; unfold test_inv.
          assert (i = 3) by omega; subst.
          destruct ls as [|b1]; [discriminate|].
          destruct ls as [|b2]; [discriminate|].
          destruct ls as [|b3]; [discriminate|].
          destruct ls; [|discriminate].
          iExists b1, b2, b3, h; simpl in *; iFrame; auto. }
        iApply (fupd_mask_mono (↑escrowN)).
        { apply namespaces.ndisj_subseteq_difference; last done.
          do 2 apply namespaces.ndot_preserve_disjoint_r.
          apply namespaces.ndot_ne_disjoint; done. }
        iMod "HX" as "#HX"; iModIntro.
        unfold onceP; iFrame "#"; auto. }
      iNext; iIntros "[? _]".
      iApply "Hpost"; done.
  Qed.

  Lemma Qp_ext : forall q1 q2, Qp_car q1 = Qp_car q2 -> @eq Qp q1 q2.
  Proof.
    intros [] []; simpl; intro; subst.
    f_equal; apply proof_irrel.
  Qed.

  Lemma Qc_div_div : forall q p1 p2, (q / p1 / p2 = q / (p1 * p2))%Qp.
  Proof.
    intros.
    apply Qp_ext.
    unfold Qp_div; simpl.
    rewrite Pos2Z.inj_mul Z2Qc_inj_mul Qcanon.Qcinv_mult_distr.
    rewrite Qcanon.Qcmult_assoc; done.
  Qed.

  Lemma big_sepL_sep_persistent {A} (Φ : nat → A → vProp Σ) l P {_ : Persistent P}:
    P ∗ ([∗ list] k↦x ∈ l, Φ k x) -∗ [∗ list] k↦x ∈ l, P ∗ Φ k x.
  Proof.
    iIntros "[??]".
    rewrite big_sepL_sepL; iFrame.
    induction l; simpl; auto.
  Qed.

  Instance onceP_persistent g P b l s v: Persistent (onceP g P b l s v).
  Proof.
    destruct s; apply _.
  Qed.

  Lemma drop_lookup: forall {A} i l (d : A), (i < length l)%nat ->
    drop i l = nth i l d :: drop (S i) l.
  Proof.
    induction i; destruct l; simpl; auto; intros; try omega.
    apply IHi; omega.
  Qed.

  Opaque init_table test seq.

  Lemma main_spec :
    {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ ([∗ list] i↦_ ∈ replicate size 0, ZplusPos i m_entries ↦ ?) ∗
        ([∗ list] i ∈ seq 0 3, ZplusPos (Z.of_nat i) results ↦ ?) }}}
    main #() {{{ v, RET v; ⌜v = #3⌝ }}}.
  Proof using mG linG histG hashG gpGv gpGk gpGb cG agreeGv agreeGk agreeGb ZG.
    intros; iIntros "(#Hctx & #Hctx' & Hentries & Hresults) Hpost".
    unfold main.
    unfold of_val at 1.
    wp_rec'.
    match goal with |-context[App (Rec _ ?b ?e) _] => assert (Closed (b :b: nil) e) as Hclosed by apply I end.
    wp_bind (init_table _)%E; clear Hclosed.
    iApply (init_table_spec with "[$Hentries]"); first auto.
    iNext; iIntros (?) "H"; iDestruct "H" as (entries g lgk lgv) "(% & Hentries & #HA & HH)".
    wp_rec'.
    match goal with |-context[App (Rec _ ?b ?e) _] => assert (Closed (b :b: nil) e) as Hclosed by apply I end.
    wp_bind (malloc _)%E; clear Hclosed.
    iApply wp_mask_mono; first auto.
    iApply (malloc_spec with "Hctx"); [omega|done|].
    iNext; iIntros (t) "Ht".
    wp_rec'.
    rewrite is_closed_nil_subst; last done.
    iMod (own_alloc (Excl ())) as (gt1) "Htok1"; first done.
    iMod (own_alloc (Excl ())) as (gt2) "Htok2"; first done.
    iMod (own_alloc (Excl ())) as (gt3) "Htok3"; first done.
    set (gts := gt1 :: gt2 :: gt3 :: nil).
    iMod (own_alloc (●! ∅ ⋅ ◯! ∅)) as (gh) "[Href Hhist]"; first done.
    set (N := namespaces.ndot namespaces.nroot "N").
    iMod (inv_alloc N top (∃H, hashtable_inv gh g lgk lgv H)%I with "[HH Href]") as "#Hinv".
    { iNext.
      iApply objective_objectively.
      iExists (λ _ : Z, None); iFrame.
      iExists nil; iSplit; first auto.
      iExists ∅; rewrite fmap_empty; auto. }

    match goal with |-context[App (Rec _ ?b ?e) _] => assert (Closed (b :b: nil) e) as Hclosed by apply I end.
    wp_bind (for_loop _ _ _ _); clear Hclosed.
    iPoseProof (big_sepL_sep_persistent _ _ ⎡PSCtx⎤%I with "[$Hctx $Hentries]") as "Hentries".
    iPoseProof (big_sepL_mono with "Hentries") as "Hentries".
    { intros; iIntros "[Hctx H]".
      iApply (na_frac_from_non with "[$Hctx $H]"); auto. }
    iMod (fupd_big_sepL with "Hentries") as "Hentries".
    iApply (wp_for_simple with "[] [Hentries Ht Hhist Hresults]").
    { instantiate (1 := (fun i : Z => ⌜0 <= i <= 3⌝ ∗
        ([∗ list] j↦x ∈ entries, ZplusPos j m_entries ↦{1/Z.to_pos (2^i)} Z.pos x) ∗
        ghost_hist gh (◯!{1/Z.to_pos (2^i)} ∅) ∗
        ([∗ list] i ∈ seq 0 (Z.to_nat i), [PP ZplusPos (Z.of_nat i) t in false |
           onceP (nth i gts undef) (test_inv (1/Z.to_pos (2^(Z.of_nat i + 1))) gh (Z.of_nat i) results) ]) ∗
        ([∗ list] i ∈ seq (Z.to_nat i) (3 - Z.to_nat i), ZplusPos (Z.of_nat i) t ↦ ?) ∗
        ([∗ list] i ∈ seq (Z.to_nat i) (3 - Z.to_nat i), ZplusPos (Z.of_nat i) results ↦ ?))%I).
      iIntros "!#" (i ?) "!# (% & % & Hentries & Hhist & Hprots & Ht & Hresults) Hpost".
      wp_let.
      wp_op.
      wp_let.
      wp_bind ([_]_na <- _)%E.
      destruct (3 - Z.to_nat i)%nat eqn: Hthree.
      { lapply (Nat.sub_gt 3 (Z.to_nat i)); first done.
        change 3%nat with (Z.to_nat 3); apply Z2Nat.inj_lt; omega. }
      change (seq _ (S n)) with (Z.to_nat i :: seq (S (Z.to_nat i)) n); simpl.
      iDestruct "Ht" as "[Hi Ht]"; iDestruct "Hresults" as "[Hres Hresults]".
      rewrite Z2Nat.id; last omega.
      iApply (GPS_PP_Init (p:=tt) (onceP (nth (Z.to_nat i) gts undef)
        (test_inv ((1/Z.to_pos (2^(i + 1)))) gh i results)) false with "[$Hctx $Hctx' $Hi]");
        [done|done|done|].
      iNext; iIntros "#Hprot".
      wp_seq.
      iAssert ((([∗ list] j↦x ∈ entries, ZplusPos j m_entries ↦{1 / Z.to_pos (2 ^ (i + 1))} Z.pos x) ∗
        ([∗ list] j↦x ∈ entries, ZplusPos j m_entries ↦{1 / Z.to_pos (2 ^ (i + 1))} Z.pos x)) ∗
        ghost_hist gh (◯!{1 / Z.to_pos (2 ^ (i + 1))} ∅) ∗
        ghost_hist gh (◯!{1 / Z.to_pos (2 ^ (i + 1))} ∅))%I with "[Hentries Hhist]"
        as "[[? Hentries] [? Hhist]]".
      { assert (1 / Z.to_pos (2 ^ (i + 1)) + 1 / Z.to_pos (2 ^ (i + 1)) = 1 / Z.to_pos (2 ^ i))%Qp
          as Heq.
        { rewrite Z.pow_add_r; try omega.
          rewrite Z.pow_1_r Z2Pos.inj_mul; try omega; simpl.
          rewrite -Qc_div_div Qp_div_2; done.
          { apply Z.pow_pos_nonneg; omega. } }
        iSplitL "Hentries".
        * rewrite -big_sepL_sepL.
          iApply (big_sepL_mono with "Hentries").
          simpl; intros; iIntros "H".
          iApply na_frac_split; rewrite Heq; done.
        * iDestruct "Hhist" as "[??]".
          rewrite Z.pow_add_r; try omega.
          rewrite Z.pow_1_r Z2Pos.inj_mul; try omega; simpl.
          rewrite -Qc_div_div; iFrame.
          { apply Z.pow_pos_nonneg; omega. } }
      wp_op.
      wp_bind ([_]_na <- _)%E.
      iApply (na_write with "[$Hctx $Hres]"); first done.
      iNext; iIntros "Hres".
      wp_seq.
      iApply wp_fork; first done.
      iSplit; first done.
      iSplitL "Hpost Hentries Hhist Hprots Ht Hresults".
      + iApply "Hpost".
        iSplit; first by iPureIntro; omega.
        rewrite Z2Nat.inj_add; try omega.
        rewrite Nat.sub_add_distr Hthree Nat.add_1_r seq_S big_sepL_app; simpl.
        rewrite Nat.sub_0_r; iFrame.
        rewrite Z2Nat.id; [auto | omega].
      + iApply (test_spec with "[-]"); [ | | | by iFrame "# ∗" |]=>//.
        * unfold base_mask; rewrite !disjoint_union_l.
          split; first by split; apply namespaces.ndot_ne_disjoint.
          apply namespaces.ndot_preserve_disjoint_l, namespaces.ndot_ne_disjoint; done.
        * iIntros "!> !>" (?) "?"; done. }
    { simpl.
      change (seq 0 (Z.to_nat 0)) with (@nil nat); simpl.
      iSplit; first by iPureIntro; omega.
      rewrite Qp_div_1.
      iFrame. }
    iNext; iIntros "H"; iDestruct "H" as (i) "(% & % & Hentries & Hhist & #Hprots & _ & _)".
    assert (i = 3) by omega; subst.
    wp_rec'.

    match goal with |-context[App (Rec _ ?b ?e) _] => assert (Closed (b :b: nil) e) as Hclosed by apply I end.
    wp_bind (Alloc); clear Hclosed.
    iApply (alloc with "Hctx"); first done.
    iNext; iIntros (total) "Htotal".
    wp_rec'.
    match goal with |-context[App (Rec _ ?b ?e) _] => assert (Closed (b :b: nil) e) as Hclosed by apply I end.
    wp_bind ([_]_na <- _)%E; clear Hclosed.
    iApply (na_write with "[$Hctx $Htotal]"); first done.
    iNext; iIntros "Htotal".
    wp_rec'.
    wp_bind (for_loop _ _ _ _).
    iApply (wp_for_simple with "[] [Htok1 Htok2 Htok3 Hentries Htotal]").
    { instantiate (1 := (fun i : Z => ⌜0 <= i <= 3⌝ ∗
        ([∗ list] x ∈ drop (Z.to_nat i) gts, ⎡own x (Excl ())⎤) ∗
        ∃ lr, ⌜length lr = Z.to_nat i⌝ ∗
          ([∗ list] j↦x ∈ lr, let '(h, b1, b2, b3) := x in
            ⌜add_events ∅ (HAdd 1 1 b1 :: HAdd 2 1 b2 :: HAdd 3 1 b3 :: nil) h⌝ ∗
            ghost_hist gh (◯!{1/Z.to_pos (2^(Z.of_nat j + 1))} (Excl <$> h))) ∗
          total ↦ sum_list_with (fun '(_, b1, b2, b3) =>
            length (filter (fun b => b = true) (b1 :: b2 :: b3 :: nil))) lr)%I).
      iIntros "!#" (i ?) "!# (% & % & Htoks & H) Hpost"; iDestruct "H" as (lr) "(% & Hhists & Htotal)".
      wp_let.
      match goal with |-context[(WP (_ ?e) {{ _ }})%I] => wp_bind e end.
      iLöb as "IH".
      iApply wp_repeat.
      wp_op.
      assert (Z.to_nat i < 3)%nat.
      { match goal with H : i < 3 |- _ => apply Z2Nat.inj_lt in H; try omega end; done. }
      iPoseProof (big_sepL_lookup with "Hprots") as "Hprot".
      { apply (lookup_seq _ _ (Z.to_nat i)); done. }
      rewrite Nat.add_0_l Z2Nat.id; last omega.
      rewrite (drop_lookup _ _ undef); last done.
      iDestruct "Htoks" as "[Htok Htoks]".
      iApply (GPS_PP_Read (p := tt) (onceP _ _) emp%I (fun s v => ⌜v = Z.b2z s⌝ ∗
        if s then test_inv (1/Z.to_pos (2^(i + 1))) gh i results
        else ⎡own (nth (Z.to_nat i) gts undef) (Excl ())⎤)%I
        with "Hctx [Htok $Hprot]"); [auto | auto | ..].
      { iSplitL "Htok".
        * iNext; iIntros (s) "_".
          iLeft; iIntros (v) "[[% H] _]".
          destruct s; last by iModIntro; auto.
          iPoseProof (escrow_apply with "[] [$H $Htok]") as "H".
          { iIntros "[H1 H2]".
           iPoseProof (own_valid_2 with "H1 H2") as "%"; done. }
          iPoseProof (fupd_mask_mono (↑escrowN) with "H") as ">H".
          { apply namespaces.ndisj_subseteq_difference; last done.
            do 2 apply namespaces.ndot_preserve_disjoint_r.
            apply namespaces.ndot_ne_disjoint; done. }
          iMod "H"; iModIntro; auto.
        * iNext; do 2 (rewrite monPred_objectively_forall; iIntros (?)).
          iApply objectively_entails.
          iIntros "[? #?]"; iModIntro; iFrame "#". }
      iNext; iIntros (b ?) "(% & _ & % & Hcond)"; subst.
      iExists (Z.b2z b); iSplit; first auto.
      destruct b; simpl.
      + iClear "IH".
        iNext; iNext. wp_seq.
        wp_bind ([_]_na)%E.
        iApply (na_read with "[$Hctx $Htotal]"); first auto.
        iNext; iIntros (?) "[% Htotal]".
        wp_op.
        wp_bind ([_]_na)%E.
        iDestruct "Hcond" as (b1 b2 b3 h) "(% & ? & Hres)".
        iApply (na_read with "[$Hctx $Hres]"); first auto.
        iNext; iIntros (?) "[% Hres]"; subst.
        wp_op.
        iApply (na_write with "[$Hctx $Htotal]"); first auto.
        iNext; iIntros "Htotal".
        iApply "Hpost".
        iSplit; first by iPureIntro; omega.
        rewrite Z2Nat.inj_add; try omega.
        rewrite Nat.add_1_r; simpl; iFrame.
        iExists (lr ++ (h, b1, b2, b3) :: nil)%list.
        rewrite app_length; simpl; iSplit; first by iPureIntro; omega.
        rewrite big_sepL_app; simpl; iFrame.
        iSplitR "Htotal".
        * iSplit; last auto.
          iSplit; first auto.
          replace (length lr) with (Z.to_nat i); rewrite Nat.add_0_r Z2Nat.id; auto; omega.
        * rewrite sum_list_with_app; simpl.
          rewrite Nat.add_0_r Nat2Z.inj_add; done.
      + iNext.
        iApply ("IH" with "[$Hcond $Htoks] Hhists Htotal Hpost"). }
    { simpl.
      iSplit; first auto; iFrame.
      iExists nil; simpl; iFrame; auto. }
    iNext; iIntros "H"; iDestruct "H" as (i) "(% & % & _ & H)".
    iDestruct "H" as (lr) "(% & Hhists & Htotal)".
    wp_seq.
    iMod (inv_open with "Hinv") as "[Href Hclose]"; first auto.
    iApply (na_read with "[$Hctx $Htotal]"); first auto.
    { apply namespaces.ndisj_subseteq_difference; last auto.
      apply namespaces.ndot_ne_disjoint; done. }
    iNext; iIntros (?) "[% Htotal]"; subst.
    iDestruct (monPred_objectively_elim with "Href") as (HT) "[HH Href]".
    iDestruct "Href" as (hr) "[% Href]".
    iDestruct "Href" as (mr) "[% Href]".
    assert (i = 3) by omega; subst.
    destruct lr as [|(((?, ?), ?), ?)]; [discriminate|].
    destruct lr as [|(((?, ?), ?), ?)]; [discriminate|].
    destruct lr as [|(((?, ?), ?), ?)]; [discriminate|].
    destruct lr; [simpl | discriminate].
    iDestruct "Hhists" as "([% Hhist1] & [% Hhist2] & [% Hhist3] & _)".
    iCombine "Hhist Hhist1 Hhist2 Hhist3" as "Hhist".
    replace (_ ⋅ _) with 1%Qp.
    iPoseProof (own_valid_2 with "Href Hhist") as "%".
    iMod ("Hclose" with "[HH Href]").
    { iNext; iApply objective_objectively.
      iExists HT; iFrame; auto. }
    iModIntro; iApply "Hpost".
    iPureIntro.
    match goal with H : valid _ |- _ => apply frac_auth_agreeL in H as Hmr end.
    rewrite Nat.add_0_r -!app_length -!filter_app; simpl.
    do 2 apply f_equal.
    apply Z2Nat.inj; try omega.
    rewrite Nat2Z.id; eapply add_three; eauto.
    rewrite Hmr; done.
    { apply leibnizC_leibniz.
      rewrite (comm _ (1 / _)%Qp) -!assoc Pos.mul_1_r -!Qc_div_div.
      unfold op, cmra_op; simpl; unfold frac_op.
      rewrite !Qp_div_2; done. }
  Qed.

End proof.

End hashtable.
