From igps.gps Require Export shared.
From igps.examples Require Import snapshot.

Canonical Structure histProtocol : protocolT := ProtocolT _ (map_subseteq(M := gmap nat)(A := Z)).

Instance histPrt_facts : protocol_facts histProtocol.
Proof. esplit; try apply _.
  - intro. apply map_included_preorder, _.
  - intros ???. apply map_included_preorder, _.
Qed.

Section gmaps.

Context {K} {E : EqDecision K} {C : Countable K} {A : ofeT} {L : LeibnizEquiv A}.

(* The existing version of this requires A to be a cmra, which is completely unnecessary. *)
Global Instance gmap_unit : Unit (gmap K A) := (∅ : gmap K A).

Definition compatible (m1 m2 : gmap K A) :=
  map_relation eq (fun _ => True) (fun _ => True) m1 m2.

Lemma compatible_union (m1 m2 : gmap K A) :
  compatible m1 m2 -> forall i x, (m1 ∪ m2) !! i = Some x ↔
    m1 !! i = Some x ∨ (m1 !! i = None ∧ m2 !! i = Some x).
Proof.
  intros; unfold union, map_union, union_with, map_union_with. rewrite (lookup_merge _).
  destruct (m1 !! i), (m2 !! i); compute; intuition congruence.
Qed.

Lemma compatible_union_assoc (m1 m2 m3 : gmap K A) :
  compatible m1 m2 -> compatible (m1 ∪ m2) m3 -> compatible m1 (m2 ∪ m3).
Proof.
  repeat intro.
  specialize (H i); specialize (H0 i).
  unfold union, map_union in *; rewrite !lookup_union_with in H H0 |- *.
  destruct (m1 !! i), (m2 !! i), (m3 !! i); done.
Qed.

Lemma compatible_comm1 (m1 m2 : gmap K A) : compatible m1 m2 -> compatible m2 m1.
Proof.
  repeat intro.
  specialize (H i).
  destruct (m1 !! i), (m2 !! i); auto.
  inversion H; subst; auto.
Qed.

Global Instance compatible_comm : Comm iff compatible.
Proof.
  split; apply compatible_comm1.
Qed.

Global Instance compatible_refl : Reflexive compatible.
Proof.
  repeat intro.
  destruct (x !! i); simpl; auto.
Qed.

Lemma compatible_subset : forall m1 m2 m (Hcompat : compatible m2 m) (Hsub : m1 ⊆ m2),
  compatible m1 m.
Proof.
  repeat intro.
  specialize (Hsub i); specialize (Hcompat i).
  destruct (m1 !! i), (m2 !! i), (m !! i); auto; simpl in *; subst; done.
Qed.

Corollary subset_compatible : forall m1 m2 (Hsub : m1 ⊆ m2), compatible m1 m2.
Proof.
  intros; eapply compatible_subset; eauto; reflexivity.
Qed.

Global Instance gmapC_union : Union (gmapC K A) := map_union.

Lemma compatible_union_comm : forall m1 m2 (Hcompat : compatible m1 m2),
  m1 ∪ m2 = m2 ∪ m1.
Proof.
  intros; apply map_leibniz; intro.
  specialize (Hcompat i); unfold union, gmapC_union, map_union, union_with, map_union_with.
  rewrite !lookup_merge.
  destruct (m1 !! i), (m2 !! i); auto; inversion Hcompat; auto.
Qed.

Lemma map_subseteq_equiv (m1 m2 : gmap K A) : m1 ⊆ m2 -> m2 ⊆ m1 -> map_equiv m1 m2.
Proof.
  repeat intro.
  destruct (m1 !! i) eqn: Hm1.
  - setoid_rewrite map_subseteq_spec in H; rewrite (H _ _ Hm1); auto.
  - destruct (m2 !! i) eqn: Hm2; auto.
    setoid_rewrite map_subseteq_spec in H0; rewrite (H0 _ _ Hm2) in Hm1; discriminate.
Qed.

Global Instance map_subset_order : PartialOrder (map_subseteq(M := gmap K)(A := A)).
Proof.
  constructor; first apply _.
  repeat intro.
  apply gmapC_leibniz; first apply _.
  apply map_subseteq_equiv; auto.
Qed.

End gmaps.

Instance hist_join (A : ofeT) {D : OfeDiscrete A} {L : LeibnizEquiv A} {U : Unit A} `{EqDecision A} : Join
  (fun (a b : gmapC nat A) => map_subseteq(M := gmap nat)(A := A) a b) :=
  { join (m1 m2 : gmapC nat A) := if decide (compatible m1 m2)
      then Some (union m1 m2) else None }.
Proof.
  - repeat intro.
    rewrite lookup_empty; destruct (a !! i) eqn: Ha; rewrite Ha; done.
  - intros ?????; destruct (decide _); last discriminate.
    destruct (decide _); last discriminate.
    intros Ha Hb; inversion Ha; inversion Hb; subst.
    exists (a ∪ d).
    destruct (decide _).
    split; auto.
    assert (compatible b a) by (rewrite comm; done).
    destruct (decide _).
    rewrite fin_maps.Assoc_instance_0 (compatible_union_comm b a); auto.
    { contradiction n; apply compatible_union_assoc; auto.
      rewrite compatible_union_comm; auto. }
    { contradiction n.
      eapply compatible_subset; eauto.
      apply (map_union_subseteq_l a b). }
  - repeat intro.
    destruct (decide _), (decide _); auto; first (rewrite compatible_union_comm; auto);
      contradiction n; rewrite compatible_comm; auto.
  - intros.
    destruct (decide _); inversion H; subst.
    apply (map_union_subseteq_l a b).
  - intros.
    destruct (decide _); [|contradiction n; apply subset_compatible; auto].
    apply f_equal, (map_subseteq_union a b); auto.
Qed.
