From iris.proofmode Require Import tactics.
From iris.base_logic Require Import iprop own.
From iris.algebra Require Import excl.

From igps Require Export malloc repeat notation wp_tactics_vProp escrows.
From igps Require Import persistor viewpred proofmode weakestpre protocols.
From igps Require Import fork.
From igps.gps Require Import plain.
From igps.examples Require Import unit_token.

Definition message_passing : base.expr :=
  let: "x" := Alloc in
  let: "y" := Alloc in
    ["x"]_na <- #0 ;;
    ["y"]_na <- #0 ;;
    Fork (["x"]_na <- #37 ;; ["y"]_at <- #1) ;;
    repeat (["y"]_at) ;;
    ["x"]_na.

Section Interpretation.
  Context `{foundationG Σ} `{!uTokG Σ}.

  Set Default Proof Using "Type".

  Definition XE (x : loc) (γ: gname) : vProp Σ
    := [es Tok γ ⇝ x ↦ 37 ]%I.

  Definition YP x (γ: gname) (s: pr_state boolProtocol) (v: Z) : vProp Σ :=
    (match s with
    | false => ⌜v = 0%nat⌝
    | true => ⌜v = 1%nat⌝ ∗ XE x γ
    end)%I.

  Global Instance YP_persistent x γ s v: Persistent (YP x γ s v).
  Proof. rewrite /Persistent. destruct s; by iIntros "#YP". Qed.

  Definition mpInt x γ : interpC Σ boolProtocol
    := (λ b _ s v, if b then True else YP x γ s v)%I.

  Global Instance interp_persistent : Persistent (mpInt x γ b l s v V).
  Proof. destruct b; intros; apply _. Qed.
End Interpretation.

Section proof.
  Context `{fG : foundationG Σ}
          `{mG : uTokG Σ}
          `{persG : persistorG Σ}
          `{agreeG : !gps_agreeG Σ boolProtocol}
          `{gG: !gpsG Σ boolProtocol _}.

  Set Default Proof Using "Type*".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (vProp Σ).

  Lemma message_passing_spec:
    {{{⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ }}}
        (message_passing)
    {{{ (v: Z), RET #v; ⌜v = 37%nat⌝ }}}.
  Proof.
    intros. iIntros "[#kI #kIp] Post". rewrite /message_passing.
    wp_bind Alloc.
    iApply (alloc with "kI"); [done|].
    iNext. iIntros (x) "Nax".
    wp_seq.
    wp_bind Alloc.
    iApply (alloc with "kI"); [done|].
    iNext. iIntros (y) "Nay".
    wp_seq.
    wp_bind ([_]_na <- _)%E.

    iApply (na_write with "[$kI $Nax]"); [done|].
    iNext. iIntros "Nax".
    wp_seq.
    wp_bind ([_]_na <- _)%E.

    iMod (own_alloc (Excl ())) as (γ) "Tok"; first done.

    iApply (GPS_PP_Init_default (mpInt x γ) false with "[kI $kIp] [$Nay]");
    [done|done|done|..].
    { by iSplitL "". }

    iNext. iIntros "#yPRT".
    wp_seq.

    wp_bind (Fork _)%E.
    iApply (wp_mask_mono (↑physN)); first done.
    iCombine "Nax" "yPRT" as "SndT".
    iApply (wp_fork with "[$kI SndT Post Tok]"); [done|].
    iSplitL "Post Tok".
    - iNext.
      wp_seq. wp_bind ((rec: "f" <> := _ ) _)%E.
      (iLöb as "IH" forall "yPRT ∗").
      iApply wp_repeat; first auto.
      iApply (GPS_PP_Read (p:=()) (mpInt x γ) True%I (YP x γ)
        with "kI [$yPRT]"); [done|done|..].
      + iSplitL.
        * iNext. iIntros (?). iIntros "_".
          iLeft. iIntros (?) "[YP  _]". by simpl.
        * rewrite monPred_objectively_unfold /=. iIntros (?????) "!> [_ #$] //".
      + iNext. iIntros (s v) "(_ & #yPRT5 & #YP)".
        destruct s.
        * iDestruct "YP" as "(% & XE)".
          iExists (1). iSplit; first by rewrite H.
          iDestruct (escrow_apply (Tok γ) (x ↦ 37))%I as "EA".
          iSpecialize ("EA" with "[]").
          { iIntros "?". by iApply unit_tok_exclusive. }
          iSpecialize ("EA" with "[$XE $Tok]").
          iNext. iNext.
          iApply fupd_wp. iMod (fupd_mask_mono with "EA") as "Nax"; [done|].
          iModIntro.
          wp_seq.
          iApply (na_read with "[$kI $Nax]"); [done|].
          iNext. iIntros (z) "(% & Nax)".
          subst z. by iApply ("Post" with "[%]").
        * iDestruct "YP" as "%".
          iExists 0. iSplit; first by rewrite H.
          erewrite bool_decide_true; last done.
          iNext. iNext.
          iApply ("IH" with "[yPRT5] Post Tok").
          iIntros "!#"; iFrame "yPRT5".
    - iNext. iDestruct "SndT" as "(Nax & #yPRT')".
      wp_bind ([_]_na <- _)%E.
      iApply (na_write with "[$kI $Nax]"); [done|].
      iNext. iIntros "Nax".
      wp_seq.

      (* allocate escrow *)
      iDestruct (escrow_alloc (Tok γ) (x ↦ 37)%I with "[$Nax]") as "XE".

      iMod (fupd_mask_mono _ ⊤ with "XE") as "XE"; first auto.

      (* write release to final state *)
      iApply (GPS_PP_Write (p:=()) (mpInt x γ) true (XE x γ) (λ _, True%I)
              with "[$kI yPRT' XE]"); [done|done|move=>[|]; done|..].
      { iSplitL ""; [|by iFrame "XE"].
        iNext. iIntros "[oL _]". iModIntro.
        iSplitL ""; [done|]. iSplit; iNext; auto. by iFrame "oL". }

      (* trivial post cond *)
      iModIntro. auto.
  Qed.

End proof.
