From iris.proofmode Require Import tactics.
From iris.algebra Require Import auth frac excl gmap agree.
From iris.bi Require Import big_op fractional.
From iris.base_logic.lib Require Import cancelable_invariants.
From stdpp Require Export namespaces.

Import uPred.

From igps Require Import lang.
From igps.base Require Import accessors.

Definition fracN : namespace := nroot .@ "fractor".

Section Fractor.
  Context `{!foundationG Σ, !cinvG Σ}.

  Implicit Types (l: loc) (φ: loc -c> InfoT -c> iProp Σ)
                          (Ψ: loc -c> frac -c> InfoT -c> iProp Σ).

  Definition fractored l φ Ψ p: iProp Σ :=
    (∃ X, Ψ l p X
      ∗ ∃ γ, Info l p (to_agree (encode (X,γ)))
      ∗ cinv_own γ p ∗ cinv (fracN .@ l) γ (φ l X))%I.

  Global Instance fractored_proper l : Proper ((≡) ==> (≡) ==> (=) ==> (≡)) (fractored l).
  Proof.
    intros ? ? H1 ? ? H2 ??->. rewrite /fractored. repeat f_equiv. exact: H2. exact: H1.
  Qed.
  Global Instance fractored_ne l n :
    Proper ((dist n) ==> (dist n) ==> (=) ==> (dist n)) (fractored l).
  Proof.
    intros ?? H1 ? ? H2 ??->. rewrite /fractored. repeat f_equiv. exact: H2. exact: H1.
  Qed.
  Global Instance fractored_contractive l n :
    Proper (dist_later n ==> (=) ==> (=) ==> dist n) (fractored l).
  Proof.
    intros ?? H1 ??-> ??->. rewrite /fractored. repeat (f_contractive || f_equiv).
    exact: H1.
  Qed.
  Global Instance fractored_mono l φ :
    Proper (pointwise_relation _ (pointwise_relation _ (pointwise_relation _ (⊢)))
                                    ==> (=) ==> (⊢))
           (fractored l φ).
  Proof. unfold fractored, pointwise_relation=>?????->. repeat f_equiv. auto. Qed.
  Global Instance fractored_mono_flip l φ :
    Proper (flip (pointwise_relation _ (pointwise_relation _ (pointwise_relation _ (⊢))))
                                    ==> (=) ==> flip (⊢))
           (fractored l φ).
  Proof. solve_proper. Qed.

  Lemma fractor_mono l φ Ψ Ψ' p:
      (∀ X, Ψ l p X → Ψ' l p X) ∗ fractored l φ Ψ p ⊢ fractored l φ Ψ' p.
  Proof.
    iIntros "[Imp Frac]".
    iDestruct "Frac" as (X) "[Pred Inv]".
    iExists X. iFrame "Inv".
    by iApply "Imp".
  Qed.

  Lemma fractor_drop l φ Ψ p:
      fractored l φ Ψ p ⊢ ∃ X, Ψ l p X.
  Proof.
    iIntros "Frac". iDestruct "Frac" as (X) "[Frac _]".
    by iExists X.
  Qed.

  Lemma Info_join l p1 p2 v1 v2:
    Info l p1 (to_agree v1) ∗ Info l p2 (to_agree v2)
    ⊢ ⌜v1 = v2⌝ ∗ Info l (p1 + p2) (to_agree v1).
  Proof.
    iIntros "[Info1 Info2]".
    iCombine "Info1" "Info2" as "Info".
    iDestruct (own_valid with "Info") as %Valid.
    apply singleton_valid in Valid.
    destruct Valid as [_ Valid2]. simpl in Valid2.
    apply agree_op_inv, (inj to_agree) in Valid2. setoid_subst.
    iSplitR "Info"; auto. by rewrite agree_idemp.
  Qed.

  Lemma fractor_splitjoin l φ Ψ1 Ψ2 p1 p2:
      fractored l φ Ψ1 p1 ∗ fractored l φ Ψ2 p2
      ⊣⊢ fractored l φ (λ l p X, Ψ1 l p1 X ∗ Ψ2 l p2 X) (p1 + p2).
  Proof.
    iSplit.
    - iIntros "[Frac1 Frac2]".
      iDestruct "Frac1" as (X1) "[Frac1 Inv1]".
      iDestruct "Inv1" as (γ1) "[Info1 [FOwn1 #FCtx1]]".
      iDestruct "Frac2" as (X2) "[Frac2 Inv2]".
      iDestruct "Inv2" as (γ2) "[Info2 [FOwn2 #FCtx2]]".
      iPoseProof (Info_join with "[Info1 Info2]") as "Info".
        { iSplitL "Info1"; auto. }
      iDestruct "Info" as "[Eq Info]". iDestruct "Eq" as %Eq%encode_inj.
      inversion Eq. subst.
      iExists X2. iFrame. iExists γ2.
      iCombine "FOwn1" "FOwn2" as "FOwn".
      rewrite fractional. by iFrame.
    - iIntros "Frac".
      iDestruct "Frac" as (X) "[[Pred1 Pred2] Frac]".
      iDestruct "Frac" as (γ) "[Info [FOwn #FCtx]]".
      rewrite fractional (_:Info l (p1+p2) _ ≡ (Info l p1 _ ∗ Info l p2 _)%I);
        last first.
      { rewrite /Info /Infos -own_op -auth_frag_op op_singleton pair_op.
        do 2 f_equiv. apply: fin_maps.singleton_proper.
        by rewrite -[to_agree _]agree_idemp. }
      iDestruct "Info" as "[Info1 Info2]".
      iDestruct "FOwn" as "[FOwn1 FOwn2]".
      iSplitL "Pred1 Info1 FOwn1".
      + iExists X. iFrame. iExists γ. iFrame. auto.
      + iExists X. iFrame. iExists γ. iFrame. auto.
  Qed.

  Lemma fractor_join_l l φ1 φ2 Ψ1 Ψ2 p1 p2:
      fractored l φ1 Ψ1 p1 ∗ fractored l φ2 Ψ2 p2
      ⊢ fractored l φ1 (λ l p X, Ψ1 l p1 X ∗ Ψ2 l p2 X) (p1 + p2).
  Proof.
    iIntros "[Frac1 Frac2]".
    iDestruct "Frac1" as (X1) "[Frac1 Inv1]".
    iDestruct "Inv1" as (γ1) "[Info1 [FOwn1 #FCtx1]]".
    iDestruct "Frac2" as (X2) "[Frac2 Inv2]".
    iDestruct "Inv2" as (γ2) "[Info2 [FOwn2 #FCtx2]]".
    iPoseProof (Info_join with "[Info1 Info2]") as "Info".
        { iSplitL "Info1"; auto. }
      iDestruct "Info" as "[Eq Info]". iDestruct "Eq" as %Eq%encode_inj.
      inversion Eq. subst.
    iCombine "FOwn1" "FOwn2" as "FOwn". rewrite fractional.
    iExists X2. iFrame. iExists γ2. rewrite fractional. by iFrame.
  Qed.

  Lemma fractor_join_r l φ1 φ2 Ψ1 Ψ2 p1 p2:
    fractored l φ1 Ψ1 p1 ∗ fractored l φ2 Ψ2 p2
    ⊢ fractored l φ2 (λ l p X, Ψ1 l p1 X ∗ Ψ2 l p2 X) (p1 + p2).
  Proof.
    iIntros "[Frac1 Frac2]".
    iDestruct "Frac1" as (X1) "[Frac1 Inv1]".
    iDestruct "Inv1" as (γ1) "[Info1 [FOwn1 #FCtx1]]".
    iDestruct "Frac2" as (X2) "[Frac2 Inv2]".
    iDestruct "Inv2" as (γ2) "[Info2 [FOwn2 #FCtx2]]".
    iPoseProof (Info_join with "[Info1 Info2]") as "Info".
        { iSplitL "Info1"; auto. }
      iDestruct "Info" as "[Eq Info]". iDestruct "Eq" as %Eq%encode_inj.
      inversion Eq. subst.
    iCombine "FOwn1" "FOwn2" as "FOwn". rewrite fractional.
    iExists X2. iFrame "Frac1 Frac2". iExists γ2. rewrite fractional. by iFrame.
  Qed.


  Lemma fractor_join_later_rl l φ1 φ2 Ψ1 Ψ2 p1 p2 E:
      fractored l φ1 Ψ1 p1 ∗ ▷ fractored l φ2 Ψ2 p2
      ={E}=∗ fractored l φ1 (λ l p X, Ψ1 l p1 X ∗ ▷ Ψ2 l p2 X) (p1 + p2).
    Proof.
      iIntros "[Frac1 Frac2]".
      iDestruct "Frac1" as (X1) "[Frac1 Inv1]".
      iDestruct "Inv1" as (γ1) "[Info1 [FOwn1 #FCtx1]]".
      iDestruct "Frac2" as (X2) "[Frac2 Inv2]".
      iDestruct "Inv2" as (γ2) "[>Info2 [>FOwn2 #FCtx2]]".
      iPoseProof (Info_join with "[Info1 Info2]") as "Info".
        { iSplitL "Info1"; auto. }
      iDestruct "Info" as "[Eq Info]". iDestruct "Eq" as %Eq%encode_inj.
      inversion Eq. subst.
      iCombine "FOwn1" "FOwn2" as "FOwn". rewrite fractional.
      iModIntro. iExists X2. iFrame "Frac1 Frac2". iExists γ2.
      rewrite fractional. by iFrame.
    Qed.

  Lemma fractor_alloc (E : coPset) v X l φ Ψ:
    ↑physN ⊆ E →
    PSCtx ∗ Info l 1 v ∗ ▷ φ l X ∗ Ψ l 1%Qp X
      ={E}=∗ fractored l φ Ψ 1.
  Proof.
    intros Sub.
    iIntros "[#PS [Info [Inv Pred]]]".
    iMod ((cinv_alloc E (fracN .@ l) (φ l X)) with "Inv") as (γ) "[FOwn Ctx]".
    iAssert (|={E}=> Info l 1 (to_agree (encode (X,γ))))%I
      with "[Info]" as ">Info".
      { iApply PSInv_Info_update => //; auto. }
    iModIntro.
    iExists X. iFrame "Pred". iExists γ. by iFrame.
  Qed.

  Lemma fractor_dealloc (E : coPset) l φ Ψ:
    ↑fracN .@ l ⊆ E →
    fractored l φ Ψ 1 ={E}=∗ ∃ X v, Info l 1 v ∗ Ψ l 1%Qp X ∗ ▷ φ l X.
  Proof.
    intros Sub.
    iIntros "Frac". iDestruct "Frac" as (X) "[Pred Frac]".
    iDestruct "Frac" as (γ) "[Info [FOwn #Ctx]]".
    iMod ((cinv_cancel _ _ _ _ Sub) with "Ctx FOwn") as "Inv".
    iModIntro.
    iExists X, (to_agree (encode (X,γ))). by iFrame.
  Qed.

  Lemma fractor_open' E l φ Ψ p:
    ↑ fracN .@ l ⊆ E →
      fractored l φ Ψ p ={E,E∖↑fracN .@ l}=∗
      ∃ X, ▷ φ l X ∗ Ψ l p X
             ∗ (∀ Ψ', Ψ' l p X -∗ fractored l φ Ψ' p)
             ∗ (▷ φ l X ={E∖↑fracN .@ l, E}=∗ True).
  Proof.
    intros Sub.
    iIntros "Frac". iDestruct "Frac" as (X) "[Pred Frac]".
    iDestruct "Frac" as (γ) "[Info [FOwn #Ctx]]".
    iMod ((cinv_open _ _ _ _ _ Sub) with "Ctx FOwn") as "[Inv [FOwn Hclose]]".
    iModIntro. iExists X. iFrame.
    iIntros (Ψ') "Pred".
    iExists X. iFrame. iExists γ. by iFrame.
  Qed.

  Lemma fractor_open E l φ Ψ p:
    ↑ fracN .@ l ⊆ E →
      fractored l φ Ψ p ={E,E∖↑fracN .@ l}=∗
      ∃ X, ▷ φ l X ∗ Ψ l p X
           ∗ ∀ Ψ',  ▷ φ l X ∗ Ψ' l p X ={E∖↑fracN .@ l,E}=∗ fractored l φ Ψ' p.
  Proof.
    intros Sub.
    iIntros "Frac".
    iMod (fractor_open' with "Frac") as (X) "(Inv & Pred & HT & HClose)"; [done|].
    iModIntro. iExists X. iFrame. iIntros (Ψ') "[Inv Pred]".
    iMod ("HClose" with "Inv"). by iApply "HT".
  Qed.
End Fractor.
