From iris.proofmode Require Import tactics.
From iris.base_logic Require Import iprop.
From iris.program_logic Require Import ownp.
From igps Require Import weakestpre. (* TODO: why ??? *)
From igps.gps Require Export inst_shared.
Import base.

Section Fractional.
  Context `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{persG : persistorG Σ}
          `{agreeG : gps_agreeG Σ Prtcl}
          `{CInvG : cancelable_invariants.cinvG Σ}
          {IP : interpC Σ Prtcl}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (@vProp Σ).

  Notation F := (IP true).
  Notation F_past := (IP false).

  Implicit Types
           (l : loc)
           (s : pr_state Prtcl)
           (ζ : gset (state_sort Prtcl))
           (S : pr_state Prtcl * (Z * View))
           (v : Z)
           (V : View)
           (h : History).

  Notation RA γ s := (∃ V_0 v, monPred_in V_0 ∧ ⎡Reader γ {[s, (v, V_0)]}⎤)%I.

  Implicit Types (e : state_sort Prtcl).

  Local Open Scope I.
  Definition gpsFRaw l n : iProp :=
    ∃ (γ γ_x: gname) ζ,
      ⌜n = encode(γ,γ_x)⌝ ∗ gps_inv IP l γ γ_x ∗ Writer γ ζ.

  Definition RFRaw s V l (q: frac) n : iProp :=
    ∃ (γ γ_x: gname) e_x,
      ⌜n = encode(γ,γ_x)⌝ ∗ exwr γ_x q e_x
                 ∗ ∃ Vr v, monPred_at (monPred_in Vr) V
                                 ∗ ⌜(st_view e_x !! l : option positive) ⊑ Vr !! l⌝ ∗ Reader γ {[s, (v, Vr)]}.

  Program Definition vGPS_FP_def l s (q: frac) : vProp Σ :=
    MonPred (λ V, fractored l (gpsFRaw) (RFRaw s V) q) _.
  Next Obligation.
    intros l s q V V' Mono.
    iIntros "RP". iDestruct "RP" as (X) "(RPV & ?)".
    iDestruct "RPV" as (γ γ_x e_x) "(? & ? & RA)".
    iDestruct "RA" as (Vr v) "(? & ? & ?)".
    iExists X. iFrame. iExists γ, γ_x, e_x. iFrame.
    iExists Vr, v. by iFrame.
  Qed.
  Definition vGPS_FP_aux : seal vGPS_FP_def. by eexists. Qed.
  Definition vGPS_FP := unseal vGPS_FP_aux.
  Definition vGPS_FP_eq : vGPS_FP = _ := seal_eq _.

  Local Close Scope I.

  Lemma GPS_FPs_agree l s1 s2 q1 q2 (E : coPset) (HE: ↑fracN .@ l ⊆ E):
    (vGPS_FP l s1 q1 ∗ vGPS_FP l s2 q2
     ={E}=∗  ⌜s1 ⊑ s2 ∨ s2 ⊑ s1⌝ ∗ vGPS_FP l s1 q1 ∗ vGPS_FP l s2 q2)%I.
  Proof.
    iStartProof (uPred _).
    rewrite vGPS_FP_eq /vGPS_FP_def /=.
    iIntros (?) "RP".
    iDestruct (fractor_splitjoin l with "RP") as "H".
    iMod (fractor_open with "H") as (X) "(oI & (oR1 & oR2) & HClose)";
    first auto.
    iDestruct "oR1" as (γ1 γ_x1 e_x1) "(% & oe1 & #RA1)".
    iDestruct "oR2" as (γ2 γ_x2  e_x2) "(% & oe2 & #RA2)".
    iDestruct "oI" as (γ γ_x ζ) "(>% & oI & oI2)". subst.
    apply encode_inj in H0. inversion H0. subst.
    apply encode_inj in H1. inversion H1. subst.
    iDestruct "RA1" as (V1 v1) "(? & ? & oR1)".
    iDestruct "RA2" as (V2 v2) "(? & ? & oR2)".
    iMod (Reader_singleton_agree _ _ _ _ s1 _ _ s2 with "[$oI $oR1 $oR2]") as "[oI %]".
    iMod ("HClose" $! (λ l0 _ X, RFRaw s1 _ l0 q1 X ∗ RFRaw s2 _ l0 q2 X)%I
          with "[oI oI2 oe1 oe2]") as "oR".
    - iSplitL "oI oI2"; last iSplitL "oe1"; iExists γ, γ_x.
      + iExists ζ. iFrame "oI oI2". by iNext.
      + iExists e_x1. iFrame "oe1". iSplit; first done.
        iExists V1, v1. by repeat iSplit.
      + iExists e_x2. iFrame "oe2". iSplit; first done.
        iExists V2, v2. by repeat iSplit.
    - iDestruct (fractor_splitjoin l with "oR") as "[o1 o2]".
        by iFrame "o1 o2".
  Qed.

  Lemma GPS_FP_Init l s v (E : coPset)
        (HEN : ↑physN ⊆ E):
      {{{ ⎡PSCtx⎤ ∗ own_loc l ∗ ▷ F l s v ∗ ▷ F_past l s v }}}
      ([ #(LitLoc l) ]_na <- #(LitInt v)) @ E
      {{{ RET #() ; vGPS_FP l s 1%Qp }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V); iIntros "(#kI & oL & F & Fp)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq /=; iIntros (V3 ??) "kS".
    unfold own_loc, own_loc_prim.
    iDestruct "oL" as (h i) "(Alloc & Hist & Info)".
    rewrite valloc_local_eq /valloc_local_def /=.
    iDestruct "Alloc" as %Alloc.
    iApply program_logic.weakestpre.wp_fupd.
    assert (V ⊑ V3) by (etransitivity; eauto).
    iApply (RawProto_init IP l s v _ h
               with "[$kI $kS $Hist $F $Fp]"); first auto.
    { iPureIntro; eapply alloc_local_proper; eauto. }

    iNext.
    iIntros (V' γ γ_x) "(% & kS' & oI & ex & #oR & oW)"; iFrame.

    iMod (fractor_alloc E i (encode(γ,γ_x)) l (gpsFRaw) (RFRaw s V')
          with "[-Post]") as "FP"; first auto.
    { iFrame "kI Info". iSplitL "oI oW"; iExists γ, γ_x.
      - iExists {[s, (v, V')]}. iFrame "oI oW". by iNext.
      - iExists (s, (v, V')). iFrame "ex". iSplitL ""; first done.
        iExists V', v. iFrame "oR". rewrite monPred_in_eq /=. by iSplit. }
    iModIntro; iApply ("Post" $! _ with "[%]").
    { etrans; eauto. }
    rewrite vGPS_FP_eq /=; done.
  Qed.

  Lemma GPS_FP_Read' l q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
        s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
      {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
                  (∀ v, F_past l s' v ∗ P ∗ vGPS_FP l s' q ={E}=∗ Q s' v)
                  ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                          ={E ∖ ↑fracN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ vGPS_FP l s q  }}}
      ([ #(LitLoc l) ]_at) @ E
      {{{ s' v, RET #(LitInt v) ; ⌜s ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V); iIntros "(#kI & VS & VSDup & oP & oR)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_FP_eq /=; iIntros (V3 ??) "kS".
    iMod (fractor_open with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.
    iDestruct "RRaw" as (γ γ_x e_x) "(% & Hex & RA)".
    rewrite monPred_in_eq /=.
    iDestruct "gpsRaw" as (γ' γ_x' ζ)
                          "(>% & oI & Writer)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.

    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    assert (V ⊑ V3) by (etransitivity; eauto).
    iApply (RawProto_Read IP l P (fun s v => |={E ∖ ↑fracN .@ l, E}=> Q s v)%I s _ _ V_r v_r γ γ_x
            with "[VS $oI $kI $kS $oP $oR VSDup Writer Hex HClose]"); [solve_ndisj|auto|..].
    { etransitivity; eauto. }
    { iSplitR "VSDup"; iNext; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ? Hs').
      iDestruct ("VS" $! _ Hs') as "[VS|VS]"; last first.
      { iRight. iIntros (?????). by iApply ("VS" $! _ _ with "[%]"). }
      iLeft; iIntros (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (RFRaw s' j0) with "[Writer oI oR' Hex]").
      { iSplitL "oI Writer".
        + iNext. iExists γ, γ_x, ζ. iFrame; auto.
        + iExists γ, γ_x, e_x. iFrame "Hex". iSplitL ""; first auto.
          iExists Vr, v. rewrite monPred_in_eq /=.
          iSplit; last iFrame "oR'"; iPureIntro; auto.
          do 4 (etrans; eauto). }
      iMod ("VS" $! _ j0 with "[% ] [$]") as "$".
      { repeat (etrans; eauto). }
      iApply fupd_intro_mask'. solve_ndisj. }
    iNext.
    iIntros (s' v Vr V') "(% & kS' & #Hs' & oQ & #oR' & % & %)".
    iFrame.
    iApply ("Post" $! _ _ _ with "[%]").
    { etrans; eauto. }
    by iFrame "Hs'".
  Qed.

  Lemma GPS_FP_Read l q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
        s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
    ⎡PSCtx⎤ ⊢ {{{ (▷ ∀ s', ⌜s ⊑ s'⌝ →
                  (∀ v, F_past l s' v ∗ P ={E ∖ ↑fracN .@ l}=∗ Q s' v)
                  ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                          ={E ∖ ↑fracN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ vGPS_FP l s q  }}}
      ([ #(LitLoc l) ]_at) @ E
      {{{ s' v, RET #(LitInt v) ;
           ⌜s ⊑ s'⌝ ∗ vGPS_FP l s' q ∗ Q s' v }}}.
  Proof.
    iIntros "#kI !#" (?) "(VS & VSDup & oP & oR) Post".
    iApply (GPS_FP_Read' _ _ P with "[$kI VS $VSDup $oP $oR]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_FP_Read_abs'
      l q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
      s (E : coPset) V (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
      {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
                  (∀ v, F_past l s' v ∗ P ∗ ⎡vGPS_FP l s q V⎤ ={E}=∗ Q s' v)
                  ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                          ={E ∖ ↑fracN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ ⎡vGPS_FP l s q V⎤ ∗ monPred_in V }}}
      ([ #(LitLoc l) ]_at) @ E
      {{{ s' v, RET #(LitInt v) ;
           ⌜s ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V') "(#kI & VS & VSDup & oP & oR & %)"; iIntros (Vb ?) "Post".
    rewrite WP_Def.wp_eq vGPS_FP_eq /=; iIntros (Vc ??) "kS".
    iMod (fractor_open with "[$oR]") as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.

    iDestruct "RRaw" as (γ γ_x e_x) "(% & Hex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' ζ) "(>% & oI & Writer)".
    subst X. apply encode_inj in H1. inversion H1. subst γ' γ_x'.

    rewrite monPred_in_eq /=.
     iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    iApply (RawProto_Read IP l P (fun s v => |={E ∖ ↑fracN .@ l, E}=> Q s v)%I _ _ _ V_r v_r γ γ_x
            with "[VS $oI $kI $kS $oP $oR VSDup Writer Hex HClose]");
      [solve_ndisj|solve_jsl|..].
    { iSplitR "VSDup"; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ? Hs').
      iDestruct ("VS" $! _ with "[%//]") as "[VS|VS]"; last by auto.
      iLeft; iIntros "!>" (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (RFRaw s V) with "[Writer oI oR' Hex]").
      { iSplitL "oI Writer".
        + iNext. iExists γ, γ_x, ζ. iFrame; auto.
        + iExists γ, γ_x, e_x. iFrame "Hex". iSplitL ""; first auto.
          iExists V_r, v_r. rewrite monPred_in_eq /=; auto. }
      iMod ("VS" $! _ j0 with "[%] [$]") as "$"; [solve_jsl|].
      iApply fupd_intro_mask'. solve_ndisj. }
    iNext.
    iIntros (s' v Vr V'') "(% & kS' & #Hs' & oQ & #oR' & % & %)".
    iFrame.
    iApply ("Post" $! _ _ _ with "[%]"); [solve_jsl|].
    by iFrame "Hs'".
  Qed.

  Lemma GPS_FP_Read_abs l q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
      s (E : coPset) V (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
    ⎡PSCtx⎤ ⊢ {{{ (▷ ∀ s', ⌜s ⊑ s'⌝ →
                  (∀ v, F_past l s' v ∗ P ={E ∖ ↑fracN .@ l}=∗ Q s' v)
                  ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                          ={E ∖ ↑fracN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ ⎡vGPS_FP l s q V⎤ ∗ monPred_in V }}}
      ([ #(LitLoc l) ]_at) @ E
      {{{ s' v, RET #(LitInt v) ;
           ⌜s ⊑ s'⌝ ∗ ⎡vGPS_FP l s q V⎤ ∗ Q s' v }}}.
  Proof.
    iIntros "#kI !#" (?) "(VS & VSDup & oP & oR & Seen) Post".
    iApply (GPS_FP_Read_abs' _ _ P with "[$kI VS $VSDup $oP $oR $Seen]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_FP_Write l q s s' v (E : coPset)
        (P: vPred) (Q : pr_state Prtcl → vPred)
        (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E) (HFinal : final_st s'):
      {{{ ⎡PSCtx⎤ ∗ ▷ (P ∗ vGPS_FP l s' q ={E ∖ ↑fracN .@ l}=∗
                                                 Q s' ∗ ▷ (F l s' v) ∗ ▷ F_past l s' v)
           ∗ P ∗ vGPS_FP l s q }}}
      ([ #(LitLoc l) ]_at <- #(LitInt v)) @ E
      {{{ RET #() ; Q s' }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V); iIntros "(#kI & VS & oP & oW)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_FP_eq /=; iIntros (V3 ??) "kS".
    iMod (fractor_open' with "[$oW]")
      as (X) "(gpsRaw & RRaw & HT & HClose)"; first auto.
    iDestruct "RRaw" as (γ γ_x e_x) "(% & Hex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' ζ) "(>% & oI & >wTok)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.

    rewrite monPred_in_eq /=.
    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".
    destruct e_x as [s_x [v_x V_x]].

    assert (V_r ⊑ V3) by (do 2 (etrans; eauto)).
    iApply (RawProto_Write_non_ex IP l
                                  s s_x s' _ _ V_r V_x q v v_r v_x γ γ_x _ ζ
            with "[$kI $kS $oI $Hex $oR $wTok]");
    [solve_ndisj|auto|auto|auto|..].
    iNext.
    iIntros (V') "(% & kS' & oI & ex & #oR' & wTok)".
    iFrame; iApply ("Post" $! _ with "[%]").
    { etrans; eauto. }
    iAssert (monPred_at (vGPS_FP l s' q) V') with "[HT ex]" as "oW".
    { rewrite vGPS_FP_eq /=.
      iApply "HT".
      iExists γ, γ_x, (s_x, (v_x, V_x)). iFrame "ex".
      iSplit; first done.
      iExists V', v. iFrame "oR'". rewrite monPred_in_eq /=. iSplit; first done.
      iPureIntro. do 4 (etrans; eauto).
    }
    assert (V ⊑ V') by (do 2 (etrans; eauto)).
    iMod ("VS" $! _ with "[%] [oP oW]") as "($ & oF & oFp)"; first done.
    { rewrite vGPS_FP_eq /=; iFrame. }
    iSpecialize ("oI" with "[$oF $oFp]").
    iMod ("HClose" with "[-]") as "$"; [|done].
    iNext. iExists γ, γ_x, ({[s', (v, V')]} ∪ ζ).
    iFrame "wTok". by iSplit.
  Qed.

  Lemma GPS_FP_FAI C l q s (E : coPset)
        (P : vPred) (Q: pr_state Prtcl → Z → vPred)
        (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E) :
    let mod1C (v: Z) := ((v + 1) `mod` Z.pos C)%Z in
      {{{ ⎡PSCtx⎤ ∗ ▷ (∀ v s', ⌜s ⊑ s'⌝ ∗ F l s' v ∗ P
                     ={E ∖ ↑fracN .@ l}=∗ ∃ s'', ⌜s' ⊑ s''⌝
                                                   ∗ Q s'' v
                                                   ∗ ▷ (F l s'' (mod1C v)) ∗ ▷ F_past l s'' (mod1C v))
            ∗ P ∗ vGPS_FP l s q }}}
      (FAI C #(LitLoc l)) @ E
      {{{ s'' (v: Z), RET #(LitInt v) ;
           ⌜s ⊑ s''⌝ ∗ vGPS_FP l s'' q ∗ Q s'' v }}}.
  Proof.
    intros. iStartProof (uPred _).
    rewrite vGPS_FP_eq WP_Def.wp_eq /=.
    iIntros (V); iIntros "(#kI & VS & oP & oW)" (V2 ?) "Post".
    iIntros (V3 ??) "kS".
    iMod ((fractor_open) with "[$oW]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.
    iDestruct "RRaw" as (γ γ_x e_x) "(% & ex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' ζ) "(>% & oI & >wTok)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.

    destruct e_x as [s_x [v_x V_x]].
    rewrite monPred_in_eq /=.
    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    assert (V ⊑ V3) by (etrans; eauto).
    iApply (RawProto_FAI_non_ex IP _ _
                                s s_x _ _ V_r V_x q%Qp v_r v_x γ γ_x _ P Q
            with "[VS $kI $kS $oI $oP $ex $oR wTok]");
    [auto|auto|solve_ndisj| |].
    { etrans; eauto. }
    { iSplitR "wTok"; last by iExists _.
      iNext.
      iIntros (?? j ?) "Hs'".
      iApply ("VS" $! _ _ _ with "[%] [-//]").
      etrans; eauto. }
    iNext.
    iIntros (s'' v Vr V')
            "(% & kS' & Hs' & oI & Q & ex & #oR' & % & % & wTok)".
    iFrame; iApply ("Post" $! _ _ _ with "[%]").
    { etrans; eauto. }
    iFrame "Hs' Q".
    iApply "HClose". iSplitL "oI wTok".
    - iNext. iDestruct "wTok" as (ζ') "wTok".
      iExists γ, γ_x, _.
      by iFrame "oI wTok".
    - iExists γ, γ_x, (s_x, (v_x, V_x)). iFrame "ex".
      repeat iSplit; first done.
      iExists Vr, (mod1C v). rewrite monPred_in_eq /=. repeat iSplit; [auto| |auto].
      iPureIntro. do 4 (etrans; eauto).
  Qed.

  Lemma GPS_FP_CAS' l q s v_o v_n (E : coPset)
        (P : vPred) (Q: pr_state Prtcl → vPred) (R : pr_state Prtcl → Z → vPred)
        (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E) :
      {{{ ⎡PSCtx⎤ ∗ ▷ (∀ s', ⌜s ⊑ s'⌝ ∗ F l s' v_o ∗ P
                  ={E ∖ ↑fracN .@ l}=∗ ∃ s'', ⌜s' ⊑ s''⌝ ∗ ▷ (F l s'' v_n) ∗ ▷ F_past l s'' v_n
                    ∗ (vGPS_FP l s'' q ={E}=∗ Q s''))
           ∗ ▷ (∀ s' v, ⌜s ⊑ s'⌝ ∗ ⌜v ≠ v_o⌝ ∗ ▷ F_past l s' v ∗ P ∗ vGPS_FP l s' q
                        ={E}=∗ R s' v)
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v ={E ∖ ↑fracN .@ l}=∗
                                                                       F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ vGPS_FP l s q }}}
      (CAS #(LitLoc l) #(LitInt v_o) #(LitInt v_n)) @ E
      {{{ s'' (b: bool) v, RET #(lit_of_bool b) ;
           ⌜s ⊑ s''⌝ ∗ (if b then Q s'' else R s'' v) }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V); iIntros "(#kI & VS & VSF & VSDup & oP & oW)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_FP_eq /=; iIntros (V3 ??) "kS".
    iMod ((fractor_open) with "[$oW]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.
    iDestruct "RRaw" as (γ γ_x e_x) "(% & ex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' ζ) "(>% & oI & >wTok)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.

    destruct e_x as [s_x [v_x V_x]].
    rewrite monPred_in_eq /=.
    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    assert (V_r ⊑ V3) by (do 2 (etrans; eauto)).
    assert (V ⊑ V3) by (do 2 (etrans; eauto)).
    iApply (RawProto_CAS_non_ex_weak IP l
              s s_x _ _ V_r V_x q v_o v_n v_r v_x γ γ_x _
              (P ∗ ⎡(∀ Ψ', ▷ gpsFRaw l (encode (γ, γ_x)) ∗ Ψ' l q (encode (γ, γ_x)) ={
                 E ∖ ↑fracN.@l,E}=∗ fractored l gpsFRaw Ψ' q)⎤)%I
              (fun s => |={E ∖ ↑fracN .@ l, E}=> Q s)%I
              (fun s v => |={E ∖ ↑fracN .@ l, E}=> R s v)%I
            with "[VS VSF $kI $kS oI oP $ex $oR wTok VSDup HClose]");
      [solve_ndisj|solve_ndisj|solve_ndisj| |].
    { iSplitL "VS"; last iSplitL "VSDup"; last iSplitL "VSF"; last (by iFrame; iExists _); iNext.
      - iIntros (s' j ?) "(Hs' & F & oP & HClose)".
        iMod ("VS" $! _ _ with "[%] [$Hs' $F $oP]") as (s'') "(? & ? & ? & Q)".
        { etrans; eauto. }
        iExists s''; iFrame.
        iModIntro; iIntros (Vr j0 Hj) "(% & % & oI & ex & Writer & oR')".
        iDestruct "Writer" as (ζ') "Writer".
        iMod ("HClose" $! (RFRaw s'' j0) with "[oI ex Writer oR']") as "oW".
        { iSplitL "oI Writer".
          + iNext. iExists γ, γ_x, ζ'. iFrame; auto.
          + iExists γ, γ_x, (s_x, (v_x, V_x)). iSplit; first auto. iFrame.
            iExists Vr, v_n. rewrite monPred_in_eq /=.
            iSplit; last iFrame "oR'"; iPureIntro; auto.
            do 4 (etrans; eauto). }
        iMod ("Q" $! j0 with "[//] [$oW]") as "$".
        iApply fupd_intro_mask'. solve_ndisj.
      - by iSpecialize ("VSDup" $! _).
      - iIntros (s' v Vr j ?) "(Hs' & Hv & Fp & (oP & HClose) & % & % & oI & ex & Writer & oR')".
        iDestruct "Writer" as (ζ') "Writer".
        iMod ("HClose" $! (RFRaw s' j) with "[oI ex Writer oR']") as "oW".
        { iSplitL "oI Writer".
          + iNext. iExists γ, γ_x, ζ'. iFrame; auto.
          + iExists γ, γ_x, (s_x, (v_x, V_x)). iFrame "ex". iSplitL ""; first auto.
            iExists Vr, v. rewrite monPred_in_eq /=.
            iSplit; last iFrame "oR'"; iPureIntro; auto.
            do 4 (etrans; eauto). }
        iMod ("VSF" $! _ _ _ with "[%] [$Hs' $Hv $Fp $oW $oP]") as "$".
        { etrans; eauto. }
        iApply fupd_intro_mask'. solve_ndisj. }

    iNext.
    iIntros (s'' b v Vr V')
            "(% & kS' & Hs' & IF & #oR' & % & %)".
    iFrame.
    iSpecialize ("Post" $! s'' b v _ with "[%]").
    { etrans; eauto. }
    iApply "Post". iFrame "Hs'". by destruct b.
  Qed.

  Lemma GPS_FP_CAS l q s v_o v_n (E : coPset)
        (P : vPred) (Q: pr_state Prtcl → vPred) (R : pr_state Prtcl → Z → vPred)
        (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E) :
    ⎡PSCtx⎤ ⊢ {{{ ▷ (∀ s', ⌜s ⊑ s'⌝ ∗ F l s' v_o ∗ P
                  ={E ∖ ↑fracN .@ l}=∗ ∃ s'', ⌜s' ⊑ s''⌝
                                                ∗ Q s'' ∗ ▷ (F l s'' v_n) ∗ ▷ F_past l s'' v_n)
           ∗ ▷ (∀ s' v, ⌜s ⊑ s'⌝ ∗ ⌜v ≠ v_o⌝ ∗ ▷ F_past l s' v ∗ P
                        ={E ∖ ↑fracN .@ l}=∗ R s' v)
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v ={E ∖ ↑fracN .@ l}=∗
                                                                       F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ vGPS_FP l s q }}}
      (CAS #(LitLoc l) #(LitInt v_o) #(LitInt v_n)) @ E
      {{{ s'' (b: bool) v, RET #(lit_of_bool b) ;
           ⌜s ⊑ s''⌝
             ∗ vGPS_FP l s'' q ∗ (if b then Q s'' else R s'' v) }}}.
  Proof.
    intros; iIntros "#kI !#" (?) "(VS & VSF & VSDup & oP & oW) Post".
    iApply (GPS_FP_CAS' _ _ _ _ _ _ P (fun s => vGPS_FP l s q ∗ Q s)%I
      (fun s v => vGPS_FP l s q ∗ R s v)%I with "[$kI VS VSF $VSDup $oP $oW]");
      [done | done | .. |].
    { iSplitL "VS".
      - iIntros "!>" (?) "Hs'"; iMod ("VS" with "Hs'") as (s'') "(? & ? & ? & ?)".
        iExists s''; iFrame.
        iModIntro; auto.
      - iIntros "!>" (??) "(? & ? & ? & ? & ?)"; iFrame.
        iApply fupd_mask_mono; last by iApply "VSF"; iFrame.
        solve_ndisj. }
    iIntros "!>" (s'' b v) "(? & ?)".
    destruct b; iApply ("Post" $! s'' _ v); iFrame.
  Qed.

  Global Instance GPS_FP_fractional l s:
    fractional.Fractional (λ q, vGPS_FP l s q).
  Proof.
    intros q1 q2.
    iStartProof (uPred _).
    iIntros (?).
    rewrite vGPS_FP_eq /= /vGPS_FP_def.
    rewrite fractor_splitjoin; last auto.
    iSplit; iIntros "FR";
    iApply (fractor_mono with "[$FR]"); iIntros (X).
    - iIntros "R". iDestruct "R" as (γ γ_x e_x) "(% & (ex1 & ex2) & #R)".
      iSplitL "ex1"; iExists γ, γ_x, e_x; iFrame "#"; by iFrame.
    - iIntros "[R1 R2]".
      iDestruct "R1" as (? ? ?) "(% & ex1 & R1)".
      iDestruct "R2" as (γ γ_x e_x) "(% & ex2 & R2)".
      subst X. apply encode_inj in H3. inversion H3. subst.
      iExists γ, γ_x, e_x. iFrame "R2". iSplit; first auto.
      iCombine "ex1" "ex2" as "ex".
      iDestruct (own.own_valid with "ex") as %Val%auth_valid_discrete.
      move: Val => [_ /= /agree_op_inv /to_agree_inj Eqe]. setoid_subst.
      iDestruct "ex" as "[ex1 ex2]". rewrite agree_idemp. iSplitL "ex1"; iFrame.
  Qed.

  Lemma GPS_FP_splitjoin l s q1 q2:
    vGPS_FP l s q1 ∗ vGPS_FP l s q2 ⊣⊢ vGPS_FP l s (q1 + q2).
  Proof. by rewrite GPS_FP_fractional. Qed.

  Lemma GPS_FP_dealloc l s (E: coPset) (HNl: ↑fracN .@ l ⊆ E):
    vGPS_FP l s 1%Qp ={E}=∗ own_loc l.
  Proof.
    rewrite vGPS_FP_eq /vGPS_FP_def /own_loc /own_loc_prim
            valloc_local_eq /valloc_local_def /alloc_local /=.
    iStartProof (uPred _).
    iIntros (?) "FP".
    iMod (fractor_dealloc with "FP") as (n i) "(Info & RR & Inv)"; first auto.
    iDestruct "RR" as (γ γ_x e_x) "(% & ex & RA)".
    rewrite monPred_in_eq /monPred_in_def /=.
    iDestruct "Inv" as (γ' γ_x' ζ) "(>% & oI & _)".
    subst n. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "oI" as (ζ' h e') "(>oA & _ & >Hist & ? & ? & ? & ? & >% & ?)".
    iDestruct "RA" as (Vr v) "(% & _ & RA)".
    iDestruct (Reader_Auth_sub with "[$oA $RA]")
      as %Inζ%elem_of_subseteq_singleton.
    iExists h, i. iFrame "Hist Info". iPureIntro.
    exists (VInj v), Vr. repeat split; [|done|done].
    rewrite -H elem_of_map_gset. by exists (s, (v, Vr)).
  Qed.

End Fractional.

Arguments vGPS_FP [_ _ _ _ _ _] _ _ _ _.
Arguments GPS_FP_Init [_ _ _ _ _ _] _ [_] _ [_] _ _ _.
Arguments GPS_FP_Read' [_ _ _ _ _ _] _ [_ _] _ _ [_ _] _ _ [_].
Arguments GPS_FP_Read [_ _ _ _ _ _] _ [_ _] _ _ [_ _] _ _.
Arguments GPS_FP_Write [_ _ _ _ _ _] _ [_ _ _ _] [_ _] _ _ _ _ _ [_].
Arguments GPS_FP_FAI [_ _ _ _ _ _] _ [_ _] [_ _ _] _ _ _ _ [_].
Arguments GPS_FP_CAS' [_ _ _ _ _ _] _ [_ _] [_ _ _ _] _ _ _ _ _ [_].
Arguments GPS_FP_CAS [_ _ _ _ _ _] _ [_ _] [_ _ _ _] _ _ _ _ _.
Arguments GPS_FPs_agree [_ _ _ _ _ _ _ _ _ _ _ _ _] _.
Arguments GPS_FP_splitjoin [_ _ _ _ _ _ _ _ _ _ _].

Notation "'[FP' l 'in' s | p ]_ q" :=
  (vGPS_FP p l s q)
    (at level 0,
     format "[FP  l  in  s  |  p  ]_ q").

(* TODO: fix instances to compile *)
Global Instance vGPS_FP_contractive `{aG : cinvG Σ} `{foundationG Σ} `{gpsG Σ Prtcl} `{cancelable_invariants.cinvG Σ}:
  Contractive (λ F : interpC Σ Prtcl, [FP l in s1 | F ]_q).
Proof.
  repeat intro.
  rewrite !vGPS_FP_eq /= /fractored.
  constructor => ?.
  apply bi.exist_ne => ?. apply bi.sep_ne; [done|].
  apply bi.exist_ne => ?. repeat (apply bi.sep_ne; [done|]).
  f_contractive. repeat apply bi.exist_ne => ?.
  apply bi.sep_ne; [done|].
  apply bi.sep_ne; [by apply gps_inv_ne|done].
Qed.

Global Instance vGPS_FP_proper {Σ Prtcl fG PF gpsG persG}:
  Proper ((≡) ==> (=) ==> (=) ==> (≡)) (@vGPS_FP Σ Prtcl fG PF gpsG persG q).
Proof. solve_proper. Qed.
