From iris.proofmode Require Import tactics.
From iris.base_logic Require Import iprop.
From iris.program_logic Require Import ownp.
From igps Require Import weakestpre. (* TODO: why ??? *)
From igps.gps Require Export inst_shared.
Import base.

Section Plain.
  Context {Σ : gFunctors} {Prtcl : protocolT} `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{persG : persistorG Σ}
          `{agreeG : gps_agreeG Σ Prtcl}
          `{pCount : Countable Param}
          {IP : interpC Σ Prtcl}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (vProp Σ).

  Notation F := (IP true).
  Notation F_past := (IP false).

  Implicit Types
           (l : loc)
           (s : pr_state Prtcl)
           (ζ : gset (state_sort Prtcl))
           (S : pr_state Prtcl * (Z * View))
           (v : Z)
           (V : View)
           (h : History)
           (e : state_sort Prtcl)
           (p : Param).

  Definition gpsPRaw p l n : iProp :=
    (∃ (γ γ_x γx: gname) (e_x: state_sort Prtcl) ζ,
      ⌜n = encode (p,γ,γ_x,γx)⌝ ∗ gps_inv IP l γ γ_x
      ∗ Writer γ ζ ∗ exwr γ_x 1%Qp e_x ∗ own.own γx (to_agree e_x))%I.

  Definition RPRaw p s V l n : iProp :=
    (∃ (γ γ_x γx: gname) e_x,
      ⌜n = encode (p,γ,γ_x,γx)⌝ ∗ own.own γx (to_agree e_x)
      ∗ ∃ Vr v, ⌜Vr ⊑ V⌝
              ∗ ⌜st_view e_x !! l ⊑ (Vr !! l : option positive)⌝ ∗ Reader γ {[s, (v, Vr)]})%I.

  Program Definition vGPS_PP_def l p s : vProp Σ :=
    MonPred (λ V, persisted l (gpsPRaw p) (RPRaw p s V)) _.
  Next Obligation.
    intros. intros V V' Mono.
    iIntros "RP". iDestruct "RP" as (X) "(RPV & ?)".
    iDestruct "RPV" as (γ γ_x γx e_x) "(% & ? & RA)".
    iDestruct "RA" as (Vr v) "(% & % & ?)".
    iExists X. iFrame. iExists γ, γ_x, γx, e_x. iFrame "%∗".
    iExists Vr, v. iFrame "%∗". iPureIntro. by etrans.
  Qed.
  Definition vGPS_PP_aux : seal vGPS_PP_def. by eexists. Qed.
  Definition vGPS_PP := unseal vGPS_PP_aux.
  Definition vGPS_PP_eq : vGPS_PP = _ := seal_eq _.

  Global Instance vGPS_PP_persistent l p s : Persistent (vGPS_PP l p s).
  Proof.
    rewrite vGPS_PP_eq /Persistent /bi_persistently. split => ? /=.
    rewrite /sbi_persistently /= monPred_persistently_eq /=.
    apply persistor_persistent => ?. apply _.
  Qed.

  Local Instance :
    Frame false (gps_inv IP l γ γ_x) (gpsPRaw p l (encode (p,γ,γ_x,γx)))
          (Writer γ ζ ∗ exwr γ_x 1%Qp e_x ∗ own.own γx (to_agree e_x)).
  Proof.
    intros. rewrite /Frame. iIntros "[? ?]". iExists _, _.
    iFrame "∗". iExists _, _, _. cbn. by iFrame "∗".
  Qed.

  Lemma GPS_PPs_agree l p s1 s2 (E : coPset) (HE: ↑persist_locN .@ l ⊆ E):
    vGPS_PP l p s1 ∗ vGPS_PP l p s2 ={E}=∗ ⌜s1 ⊑ s2 ∨ s2 ⊑ s1⌝.
  Proof.
    iStartProof (uPred _). rewrite vGPS_PP_eq /vGPS_PP_def /=.
    iIntros (?) "RP". rewrite persistor_splitjoin.
    iMod (persistor_open with "RP") as (X) "(oI & (oR1 & oR2) & HClose)";
      first auto.
    iDestruct "oR1" as (γ1 γ_x1 γx1 e_x1) "(% & #oD1 & #RA1)".
    iDestruct "oR2" as (γ2 γ_x2 γx2 e_x2) "(% & #oD2 & #RA2)".
    iDestruct "oI" as (γ γ_x γx e_x ζ) "(>% & oI & oI2)". subst.
    apply encode_inj in H0. inversion H0. subst.
    apply encode_inj in H1. inversion H1. subst.
    iDestruct "RA1" as (V1 v1) "(VV1 & Le1 & oR1)".
    iDestruct "RA2" as (V2 v2) "(VV2 & Le2 & oR2)".
    iMod (Reader_singleton_agree _ _ _ _ s1 _ _ s2 with "[$oI $oR1 $oR2]") as "[oI %]".
    iMod ("HClose" $! (λ l0 X, RPRaw p s1 _ l0 X ∗ RPRaw p s2 _ l0 X)%I
          with "[oI oI2]") as "oR"; last done.
    iSplitL "oI oI2"; last iSplitL ""; iExists γ, γ_x, γx.
    - iExists e_x, ζ. iFrame "oI oI2". by iNext.
    - iExists e_x1. iFrame "oD1". iSplitL; first auto.
      iExists V1, v1. iFrame "VV1 Le1 oR1".
    - iExists e_x2. iFrame "oD2". iSplitL; first auto.
      iExists V2, v2. iFrame "VV2 Le2 oR2".
  Qed.

  Lemma GPS_PP_Init l p s v (E : coPset)
    (HEN : ↑physN ⊆ E) (HNp : ↑persistN ⊆ E):
    {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ own_loc l ∗ ▷ F l s v ∗ ▷ F_past l s v }}}
           ([ #l ]_na <- #v) @ E
           {{{ RET #() ; vGPS_PP l p s }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V) "(#kI & #kIP & oL & F & Fp)"; iIntros (V1 ?) "Post".
    rewrite /own_loc /own_loc_prim.
    iDestruct "oL" as (h i) "(Alloc & Hist & Info)".
    rewrite valloc_local_eq /valloc_local_def /=.
    iDestruct "Alloc" as %Alloc.
    rewrite WP_Def.wp_eq /=.
    iIntros (V2 ??) "kS".
    iApply program_logic.weakestpre.wp_fupd.

    assert (V ⊑ V2) by (etransitivity; eauto).
    iApply (RawProto_init IP l s v _ _
               with "[$kI $kS $Hist $F $Fp]"); first auto.
    { iPureIntro; eapply alloc_local_proper; eauto. }

    iNext.
    iIntros (V' γ γ_x) "(% & kS' & oI & ex & #oR & oW)"; iFrame.

    iMod (own.own_alloc (to_agree ((s, (v, V')) : state_sort _))) as (γx) "#oEx"; first done.

    iMod (persistor_inv_alloc E i (encode (p,γ,γ_x,γx)) l
          (gpsPRaw p) (RPRaw p s V')
          with "[-Post $kIP]") as "PP"; [auto|auto|..].
      { iFrame "kI Info oI oW ex oEx".
        repeat iExists _. iFrame "oEx". iSplitL; [done|].
        iExists _, _; iFrame "oR". by iSplit. }
      iModIntro; iApply ("Post" $! _ with "[%]").
      { etrans; eauto. }
    rewrite vGPS_PP_eq /=; done.
  Qed.

  Lemma GPS_PP_Read' l p (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷(∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past l s' v ∗ vGPS_PP l p s' ∗ P ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P
                            ={E ∖ ↑persist_locN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                    ={E ∖ ↑persist_locN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
      ∗ P ∗ vGPS_PP l p s }}}
    [ #l ]_at @ E
  {{{ s' v, RET #v ; ⌜s ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V) "(#kI & VS & VSDup & oP & oR)"; iIntros (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_PP_eq /=; iIntros (V3 ??) "kS".
    iMod (persistor_open with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first auto.
    iDestruct "RRaw" as (γ γ_x γx e_x) "(% & #Hex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' γx' e_x' ζ)
                          "(>% & oI & Writer & >Ex & >#Hex')".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x' γx'.

    iCombine "Hex" "Hex'" as "He".
    iDestruct (own.own_valid with "He") as %He%agree_op_inv%(inj to_agree).
    setoid_subst.

    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    assert (V ⊑ V3) by (etransitivity; eauto).
    iApply (RawProto_Read IP l P (fun s v => |={E ∖ ↑persist_locN .@ l, E}=> Q s v)%I s _ _ V_r v_r γ γ_x
            with "[VS $oI $kI $kS $oP $oR VSDup Writer Ex HClose]"); [solve_ndisj|auto|..].
    { etransitivity; eauto. }
    { iSplitR "VSDup"; iNext; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ? Hs'). iDestruct ("VS" $! s' Hs') as "[VS|VS]"; last first.
      { iRight. iIntros (?????). iApply ("VS" $! _ _ with "[//]"). }
      iLeft; iIntros (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (RPRaw p s' j0) with "[Writer oI oR' Ex]").
      { iSplit.
        + iNext. iExists γ, γ_x, γx, e_x', ζ. iFrame; auto.
        + iExists γ, γ_x, γx, e_x'. iFrame "Hex". iSplitL ""; first auto.
          iExists Vr, v. iSplit; last iFrame "oR'"; iPureIntro; auto.
          do 4 (etrans; eauto). }
      iMod ("VS" $! _ j0 with "[%] [$]") as "$"; [solve_jsl|].
      iApply fupd_intro_mask'. solve_ndisj. }
    iNext.
    iIntros (s' v Vr V') "(% & kS' & % & oQ & #oR' & % & %)".
    iFrame. iApply ("Post" $! s' _ _ with "[//]"). by iFrame "%".
  Qed.

  Lemma GPS_PP_Read l p (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  ⎡PSCtx⎤ ⊢ {{{ ▷(∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past l s' v ∗ P ={E ∖ ↑persist_locN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P
                            ={E ∖ ↑persist_locN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                    ={E ∖ ↑persist_locN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
      ∗ P ∗ vGPS_PP l p s }}}
    [ #l ]_at @ E
  {{{ s' v, RET #v ; ⌜s ⊑ s'⌝ ∗ vGPS_PP l p s' ∗ Q s' v }}}.
  Proof.
    iIntros "#kI !#" (?) "(VS & VSDup & oP & oR) Post".
    iApply (GPS_PP_Read' _ _ P with "[$kI VS $VSDup $oP $oR]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_PP_Write l p s s' v (E : coPset)
      (P: vPred) (Q : pr_state Prtcl → vPred)
      (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E)
      (HFinal: final_st s'):
   {{{ ⎡PSCtx⎤ ∗ ▷ (P ∗ vGPS_PP l p s' ={E ∖ ↑persist_locN .@ l}=∗
              Q s' ∗ ▷ (F l s' v) ∗ ▷ F_past l s' v)
      ∗ ▷ P ∗ vGPS_PP l p s }}}
    ([ #l ]_at <- #v) @ E
  {{{ RET #() ;  vGPS_PP l p s' ∗ Q s' }}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V); iIntros "(#kI & VS & oP & oW)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_PP_eq /=; iIntros (V3 ??) "kS".
    iMod ((persistor_open') with "oW")
      as (X) "(gpsRaw & RRaw & #HT & HClose)"; [done|].
    iDestruct "RRaw" as (γ γ_x γx e_x) "(% & #Hex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' γx' e_x' ζ)
                          "(>% & oI & >wTok & >ex & >#Hex')".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x' γx'.

    iCombine "Hex" "Hex'" as "He".
    iDestruct (own.own_valid with "He") as %He%agree_op_inv%(inj to_agree). setoid_subst.
    destruct e_x' as [s_x [v_x V_x]].

    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    iApply (RawProto_Write_non_ex IP l
              s s_x s' _ _ V_r V_x 1%Qp v v_r v_x γ γ_x _ ζ
            with "[$oI $kI $kS $ex $oR $wTok]"); [solve_ndisj|auto|auto|auto|].
    { do 2 (etransitivity; eauto). }
    iNext.
    iIntros (V') "(% & kS' & oI & ex & #oR' & wTok)".
    iFrame.
    iApply ("Post" $! _ with "[%]").
    { etrans; eauto. }
    iAssert (persisted l (gpsPRaw p) (RPRaw p s' V')) with "[#]" as "#oPP".
    { iApply "HT".
      iExists γ, γ_x, γx, (s_x, (v_x, V_x)).
      repeat iSplit; [auto|auto|].
      iExists V', v. repeat iSplit; [auto| |auto].
      iPureIntro. do 4 (etransitivity; eauto). }
    iMod ("VS" $! V' with "[%] [$oP $oPP]") as "($ & oF & oFp)"; [solve_jsl|].
    iSpecialize ("oI" with "[$oF $oFp]").
    iFrame "oPP". iApply "HClose".
    iNext. iExists γ, γ_x, γx, (s_x, (v_x, V_x)), ({[s', (v, V')]} ∪ ζ).
    by iFrame "oI wTok ex Hex".
  Qed.

  Lemma GPS_PP_CAS' l p s v_o v_n (E : coPset)
    (P : vPred) (Q: pr_state Prtcl → vPred) (R : pr_state Prtcl → Z → vPred)
    (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E) :
  {{{  ⎡PSCtx⎤ ∗ ▷ (∀ s', ⌜s ⊑ s'⌝ ∗ F l s' v_o ∗ P ={E ∖ ↑persist_locN .@ l}=∗
                  ∃ s'', ⌜s' ⊑ s''⌝ ∗ ▷ (F l s'' v_n) ∗ ▷ F_past l s'' v_n ∗
                  (vGPS_PP l p s'' ={E}=∗ Q s''))
      ∗ ▷(∀ s' v, ⌜s ⊑ s'⌝ ∗ ⌜v ≠ v_o⌝ ∗ ▷ F_past l s' v ∗ vGPS_PP l p s' ∗ P
                      ={E}=∗ R s' v)
      ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                  ={E ∖ ↑persist_locN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
      ∗ P ∗ vGPS_PP l p s }}}
    (CAS #l #v_o #v_n) @ E
  {{{ s'' (b: bool) v, RET #(lit_of_bool b) ;
      ⌜s ⊑ s''⌝ ∗ (if b then Q s'' else R s'' v )}}}.
  Proof.
    iStartProof (uPred _).
    iIntros (? V); iIntros "(#kI & VS & VSF & VSDup & oP & oW)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_PP_eq /=; iIntros (V3 ??) "kS".
    iMod ((persistor_open) with "[$oW]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.
    iDestruct "RRaw" as (γ γ_x γx e_x) "(% & #Hex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' γx' e_x' ζ)
                          "(>% & oI & >wTok & >ex & >#Hex')".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x' γx'.

    iCombine "Hex" "Hex'" as "He".
    iDestruct (own.own_valid with "He") as %He%agree_op_inv%(inj to_agree).
    setoid_subst. destruct e_x' as [s_x [v_x V_x]].

    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    assert (V_r ⊑ V3) by (do 2 (etransitivity; eauto)).
    assert (V ⊑ V3) by (do 2 (etransitivity; eauto)).
    iApply (RawProto_CAS_non_ex_weak IP l
              s s_x _ _ V_r V_x 1%Qp v_o v_n v_r v_x γ γ_x _
              (P ∗ ⎡(
               (∀ Ψ', ▷ gpsPRaw p l (encode (p, γ, γ_x, γx)) ∗ Ψ' l (encode (p, γ, γ_x, γx)) ={
                 E ∖ ↑persist_locN.@l,E}=∗ persisted l (gpsPRaw p) Ψ'))⎤)%I
              (fun s => |={E ∖ ↑persist_locN .@ l, E}=> Q s)%I
              (fun s v => |={E ∖ ↑persist_locN .@ l, E}=> R s v)%I
            with "[VS VSF $kI $kS oI oP $ex $oR wTok VSDup HClose]");
      [solve_ndisj|solve_ndisj|solve_ndisj| |].
    { iSplitL "VS"; last iSplitL "VSDup"; last iSplitL "VSF"; last (by iFrame; iExists _); iNext.
      - iIntros (s' j ?) "(Hs' & F & oP & HClose)".
        iMod ("VS" $! _ _ with "[%] [$Hs' $F $oP]") as (s'') "(? & ? & ? & Q)".
        { etrans; eauto. }
        iExists s''; iFrame.
        iModIntro; iIntros (Vr j0 Hj) "(% & % & oI & ex & Writer & oR')".
        iDestruct "Writer" as (ζ') "Writer".
        iMod ("HClose" $! (RPRaw p s'' j0) with "[oI ex Writer oR']") as "oW".
        { iSplit.
          + iNext. iExists γ, γ_x, γx, (s_x, (v_x, V_x)), ζ'. iFrame; auto.
          + iExists γ, γ_x, γx, (s_x, (v_x, V_x)). iFrame "Hex". iSplitL ""; first auto.
            iExists Vr, v_n.
            iSplit; last iFrame "oR'"; iPureIntro; auto.
            do 4 (etrans; eauto). }
        iMod ("Q" $! _ with "[//] [oW //]") as "$".
        iApply fupd_intro_mask'. solve_ndisj.
      - by iSpecialize ("VSDup" $! _).
      - iIntros (s' v Vr j ?) "(Hs' & Hv & Fp & (oP & HClose) & % & % & oI & ex & Writer & oR')".
        iDestruct "Writer" as (ζ') "Writer".
        iMod ("HClose" $! (RPRaw p s' j) with "[oI ex Writer oR']") as "oW".
        { iSplit.
          + iNext. iExists γ, γ_x, γx, (s_x, (v_x, V_x)), ζ'. iFrame; auto.
          + iExists γ, γ_x, γx, (s_x, (v_x, V_x)). iFrame "Hex". iSplitL ""; first auto.
            iExists Vr, v.
            iSplit; last iFrame "oR'"; iPureIntro; auto.
            do 4 (etrans; eauto). }
        iMod ("VSF" $! _ _ _ with "[%] [$Hs' $Hv $Fp $oW $oP]") as "$".
        { etrans; eauto. }
        iApply fupd_intro_mask'. solve_ndisj. }

    iNext.
    iIntros (s'' b v Vr V')
            "(% & kS' & Hs' & IF & #oR' & % & %)".
    iFrame.
    iSpecialize ("Post" $! s'' b v _ with "[%]").
    { etrans; eauto. }
    iApply "Post". iFrame. by destruct b.
  Qed.

  Lemma GPS_PP_CAS l p s v_o v_n (E : coPset)
    (P : vPred) (Q: pr_state Prtcl → vPred) (R : pr_state Prtcl → Z → vPred)
    (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E) :
  ⎡PSCtx⎤ ⊢ {{{  ▷ (∀ s', ⌜s ⊑ s'⌝ ∗ F l s' v_o ∗ P
                      ={E ∖ ↑persist_locN .@ l}=∗ ∃ s'', ⌜s' ⊑ s''⌝
                        ∗ Q s'' ∗ ▷ (F l s'' v_n) ∗ ▷ F_past l s'' v_n)
      ∗ ▷(∀ s' v, ⌜s ⊑ s'⌝ ∗ ⌜v ≠ v_o⌝ ∗ ▷ F_past l s' v ∗ P
                      ={E ∖ ↑persist_locN .@ l}=∗ R s' v)
      ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
                  ={E ∖ ↑persist_locN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
      ∗ P ∗ vGPS_PP l p s }}}
    (CAS #l #v_o #v_n) @ E
  {{{ s'' (b: bool) v, RET #(lit_of_bool b) ;
      ⌜s ⊑ s''⌝ ∗ vGPS_PP l p s'' ∗ (if b then Q s'' else R s'' v )}}}.
  Proof.
    intros; iIntros "#kI !#" (?) "(VS & VSF & VSDup & oP & oW) Post".
    iApply (GPS_PP_CAS' _ _ _ _ _ _ P (fun s => vGPS_PP l p s ∗ Q s)%I
      (fun s v => vGPS_PP l p s ∗ R s v)%I with "[$kI VS VSF $VSDup $oP $oW]");
      [done | done | .. |].
    { iSplitL "VS".
      - iIntros "!>" (?) "Hs'"; iMod ("VS" with "Hs'") as (s'') "(? & ? & ? & ?)".
        iExists s''; iFrame.
        iModIntro; auto.
      - iIntros "!>" (??) "(? & ? & ? & ? & ?)"; iFrame.
        iApply fupd_mask_mono; last by iApply "VSF"; iFrame.
        solve_ndisj. }
    iIntros "!>" (s'' b v) "(? & ?)".
    destruct b; iApply ("Post" $! s'' _ v); iFrame.
  Qed.

  Lemma GPS_PP_FAI C l p s (E : coPset)
    (P : vPred) (Q: pr_state Prtcl → Z → vPred)
    (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E) :
  let mod1C (v: Z) := ((v + 1) `mod` Z.pos C)%Z in
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ v s', ⌜s ⊑ s'⌝ ∗ F l s' v ∗ P
                      ={E ∖ ↑persist_locN .@ l}=∗ ∃ s'', ⌜s' ⊑ s''⌝
                        ∗ Q s'' v
                        ∗ ▷ (F l s'' (mod1C v)) ∗ ▷ F_past l s'' (mod1C v))
      ∗ P ∗ vGPS_PP l p s }}}
    (FAI C #l) @ E
  {{{ s'' (v: Z), RET #v ;
      ⌜s ⊑ s''⌝ ∗ vGPS_PP l p s'' ∗ Q s'' v }}}.
  Proof.
    intros. iStartProof (uPred _).
    iIntros (V); iIntros "(#kI & VS & oP & oW)" (V2 ?) "Post".
    rewrite WP_Def.wp_eq vGPS_PP_eq /=; iIntros (V3 ??) "kS".
    iMod ((persistor_open) with "[$oW]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.
    iDestruct "RRaw" as (γ γ_x γx e_x) "(% & #Hex & RA)".
    iDestruct "gpsRaw" as (γ' γ_x' γx' e_x' ζ)
                          "(>% & oI & >wTok & >ex & >#Hex')".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x' γx'.

    iCombine "Hex" "Hex'" as "He".
    iDestruct (own.own_valid with "He") as %He%agree_op_inv%(inj to_agree). setoid_subst.
    destruct e_x' as [s_x [v_x V_x]].

    iDestruct "RA" as (V_r v_r) "(% & % & #oR)".

    assert (V ⊑ V3) by (do 2 (etrans; eauto)).
    iApply (RawProto_FAI_non_ex IP _ _
              s s_x _ _ V_r V_x 1%Qp v_r v_x γ γ_x _ P Q
            with "[VS $kI $kS $oI $oP $ex $oR wTok]");
      [auto|auto|solve_ndisj| |].
    { etrans; eauto. }
    { iSplitR "wTok"; last by iExists _.
      iNext; iIntros (?? j ?) "?".
      iMod ("VS" $! _ _ j with "[%] [-//]"); [solve_jsl|done]. }
    iNext.
    iIntros (s'' v Vr V')
            "(% & kS' & Hs' & oI & Q & ex & #oR' & % & % & wTok)".
    iFrame; iApply ("Post" $! _ _ _ with "[%]"); [solve_jsl|].
    iFrame.
    iApply "HClose". iSplitL "oI ex wTok".
    + iNext. iDestruct "wTok" as (ζ') "wTok".
      iExists γ, γ_x, γx, (s_x, (v_x, V_x)), _.
      by iFrame "oI wTok ex Hex".
    + iExists γ, γ_x, γx, (s_x, (v_x, V_x)).
      repeat iSplit; [auto|auto|].
      iExists Vr, _. repeat iSplit; [auto| |auto].
      iPureIntro. do 4 (etrans; eauto).
  Qed.

End Plain.

Section Plain_Default.
  Context {Σ : gFunctors} {Prtcl : protocolT}
          `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{persG : persistorG Σ}
          `{agreeG : gps_agreeG Σ Prtcl}
          {IP : interpC Σ Prtcl}.
  Set Default Proof Using "Type".

  Notation F := (IP true).
  Notation F_past := (IP false).

  Lemma GPS_PP_Init_default l s v (E : coPset)
    (HEN : ↑physN ⊆ E) (HNp : ↑persistN ⊆ E):
    ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ⊢ {{{ own_loc l ∗ ▷ F l s v ∗ ▷ F_past l s v }}}
    ([ #l ]_na <- #v) @ E
    {{{ RET #() ; vGPS_PP (IP := IP) l () s }}}.
  Proof.
    iIntros "#[??] !#" (?) "(? & ? & ?) Post".
    iApply (GPS_PP_Init _ () with "[-Post]"); [done | done | | done].
    iFrame "#"; iFrame.
  Qed.

End Plain_Default.

Lemma GPS_PPs_param_agree
  {Σ : gFunctors} {Prtcl : protocolT}
  `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
  `{persG : persistorG Σ}
  `{agreeG : gps_agreeG Σ Prtcl}
  `{pCount : Countable Param}
  {IP1 IP2 : interpC Σ Prtcl} l p1 p2 s1 s2:
  vGPS_PP (IP:=IP1) l p1 s1 ∗ vGPS_PP (IP:= IP2) l p2 s2 ⊢ ⌜p1 = p2⌝.
Proof.
  split => ?.
  rewrite !vGPS_PP_eq /vGPS_PP_def /=.
  iIntros "RP". rewrite /bi_sep /= /sbi_sep /= monPred_sep_eq /= persistor_join_l persistor_drop.
  iDestruct "RP" as (X) "[oR1 oR2]".
  iDestruct "oR1" as (γ1 γ_x1 γx1 e_x1) "(% & _)".
  iDestruct "oR2" as (γ2 γ_x2 γx2 e_x2) "(% & _)".
  rewrite H0 in H. apply encode_inj in H. by inversion H.
Qed.

Arguments vGPS_PP [_ _ _ _ _ _ _ _ _ _] _ _ _.
Arguments GPS_PP_Init [_ _ _ _ _ _ _ _ _ _] _ [_ _] _ [_ _] _ _.
Arguments GPS_PP_Read [_ _ _ _ _ _ _ _ _ _] _ [_ _] _ _ [_ _] _ _.
Arguments GPS_PP_Read' [_ _ _ _ _ _ _ _ _ _] _ [_ _] _ _ [_ _] _ _.
Arguments GPS_PP_Write [_ _ _ _ _ _ _ _ _ _] _ [_ _ _] _ [_ _] _ _ _ _ _.
Arguments GPS_PP_FAI [_ _ _ _ _ _ _ _ _ _] _ [_ _ _ _ _] _ _ _ _.
Arguments GPS_PP_CAS [_ _ _ _ _ _ _ _ _ _] _ [_ _ _ _ _ _] _ _ _ _ _.
Arguments GPS_PP_CAS' [_ _ _ _ _ _ _ _ _ _] _ [_ _ _ _ _ _] _ _ _ _ _.
Arguments GPS_PPs_agree [_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _].
Arguments GPS_PPs_param_agree [_ _ _ _ _ _ _ _ _ _ _ _ _].
Arguments GPS_PP_Init_default [_ _ _ _ _ _ _] _ [_ ] _ [_ _] _ _.

Notation "'[PP' l 'in' s | F ]" :=
  (vGPS_PP F l () s)
    (format "[PP  l  in  s  |  F  ]").

Notation "'[PP' l 'in' s '@' p | F ]" :=
  (vGPS_PP F l p s)
    (format "[PP  l  in  s  @  p  |  F  ]").

Global Instance vGPS_PP_contractive `{fG : foundationG Σ}
       `{gpsG Σ Prtcl}
       `{pG : persistorG Σ} `{aG : gps_agreeG Σ Prtcl} `{pC: Countable Param}:
  Contractive (λ F : interpC Σ Prtcl, [PP l in s1 @ (p : Param) | F ]).
Proof.
  repeat intro.
  rewrite !vGPS_PP_eq /= /persisted.
  constructor => ?.
  apply bi.exist_ne => ?.
  apply bi.sep_ne; [done|].
  apply bi.sep_ne; [|done].
  apply invariants.inv_contractive. destruct n; first done.
  repeat apply bi.exist_ne => ?.
  apply bi.sep_ne; [done|].
  apply bi.sep_ne; [by apply gps_inv_ne|done].
Qed.

Global Instance gpsPPRaw_proper {Σ Prtcl fG PF gpsG agreeG PType EqPT CountPT}:
  Proper ((≡) ==> (=) ==> (=) ==> (=) ==> (≡))
         (@gpsPRaw Σ Prtcl fG PF gpsG agreeG PType EqPT CountPT).
Proof.
  intros ? ? H1 ? ? H2 ? ? H3 ?? H4. subst.
  unfold gpsPRaw.
  repeat (apply (_ : Proper (_ ==> (≡)) _); repeat intro).
  repeat (apply (_ : Proper (_ ==> _ ==> (≡)) _); repeat intro); auto.
  by apply gps_inv_proper.
Qed.

Global Instance vGPS_PP_proper {Σ Prtcl fG PF gpsG persG agreeG PType EqPT CountPT}:
  Proper ((≡) ==> (=) ==> (=) ==> (=) ==> (≡))
         (@vGPS_PP Σ Prtcl fG PF gpsG persG agreeG PType EqPT CountPT).
Proof.
  intros ? ? H1 ? ? H2 ? ? H3 ? ? H4. constructor => ?. subst.
  rewrite !vGPS_PP_eq /=. apply persisted_proper.
  - repeat intro. exact: gpsPPRaw_proper.
  - repeat intro. by f_equiv.
Qed.
