From iris.proofmode Require Import tactics.
From iris.base_logic.lib Require Import cancelable_invariants sts.
From iris.algebra Require Export cmra_big_op ofe.

Import uPred.

From igps Require Import persistor fractor proofmode weakestpre.
From igps Require Export notation na.
From igps.gps Require Import shared inst_shared.

Section RecInterp.
  Context {Σ : gFunctors} (Prtcl : protocolT).
  Set Default Proof Using "Type".

  Global Instance IntType_cofe : Cofe (interpC Σ Prtcl).
  Proof. by apply _. Qed.
  Global Instance intpair_inhabited : Inhabited (interpC Σ Prtcl).
  Proof. constructor; repeat intro; exact True%I. Qed.
End RecInterp.

Structure OTag := otag { otagged :> ofeT }.
Definition oparam o := otag o.
Canonical ointerp o := oparam o.
Structure TTag := ttag { ttagged :> Type }.
Definition tparam T := ttag T.
Canonical tinterp T := tparam T.

Structure type_ofe :=
  mkto {
      ofe_of_ofe :> OTag;
      type_of_ofe : TTag ;
      _ : ttagged (type_of_ofe) = ofe_car ofe_of_ofe
    }.
(* Arguments type_of_ofe _. *)
Definition type_of_eq (to : type_ofe) : ttagged (type_of_ofe to) = ofe_car (ofe_of_ofe to) := ltac:(by destruct to).
Program Canonical type_of_interp Σ Prtcl : type_ofe :=
  mkto (ointerp (interpC Σ Prtcl)) (tinterp (bool -c> loc -c> pr_state Prtcl -c> Z -c> vProp Σ)) (Logic.eq_refl).
Program Canonical type_of_param X (t : type_ofe) : type_ofe :=
  mkto (oparam (X -c> ofe_of_ofe t)) (tparam (X -> type_of_ofe t)) (f_equal (λ Y, X -> Y) (type_of_eq t)).

(* Global Instance Dist_type_of_ofe o (to : type_ofe o) : Dist (type_of_ofe to). *)
(* Proof. *)
(*   destruct to. *)
(*   intros ?. red. rewrite e. exact (@dist o _ H). *)
(* Defined. *)

Definition type_ofe_conv (to : type_ofe) : ∀ (F : type_of_ofe to -> type_of_ofe to), ofe_of_ofe to -c> ofe_of_ofe to :=
  ltac:(by rewrite type_of_eq).

Structure funC_of :=
  mk_funC_of { funC_ofe : type_ofe;
               _ : Cofe (funC_ofe);
               _ : Inhabited (funC_ofe);
               _ : ∀ F, Contractive (type_ofe_conv funC_ofe F) → ofe_of_ofe funC_ofe;
             }.
Definition funC_Cofe f : Cofe (funC_ofe f) := ltac:(by destruct f).
Definition funC_Inhabited f : Inhabited (funC_ofe f) := ltac:(by destruct f).
Definition funC_fix (f : funC_of) : ∀ F, Contractive (type_ofe_conv (funC_ofe f) F) → funC_ofe f := ltac:(by destruct f).

Canonical funC_interpC Σ Prtcl :=
  mk_funC_of (type_of_interp Σ Prtcl)
             _
             _
             (λ F HF, fixpoint (type_ofe_conv _ F) (Hf := HF)).
Local Existing Instances funC_Cofe funC_Inhabited.
Canonical funC_param (f : funC_of) (X : ofeT) :=
  mk_funC_of (type_of_param X (funC_ofe f))
             _ _
             (λ F, @fixpoint _ _ _ (type_ofe_conv _ F)).

Tactic Notation "exvar" constr(type) tactic(cnt) :=
  let name := fresh in
  evar (name : type);
  let var := (eval unfold name in name) in
  clear name;
  cnt var.
Tactic Notation "recproto" uconstr(proto) tactic(cont) :=
  (* "at" integer_list(occs) := *)
  exvar (funC_of) ltac:(
    fun f =>
      exvar (type_of_ofe (funC_ofe f) → type_of_ofe (funC_ofe f)) ltac:(
        fun F =>
          exvar (Contractive (type_ofe_conv (funC_ofe f) F)) ltac:(
            fun C =>
              let t := type of proto in
              let delete := fresh "delete" in
              (assert (delete : type_of_ofe (funC_ofe f) = tinterp t) by exact: eq_refl);
              clear delete;
              (assert (delete : proto = (funC_fix f F C)) by exact: eq_refl);
              clear delete;
              let F := (eval lazy delta [id] in F) in
              let f := (eval lazy delta [id] in f) in
              let unf := uconstr:(@fixpoint_unfold (funC_ofe f) (funC_Cofe f) (funC_Inhabited f) F (C)) in
              let f' := f in
              let rec go f F unf proto :=
                  (* idtac f; idtac F; *)
                  lazymatch f with
                  | funC_interpC ?Σ ?Prtcl =>
                    cont uconstr:(unf _ _ _ _ : proto _ _ _ _ ≡ F _ _ _ _)
                  | funC_param ?f ?X =>
                    go f uconstr:(F _) uconstr:(unf _) uconstr:(proto _)
                  end
              in
              go f (F proto) unf proto
          )
      )
  ).

Ltac spine_of int :=
  match int with
  | funC_fix ?f ?F ?C => uconstr:((f,F,C))
  | _ => let int' := (eval red in int) in
         let r := spine_of int' in
         uconstr:(r)
  end.

Ltac unfold_lemma int :=
  let orig_int := int in
  let rec unfold_lemma int :=
  match int with
  | funC_fix ?f ?F ?C =>
    let rec go f F unf int :=
        (* idtac f; idtac F; *)
        lazymatch f with
        | funC_interpC ?Σ ?Prtcl =>
          uconstr:(unf _ _ _ _ : int _ _ _ _ ≡ F _ _ _ _)
        | funC_param ?f ?X =>
          go f uconstr:(F _) uconstr:(unf _) uconstr:(int _)
        end
    in
    let unf := uconstr:(@fixpoint_unfold (funC_ofe f) (funC_Cofe f) (funC_Inhabited f) (F) (C)) in
    go (f) (F orig_int) unf orig_int
  | _ => let int' := (eval red in int) in
         let r := unfold_lemma int' in
         uconstr:(r)
  end
    in unfold_lemma int.

Notation unf_int int := (ltac:(let e := unfold_lemma int in refine e)) (only parsing).

Notation "'[rec' F ]" := (funC_fix _ (F : _ -> _) _).

(* Section Example. *)
(*   Definition prot : protocolT := ProtocolT bool (implb). *)
(*   Context `{Σ : gFunctors}. *)
(*   Context `{fG : foundationG Σ} `{GPSG : !@gpsG Σ prot PF} `{cinvG Σ}. *)

(*   (* Typeclasses eauto := debug. *) *)
(*   Program Definition int : View -> View -> interpC Σ prot := *)
(*     [rec (fun F V1 V2 b l s v => [FP l in s | F ∅ V1 ]_1)]. *)
(*   Next Obligation. cbn. *)
(*     repeat intro. cbn. apply vGPS_FP_contractive. *)
(*     destruct n => //. *)
(*     by apply (H0). *)
(*   Qed. *)

(*   (* Demonstration of recproto_unfold *) *)
(*   Lemma int_Exclusive V1 V1' V2 V2' b b' l s s' v v' V V' : *)
(*     (int V1 V2 b l s v V) ∗ int V1' V2' b' l s' v' V' -∗ False. *)
(*   Proof. *)
(*     iIntros "?". *)
(*     rewrite !(unf_int int). *)
(*   Abort. *)
(* End Example. *)
