From iris.base_logic Require Import iprop lib.cancelable_invariants.
From iris.proofmode Require Import tactics.
From igps.gps Require Export inst_shared.
From igps Require Import bigop.
From iris.bi Require Import fractional.

Section Gname_StrongSW.
  Context {Σ : gFunctors} {Prtcl : protocolT}
          `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{CInvG : cinvG Σ}
          {IP : gname → interpC Σ Prtcl}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (vProp Σ).

  Notation F γ := (IP γ true).
  Notation F_past γ := (IP γ false).

  Definition WAEx l γ γ_x s : vProp Σ
    := (∃ v V_0 ζ, (monPred_in V_0)
        ∧ ⎡Writer γ ζ ∗ exwr γ_x 1%Qp (s, (v, V_0))
                        ∗ Reader γ {[s, (v, V_0)]}
                        ∗ ⌜maxTime l V_0 ζ⌝⎤)%I.
  Arguments WAEx _ _ _ / _.

  Implicit Types
      (l : loc)
      (s : pr_state Prtcl)
      (ζ : gset (state_sort Prtcl))
      (S : pr_state Prtcl * (Z * View))
      (v : Z)
      (V : View)
      (h : History).

  Local Open Scope I.
  Definition gpsSWnRaw γ l n : iProp :=
    ∃ (γ_x: gname),
      ⌜n = encode (γ,γ_x)⌝ ∗ gps_inv (IP γ) l γ γ_x.

  Definition ReaderSWnRaw γ s V l n : iProp :=
    ∃ (γ_x: gname),
      ⌜n = encode (γ,γ_x)⌝ ∗ monPred_at (PrtSeen γ s) V.

  Definition WriterSWnRaw γ s V l n : iProp :=
    ∃ (γ_x: gname),
      ⌜n = encode (γ,γ_x)⌝ ∗ monPred_at (WAEx l γ γ_x s) V.

  Program Definition vGPS_nFRP_def γ l s (q: frac) : vProp Σ :=
    MonPred (λ V, fractored l (gpsSWnRaw γ) (λ l q, ReaderSWnRaw γ s V l) q) _.
  Next Obligation.
    intros. intros V V' Mono.
    iIntros "RP". iDestruct "RP" as (X) "(RPV & ?)".
    iDestruct "RPV" as (γ_x) "(? & RA)".
    iExists X. iFrame. iExists γ_x. by iFrame.
  Qed.
  Definition vGPS_nFRP_aux : seal vGPS_nFRP_def. by eexists. Qed.
  Definition vGPS_nFRP := unseal vGPS_nFRP_aux.
  Definition vGPS_nFRP_eq : vGPS_nFRP = _ := seal_eq _.

  Program Definition vGPS_nFWP_def γ l s (q: frac) : vProp Σ :=
    MonPred (λ V, fractored l (gpsSWnRaw γ) (λ l q, WriterSWnRaw γ s V l) q) _.
  Next Obligation.
    intros. intros V V' Mono.
    iIntros "WP". iDestruct "WP" as (X) "(WPV & ?)".
    iDestruct "WPV" as (γ_x) "(? & WA)".
    iExists X. iFrame. iExists γ_x. by iFrame.
  Qed.
  Definition vGPS_nFWP_aux : seal vGPS_nFWP_def. by eexists. Qed.
  Definition vGPS_nFWP := unseal vGPS_nFWP_aux.
  Definition vGPS_nFWP_eq : vGPS_nFWP = _ := seal_eq _.

  Close Scope I.

  Local Instance :
    Frame false (gps_inv (IP γ) l γ γ_x)
          (gpsSWnRaw γ l (encode (γ, (γ_x)))) True.
  Proof. intros. rewrite /Frame /=. iIntros "[? _]". iExists _. by iFrame "∗". Qed.

  Lemma GPS_nFWP_Init_strong l (s: gname -> pr_state Prtcl) v
    (Q : gname -> vPred) (E : coPset) (HEN : ↑physN ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ own_loc l ∗
        (∀ γ,
   |={E ∖ ↑persist_locN.@l}=> ▷ F γ l (s γ) v ∗ ▷ F_past γ l (s γ) v ∗ Q γ) }}}
    ([ #l ]_na <- #v) @ E
  {{{ γ, RET #() ; Q γ ∗ vGPS_nFWP γ l (s γ) 1%Qp }}}.
  Proof.
    iStartProof (uPred _).
    rewrite vGPS_nFWP_eq /vGPS_nFWP_def WP_Def.wp_eq /=.
    iIntros (? V); iIntros "(#kI & oL & F)" (V2 ?) "Post".
    iIntros (V3 ? ?) "kS".
    rewrite /own_loc /own_loc_prim valloc_local_eq /valloc_local_def /=.
    iDestruct "oL" as (h i) "(% & Hist & Info)".
    iApply program_logic.weakestpre.wp_fupd.
    assert (V ⊑ V3) by (etrans; eauto).
    iApply (RawProto_init_strong IP l s
               with "[$kI $kS $Hist]"); first auto.
    { iPureIntro; eapply alloc_local_proper; eauto. }
    iNext.
    iIntros (V' γ γ_x) "(% & kS' & oI & ex & #oR & oW)".

    iSpecialize ("F" $! γ).
    iMod (fupd_mask_mono with "F") as "(F & Fp & Q)"; first solve_ndisj.
    assert (V2 ⊑ V') by (etrans; eauto).
    assert (V ⊑ V') by (etrans; eauto).
    iMod (fractor_alloc E i (encode(γ,γ_x)) l
                        (gpsSWnRaw γ) (λ l q, WriterSWnRaw γ (s γ) V' l)
           with "[-Post kS' Q]") as "FP"; first auto.
      { iFrame "kI Info". iSplitL "F Fp oI"; iExists γ_x.
        - iSplitL ""; [by iNext|]. iApply "oI". by iFrame.
        - iSplitL ""; first done.
          iExists v, V', {[(s γ), (v, V')]}. iFrame "ex oW oR".
          iSplit; first auto. iPureIntro.
          by move => ? /elem_of_singleton ->. }
    iFrame; by iApply ("Post" $! _ V' with "[%] [$Q $FP]").
  Qed.

  Lemma GPS_nFWP_ExRead γ l s q (E : coPset)
      (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ <obj> (∀ v, F_past γ l s v ={E ∖ ↑fracN .@ l}=∗
                   F_past γ l s v ∗ F_past γ l s v ) ∗ vGPS_nFWP γ l s q }}}
    ([ #l ]_at) @ E
  {{{ v, RET #v ;
      vGPS_nFWP γ l s q ∗ F_past γ l s v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite vGPS_nFWP_eq WP_Def.wp_eq /=.
    iIntros (? V); iIntros "(#kI & VSDup & oW)" (V2 ?) "Post".
    iIntros (V3 ??) "kS".
    iMod (fractor_open with "[$oW]")
      as (X) "(gpsRaw & WRaw & HClose)"; first auto.
    iDestruct "WRaw" as (γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & oR & %)".

    assert (V_x ⊑ V3) by (do 2 (etrans; eauto)).
    iApply (RawProto_Read_ex (IP γ)
            with "[$kI $kS $oI $ex $oR $wTok VSDup]"); [solve_ndisj|done|done|..].
    { iNext. by iSpecialize ("VSDup" $! _ with ""). }
    iNext.
    iIntros (v V') "(% & kS & oI & Fp & ex & oR' & wTok)".
    iFrame; iApply ("Post" $! _ V' with "[%]").
    { etrans; eauto. }
    iFrame "Fp".
    iApply "HClose". iSplitL "oI".
    + iNext. iExists γ_x. by iFrame.
    + iExists γ_x. iSplitL ""; first auto.
      iExists v_x, V_x,_. iFrame "wTok ex oR'". iSplit; last auto.
      iPureIntro. etrans; eauto.
  Qed.

  Lemma GPS_nFWP_Writer_Reader γ l s q1 q2:
    vGPS_nFWP γ l s (q1 + q2) ⊢ vGPS_nFWP γ l s q1 ∗ vGPS_nFRP γ l s q2.
  Proof.
    rewrite vGPS_nFWP_eq vGPS_nFRP_eq /vGPS_nFWP_def /vGPS_nFRP_def /=.
    iStartProof (uPred _). iIntros (V) "oW".
    rewrite /gpsSWnRaw /WriterSWnRaw /=.
    iApply fractor_splitjoin.
    iApply (fractor_mono with "[$oW]").
    iIntros (X) "oW".
    iDestruct "oW" as (γ_x) "[#? WA]".
    iDestruct "WA" as (v V0 ζ) "(#? & ? & ? & #oR & #?)".
    iSplitR ""; iExists γ_x; (iSplitL ""; first auto).
    - iExists v, V0, ζ. iFrame. iSplit; [auto|by iSplitL ""].
    - rewrite /PrtSeen. iExists V0, v. by iFrame "oR".
  Qed.

  Lemma GPS_nFWP_Writer_Reader_join γ l s1 s2 q1 q2:
    vGPS_nFWP γ l s1 q1 ∗ vGPS_nFRP γ l s2 q2 ⊢ vGPS_nFWP γ l s1 (q1 + q2).
  Proof.
    iStartProof (uPred _). iIntros (V) "(o1 & o2)". iCombine "o1" "o2" as "oW".
    rewrite vGPS_nFWP_eq vGPS_nFRP_eq fractor_splitjoin; last auto.
    iApply (fractor_mono with "[$oW]").
    by iIntros (X) "(oW & _)".
  Qed.

  Lemma GPS_nFRP_gname_agree γ1 γ2 l s1 s2 q1 q2:
    vGPS_nFRP γ1 l s1 q1 ∗ vGPS_nFRP γ2 l s2 q2 ⊢ ⌜γ1 = γ2⌝.
  Proof.
    iStartProof (uPred _). iIntros (V) "oR".
    rewrite vGPS_nFRP_eq /= fractor_join_l.
    iDestruct (fractor_drop with "oR") as (X) "[oR1 oR2]".
    iDestruct "oR1" as (γ_x1) "[% _]".
    iDestruct "oR2" as (γ_x2) "[% _]".
    subst X. apply encode_inj in H0. by inversion H0.
  Qed.

  Lemma GPS_nFWP_gname_agree γ1 γ2 l s1 s2 q1 q2:
    vGPS_nFWP γ1 l s1 q1 ∗ vGPS_nFRP γ2 l s2 q2 ⊢ ⌜γ1 = γ2⌝.
  Proof.
    iStartProof (uPred _). iIntros (V) "oW".
    rewrite vGPS_nFWP_eq vGPS_nFRP_eq /= fractor_join_l.
    iDestruct (fractor_drop with "oW") as (X) "[oW oR]".
    iDestruct "oW" as (γ_x1) "[% _]".
    iDestruct "oR" as (γ_x2) "[% _]".
    subst X. apply encode_inj in H0. by inversion H0.
  Qed.

  Lemma GPS_nFRP_join_l γ l s1 s2 q1 q2 :
    vGPS_nFRP γ l s1 q1 ∗ vGPS_nFRP γ l s2 q2 ⊢ vGPS_nFRP γ l s1 (q1 + q2).
  Proof.
    iStartProof (uPred _). iIntros (V) "oR". rewrite vGPS_nFRP_eq fractor_join_l.
    iApply (fractor_mono with "[$oR]"). by iIntros (X) "[R1 R2]".
  Qed.

  Lemma GPS_nFWP_join_writer γ l s1 s2 q1 q2:
    vGPS_nFWP γ l s1 q1 ∗ vGPS_nFRP γ l s2 q2 ⊢ vGPS_nFWP γ l s1 (q1 + q2).
  Proof.
    exact: GPS_nFWP_Writer_Reader_join.
  Qed.

  Lemma GPS_nFWP_writer_exclusive γ1 γ2 l s1 s2 q1 q2 :
    vGPS_nFWP γ1 l s1 q1 ∗ vGPS_nFWP γ2 l s2 q2 ⊢ False.
  Proof.
    iStartProof (uPred _). iIntros (V) "oW". rewrite vGPS_nFWP_eq fractor_join_l. 
    iDestruct (fractor_drop with "oW") as (X) "[oW1 oW2]".
    iDestruct "oW1" as (γ_x1) "[% oW1]".
    iDestruct "oW2" as (γ_x2) "[% oW2]".
    subst X. apply encode_inj in H0. inversion H0.
    iDestruct "oW1" as (???) "(_ & oW1 & _)".
    iDestruct "oW2" as (???) "(_ & oW2 & _)".
    iApply (Writer_Exclusive with "[$oW1 $oW2]").
  Qed.

  Lemma GPS_nFWP_join_writer_2 γ1 γ2 l s1 s2 q1 q2:
    vGPS_nFWP γ1 l s1 q1 ∗ vGPS_nFRP γ2 l s2 q2 ⊢ vGPS_nFWP γ1 l s1 (q1 + q2).
  Proof.
    iStartProof (uPred _). iIntros (V) "oW".
    iDestruct (GPS_nFWP_gname_agree with "oW") as %?.
    subst γ2. iApply (GPS_nFWP_join_writer with "oW").
  Qed.

  Global Instance GPS_nFRP_fractional γ l s:
    Fractional (vGPS_nFRP γ l s).
  Proof.
    intros q1 q2. iStartProof (uPred _). iIntros (V). rewrite vGPS_nFRP_eq. iSplit.
    - iIntros "oR". iApply fractor_splitjoin.
      iApply (fractor_mono with "[$oR]").
      iIntros (?) "#?". by iSplit.
    - iIntros "(oR1 & oR2)".
      iCombine "oR1" "oR2" as "oR".
      rewrite fractor_splitjoin; last auto.
      iApply (fractor_mono with "[$oR]"). by iIntros (X) "[R1 R2]".
  Qed.

  Global Instance GPS_nFRP_exists_fractional l γ:
    Fractional (λ q, (∃ s, vGPS_nFRP γ l s q)%I).
  Proof.
    intros q1 q2. iSplit.
    - iIntros "oR". iDestruct "oR" as (s) "oR".
      rewrite GPS_nFRP_fractional. iDestruct "oR" as "[oR1 oR2]".
      iSplitL "oR1"; by iExists _.
    - iIntros "(oR1 & oR2)".
      iDestruct "oR1" as (s1) "oR1".
      iDestruct "oR2" as (s2) "oR2".
      iPoseProof (GPS_nFRP_join_l γ l s1 s2 q1 q2 with "[$oR1 $oR2]") as "oR".
      by iExists _.
  Qed.

  Lemma GPS_nFWP_dealloc γ l s (E: coPset) (HNl: ↑fracN .@ l ⊆ E):
    vGPS_nFWP γ l s 1%Qp ={E}=∗ own_loc l.
  Proof.
    iStartProof (uPred _). iIntros (V) "FP".
    rewrite vGPS_nFWP_eq.
    iMod (fractor_dealloc with "FP") as (n i) "(Info & WR & Inv)"; first auto.
    iDestruct "WR" as (γ_x) "(% & WA)".
    iDestruct "Inv" as (γ_x') "(>% & oI)".
    subst n. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "oI" as (ζ' h e') "(>oA & _ & >Hist & ? & ? & ? & ? & >% & ?)".

    iDestruct "WA" as (v Vr ζ) "(% & _ & _ & RA & _)".
    iDestruct (Reader_Auth_sub with "[$oA $RA]")
      as %Inζ%elem_of_subseteq_singleton.

    iExists h, i. iFrame "Hist Info". rewrite valloc_local_eq; iPureIntro.
    exists (VInj v), Vr. repeat split; [|auto|auto].
    rewrite -H elem_of_map_gset. by exists (s, (v, Vr)).
  Qed.

  Lemma GPS_nFRP_exists_mult_joinL l (n: positive) q γ:
       ([∗ list] _ ∈ seq 0 (Pos.to_nat n), ∃ s, vGPS_nFRP γ l s q)
    ⊢ ∃ s, vGPS_nFRP γ l s (Pos2Qp n * q).
  Proof.
    rewrite -(big_sepL_mult_fractional
                (Φ := (λ q, (∃ s, vGPS_nFRP γ l s q)%I))); last lia.
    by rewrite /bi_exist /= /sbi_exist /= monPred_exist_eq (Pos2Nat.id n).
  Qed.

  Lemma GPS_nFRP_join γ l s q1 q2:
    vGPS_nFRP γ l s q1 ∗ vGPS_nFRP γ l s q2 ⊢ vGPS_nFRP γ l s (q1 + q2).
  Proof. by rewrite fractional. Qed.

  Lemma GPS_nFRP_split γ l s q1 q2:
     vGPS_nFRP γ l s (q1 + q2) ⊢ vGPS_nFRP γ l s q1 ∗ vGPS_nFRP γ l s q2.
  Proof. by rewrite fractional. Qed.

  Lemma GPS_nFRP_mult_splitL γ l s (n: positive) q:
    vGPS_nFRP γ l s (Pos2Qp n * q)
      ⊢ [∗ list] _ ∈ seq 0 (Pos.to_nat n), vGPS_nFRP γ l s q.
  Proof. rewrite -{1}(Pos2Nat.id n) big_sepL_mult_fractional; [auto|lia]. Qed.

  Lemma GPS_nFRP_mult_joinL γ l s (n: positive) q:
    ([∗ list] _ ∈ seq 0 (Pos.to_nat n), vGPS_nFRP γ l s q)
      ⊢ vGPS_nFRP γ l s (Pos2Qp n * q).
  Proof. rewrite -{2}(Pos2Nat.id n) big_sepL_mult_fractional; [auto|lia]. Qed.

  Lemma GPS_nFWP_mult_splitL_1 γ l s n:
    vGPS_nFWP γ l s 1
    ⊢ vGPS_nFWP γ l s (1/(n+1))
      ∗ [∗ list] _ ∈ seq 0 (Pos.to_nat n), vGPS_nFRP γ l s (1/(n+1)).
  Proof.
    rewrite {1}(Pos2Qp_plus_mult n).
    iIntros "FWP".
    iDestruct (GPS_nFWP_Writer_Reader γ l s (1 / (n + 1)) (Pos2Qp n * (1 / (n + 1))) with "FWP") as "[$ FRP]".
    by iApply GPS_nFRP_mult_splitL.
  Qed.

  Lemma GPS_nFWP_ExWrite γ l s s' q v (E : coPset)
      (P: vPred) (Q : pr_state Prtcl → vPred)
      (Le: s ⊑ s') (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ v', P ∗ ▷ (F γ l s v')
            ={E ∖ ↑fracN .@ l}=∗ Q s' ∗ ▷ (F γ l s' v) ∗ ▷ F_past γ l s' v)
      ∗ P ∗ vGPS_nFWP γ l s q }}}
    ([ #l ]_at <- #v) @ E
  {{{ RET #() ; vGPS_nFWP γ l s' q ∗ Q s' }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq /=.
    iIntros (? V) "(#kI & VS & oP & oW)"; iIntros (V1 ?) "Post"; iIntros (V2 ??) "kS".
    rewrite vGPS_nFWP_eq.
    iMod (fractor_open with "[$oW]")
      as (X) "(gpsRaw & WRaw & HClose)"; first auto.
    iDestruct "WRaw" as (γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & #oR & %)".

    assert (V_x ⊑ V2) by (do 2 (etrans; eauto)).
    iApply (RawProto_Write_ex (IP γ) l s s'
            with "[$oI $kI $kS $ex $oR $wTok]");
      [solve_ndisj|auto|auto|auto|..].
    iNext.
    iIntros (V') "(% & kS' & oI & oF & ex & #oR' & wTok & #Max)".
    assert (V1 ⊑ V') by (etrans; eauto).
    iFrame; iApply ("Post" $! V' with "[%]"); first done.
    assert (V ⊑ V') by (etrans; eauto).
    iMod ("VS" $! _ V' with "[%] [$oP $oF]") as "(oQ & oF & oFp)"; first done.
    iFrame "oQ". iApply ("HClose"). iSplitL "oI oF oFp".
    - iExists γ_x. iSplitL ""; first auto.
      iApply ("oI" with "[$oF $oFp]").
    - iExists γ_x. iSplitL ""; first auto.
      iExists v, V', _. by iFrame "wTok ex oR' Max".
  Qed.

  Lemma GPS_nFRP_PrtSeen γ l s q:
    vGPS_nFRP γ l s q ⊢ PrtSeen γ s.
  Proof.
    iStartProof (uPred _). iIntros (V).
    iIntros "oR". rewrite vGPS_nFRP_eq. iDestruct "oR" as (X) "[oR ?]".
    iDestruct "oR" as (γ_x) "[? H]". iDestruct "H" as (V') "H". eauto.
  Qed.

  Lemma GPS_nFWP_PrtSeen γ l s q:
    vGPS_nFWP γ l s q ⊢ PrtSeen γ s.
  Proof.
    iStartProof (uPred _). iIntros (V) "oR". rewrite vGPS_nFWP_eq.
    iDestruct "oR" as (X) "[oR ?]". iDestruct "oR" as (γ_x) "[? WA]".
    iDestruct "WA" as (v V_0 ζ) "(vS & oW & _ & oW3 & oW4)".
    iExists V_0, v. iFrame "vS oW3".
  Qed.

  Lemma GPS_nFRP_PrtSeen_update γ l q s1 s2 (Le: s1 ⊑ s2):
    vGPS_nFRP γ l s1 q ∗ PrtSeen γ s2 ⊢ vGPS_nFRP γ l s2 q.
  Proof.
    iStartProof (uPred _). iIntros (V) "[oR oS]". rewrite vGPS_nFRP_eq.
    iDestruct "oR" as (X) "[oR Inv]". iDestruct "oR" as (γ_x) "[oX oS1]".
    iExists X. iFrame "Inv". iExists γ_x. by iSplit.
  Qed.

  Lemma GPS_nFRP_Read' l γ q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ∗ vGPS_nFRP γ l s' q ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past γ l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
      ∗ P ∗ vGPS_nFRP γ l s q  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq /=.
    iIntros (? V) "(#kI & VS & VSDup & oP & oR)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    setoid_rewrite H1; setoid_rewrite H2.
    rewrite vGPS_nFRP_eq. iMod (fractor_open with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.

    iDestruct "RRaw" as (γ_x) "(% & RA)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "RA" as (V_r v_r) "(% & #oR)".

    iApply (RawProto_Read (IP γ) l P (fun s v => |={E ∖ ↑fracN.@l, E}=> Q s v)%I s _ _ V_r v_r γ γ_x
            with "[VS $oI $kI $kS $oP $oR VSDup HClose]"); [solve_ndisj|auto|..].
    { iSplitR "VSDup"; last by iSpecialize ("VSDup" $! _).
      iIntros "!>" (s' j ? Hs').
      iDestruct ("VS" $! _ Hs') as "[VS|VS]"; last by auto with iFrame.
      iLeft; iIntros (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (fun l _ p => ReaderSWnRaw γ s' j0 l p) with "[oI oR']").
      { iSplit.
        + iNext. iExists γ_x. by iFrame "oI".
        + iExists γ_x. iSplit; first done. iExists Vr, v. by iFrame "oR'". }
      iMod ("VS" $! _ _ with "[%] [$]") as "$"; first solve_jsl.
      iApply fupd_intro_mask'. solve_ndisj. }
    iIntros "!>" (s' v Vr V') "(% & kS' & Hs' & oQ & oR' & % & %)".
    iFrame; iApply ("Post" $! _ _ _ with "[%]"); first done.
    by iFrame "Hs'".
  Qed.

  Lemma GPS_nFRP_Read l γ q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ={E ∖ ↑fracN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past γ l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
      ∗ P ∗ vGPS_nFRP γ l s q  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ vGPS_nFRP γ l s' q ∗ Q s' v }}}.
  Proof.
    intros.
    iIntros "(#kI & VS & VSDup & oP & oR) Post".
    iApply (GPS_nFRP_Read' _ _ _ P with "[$kI VS $VSDup $oP $oR]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_nFRP_Read_PrtSeen_abs'
    l γ q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s1 s2 V (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s1 ⊑ s' ∧ s2 ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ∗ ⎡vGPS_nFRP γ l s1 q V⎤ ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s1 ⊑ s' ∧ s2 ⊑ s'⌝ ∗ F_past γ l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
      ∗ P ∗ monPred_in V ∗ ⎡vGPS_nFRP γ l s1 q V⎤ ∗ PrtSeen γ s2  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s1 ⊑ s' ∧ s2 ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq vGPS_nFRP_eq /=.
    iIntros (? V'); iIntros "(#kI & VS & VSDup & oP & % & oR & oS2)" (Vb ?) "Post";
    iIntros (Vc ??) "kS".

    iMod (fractor_open with "[$oR]") as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.

    iDestruct "RRaw" as (γ_x) "(% & oS1)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H1. inversion H1. subst γ_x'. clear H1.

    iDestruct "oS1" as (V1 v1) "(% & #oR1)".
    iDestruct "oS2" as (V2 v2) "(% & #oR2)".

    iMod (Reader_singleton_agree _ _ _ _ s1 v1 V1 s2 with "[$oI oR1 $oR2 //]") as "[oI %]".
    destruct H2 as [Hs2|Hs1].
    - iApply (RawProto_Read (IP γ) l P (fun s v => |={E ∖ ↑fracN .@ l, E}=> Q s v)%I s2 _ _ V2 v2 γ γ_x
                with "[$oI $kI $kS $oP $oR2 VS VSDup HClose]"); [solve_ndisj|solve_jsl|..].
      { iSplitR "VSDup".
        - iIntros (s' j ??).
          iDestruct ("VS" $! _ with "[%]") as "[VS|VS]"; last by auto with iFrame.
          { split; solve_jsl. }
          iLeft; iIntros "!>" (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
          iMod ("HClose" $! (fun l _ => ReaderSWnRaw γ s1 V l) with "[oI oR']").
          { iSplitL "oI".
            + iNext. iExists γ_x. iFrame; auto.
            + iExists γ_x. iSplit; first done. iExists V1, v1. by iFrame "oR1". }
          iMod ("VS" $! _ _ with "[%] [-]") as "$"; first solve_jsl.
          { iFrame. }
          iApply fupd_intro_mask'. solve_ndisj.
        - iIntros "!>" (s' v ??) "[% Fp]".
          iApply ("VSDup" $! _ s' v _ with "[%] [$Fp]"); first reflexivity.
          iPureIntro. split; [by etrans|auto]. }
      iNext. iIntros (s' v Vr V'') "(% & kS & % & Q & oR' & % & %)".
      iFrame; iApply ("Post" $! s' _ V'' with "[%]"); [solve_jsl|].
      iMod "Q" as "$".
      iPureIntro; split; [by etrans|auto].
    - iApply (RawProto_Read (IP γ) l P (fun s v => |={E ∖ ↑fracN .@ l, E}=> Q s v)%I s1 _ _ V1 v1 γ γ_x
                with "[$oI $kI $kS $oP $oR1 VS VSDup HClose]"); [solve_ndisj|solve_jsl|..].
      { iSplitR "VSDup".
        - iIntros (s' j ??).
          iDestruct ("VS" $! _ with "[%]") as "[VS|VS]"; last by auto with iFrame.
          { split; solve_jsl. }
          iNext. iLeft; iIntros (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
          iMod ("HClose" $! (fun l _ => ReaderSWnRaw γ s1 V l) with "[oI oR']").
          { iSplitL "oI".
            + iNext. iExists γ_x. iFrame; auto.
            + iExists γ_x. iSplit; first done. iExists V1, v1. by iFrame "oR1". }
          iMod ("VS" $! _ _ with "[%] [-]") as "$"; first solve_jsl.
          { iFrame. }
          iApply fupd_intro_mask'. solve_ndisj.
        - iIntros "!>" (s' v ??) "[% Fp]".
          iApply ("VSDup" $! _ s' v _ with "[%] [$Fp]"); first reflexivity.
          iPureIntro. split; solve_jsl. }
      iNext. iIntros (s' v Vr V'') "(% & kS & % & Q & oR' & % & %)".
      iFrame; iApply ("Post" $! _ _ V'' with "[%]").
      { solve_jsl. }
      iMod "Q"; iFrame.
      iPureIntro; split; solve_jsl.
  Qed.

  Lemma GPS_nFRP_Read_PrtSeen_abs l γ q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s1 s2 V (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  ⎡PSCtx⎤ ⊢ {{{ (▷ ∀ s', ⌜s1 ⊑ s' ∧ s2 ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ={E ∖ ↑fracN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s1 ⊑ s' ∧ s2 ⊑ s'⌝ ∗ F_past γ l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
      ∗ P ∗ monPred_in V ∗ ⎡vGPS_nFRP γ l s1 q V⎤ ∗ PrtSeen γ s2  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s1 ⊑ s' ∧ s2 ⊑ s'⌝ ∗ ⎡vGPS_nFRP γ l s1 q V⎤ ∗ Q s' v }}}.
  Proof.
    iIntros "#kI !#" (?) "(VS & H) Post".
    iApply (GPS_nFRP_Read_PrtSeen_abs' _ _ _ P with "[$kI VS $H]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_nFRP_Read_abs' l γ q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s V (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ∗ ⎡vGPS_nFRP γ l s q V⎤ ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past γ l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
      ∗ P ∗ monPred_in V ∗ ⎡vGPS_nFRP γ l s q V⎤  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite vGPS_nFRP_eq WP_Def.wp_eq /=.
    iIntros (? V'); iIntros "(#kI & VS & VSDup & oP & % & oR)" (Vb ?) "Post";
      iIntros (Vc ??) "kS".

    iMod (fractor_open with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.

    iDestruct "RRaw" as (γ_x) "(% & oS)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H1. inversion H1. subst γ_x'. clear H1.

    iDestruct "oS" as (V1 v1) "(% & #oR)".

    iApply (RawProto_Read (IP γ) l P (fun s v => |={E ∖ ↑fracN .@ l, E}=> Q s v)%I s _ _ V1 v1 γ γ_x
                with "[$oI $kI $kS $oP $oR VS VSDup HClose]"); [solve_ndisj|auto|..].
    { solve_jsl. }
    { iSplitR "VSDup"; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ??).
      iDestruct ("VS" $! _ with "[% //]") as "[VS|VS]"; last by auto with iFrame.
      iLeft; iIntros "!>"(v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (fun l _ => ReaderSWnRaw γ s V l) with "[oI oR']").
      { iSplitL "oI".
        + iNext. iExists γ_x. iFrame; auto.
        + iExists γ_x. iSplit; first done. iExists V1, v1. by iFrame "oR". }
      iMod ("VS" $! _ _ with "[%] [-]") as "$"; first solve_jsl.
      { iFrame. }
      iApply fupd_intro_mask'. solve_ndisj. }
      iNext. iIntros (s' v Vr V'') "(% & kS & % & Q & oR' & % & %)".
      iFrame; iApply ("Post" $! _ _ V'' with "[%]"). { solve_jsl. }
      iMod "Q"; iFrame; auto.
  Qed.

  Lemma GPS_nFRP_Read_abs l γ q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s V (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  ⎡PSCtx⎤ ⊢ {{{ (▷ ∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ={E ∖ ↑fracN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past γ l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
      ∗ P ∗ monPred_in V ∗ ⎡vGPS_nFRP γ l s q V⎤ }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ ⎡vGPS_nFRP γ l s q V⎤ ∗ Q s' v }}}.
  Proof.
    iIntros "#kI !#" (?) "(VS & H) Post".
    iApply (GPS_nFRP_Read_abs' _ _ _ P with "[$kI VS $H]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.
End Gname_StrongSW.

Section Gname_StrongSW_2.
  Context {Σ : gFunctors} {Prtcl : protocolT}
          `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{CInvG : cinvG Σ}
          {IP1 IP2 : gname → interpC Σ Prtcl}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (vProp Σ).

  Lemma GPS_nFWP_writer_exclusive2 γ1 γ2 l s1 s2 q1 q2 V:
    vGPS_nFWP (IP := IP1) γ1 l s1 q1 V ∗ vGPS_nFWP (IP := IP2) γ2 l s2 q2 V
    ⊢ False.
  Proof.
    rewrite !vGPS_nFWP_eq fractor_join_l. iIntros "oW".
    iDestruct (fractor_drop with "oW") as (X) "[oW1 oW2]".
    iDestruct "oW1" as (γ_x1) "[% oW1]".
    iDestruct "oW2" as (γ_x2) "[% oW2]".
    subst X. apply encode_inj in H0. inversion H0.
    iDestruct "oW1" as (???) "(_ & oW1 & _)".
    iDestruct "oW2" as (???) "(_ & oW2 & _)".
    iApply (Writer_Exclusive with "[$oW1 $oW2]").
  Qed.

End Gname_StrongSW_2.

Section Gname_StrongSW_Param.
  Context {Σ : gFunctors} {Prtcl : protocolT}
          `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{persG : persistorG Σ}
          `{agreeG : gps_agreeG Σ Prtcl}
          `{pC: Countable Param}
          {IP : gname → interpC Σ Prtcl}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (@vProp Σ).

  Notation F γ := (IP γ true).
  Notation F_past γ := (IP γ false).

  Implicit Types
      (l : loc)
      (s : pr_state Prtcl)
      (ζ : gset (state_sort Prtcl))
      (S : pr_state Prtcl * (Z * View))
      (v : Z)
      (V : View)
      (h : History)
      (p : Param).

  Local Open Scope I.
  Definition gpsSWpnRaw p γ l n : iProp :=
    ∃ (γ_x: gname),
      ⌜n = encode (p,γ,γ_x)⌝ ∗ gps_inv (IP γ) l γ γ_x.

  Definition ReaderSWpnRaw p γ s l n : vProp Σ :=
    ∃ (γ_x: gname),
      ⌜n = encode (p,γ,γ_x)⌝ ∗ PrtSeen γ s.

  Definition WriterSWpnRaw p γ s l n : vProp Σ :=
    ∃ (γ_x: gname),
      ⌜n = encode (p,γ,γ_x)⌝ ∗ WAEx l γ γ_x s.

  Program Definition vGPS_nRSP_def γ l p s : vProp Σ :=
    MonPred (λ V, persisted l (gpsSWpnRaw p γ) (λ l X, ReaderSWpnRaw p γ s l X V)) _.
  Next Obligation. solve_proper. Qed.
  Definition vGPS_nRSP_aux : seal vGPS_nRSP_def. by eexists. Qed.
  Definition vGPS_nRSP := unseal vGPS_nRSP_aux.
  Definition vGPS_nRSP_eq : vGPS_nRSP = _ := seal_eq _.

  Global Instance vGPS_nRSP_persistent γ l p s
    : Persistent (vGPS_nRSP γ l p s).
  Proof.
    rewrite /Persistent vGPS_nRSP_eq. iStartProof (uPred _). iIntros (V).
    iApply persistent.
  Qed.

  Program Definition vGPS_nWSP_def γ l p s : vProp Σ :=
    MonPred (λ V, persisted l (gpsSWpnRaw p γ) (λ l X, WriterSWpnRaw p γ s l X V)) _.
  Next Obligation. solve_proper. Qed.
  Definition vGPS_nWSP_aux : seal vGPS_nWSP_def. by eexists. Qed.
  Definition vGPS_nWSP := unseal vGPS_nWSP_aux.
  Definition vGPS_nWSP_eq : vGPS_nWSP = _ := seal_eq _.

  Lemma GPS_nSW_Init_strong l p (s: gname -> pr_state Prtcl) v
      (Q : gname -> vPred) (E : coPset)
      (HEPers : ↑persistN ⊆ E) (HNPhys : ↑physN ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ own_loc l ∗
       (∀ γ, vGPS_nWSP γ l p (s γ) ={E ∖ ↑persist_locN.@l}=∗
              ▷ F γ l (s γ) v ∗ ▷ F_past γ l (s γ) v ∗ Q γ) }}}
    ([ #l ]_na <- #v) @ E
  {{{ γ, RET #() ; Q γ }}}%stdpp.
  Proof.
    iStartProof (uPred _).
    rewrite valloc_local_eq WP_Def.wp_eq vGPS_nWSP_eq /=.
    iIntros (Φ V) "(#kI & #kIp & oL & VS)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    setoid_rewrite H1; setoid_rewrite H2.
    iDestruct "oL" as (h i) "(Alloc & Hist & Info)".

    iApply program_logic.weakestpre.wp_fupd.
    iApply (RawProto_init_strong IP _ s
            with "[$kI $kS $Hist $Alloc]"); first auto.

    iNext.
    iIntros (V' γ γ_x) "(#Le & kS & oI & ex & #oR & oW)".
    iMod (persistor_inv_alloc' E i (encode (p,γ,γ_x)) l
          (gpsSWpnRaw p γ) (λ l X, WriterSWpnRaw p γ (s γ) l X V')
          with "[$Info $kI $kIp]") as "[WP HClose]"; [auto|auto|..].
    iMod ("VS" $! γ _ with "Le [- Post HClose oI kS]") as "(F & Fp & Q)".
    { iApply "WP".
      iExists γ_x. iSplitL ""; first done.
      iExists v, V', {[s γ, (v, V')]}. iFrame "ex oW oR".
      iSplit; first auto. iPureIntro.
      by move => ? /elem_of_singleton ->. }
    iMod ("HClose" with "[F Fp oI]").
    { iExists γ_x. iSplit; [iNext; done|]. iApply "oI". iFrame. }
    iFrame; iApply ("Post" $! _ _ with "Le Q").
  Qed.

  Lemma GPS_nSW_ExWrite l γ p s s' v (E : coPset)
      (P: vPred) (Q : gname → pr_state Prtcl → vPred)
      (Le: s ⊑ s') (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ v', P ∗ F γ l s v' ∗ vGPS_nWSP γ l p s'
            ={E ∖ ↑persist_locN .@ l}=∗ Q γ s' ∗ ▷ (F γ l s' v) ∗ ▷ F_past γ l s' v)
      ∗ P ∗ vGPS_nWSP γ l p s }}}
    ([ #l ]_at <- #v) @ E
  {{{ RET #() ; vGPS_nRSP γ l p s' ∗ Q γ s' }}}%stdpp.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq vGPS_nWSP_eq vGPS_nRSP_eq /=.
    iIntros (? V) "(#kI & VS & oP & oW)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod ((persistor_open') with "[$oW]")
      as (X) "(gpsRaw & WRaw & #HT & HClose)"; first auto.
    iDestruct "WRaw" as (γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & #oR & %)".

    iApply (RawProto_Write_ex (IP γ) l s s' _ _ v_x V_x γ γ_x v _ ζ
            with "[$oI $kI $kS $ex $oR $wTok]");
      [solve_ndisj|auto|auto|auto|..].
    { do 2 (etrans; eauto). }
    iNext.
    iIntros (V') "(% & kS' & oI & oF & ex & #oR' & wTok & #?)".
    setoid_rewrite H1; setoid_rewrite H2; setoid_rewrite H4.
    iFrame; iApply ("Post" $! V' with "[%]"); [done|].
    iMod ("VS" $! _ _ with "[%] [$oP $oF wTok ex]") as "(oQ & oF & oFp)"; [done|..].
    - iApply "HT".
      iExists γ_x. iSplitL ""; first auto.
      iExists v, V',_. iFrame "wTok ex oR'". by iSplit; auto.
    - iFrame "oQ".
      iMod ("HClose" with "[-]").
      iExists γ_x. iSplitL ""; first auto.
      iApply ("oI" with "[$oF $oFp]").
      iModIntro. iApply "HT".
      iExists γ_x. iSplitL ""; first auto.
      iExists V',_. by iFrame "oR'".
  Qed.

  Lemma GPS_nSW_ExRead l γ p s (E : coPset)
      (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ <obj> (∀ v, F_past γ l s v ={E ∖ ↑persist_locN .@ l}=∗
                   F_past γ l s v ∗ F_past γ l s v ) ∗ vGPS_nWSP γ l p s }}}
    ([ #l ]_at) @ E (* TODO: can be non-atomic *)
  {{{ v, RET #v ;
      vGPS_nWSP γ l p s ∗ F_past γ l s v }}}%stdpp.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq /= vGPS_nWSP_eq.
    iIntros (? V) "(#kI & VSDup & oW)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod ((persistor_open) with "[$oW]")
      as (X) "(gpsRaw & WRaw & HClose)"; first auto.
    iDestruct "WRaw" as (γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & #oR & %)".

    iApply (RawProto_Read_ex (IP γ)
              with "[$kI $kS $oI $ex $oR $wTok VSDup]"); [solve_ndisj| |done|..].
    { do 2 (etrans; eauto). }
    { iNext. by iSpecialize ("VSDup" $! _ with ""). }
    iNext.
    iIntros (v V') "{oR} (% & kS & oI & Fp & ex & #oR' & wTok)".
    assert (V1 ⊑ V') by (etrans; eauto).
    iFrame; iApply ("Post" $! _ V' with "[%]"); first done.
    iFrame "Fp".
    iApply "HClose". iSplitL "oI".
    + iNext. iExists γ_x. by iFrame.
    + iExists γ_x. iSplitL ""; first auto.
      iExists v_x, V_x,_. iFrame "wTok ex oR'". iSplit; last auto.
      iPureIntro. do 2 (etrans; eauto).
  Qed.

  Lemma GPS_nSW_Read' l γ p (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ∗ vGPS_nRSP γ l p s' ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P
                                       ={E ∖ ↑persist_locN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past γ l s' v
              ={E ∖ ↑persist_locN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
           ∗ P ∗ vGPS_nRSP γ l p s }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ Q s' v }}}%stdpp.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq /= vGPS_nRSP_eq.
    iIntros (? V) "(#kI & VS & VSDup & oP & oR)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod ((persistor_open) with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first auto.
    iDestruct "RRaw" as (γ_x) "(% & RA)".
    iDestruct "gpsRaw" as (γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ_x'.
    iDestruct "RA" as (V_r v_r) "(% & #oR)".

    assert (V ⊑ V2) by (etrans; eauto).
    iApply (RawProto_Read (IP γ) l P (fun s v => |={E ∖ ↑persist_locN.@l, E}=> Q s v)%I s _ _ V_r v_r γ γ_x
              with "[VS $oI $kI $kS $oP $oR VSDup HClose]"); [solve_ndisj|auto|..].
    { etrans; eauto. }
    { iSplitR "VSDup"; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ? Hs').
      iDestruct ("VS" $! _ Hs') as "[VS|VS]"; last by auto with iFrame.
      iLeft; iIntros "!>" (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (λ l X, ReaderSWpnRaw p γ s' l X j0) with "[oI oR']").
      { iSplit.
        + iNext. iExists γ_x. by iFrame "oI".
        + iExists γ_x. iSplit; first done. iExists Vr, v. by iFrame "oR'". }
      iMod ("VS" $! _ _ with "[%] [$]") as "$"; [solve_jsl|].
      iApply fupd_intro_mask'. solve_ndisj. }
    iNext.
    iIntros (s' v Vr V') "(% & kS' & Hs' & >oQ & oR' & % & %)".
    iFrame; iApply ("Post" $! _ _ _ with "[%]"); [solve_jsl|auto].
  Qed.

  Lemma GPS_nSW_Read l γ p (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past γ l s' v ∗ P ={E ∖ ↑ persist_locN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F γ l s'' v ∗ P
                                       ={E ∖ ↑persist_locN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past γ l s' v
              ={E ∖ ↑persist_locN .@ l}=∗ F_past γ l s' v ∗ F_past γ l s' v )
           ∗ P ∗ vGPS_nRSP γ l p s }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ vGPS_nRSP γ l p s' ∗ Q s' v }}}%stdpp.
  Proof.
    intros.
    iIntros "(#kI & VS & H) Post".
    iApply (GPS_nSW_Read' _ _ _ P with "[$kI VS $H]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_nSW_Writer_Reader γ l p s:
    vGPS_nWSP γ l p s ⊢ vGPS_nWSP γ l p s ∗ vGPS_nRSP γ l p s.
  Proof.
    iStartProof (uPred _). iIntros (V) "oW". rewrite vGPS_nWSP_eq vGPS_nRSP_eq.
    iDestruct "oW" as (X) "[oW #?]".
    iDestruct "oW" as (γ_x) "[#? WA]".
    iDestruct "WA" as (v V0 ζ) "(#? & ? & ? & #oR & #?)".
    iSplitR ""; iExists X; (iSplitR ""; last auto);
      iExists γ_x; (iSplitL ""; first auto).
    - iExists v, V0, ζ. iFrame. iSplit; [auto|by iSplitL ""].
    - iExists V0, v. by iFrame "oR".
  Qed.

  Lemma GPS_nSW_nFRP_clash `{CInvG : cinvG Σ} {IP2 : gname → interpC Σ Prtcl}
    γ l p s1 s2 q (E : coPset)
    (HNl1: ↑persist_locN .@ l ⊆ E) (HNl2: ↑fracN .@ l ⊆ E):
    vGPS_nRSP γ l p s1 -∗ vGPS_nFRP (IP:=IP2) γ l s2 q ={E}=∗ False.
  Proof.
    rewrite vGPS_nFRP_eq vGPS_nRSP_eq. iStartProof (uPred _).
    iIntros (?) "oR1 % -> oR2".
    iMod (persistor_open with "oR1") as (X1) "(oI1 & oR1 & HClose1)";
      first auto.
    iMod (fractor_open with "oR2")
      as (X2) "(oI2 & oR2 & HClose2)"; first solve_ndisj.
    iDestruct "oI1" as (?) "(_ & oI1)".
    iDestruct "oI1" as (???) "(_&_&>oH1&_)".
    iDestruct "oI2" as (?) "(_ & oI2)".
    iDestruct "oI2" as (???) "(_&_&>oH2&_)".
    by iDestruct (Hist_Exclusive with "[$oH1 $oH2]") as %?.
  Qed.

  Lemma GPS_nSW_Reader_PrtSeen γ l p s:
    vGPS_nRSP γ l p s ⊢ PrtSeen γ s.
  Proof.
    iStartProof (uPred _). iIntros (V) "oR". rewrite vGPS_nRSP_eq.
    iDestruct "oR" as (X) "[oR ?]". iDestruct "oR" as (γ_x) "[? $]".
  Qed.

  Lemma GPS_nSW_Reader_PrtSeen_update γ l p s1 s2 (Le: s1 ⊑ s2):
    vGPS_nRSP γ l p s1 ∗ PrtSeen γ s2 ⊢ vGPS_nRSP γ l p s2.
  Proof.
    rewrite vGPS_nRSP_eq. iStartProof (uPred _). iIntros (V) "[oR oS]".
    iDestruct "oR" as (X) "[oR Inv]". iDestruct "oR" as (γ_x) "[oX oS1]".
    iExists X. iFrame "Inv". iExists γ_x. iFrame "oX oS".
  Qed.
End Gname_StrongSW_Param.


Section Gname_StrongSW_Param_Default.
  Context {Σ : gFunctors} {Prtcl : protocolT}
          `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
          `{persG : persistorG Σ}
          `{agreeG : gps_agreeG Σ Prtcl}
          {IP : gname → interpC Σ Prtcl}.
  Set Default Proof Using "Type".

  Notation F γ := (IP γ true).
  Notation F_past γ := (IP γ false).

  Set Default Proof Using "Type".

  Lemma GPS_nSW_Init_default l (s: gname -> pr_state Prtcl) v
      (Q : gname -> vProp _) (E : coPset)
      (HEPers : ↑persistN ⊆ E) (HNPhys : ↑physN ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ own_loc l ∗
       (∀ γ, vGPS_nWSP (IP:=IP) γ l () (s γ) ={E ∖ ↑persist_locN.@l}=∗
              ▷ F γ l (s γ) v ∗ ▷ F_past γ l (s γ) v ∗ Q γ) }}}
    ([ #l ]_na <- #v) @ E
  {{{ γ, RET #() ; Q γ }}}.
  Proof. by apply GPS_nSW_Init_strong. Qed.

End Gname_StrongSW_Param_Default.

Lemma GPS_nSWs_param_agree
  {Σ : gFunctors} {Prtcl : protocolT}
  `{fG : !foundationG Σ} `{GPSG : !gpsG Σ Prtcl PF}
  `{persG : persistorG Σ}
  `{agreeG : gps_agreeG Σ Prtcl}
  `{pC: Countable Param}
  {IP1 IP2 : gname → interpC Σ Prtcl} l γ1 γ2 p1 p2 s1 s2 V:
  vGPS_nRSP (IP:=IP1) γ1 l p1 s1 V
  ∗ vGPS_nRSP (IP:=IP2) γ2 l p2 s2 V ⊢ ⌜p1 = p2 ∧ γ1 = γ2⌝.
Proof.
  iIntros "RP". rewrite !vGPS_nRSP_eq persistor_join_l persistor_drop.
  iDestruct "RP" as (X) "[oR1 oR2]".
  iDestruct "oR1" as (γx1) "(% & _)".
  iDestruct "oR2" as (γx2) "(% & _)".
  rewrite H0 in H. apply encode_inj in H. by inversion H.
Qed.

Section SingleWriter.
  Context {Σ : gFunctors} {Prtcl : protocolT} `{fG : !foundationG Σ}
          `{GPSG : !gpsG Σ Prtcl PF}
          `{persG : persistorG Σ}
          `{agreeG : gps_agreeG Σ Prtcl}
          `{CInvG : cinvG Σ}
          {IP : interpC Σ Prtcl}.
  Set Default Proof Using "Type".
  Local Notation iProp := (iProp Σ).
  Local Notation vPred := (vProp Σ).

  Notation F := (IP true).
  Notation F_past := (IP false).

  Implicit Types
      (l : loc)
      (s : pr_state Prtcl)
      (ζ : gset (state_sort Prtcl))
      (S : pr_state Prtcl * (Z * View))
      (v : Z)
      (V : View)
      (h : History).

  Implicit Types (e : state_sort Prtcl).

  Local Open Scope I.
  Definition gpsSWRaw l n : iProp :=
    ∃ (γ γ_x: gname),
      ⌜n = encode (γ,γ_x)⌝ ∗ gps_inv IP l γ γ_x.

  Definition ReaderSWRaw s l n : vProp Σ :=
    ∃ (γ γ_x: gname),
      ⌜n = encode (γ,γ_x)⌝ ∗ PrtSeen γ s.

  Definition WriterSWRaw s l n : vProp Σ :=
    ∃ (γ γ_x: gname),
      ⌜n = encode (γ,γ_x)⌝ ∗ WAEx l γ γ_x s.
  Local Close Scope I.

  Program Definition vGPS_RSP_def l s : vProp Σ :=
    MonPred (λ V, persisted l gpsSWRaw (λ l X, ReaderSWRaw s l X V)) _.
  Next Obligation. solve_proper. Qed.
  Definition vGPS_RSP_aux : seal vGPS_RSP_def. by eexists. Qed.
  Definition vGPS_RSP := unseal vGPS_RSP_aux.
  Definition vGPS_RSP_eq : vGPS_RSP = _ := seal_eq _.

  Global Instance vGPS_RSP_persistent l s : Persistent (vGPS_RSP l s).
  Proof.
    rewrite /Persistent /bi_persistently vGPS_RSP_eq. iStartProof (uPred _).
    iIntros (V). iApply persistent.
  Qed.

  Program Definition vGPS_WSP_def l s : vProp Σ :=
    MonPred (λ V, persisted l gpsSWRaw (λ l X, WriterSWRaw s l X V)) _.
  Next Obligation. solve_proper. Qed.
  Definition vGPS_WSP_aux : seal vGPS_WSP_def. by eexists. Qed.
  Definition vGPS_WSP := unseal vGPS_WSP_aux.
  Definition vGPS_WSP_eq : vGPS_WSP = _ := seal_eq _.

  Program Definition vGPS_FRP_def l s (q: frac) : vProp Σ :=
    MonPred (λ V, fractored l gpsSWRaw (λ l q X, ReaderSWRaw s l X V) q) _.
  Next Obligation. solve_proper. Qed.
  Definition vGPS_FRP_aux : seal vGPS_FRP_def. by eexists. Qed.
  Definition vGPS_FRP := unseal vGPS_FRP_aux.
  Definition vGPS_FRP_eq : vGPS_FRP = _ := seal_eq _.

  Program Definition vGPS_FWP_def l s (q: frac) : vProp Σ :=
    MonPred (λ V, fractored l gpsSWRaw (λ l q X, WriterSWRaw s l X V) q) _.
  Next Obligation. solve_proper. Qed.
  Definition vGPS_FWP_aux : seal vGPS_FWP_def. by eexists. Qed.
  Definition vGPS_FWP := unseal vGPS_FWP_aux.
  Definition vGPS_FWP_eq : vGPS_FWP = _ := seal_eq _.

  Local Instance :
    Frame false (gps_inv IP l γ γ_x) (gpsSWRaw l (encode (γ, (γ_x)))) True.
  Proof.
    intros. unfold Frame, bi_intuitionistically_if.
    iIntros "[H _]". iExists _, _. by iFrame "∗".
  Qed.
  Local Close Scope I.

  Lemma GPS_SW_Readers_agree l s1 s2 (E : coPset) (HE: ↑persist_locN .@ l ⊆ E):
    vGPS_RSP l s1 ∗ vGPS_RSP l s2 ={E}=∗ ⌜s1 ⊑ s2 ∨ s2 ⊑ s1⌝.
  Proof.
    iStartProof (uPred _).
    iIntros (V) "#RP". rewrite vGPS_RSP_eq.
    iDestruct (persistor_splitjoin with "RP") as "H".
    iMod (persistor_open with "H") as (X) "(oI & (oR1 & oR2) & HClose)";
      first auto.
    iDestruct "oR1" as (γ1 γ_x1) "(% & #RA1)".
    iDestruct "oR2" as (γ2 γ_x2) "(% & #RA2)".
    iDestruct "oI" as (γ γ_x) "(>% & oI)". subst.
    apply encode_inj in H0. inversion H0. subst.
    apply encode_inj in H1. inversion H1. subst.
    iDestruct "RA1" as (V1 v1) "(? & #oR1)".
    iDestruct "RA2" as (V2 v2) "(? & #oR2)".
    iMod (Reader_singleton_agree _ _ _ _ s1 _ _ s2 with "[$oI $oR1 $oR2]") as "[oI %]".
    iMod ("HClose" $!
          (λ l0 X, ReaderSWRaw s1 l0 X V ∗ ReaderSWRaw s2 l0 X V)%I
          with "[oI]") as "oR".
    - iSplitL "oI"; last iSplitL ""; iExists γ, γ_x;
        (iSplitL ""; first auto).
      + by iNext.
      + iExists V1, v1. by iFrame "oR1".
      + iExists V2, v2. by iFrame "oR2".
    - by iPureIntro.
  Qed.

  Lemma GPS_SW_Writer_max l s s' (E : coPset) (HE: ↑persist_locN .@ l ⊆ E):
    vGPS_RSP l s ∗ ▷ vGPS_WSP l s' ={E}=∗ vGPS_WSP l s' ∗ ⌜s ⊑ s'⌝.
  Proof.
    iStartProof (uPred _).
    iIntros (V) "oRW". rewrite vGPS_WSP_eq vGPS_RSP_eq.
    iMod (persistor_join_later_rl with "oRW") as "oRW".
    iMod (persistor_open with "oRW") as (X) "(oI & (oR & oW) & HClose)";
      first auto.
    iDestruct "oR" as (γ1 γ_x1) "(% & #RA)".
    iDestruct "oW" as (γ2 γ_x2) "(>% & >WA)".
    iDestruct "oI" as (γ γ_x) "(>% & oI)". subst.
    apply encode_inj in H0. inversion H0. subst.
    apply encode_inj in H1. inversion H1. subst.

    iDestruct "RA" as (V1 v1) "(? & oR)".
    iDestruct "WA" as (v2 V2 ζ) "(#vS & oW & ex & #oR2 & %)".
    iDestruct (Reader_Writer_sub with "[$oW $oR]")
      as %Ins%elem_of_subseteq_singleton.
    iDestruct (Reader_Writer_sub with "[$oW $oR2]")
      as %Ins'%elem_of_subseteq_singleton.

    assert (Le := H  _ Ins).
    iMod (Writer_sorted with "[$oI $oW]") as "(oI & oW & %)".
    iMod ("HClose" $! (λ l0 X, WriterSWRaw s' l0 X V)%I
          with "[$oI oW ex]") as "oW".
    { iExists γ, γ_x. iSplitL ""; first auto.
      iExists v2, V2, ζ. by iFrame (H) "oW ex oR2 vS". }
    iFrame "oW".
    iPureIntro.
    by apply (H2 _ _ Ins Ins').
  Qed.

  Lemma GPS_SW_Read_ex l s (E : coPset)
      (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ <obj> (∀ v, F_past l s v ={E ∖ ↑persist_locN .@ l}=∗
                   F_past l s v ∗ F_past l s v ) ∗ vGPS_WSP l s }}}
    ([ #l ]_at) @ E
  {{{ v, RET #v ;
      vGPS_WSP l s ∗ F_past l s v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq /= vGPS_WSP_eq.
    iIntros (? V) "(#kI & VSDup & oW)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod ((persistor_open) with "[$oW]")
      as (X) "(gpsRaw & WRaw & HClose)"; first auto.
    iDestruct "WRaw" as (γ γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ' γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & #oR & %)".

    iApply (RawProto_Read_ex IP l s _ _ v_x V_x γ γ_x _ ζ
            with "[$kI $kS $oI $ex $oR $wTok VSDup]"); [solve_ndisj| |done|..].
    { do 2 (etrans; eauto). }
    { iNext. by iSpecialize ("VSDup" $! _ with ""). }
    iNext.
    iIntros (v V') "{oR} (% & kS & oI & Fp & ex & #oR' & wTok)".
    assert (V1 ⊑ V') by (etrans; eauto).
    iFrame; iApply ("Post" $! _ V' with "[%]"); first done.
    iFrame "Fp".
    iApply "HClose". iSplitL "oI".
    + iNext. iExists γ, γ_x. by iFrame.
    + iExists γ, γ_x. iSplitL ""; first auto.
      iExists v_x, V_x,_. iFrame "wTok ex oR'". iSplit; last auto.
      iPureIntro. do 2 (etrans; eauto).
  Qed.

  Lemma GPS_SW_Init l s v (Q : vPred) (E : coPset)
        (HEPers : ↑persistN ⊆ E) (HNPhys : ↑physN ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ own_loc l ∗
       (vGPS_WSP l s ={E ∖ ↑persist_locN.@l}=∗ ▷ F l s v ∗ ▷ F_past l s v ∗ Q) }}}
    ([ #l ]_na <- #v) @ E
  {{{ RET #() ; Q }}}.
  Proof.
    iStartProof (uPred _).
    rewrite valloc_local_eq WP_Def.wp_eq /= vGPS_WSP_eq.
    iIntros (? V) "(#kI & #kIp & oL & VS)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    setoid_rewrite H1; setoid_rewrite H2.
    iDestruct "oL" as (h i) "(Alloc & Hist & Info)".

    iApply program_logic.weakestpre.wp_fupd.
    iApply (RawProto_init_strong (λ _, IP)
            with "[$kI $kS $Hist $Alloc]"); first auto.

    iNext.
    iIntros (V' γ γ_x) "(#Le & kS & oI & ex & #oR & oW)".
    iMod (persistor_inv_alloc' E i (encode (γ,γ_x)) l
          gpsSWRaw (λ l X, WriterSWRaw s l X V')
          with "[$Info $kI $kIp]") as "[WP HClose]"; [auto|auto|..].
    iMod ("VS" $! _ with "Le [- Post HClose oI kS]") as "(F & Fp & Q)".
    { iApply "WP".
      iExists γ, γ_x. iSplitL ""; first done.
      iExists v, V', {[s, (v, V')]}. iFrame "ex oW oR".
      iSplit; first auto. iPureIntro.
      by move => ? /elem_of_singleton ->. }
    iMod ("HClose" with "[F Fp oI]").
    { iExists γ, γ_x. iSplit; [iNext; done|]. iApply "oI". iFrame. }
    iFrame; iApply ("Post" $! _ with "Le Q").
  Qed.

  Lemma GPS_SW_Read' l (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
    {{{ ⎡PSCtx⎤ ∗ ▷ (∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past l s' v ∗ P ∗ vGPS_RSP l s' ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P
                                       ={E ∖ ↑persist_locN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
              ={E ∖ ↑persist_locN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ vGPS_RSP l s }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ Q s' v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq vGPS_RSP_eq /=.
    iIntros (? V) "(#kI & VS & VSDup & oP & oR)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod ((persistor_open) with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first auto.
    iDestruct "RRaw" as (γ γ_x) "(% & RA)".
    iDestruct "gpsRaw" as (γ' γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "RA" as (V_r v_r) "(% & #oR)".

    assert (V ⊑ V2) by (etrans; eauto).
    iApply (RawProto_Read IP l P (fun s v => |={E ∖ ↑persist_locN.@l, E}=> Q s v)%I s _ _ V_r v_r γ γ_x
              with "[VS $oI $kI $kS $oP $oR VSDup HClose]"); [solve_ndisj|auto|..].
    { etrans; eauto. }
    { iSplitR "VSDup"; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ? Hs').
      iDestruct ("VS" $! _ Hs') as "[VS|VS]"; last by auto with iFrame.
      iLeft; iIntros "!>" (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (λ l X, ReaderSWRaw s' l X j0) with "[oI oR']").
      { iSplit.
        + iNext. iExists γ, γ_x. by iFrame "oI".
        + iExists γ, γ_x. iSplit; first done. iExists Vr, v. by iFrame "oR'". }
      iMod ("VS" $! _ _ with "[%] [$]") as "$"; [solve_jsl|].
      iApply fupd_intro_mask'. solve_ndisj. }
    iIntros "!>" (s' v Vr V') "(% & kS' & Hs' & oQ & oR' & % & %)".
    iFrame; iApply ("Post" $! _ _ _ with "[%]").
    { etrans; eauto. }
    by iFrame "Hs'".
  Qed.

  Lemma GPS_SW_Read l (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
    {{{ ⎡PSCtx⎤ ∗ ▷ (∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past l s' v ∗ P ={E ∖ ↑ persist_locN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P
                                       ={E ∖ ↑persist_locN .@ l}=∗ False))
           ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
              ={E ∖ ↑persist_locN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
           ∗ P ∗ vGPS_RSP l s }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ vGPS_RSP l s' ∗ Q s' v }}}.
  Proof.
    intros.
    iIntros "(#kI & VS & VSDup & oP & oR) Post".
    iApply (GPS_SW_Read' _ P with "[$kI VS $VSDup $oP $oR]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_SW_ExWrite l s s' v (E : coPset)
      (P: vPred) (Q : pr_state Prtcl → vPred)
      (Le: s ⊑ s') (HN : ↑physN ⊆ E) (HNl: ↑persist_locN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ v', P ∗ F l s v' ∗ vGPS_WSP l s'
            ={E ∖ ↑persist_locN .@ l}=∗ Q s' ∗ ▷ (F l s' v) ∗ ▷ F_past l s' v)
      ∗ P ∗ vGPS_WSP l s }}}
    ([ #l ]_at <- #v) @ E
  {{{ RET #() ; vGPS_RSP l s' ∗ Q s' }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq vGPS_RSP_eq vGPS_WSP_eq /=.
    iIntros (? V) "(#kI & VS & oP & oW)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod ((persistor_open') with "[$oW]")
      as (X) "(gpsRaw & WRaw & #HT & HClose)"; first auto.
    iDestruct "WRaw" as (γ γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ' γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & #oR & %)".

    iApply (RawProto_Write_ex IP l s s' _ _ v_x V_x γ γ_x v _ ζ
            with "[$oI $kI $kS $ex $oR $wTok]");
      [solve_ndisj|auto|auto|auto|..].
    { do 2 (etrans; eauto). }
    iNext.
    iIntros (V') "(% & kS' & oI & oF & ex & #oR' & wTok & #?)".
    setoid_rewrite H1; setoid_rewrite H2; setoid_rewrite H4.
    iFrame; iApply ("Post" $! V' with "[%]"); [done|].
    iMod ("VS" $! _ _ with "[%] [$oP $oF wTok ex]") as "(oQ & oF & oFp)"; [done|..].
    - iApply "HT".
      iExists γ, γ_x. iSplitL ""; first auto.
      iExists v, V',_. iFrame "wTok ex oR'". by iSplit; auto.
    - iFrame "oQ".
      iMod ("HClose" with "[-]").
      iExists γ, γ_x. iSplitL ""; first auto.
      iApply ("oI" with "[$oF $oFp]").
      iModIntro. iApply "HT".
      iExists γ, γ_x. iSplitL ""; first auto.
      iExists V',_. by iFrame "oR'".
  Qed.

  Lemma GPS_SW_Writer_Reader l s:
    vGPS_WSP l s ⊢ vGPS_WSP l s ∗ vGPS_RSP l s.
  Proof.
    iStartProof (uPred _). iIntros (V) "oW". rewrite vGPS_RSP_eq vGPS_WSP_eq.
    iDestruct "oW" as (X) "[oW #?]".
    iDestruct "oW" as (γ γ_x) "[#? WA]".
    iDestruct "WA" as (v V0 ζ) "(#? & ? & ? & #oR & #?)".
    iSplitR ""; iExists X; (iSplitR ""; last auto);
      iExists γ, γ_x; (iSplitL ""; first auto).
    - iExists v, V0, ζ. iFrame. iSplit; [auto|by iSplitL ""].
    - iExists V0, v. by iFrame "oR".
  Qed.

  Lemma GPS_FRPs_agree l s1 s2 q1 q2 (E : coPset) (HE: ↑fracN .@ l ⊆ E):
    (vGPS_FRP l s1 q1 ∗ vGPS_FRP l s2 q2
      ={E}=∗ ⌜s1 ⊑ s2 ∨ s2 ⊑ s1⌝ ∗ vGPS_FRP l s1 q1 ∗ vGPS_FRP l s2 q2)%stdpp.
  Proof.
    iStartProof (uPred _). rewrite vGPS_FRP_eq.
    iIntros (V) "[oR1 oR2]"; iCombine "oR1 oR2" as "RP".
    iDestruct (fractor_splitjoin l with "RP") as "H".
    iMod (fractor_open with "H") as (X) "(oI & (oR1 & oR2) & HClose)";
      first auto.
    iDestruct "oR1" as (γ1 γ_x1) "(% & #RA1)".
    iDestruct "oR2" as (γ2 γ_x2) "(% & #RA2)".
    iDestruct "oI" as (γ γ_x) "(>% & oI)". subst.
    apply encode_inj in H0. inversion H0. subst.
    apply encode_inj in H1. inversion H1. subst.
    iDestruct "RA1" as (V1 v1) "(? & oR1)".
    iDestruct "RA2" as (V2 v2) "(? & oR2)".
    iMod (Reader_singleton_agree _ _ _ _ s1 _ _ s2 with "[$oI $oR1 $oR2]") as "[oI %]".
    iMod ("HClose" $!
          (λ l0 _ X, ReaderSWRaw s1 l0 X V ∗ ReaderSWRaw s2 l0 X V)%I
          with "[oI]") as "oR".
    - iSplitL "oI"; last iSplitL ""; iExists γ, γ_x;
        (iSplitL ""; first auto).
      + by iNext.
      + iExists V1, v1. by iFrame "oR1".
      + iExists V2, v2. by iFrame "oR2".
    - iDestruct (fractor_splitjoin l _
                   (λ l0 _ X, ReaderSWRaw s1 l0 X V)
                   (λ l0 _ X, ReaderSWRaw s2 l0 X V)
                   with "oR") as "[o1 o2]".
      iModIntro; iSplit; first auto.
      iFrame "o1 o2".
  Qed.

  Lemma GPS_FWP_Init l s v (E : coPset)
    (HEN : ↑physN ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ own_loc l ∗ ▷ F l s v ∗ ▷ F_past l s v }}}
    ([ #l ]_na <- #v) @ E
  {{{ RET #() ; vGPS_FWP l s 1%Qp }}}%stdpp.
  Proof.
    iStartProof (uPred _).
    rewrite valloc_local_eq WP_Def.wp_eq vGPS_FWP_eq /=.
    iIntros (? V) "(#kI & oL & F & Fp)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    setoid_rewrite H1; setoid_rewrite H2.
    iDestruct "oL" as (h i)  "(Alloc & Hist & Info)".
    iApply program_logic.weakestpre.wp_fupd.
    iApply (RawProto_init IP l s v _ h
               with "[$kI $kS $Hist $Alloc $F $Fp]"); first auto.
    iNext.
    iIntros (V' γ γ_x) "(#Le & kS' & oI & ex & #oR & oW)".

    iMod (fractor_alloc E i (encode(γ,γ_x)) l
                        gpsSWRaw (λ l q X, WriterSWRaw s l X V')
           with "[-Post kS']") as "FP"; first auto.
      { iFrame "kI Info". iSplitL "oI"; iExists γ, γ_x.
        - iSplitL ""; [by iNext|done].
        - iSplitL ""; first done.
          iExists v, V', {[s, (v, V')]}. iFrame "ex oW oR".
          iSplit; first auto. iPureIntro.
          by move => ? /elem_of_singleton ->. }
    by iFrame; iApply ("Post" $! V' with "Le [$FP]").
  Qed.

  Lemma GPS_FRP_Read' l q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past l s' v ∗ P ∗ vGPS_FRP l s' q ={E}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
      ∗ P ∗ vGPS_FRP l s q  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ Q s' v }}}%stdpp.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq vGPS_FRP_eq /=.
    iIntros (? V) "(#kI & VS & VSDup & oP & oR)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod (fractor_open with "[$oR]")
      as (X) "(gpsRaw & RRaw & HClose)"; first exact HNl.

    iDestruct "RRaw" as (γ γ_x) "(% & RA)".
    iDestruct "gpsRaw" as (γ' γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "RA" as (V_r v_r) "(% & #oR)".

    assert (V ⊑ V2) by (etrans; eauto).
    iApply (RawProto_Read IP l P (fun s v => |={E ∖ ↑fracN.@l, E}=> Q s v)%I s _ _ V_r v_r γ γ_x
              with "[VS $oI $kI $kS $oP $oR VSDup HClose]"); [solve_ndisj|auto|..].
    { etrans; eauto. }
    { iSplitR "VSDup"; last by iSpecialize ("VSDup" $! _).
      iIntros (s' j ? Hs').
      iDestruct ("VS" $! _ Hs') as "[VS|VS]"; last by auto with iFrame.
      iLeft; iIntros "!>" (v Vr j0 Hle') "(? & ? & % & % & oI & oR')".
      iMod ("HClose" $! (fun l _ X => ReaderSWRaw s' l X j0) with "[oI oR']").
      { iSplit.
        + iNext. iExists γ, γ_x. by iFrame "oI".
        + iExists γ, γ_x. iSplit; first done. iExists Vr, v. by iFrame "oR'". }
      iMod ("VS" $! _ _ with "[%] [$]") as "$"; [solve_jsl|].
      iApply fupd_intro_mask'. solve_ndisj. }
    iIntros "!>" (s' v Vr V') "(% & kS' & Hs' & oQ & oR' & % & %)".
    iFrame; iApply ("Post" $! _ _ _ with "[%]").
    { etrans; eauto. }
    by iFrame "Hs'".
  Qed.

  Lemma GPS_FRP_Read l q (P : vPred) (Q : pr_state Prtcl → Z → vPred)
    s (E : coPset) (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ (▷ ∀ s', ⌜s ⊑ s'⌝ →
              (∀ v, F_past l s' v ∗ P ={E ∖ ↑fracN .@ l}=∗ Q s' v)
            ∨ (∀ s'' v, ⌜s' ⊑ s''⌝ → F l s'' v ∗ P ={E ∖ ↑fracN .@ l}=∗ False))
         ∗ ▷ <obj> (∀ s' v, ⌜s ⊑ s'⌝ ∗ F_past l s' v
            ={E ∖ ↑fracN .@ l}=∗ F_past l s' v ∗ F_past l s' v )
      ∗ P ∗ vGPS_FRP l s q  }}}
    ([ #l ]_at) @ E
  {{{ s' v, RET #v ;
      ⌜s ⊑ s'⌝ ∗ vGPS_FRP l s' q ∗ Q s' v }}}%stdpp.
  Proof.
    intros.
    iIntros "(#kI & VS & VSDup & oP & oR) Post".
    iApply (GPS_FRP_Read' _ _ P with "[$kI VS $VSDup $oP $oR]"); [done | done | | done].
    iNext. iIntros (s') "Hs"; iDestruct ("VS" with "Hs") as "[VS|VS]"; last by iRight.
    iLeft; iIntros (v) "(? & ? & ?)"; iFrame.
    iApply (fupd_mask_mono _ _ (Q s' v)); last by iApply "VS"; iFrame.
    solve_ndisj.
  Qed.

  Lemma GPS_FWP_ExRead l s q (E : coPset)
      (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ <obj> (∀ v, F_past l s v ={E ∖ ↑fracN .@ l}=∗
                   F_past l s v ∗ F_past l s v ) ∗ vGPS_FWP l s q }}}
    ([ #l ]_at) @ E
  {{{ v, RET #v ;
      vGPS_FWP l s q ∗ F_past l s v }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq vGPS_FWP_eq /=.
    iIntros (? V) "(#kI & VSDup & oW)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod (fractor_open with "[$oW]")
      as (X) "(gpsRaw & WRaw & HClose)"; first auto.
    iDestruct "WRaw" as (γ γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ' γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & oR & %)".

    iApply (RawProto_Read_ex IP l s _ _ v_x V_x γ γ_x _ ζ
              with "[$kI $kS $oI $ex $oR $wTok VSDup]"); [solve_ndisj| |done|..].
    { do 2 (etrans; eauto). }
    { iNext. by iSpecialize ("VSDup" $! _ with ""). }
    iNext.
    iIntros (v V') "(% & kS & oI & Fp & ex & oR' & wTok)".
    assert (V1 ⊑ V') by (etrans; eauto).
    iFrame; iApply ("Post" $! _ V' with "[%]"); first done.
    iFrame "Fp".
    iApply "HClose". iSplitL "oI".
    + iNext. iExists γ, γ_x. by iFrame.
    + iExists γ, γ_x. iSplitL ""; first auto.
      iExists v_x, V_x,_. iFrame "wTok ex oR'". iSplit; last auto.
      iPureIntro. do 2 (etrans; eauto).
  Qed.

  Lemma GPS_FWP_Writer_max l s s' q1 q2 (E : coPset) (HE: ↑fracN .@ l ⊆ E):
     (vGPS_FRP l s q1 ∗ ▷ vGPS_FWP l s' q2 ={E}=∗
        vGPS_FRP l s q1 ∗ vGPS_FWP l s' q2 ∗ ⌜s ⊑ s'⌝)%stdpp.
  Proof.
    iStartProof (uPred _).
    iIntros (V) "oRW". rewrite vGPS_FWP_eq vGPS_FRP_eq.
    iMod (fractor_join_later_rl with "oRW") as "oRW".
    iMod (fractor_open with "oRW") as (X) "(oI & (oR & oW) & HClose)";
      first auto.
    iDestruct "oR" as (γ1 γ_x1) "(% & #RA)".
    iDestruct "oW" as (γ2 γ_x2) "(>% & >WA)".
    iDestruct "oI" as (γ γ_x) "(>% & oI)". subst.
    apply encode_inj in H0. inversion H0. subst.
    apply encode_inj in H1. inversion H1. subst.

    iDestruct "RA" as (V1 v1) "(? & oR)".
    iDestruct "WA" as (v2 V2 ζ) "(#vS & oW & ex & #oR2 & %)".
    iDestruct (Reader_Writer_sub with "[$oW $oR]")
      as %Ins%elem_of_subseteq_singleton.
    iDestruct (Reader_Writer_sub with "[$oW $oR2]")
      as %Ins'%elem_of_subseteq_singleton.

    assert (Le := H  _ Ins).
    iMod (Writer_sorted with "[$oI $oW]") as "(oI & oW & %)".
    iMod ("HClose" $! (λ l0 _ X, ReaderSWRaw s l0 X V ∗ WriterSWRaw s' l0 X V)%I
          with "[$oI oW ex]") as "oW".
    { iSplitL ""; iExists γ, γ_x; (iSplitL ""; first auto).
      - iExists V1, v1. by iFrame "oR".
      - iExists v2, V2, ζ. by iFrame "vS oW ex oR2". }
    iDestruct (fractor_splitjoin l _
                   (λ l0 _ X, ReaderSWRaw s l0 X V)
                   (λ l0 _ X, WriterSWRaw s' l0 X V)
                   with "oW") as "[o1 o2]".
    iFrame "o1 o2".
    iPureIntro. by apply (H2 _ _ Ins Ins').
  Qed.

  Lemma GPS_FWP_Writer_Reader l s q1 q2:
    vGPS_FWP l s (q1 + q2) ⊢ vGPS_FWP l s q1 ∗ vGPS_FRP l s q2.
  Proof.
    iStartProof (uPred _). iIntros (V) "oW". rewrite vGPS_FWP_eq vGPS_FRP_eq.
    iApply fractor_splitjoin.
    iApply (fractor_mono with "[$oW]").
    iIntros (X) "oW".
    iDestruct "oW" as (γ γ_x) "[#? WA]".
    iDestruct "WA" as (v V0 ζ) "(#? & ? & ? & #oR & #?)".
    iSplitR ""; iExists γ, γ_x; (iSplitL ""; first auto).
    - iExists v, V0, ζ. iFrame. iSplit; [auto|by iSplitL ""].
    - iExists V0, v. by iFrame "oR".
  Qed.

  Lemma GPS_FWP_Writer_Reader_join l s q1 q2:
    vGPS_FWP l s q1 ∗ vGPS_FRP l s q2 ⊢ vGPS_FWP l s (q1 + q2).
  Proof.
    iStartProof (uPred _). iIntros (V) "(o1 & o2)". iCombine "o1" "o2" as "oW".
    rewrite vGPS_FWP_eq vGPS_FRP_eq fractor_splitjoin; last auto.
    iApply (fractor_mono with "[$oW]").
    by iIntros (X) "(oW & _)".
  Qed.

  Global Instance GPS_FRP_fractional l s:
    Fractional (vGPS_FRP l s).
  Proof.
    intros q1 q2. iStartProof (uPred _). rewrite vGPS_FRP_eq.
    iIntros (V). iSplit; iIntros "oR".
    - iApply fractor_splitjoin.
      iApply (fractor_mono with "[$oR]").
      iIntros (?) "#?". by iSplit.
    - rewrite fractor_splitjoin; last auto.
      iApply (fractor_mono with "[$oR]"). by iIntros (X) "[R1 R2]".
  Qed.

  Lemma GPS_FRP_join l s q1 q2:
    vGPS_FRP l s q1 ∗ vGPS_FRP l s q2 ⊢ vGPS_FRP l s (q1 + q2).
  Proof. by rewrite fractional. Qed.

  Lemma GPS_FRP_split l s q1 q2:
     vGPS_FRP l s (q1 + q2) ⊢ vGPS_FRP l s q1 ∗ vGPS_FRP l s q2.
  Proof. by rewrite fractional. Qed.

  Lemma GPS_FRP_mult_splitL l s (n: positive) q:
    vGPS_FRP l s (Pos2Qp n * q)
      ⊢ [∗ list] _ ∈ seq 0 (Pos.to_nat n), vGPS_FRP l s q.
  Proof.
    rewrite -{1}(Pos2Nat.id n).
    rewrite big_sepL_mult_fractional; [auto|lia].
  Qed.

  Lemma GPS_FRP_mult_joinL l s (n: positive) q:
    ([∗ list] _ ∈ seq 0 (Pos.to_nat n), vGPS_FRP l s q)
      ⊢ vGPS_FRP l s (Pos2Qp n * q).
  Proof.
    rewrite -{2}(Pos2Nat.id n).
    rewrite big_sepL_mult_fractional; [auto|lia].
  Qed.

  Lemma GPS_FWP_mult_splitL_1 l s n:
    vGPS_FWP l s 1
    ⊢ vGPS_FWP l s (1/(n+1))
      ∗ [∗ list] _ ∈ seq 0 (Pos.to_nat n), vGPS_FRP l s (1/(n+1)).
  Proof.
    rewrite {1}(Pos2Qp_plus_mult n).
    iIntros "FWP".
    iDestruct (GPS_FWP_Writer_Reader l s (1 / (n + 1)) (Pos2Qp n * (1 / (n + 1))) with "FWP") as "[$ FRP]".
    by iApply (GPS_FRP_mult_splitL l s n (1 / (n + 1))).
  Qed.

  Lemma GPS_FWP_ExWrite l s s' q v (E : coPset)
      (P: vPred) (Q : pr_state Prtcl → vPred)
      (Le: s ⊑ s') (HN : ↑physN ⊆ E) (HNl: ↑fracN .@ l ⊆ E):
  {{{ ⎡PSCtx⎤ ∗ ▷ (∀ v', P ∗ ▷ (F l s v')
            ={E ∖ ↑fracN .@ l}=∗ Q s' ∗ ▷ (F l s' v) ∗ ▷ F_past l s' v)
      ∗ P ∗ vGPS_FWP l s q }}}
    ([ #l ]_at <- #v) @ E
  {{{ RET #() ; vGPS_FWP l s' q ∗ Q s' }}}.
  Proof.
    iStartProof (uPred _).
    rewrite WP_Def.wp_eq /= vGPS_FWP_eq.
    iIntros (? V) "(#kI & VS & oP & oW)"; iIntros (V1 H1) "Post"; iIntros (V2 H2 ?) "kS".
    iMod (fractor_open with "[$oW]")
      as (X) "(gpsRaw & WRaw & HClose)"; first auto.
    iDestruct "WRaw" as (γ γ_x) "(% & WA)".
    iDestruct "gpsRaw" as (γ' γ_x') "(>% & oI)".
    subst X. apply encode_inj in H0. inversion H0. subst γ' γ_x'.
    iDestruct "WA" as (v_x V_x ζ) "(% & wTok & ex & #oR & %)".

    iApply (RawProto_Write_ex IP l s s'
            with "[$oI $kI $kS $ex $oR $wTok]");
      [solve_ndisj|auto|auto|auto|..].
    { do 2 (etrans; eauto). }
    iNext.
    iIntros (V') "(% & kS' & oI & oF & ex & #oR' & wTok & #Max)".
    setoid_rewrite H1; setoid_rewrite H2; setoid_rewrite H4.
    iFrame; iApply ("Post" $! V' with "[%]"); [done|].
    iMod ("VS" $! _ _ with "[%] [$oP $oF]") as "(oQ & oF & oFp)"; [done|..].
    iFrame "oQ". iApply ("HClose"). iSplitL "oI oF oFp".
    - iExists γ, γ_x. iSplitL ""; first auto.
      iApply ("oI" with "[$oF $oFp]").
    - iExists γ, γ_x. iSplitL ""; first auto.
      iExists v, V', _. by iFrame "wTok ex oR' Max".
  Qed.

End SingleWriter.

Arguments vGPS_WSP {_ _ _ _ _ _} _ _ _.
Arguments vGPS_RSP {_ _ _ _ _ _} _ _ _.
Arguments GPS_SW_Init {_ _ _ _ _ _} _ {_} _ {_} _ {_} _ _ {_}.
Arguments GPS_SW_Read {_ _ _ _ _ _} _ {_} _ _ {_ _} _ _ {_}.
Arguments GPS_SW_Read_ex {_ _ _ _ _ _} _ {_ _ _} _ _ {_}.
Arguments GPS_SW_ExWrite {_ _ _ _ _ _} _ {_ _} _ {_ _} _ _ _ _ _ {_}.
(* Arguments GPS_SW_FAI {_ _} {_ _ _ _ _ _ _} {_} _ {_ _} _ _ _ _ {_ _ _}. *)
(* Arguments GPS_SW_CAS {_ _} {_ _ _ _ _ _ _} {_} _ {_ _ _} _ _ _ _ _ {_ _ _}. *)
Arguments GPS_SW_Writer_Reader {_ _} {_ _ _ _ _ _ _}.
Arguments GPS_SW_Readers_agree {_ _} {_ _ _ _ _ _ _ _ _} _.

Arguments vGPS_FWP {_ _ _ _ _ _} _ _ _ _.
Arguments vGPS_FRP {_ _ _ _ _ _} _ _ _ _.
Arguments GPS_FWP_Init {_ _ _ _ _ _} _ {_} _ {_ _} _ {_}.
Arguments GPS_FRP_Read {_ _ _ _ _ _} _ {_ _} _ _ {_ _} _ _ {_}.
Arguments GPS_FWP_ExRead {_ _ _ _ _ _} _ {_ _ _ _} _ _ {_}.
Arguments GPS_FWP_ExWrite {_ _ _ _ _ _} _ {_ _} _ {_ _ _} _ _ _ _ _ {_}.

Arguments vGPS_nWSP {_ _ _ _ _ _ _ _ _} _ _ _ _.
Arguments vGPS_nRSP {_ _ _ _ _ _ _ _ _} _ _ _ _.
Arguments GPS_nSW_Init_strong {_ _ _ _ _ _ _ _ _} _ {_} _ _ {_} _ {_} _ _ {_}.
Arguments GPS_nSW_ExRead {_ _ _ _ _ _ _ _ _} _ {_ _} _ { _ _} _ _ {_}.
Arguments GPS_nSW_ExWrite {_ _ _ _ _ _ _ _ _} _ {_ _} _ {_} _ {_ _} _ _ _ _ _ {_}.
Arguments GPS_nSW_Read {_ _ _ _ _ _ _ _ _} _ {_ _} _ _ _ {_ _} _ _ {_}.
Arguments GPS_nSW_Init_default {_ _ _ _ _ _} _ { _} _ {_} _ {_} _ _ {_}.
Arguments GPS_nSWs_param_agree {_ _ _ _ _ _ _ _ _ _ _ _ _}.
Arguments GPS_nSW_nFRP_clash {_ _ _ _ _ _ _ _ _ _ _ _}.

Arguments vGPS_nFWP {_ _ _ _ _ _} _ _ _ _ _.
Arguments vGPS_nFRP {_ _ _ _ _ _} _ _ _ _ _.
Arguments GPS_nFWP_Init_strong {_ _ _ _ _ _} _ {_} _ {_} _ {_} _ {_}.
Arguments GPS_nFWP_ExRead {_ _ _ _ _ _} _ _ {_} _ { _ _} _ _ {_}.
Arguments GPS_nFWP_ExWrite {_ _ _ _ _ _} _ _ {_ _} _ {_ _ _} _ _ _ _ _ {_}.
Arguments GPS_nFRP_Read {_ _ _ _ _ _} _ {_} _ {_} _ _ {_ _} _ _ {_}.

Notation "'[XP' l 'in' s | p ']_W'" :=
  (vGPS_WSP p l s)
    (format "[XP  l  in  s  |  p  ]_W").
Notation "'[XP' l 'in' s | p ']_R'" :=
  (vGPS_RSP p l s)
    (format "[XP  l  in  s  |  p  ]_R").

Notation "'[FXP' l 'in' s | p ']_W_' q" :=
  (vGPS_FWP p l s q)
    (at level 0,
     format "[FXP  l  in  s  |  p  ]_W_ q").

Notation "'[FXP' l 'in' s | p ']_R_' q" :=
  (vGPS_FRP p l s q)
    (at level 0,
     format "[FXP  l  in  s  |  p  ]_R_ q").

Notation "'[nXP' l 'in' s '@' γ | F ']_W'" :=
  (vGPS_nWSP F γ l () s)
    (format "[nXP  l  in  s  @  γ  |  F  ]_W").

Notation "'[nXP' l 'in' s '@' γ | F ']_R'" :=
  (vGPS_nRSP F γ l () s)
    (format "[nXP  l  in  s  @  γ  |  F  ]_R").

Notation "'[nXP' l 'in' s '@' '(' p ',' γ ')' | F ']_W'" :=
  (vGPS_nWSP F γ l p s)
    (format "[nXP  l  in  s  @  ( p , γ )  |  F  ]_W").

Notation "'[nXP' l 'in' s '@' '(' p ',' γ ')' | F ']_R'" :=
  (vGPS_nRSP F γ l p s)
    (format "[nXP  l  in  s  @  ( p , γ )  |  F  ]_R").

Notation "'[nFXP' l 'in' s '@' γ | p ']_W_' q" :=
  (vGPS_nFWP p γ l s q)
    (at level 0,
     format "[nFXP  l  in  s @ γ  |  p  ]_W_ q").

Notation "'[nFXP' l 'in' s '@' γ | p ']_R_' q" :=
  (vGPS_nFRP p γ l s q)
    (at level 0,
     format "[nFXP  l  in  s @ γ  |  p  ]_R_ q").

Section proper.
Context `{foundationG Σ, gpsG Σ Prtcl}.

Global Instance gpsSWRaw_ne n :
  Proper ((dist n) ==> (=) ==> (=) ==> (dist n)) (@gpsSWRaw _ _ _ _ _).
Proof. solve_proper. Qed.
Global Instance gpsSWnRaw_ne n :
  Proper (pointwise_relation _ (dist n) ==> (=) ==> (=) ==> (=) ==> (dist n))
         (@gpsSWnRaw _ _ _ _ _).
Proof. rewrite /gpsSWnRaw. solve_proper. Qed.
Global Instance gpsSWpnRaw_ne `{EqDecision Param, Countable Param} n :
  Proper (pointwise_relation _ (dist n) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> (dist n)) (@gpsSWpnRaw _ _ _ _ _ _ _ _).
Proof. rewrite /gpsSWpnRaw. solve_proper. Qed.

Global Instance vGPS_RSP_contractive `{persistorG Σ} :
  Proper (dist_later n ==> (=) ==> (=) ==> dist n) vGPS_RSP.
Proof. split=>?. rewrite !vGPS_RSP_eq /=. solve_contractive. Qed.
Global Instance vGPS_WSP_contractive `{persistorG Σ} :
  Proper (dist_later n ==> (=) ==> (=) ==> dist n) vGPS_WSP.
Proof. split=>?. rewrite !vGPS_WSP_eq /=. solve_contractive. Qed.

Global Instance vGPS_FRP_contractive `{cinvG Σ} :
  Proper (dist_later n ==> (=) ==> (=) ==> (=) ==> dist n) vGPS_FRP.
Proof. split=>?. rewrite !vGPS_FRP_eq /=. solve_contractive. Qed.
Global Instance vGPS_FWP_contractive `{cinvG Σ} :
  Proper (dist_later n ==> (=) ==> (=) ==> (=) ==> dist n) vGPS_FWP.
Proof. split=>?. rewrite !vGPS_FWP_eq /=. solve_contractive. Qed.

Global Instance vGPS_nFRP_contractive `{cinvG Σ} :
  Proper (pointwise_relation _ (dist_later n) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> dist n) vGPS_nFRP.
Proof. split=>?. rewrite !vGPS_nFRP_eq /=. solve_contractive. Qed.
Global Instance vGPS_nFWP_contractive `{cinvG Σ} :
  Proper (pointwise_relation _ (dist_later n) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> dist n) vGPS_nFWP.
Proof. split=>?. rewrite !vGPS_nFWP_eq /=. solve_contractive. Qed.

Global Instance vGPS_nRSP_contractive `{persistorG Σ, EqDecision Param, Countable Param}:
  Proper (pointwise_relation _ (dist_later n) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> dist n) vGPS_nRSP.
Proof.
  split=>?. rewrite !vGPS_nRSP_eq /=. subst. f_contractive=>??.
  by apply gpsSWpnRaw_ne.
Qed.
Global Instance vGPS_nWSP_contractive `{persistorG Σ, EqDecision Param, Countable Param}:
  Proper (pointwise_relation _ (dist_later n) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> dist n) vGPS_nWSP.
Proof.
  split=>?. rewrite !vGPS_nWSP_eq /=. subst. f_contractive=>??.
  by apply gpsSWpnRaw_ne.
Qed.

Global Instance gpsSWRaw_proper :
  Proper ((≡) ==> (=) ==> (=) ==> (≡)) (@gpsSWRaw _ _ _ _ _).
Proof. solve_proper. Qed.
Global Instance gpsSWnRaw_proper :
  Proper (pointwise_relation _ (≡) ==> (=) ==> (=) ==> (=) ==> (≡))
         (@gpsSWnRaw _ _ _ _ _).
Proof. rewrite /gpsSWnRaw. solve_proper. Qed.
Global Instance gpsSWpnRaw_proper `{EqDecision Param, Countable Param} :
  Proper (pointwise_relation _ (≡) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> (≡)) (@gpsSWpnRaw _ _ _ _ _ _ _ _).
Proof. rewrite /gpsSWpnRaw. solve_proper. Qed.

Global Instance vGPS_WSP_proper `{persistorG Σ} :
  Proper ((≡) ==> (=) ==> (=) ==> (≡)) vGPS_WSP.
Proof. split=>?. rewrite !vGPS_WSP_eq /=. solve_proper. Qed.
Global Instance vGPS_RSP_proper `{persistorG Σ} :
  Proper ((≡) ==> (=) ==> (=) ==> (≡)) vGPS_RSP.
Proof. split=>?. rewrite !vGPS_RSP_eq /=. solve_proper. Qed.

Global Instance vGPS_FRP_proper `{cinvG Σ} :
  Proper ((≡) ==> (=) ==> (=) ==> (=) ==> (≡)) vGPS_FRP.
Proof. split=>?. rewrite !vGPS_FRP_eq /=. solve_proper. Qed.
Global Instance vGPS_FWP_proper `{cinvG Σ} :
  Proper ((≡) ==> (=) ==> (=) ==> (=) ==> (≡)) vGPS_FWP.
Proof. split=>?. rewrite !vGPS_FWP_eq /=. solve_proper. Qed.

Global Instance vGPS_nFRP_proper `{cinvG Σ} :
  Proper (pointwise_relation _ (≡) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> (≡)) vGPS_nFRP.
Proof. split=>?. rewrite !vGPS_nFRP_eq /=. solve_proper. Qed.
Global Instance vGPS_nFWP_proper `{cinvG Σ} :
  Proper (pointwise_relation _ (≡) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> (≡)) vGPS_nFWP.
Proof. split=>?. rewrite !vGPS_nFWP_eq /=. solve_proper. Qed.

Global Instance vGPS_nRSP_proper `{persistorG Σ, EqDecision Param, Countable Param}:
  Proper (pointwise_relation _ (≡) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> (≡)) vGPS_nRSP.
Proof.
  split=>?. rewrite !vGPS_nRSP_eq /=. subst. f_equiv=>??. by apply gpsSWpnRaw_proper.
Qed.
Global Instance vGPS_nWSP_proper `{persistorG Σ, EqDecision Param, Countable Param}:
  Proper (pointwise_relation _ (≡) ==>
          (=) ==> (=) ==> (=) ==> (=) ==> (≡)) vGPS_nWSP.
Proof.
  split=>?. rewrite !vGPS_nWSP_eq /=. subst. f_equiv=>??. by apply gpsSWpnRaw_proper.
Qed.
End proper.
