From iris.program_logic Require Import ectx_language ectxi_language language.
From stdpp Require Export strings gmap.
From iris.algebra Require Import ofe.

From igps Require Export machine.
From igps.lang Require Import base.

Definition expr : Type := base.expr * View.
Definition val : Type := base.val * View.
Definition ectx_item := base.ectx_item.
Definition fill_item : _ -> _ -> expr := λ Ki e, (base.fill_item Ki (e.1), e.2).
Definition of_val : val -> expr := λ v, (base.of_val (v.1), v.2).
Definition to_val : expr -> option val := λ e,  (λ v, (v, (e.2))) <$> base.to_val (e.1).


Definition is_closed (X : list string) (e : expr) : bool :=
  base.is_closed X (e.1).

Class Closed (X : list string) (e : expr) := closed : is_closed X e.
Instance closed_proof_irrel env e : ProofIrrel (Closed env e).
Proof. rewrite /Closed. apply _. Qed.
Instance closed_decision env e : Decision (Closed env e).
Proof. rewrite /Closed. apply _. Qed.

Definition subst x es (e : expr) : expr := (base.subst x es (e.1), e.2).

Definition subst' (mx : base.binder) (es : base.expr) : expr → expr :=
  match mx with base.BNamed x => subst x es | BAnon => id end.

(* Instance lit_dec_eq : ∀ l1 l2 : base.base_lit, Decision (l1 = l2) := _. *)

(* Definition state := @machine.state base.loc _ _ Z _ _. *)
(* Definition event := @machine.event base.loc Z. *)

Inductive step (V : View)
: ∀ (e : base.expr) (evt : event) (e' : base.expr) (ef : list expr), Prop :=
| ForkS e V' : (* π ≠ ρ ? *)
    step V
          (base.Fork e)
          (EFork V')
          (base.Lit base.LitUnit)
          [(e, V')]
  | AllocS l:
      step V
          base.Alloc
          (EWrite na l A)
          (base.Lit $ base.LitLoc l)
          []
| DeallocS l :
    step V
          (base.Dealloc (base.Lit $ base.LitLoc l))
          (EWrite na l D)
          (base.Lit $ base.LitUnit)
          []
| LoadS (l : loc) (v : Z) acc :
    step V
          (base.Load acc (base.Lit $ base.LitLoc l))
          (ERead acc l (VInj v))
          (base.of_val (base.LitV $ base.LitInt v))
          []
| StoreS l e v acc :
    base.to_val e = Some (base.LitV $ base.LitInt v) ->
    step V
          (base.Store acc (base.Lit $ base.LitLoc l) e)
          (EWrite acc l (VInj v))
          (base.Lit $ base.LitUnit)
        []
| CasFailS l e1 v1 e2 v2 vl :
    base.to_val e1 = Some (base.LitV $ base.LitInt v1) →
    base.to_val e2 = Some (base.LitV $ base.LitInt v2) →
    vl ≠ v1 →
    step V
          (base.CAS (base.Lit $ base.LitLoc l) e1 e2)
          (ERead at_hack l (VInj vl))
          (base.Lit $ base.lit_of_bool false)
          []
| CasSucS l e1 v1 e2 v2 :
    base.to_val e1 = Some (base.LitV $ base.LitInt v1) →
    base.to_val e2 = Some (base.LitV $ base.LitInt v2) →
    step V
          (base.CAS (base.Lit $ base.LitLoc l) e1 e2)
          (EUpdate l (VInj v1) (VInj v2))
          (base.Lit $ base.lit_of_bool true)
          []
| FAIS l C (v: Z):
    step V
          (base.FAI C (base.Lit $ base.LitLoc l))
          (EUpdate l (VInj v) (VInj ((v + 1) `mod` Z.pos C)))
          (base.Lit $ base.LitInt v)
          []
.

Inductive head_step :
∀ (e : expr) (σ : state) (e' : expr) (σ' : state) (ef : list expr), Prop :=
| pure_step σ π e e' (BaseStep : base.head_step e e')
  : head_step (e, π) σ (e', π) σ []
| impure_step σ σ' V V' e e' ef evt
              (ExprStep : step V e evt e' ef)
              (MachStep : machine_red σ evt V σ' V')
  : head_step (e, V) σ (e', V') σ' ef
.

Lemma to_of_val v : to_val (of_val v) = Some v.
Proof. destruct v. cbv -[base.to_val base.of_val]. by rewrite base.to_of_val. Qed.

Lemma of_to_val e v : to_val e = Some v → of_val v = e.
Proof.
  destruct e, v. cbv -[base.to_val base.of_val]. case C : (base.to_val e) => //.
  move => [<- <-]. f_equal. exact: base.of_to_val.
Qed.

Lemma val_stuck σ1 e1 σ2 e2 ef : head_step e1 σ1 e2 σ2 ef → to_val e1 = None.
Proof.
  inversion 1; subst; last inversion ExprStep;
    first (cbv -[base.to_val]; by erewrite base.val_stuck; last eassumption);
    reflexivity.
Qed.

Lemma fill_item_val Ki e :
  is_Some (to_val (fill_item Ki e)) → is_Some (to_val e).
Proof.
  move/fmap_is_Some/base.fill_item_val => H.
  exact/fmap_is_Some.
Qed.

Lemma fill_item_no_val_inj Ki1 Ki2 e1 e2 :
  to_val e1 = None → to_val e2 = None → fill_item Ki1 e1 = fill_item Ki2 e2 → Ki1 = Ki2.
Proof.
  move => /fmap_None H1 /fmap_None H2 [] H3 ?.
  exact: base.fill_item_no_val_inj H1 H2 H3.
Qed.

Lemma head_ctx_step_val Ki e σ e2 σ2 ef :
  head_step (fill_item Ki e) σ e2 σ2 ef → is_Some (to_val e).
Proof.
  inversion 1; subst; last inversion ExprStep; subst; simplify_option_eq;
    first
      (apply/fmap_is_Some; exact: base.head_ctx_step_val);
    apply/fmap_is_Some;
    first by destruct Ki.
  - destruct Ki; try (by eauto);
      inversion H1; subst; by eauto.
  - destruct Ki; try (by eauto);
      inversion H1; subst; by eauto.
  - destruct Ki; try (by eauto);
      inversion H1; subst; by eauto.
  - destruct Ki; try (by eauto);
      inversion H0; subst; by eauto.
  - destruct Ki; try (by eauto);
      inversion H0; subst; by eauto.
  - destruct Ki; try (by eauto);
      inversion H0; subst; by eauto.
  - destruct Ki; try (by eauto).
      inversion H1; subst; by eauto.
Qed.

Instance of_val_inj : Inj (=) (=) of_val.
Proof. by intros ?? Hv; apply (inj Some); rewrite -!to_of_val Hv. Qed.

Instance fill_item_inj Ki : Inj (=) (=) (fill_item Ki).
Proof. destruct Ki; move => [? ?] [? ?] [? ?]; simplify_eq/=; auto with f_equal. Qed.

Lemma is_closed_weaken X Y e : is_closed X e → X ⊆  Y → is_closed Y e.
Proof.
  destruct e. cbv -[base.is_closed].
  exact: base.is_closed_weaken.
Qed.

Lemma is_closed_weaken_nil X e : is_closed [] e → is_closed X e.
Proof. intros. destruct e. exact: base.is_closed_weaken_nil. Qed.

Lemma is_closed_subst X e x es : is_closed X e → x ∉ X → subst x es e = e.
Proof. intros. destruct e. cbv -[base.subst]. f_equal. exact: base.is_closed_subst. Qed.

Lemma is_closed_nil_subst e x es : is_closed [] e → subst x es e = e.
Proof. intros. destruct e. cbv -[base.subst]. f_equal. exact: base.is_closed_nil_subst. Qed.

Lemma is_closed_of_val X v : is_closed X (of_val v).
Proof. exact: base.is_closed_of_val. Qed.

Lemma head_step_mono e1 V1 σ1 e2 V2 σ2 efs : head_step (e1, V1) σ1 (e2, V2) σ2 efs → V1 ⊑ V2.
Proof.
  inversion 1; subst; first done.
  exact: machine_red_mono.
Qed.

(** Language *)
Lemma ra_ectxi_language_mixin :
  EctxiLanguageMixin of_val to_val fill_item head_step.
Proof.
  split;
  eauto using to_of_val, of_to_val,
  val_stuck, fill_item_val, fill_item_no_val_inj,
  head_ctx_step_val.
  move => ? [? ?] [? ?] [].
  intros. f_equal; [|done].
  by apply: base.fill_item_inj.
Qed.

Canonical Structure ra_ectxi_lang :=
  EctxiLanguage (expr:=prod _ _) ra_ectxi_language_mixin.

Canonical Structure ra_ectx_lang := EctxLanguageOfEctxi ra_ectxi_lang.
Canonical Structure ra_lang := LanguageOfEctx ra_ectx_lang.
