From iris.base_logic Require Import lib.invariants.
From iris.program_logic Require Import language.
From igps Require Import lang.surface viewpred weakestpre proofmode.
Export lang.base lang.surface.

Lemma pure_exec_view Φ (e1 e2 : base.expr) :
  PureExec Φ e1 e2 → ∀ V, PureExec (Λ:=ra_lang) Φ (e1, V) (e2, V).
Proof.
  intros PE.
  inversion PE; intros; econstructor.
  - intros σ HΦ. move: pure_exec_safe => /(_ (V,σ) HΦ); intros (e' & [V' σ'] & efs & HStep).
    exists (e',V'), σ', (map (fun e => (e,V')) efs).
    inversion HStep.
    exists K (e1', V) (e2', V').
    + simplify_eq. now rewrite -fill_view.
    + simplify_eq. now rewrite -fill_view.
    + now eauto.
  - intros σ1 [e2' V'] σ2 efs HΦ HStep.
    move: pure_exec_puredet => /(_ (V, σ1) e2' (V', σ2) (map (fst) efs) HΦ) H.
    destruct H as (? & ? & ?).
    + inversion HStep; cbn in *.
      destruct e1' as [e1' V1'], e2'0 as [e2'0 V2'0]. cbn in *.
      exists K (e1') (e2'0).
      * rewrite ->fill_view in *. by simplify_eq.
      * rewrite ->fill_view in *. by simplify_eq.
      * rewrite ->fill_view in *. simplify_eq.
        inversion H1.
        { simplify_eq. unfold surface.head_step. cbn. exact H1. }
        { simplify_eq. inversion ExprStep; simplify_eq; cbn; try exact H1.
          inversion MachStep. simplify_map_eq. inversion ThreadRed. simplify_eq.
          exact H1.
        }
    + simplify_eq. by destruct efs.
Qed.


Section Lifting.
Context {Σ: gFunctors} {Iris : ghosts.foundationG Σ}.
Local Set Default Proof Using "Σ Iris".
Implicit Types P Q : vProp Σ.
Implicit Types Φ : base.val → vProp Σ. (* TODO: fix imports *)

Lemma wp_pure_step_later `{Inhabited (state)} E e1 e2 φ Φ :
  (PureExec φ e1 e2) →
  φ →
  ▷ WP e2 @ E {{ Φ }} ⊢ WP e1 @ E {{ Φ }}.
Proof.
  iStartProof (uPred _). iIntros (?? V) "W".
  rewrite WP_Def.wp_eq /WP_Def.wp_def /=. iIntros (V' HV' π) "G".
  iApply program_logic.lifting.wp_pure_step_fupd; [apply pure_exec_view; eassumption|done|].
  iIntros "!>". by iApply "W".
Qed.

End Lifting.


Ltac inv_head_step :=
  repeat match goal with
         | _ => progress simplify_map_eq/= (* simplify memory stuff *)
         | _ => done
         | H : base.to_val _ = Some _ |- _ => apply base.of_to_val in H
         | H : context [base.to_val (base.of_val _)] |- _ => rewrite base.to_of_val in H
         | H : base.head_step' _ _ _ _ _ |- _ => inversion H; subst; clear H
         | H : base.head_step ?e _ |- _ =>
           try (is_var e; fail 1); (* inversion yields many goals if [e] is a variable
     and can thus better be avoided. *)
           inversion H; subst; clear H
         | H : surface.head_step ?e _ _ _ _ |- _ =>
           try (is_var e; fail 1); (* inversion yields many goals if [e] is a variable
     and can thus better be avoided. *)
           inversion H; subst; clear H
         | H : ra_lang.head_step ?e _ _ _ _ |- _ =>
           try (is_var e; fail 1); (* inversion yields many goals if [e] is a variable
     and can thus better be avoided. *)
           inversion H; subst; clear H
         | H : ra_lang.step _ ?e _ _ _ |- _ =>
           try (is_var e; fail 1); (* inversion yields many goals if [e] is a variable
     and can thus better be avoided. *)
           inversion H; subst; clear H
         end.

Class AsRec (e : base.expr) (f x : base.binder) (erec : base.expr) :=
  as_rec : e = base.Rec f x erec.
Global Instance AsRec_rec f x e : AsRec (base.Rec f x e) f x e := eq_refl.
Global Instance AsRec_rec_locked_val v f x e :
  AsRec (base.of_val v) f x e → AsRec (base.of_val (locked v)) f x e.
Proof. by unlock. Qed.


Local Ltac solve_exec_safe := intros; subst; do 2 eexists; exists nil; do 2!econstructor; eauto.
Local Ltac solve_exec_puredet := simpl; intros; inv_head_step;
  repeat match goal with | [ X : View * state |- _ ] => destruct X end;
  match goal with | [ X : [] = map _ ?efs |- _ ] => destruct efs; [|discriminate] end;
  simpl in *; simplify_eq; done.
Local Ltac solve_pure_exec :=
  unfold IntoVal,AsVal,AsRec in *; subst;
  repeat match goal with H : is_Some _ |- _ => destruct H as [??] end;
  apply ectx_language.det_head_step_pure_exec; [ solve_exec_safe | solve_exec_puredet ].

Global Instance pure_rec f x (erec e1 e2 : base.expr)
       `{!AsVal (e2), AsRec e1 f x erec, base.Closed (f :b: x :b: []) erec}:
  PureExec True (base.App e1 e2) (base.subst' x e2 (base.subst' f e1 erec)).
Proof. solve_pure_exec. Qed.

Global Instance pure_unop op e v v' `{!IntoVal (e) (v)}:
  PureExec (base.un_op_eval op v = Some v') (base.UnOp op e) (base.of_val v').
Proof. solve_pure_exec. Qed.

Global Instance pure_binop op e1 e2 v1 v2 v' `{!IntoVal (e1) (v1), !IntoVal (e2) (v2)}:
  PureExec (base.bin_op_eval op v1 v2 = Some v') (base.BinOp op e1 e2) (base.of_val v').
Proof. solve_pure_exec. Qed.

Global Instance pure_if_true e1 e2:
  PureExec True (base.If (base.Lit (base.LitInt 1)) e1 e2) (e1).
Proof. solve_pure_exec. Qed.

Global Instance pure_if_false e1 e2:
  PureExec True (base.If (base.Lit (base.LitInt 0)) e1 e2) (e2).
Proof. intros. solve_pure_exec. Qed.
