From iris.proofmode Require Import tactics.
From iris.base_logic Require Import iprop.
From iris.program_logic Require Import ownp.
From igps Require Import ghosts weakestpre invariants. (* TODO: why ??? *)
Import base.

Section Atomic.
  Context {Σ : gFunctors} `{fG : !foundationG Σ}
          {A}
          (α: A → vProp Σ) (* atomic pre-condition *)
          (β: A → val → vProp Σ) (* atomic post-condition *)
          (Ei Eo: coPset) (* inside/outside masks *)
  .

  Definition atomic_shift (Φ : val → vProp Σ) : vProp Σ :=
    (∃ P, ▷ P ∗ □ (▷ P ={Eo,Ei}=∗ ∃ x : A,
        α x ∗ ((α x ={Ei,Eo}=∗ ▷ P) ∧ ∀ y, β x y ={Ei,Eo}=∗ Φ y)))%I.

  Definition atomic_wp e : vProp Σ :=
      (∀ Φ, atomic_shift Φ -∗ WP e @ Eo {{ Φ }})%I.

  Lemma LA_WP e {Ht : ∀x, Timeless (α x)} (Hlt : Ei ⊆ Eo) x:
    atomic_wp e -∗ α x -∗ WP e {{v, β x v}}%I.
  Proof.
    iIntros "Hwp Ha".
    iApply wp_mask_mono; first auto.
    iApply "Hwp"; unfold atomic_shift.
    iExists (α x); iFrame.
    iIntros "!# Ha"; iMod "Ha".
    iMod (fupd_intro_mask' Eo Ei) as "Hupd"; first done. iModIntro.
    iExists x; iFrame; iSplit.
    - iIntros "Ha". iMod "Hupd"; auto.
    - iIntros (v) "Hv". iMod "Hupd"; auto.
  Qed.

  Lemma LA_Hoare e {Ht : ∀x, Timeless (α x)} (Hlt : Ei ⊆ Eo):
    to_val e = None → □atomic_wp e -∗ ∀x, {{{α x}}} e {{{v, RET v; β x v}}}.
  Proof.
    intros.
    iIntros "#Hwp" (?); iIntros "!#" (?) "Ha Hb".
    iPoseProof (LA_WP with "Hwp Ha") as "Hwp'"; first done.
    iPoseProof (wp_frame_step_l' with "[$Hb $Hwp']") as "Hwp'"; first done.
    iApply wp_mono; last done.
    intros; iIntros "[Hpost Hb]".
    iApply "Hpost"; done.
  Qed.

End Atomic.

Section LA_rules.
  Context {Σ : gFunctors} `{fG : !foundationG Σ}.

Lemma LA_Frame {A} (α : A -> vProp Σ) β Ei Eo e P Ei':
  Ei ⊆ Ei' ->
  atomic_wp α β Ei Eo e -∗
  atomic_wp (fun x => α x ∗ P x)%I (fun x v => β x v ∗ P x)%I Ei' Eo e.
Proof.
  unfold atomic_wp, atomic_shift; intros.
  iIntros "Ha" (Φ) "H"; iDestruct "H" as (P') "[HP' #Hshift]".
  iApply "Ha"; iExists P'; iFrame.
  iIntros "!# HP".
  iMod ("Hshift" with "HP") as (x) "[[Hx HP] Hclose]".
  iMod (fupd_intro_mask' Ei' Ei) as "Hupd"; [done|].
  iModIntro. iExists x; iFrame; iSplit.
  - iIntros "Hx".
    iMod "Hupd".
    iApply "Hclose"; iFrame.
  - iIntros (v) "Hv".
    iMod "Hupd".
    iApply "Hclose"; iFrame.
Qed.

Lemma LA_Inv {A} (α : A -> vProp Σ) β Ei Eo e N R {_ : Objective R}:
  ↑N ## Ei ->
  atomic_wp (fun x => ▷R ∗ α x)%I (fun x v => ▷R ∗ β x v)%I Ei Eo e -∗
  atomic_wp (fun x => inv N R ∗ α x)%I β (↑N ∪ Ei) Eo e.
Proof.
  unfold atomic_wp, atomic_shift; intros.
  iIntros "Ha" (Φ) "H"; iDestruct "H" as (P) "[HP #Hshift]".
  iApply "Ha"; iExists P; iFrame.
  iIntros "!# HP".
  iMod ("Hshift" with "HP") as (x) "[[#Hinv Hx] Hclose]".
  iPoseProof (inv_open (↑N ∪ Ei) with "Hinv") as "HR".
  { apply union_subseteq_l. }
  iMod "HR" as "[HR HcloseR]".
  iDestruct (monPred_objectively_elim with "HR") as "HR".
  assert (Ei = (↑N ∪ Ei) ∖ ↑N) as <-.
  { rewrite difference_union_distr_l_L difference_diag_L union_empty_l_L
            difference_disjoint_L; done. }
  iModIntro.
  iExists x; iFrame; iSplit.
  - iIntros "[HR Hx]".
    iMod ("HcloseR" with "[HR]").
    { by iApply objective_objectively. }
    iApply "Hclose"; auto.
  - iIntros (v) "[HR Hv]".
    iMod ("HcloseR" with "[HR]").
    { by iApply objective_objectively. }
    iApply "Hclose"; auto.
Qed.

Lemma LA_Csq {A} (α : A -> vProp Σ) β Ei Eo e α' β' Ei':
  Ei ⊆ Ei' ->
  □(∀x, α' x ={Ei', Ei}=∗ α x) -∗ □(∀x, α x ={Ei, Ei'}=∗ α' x) -∗
  □(∀x v, β x v ={Ei, Ei'}=∗ β' x v) -∗
  atomic_wp α β Ei Eo e -∗ atomic_wp α' β' Ei' Eo e.
Proof.
  unfold atomic_wp, atomic_shift; intros.
  iIntros "#Hpre1 #Hpre2 #Hpost Ha" (Φ) "H"; iDestruct "H" as (P) "[HP #Hshift]".
  iApply "Ha"; iExists P; iFrame.
  iIntros "!# HP".
  iMod ("Hshift" with "HP") as (x) "[Hx Hclose]".
  iMod ("Hpre1" with "Hx"); iModIntro.
  iExists x; iFrame; iSplit.
  - iIntros "Hx".
    iMod ("Hpre2" with "Hx").
    iApply "Hclose"; done.
  - iIntros (v) "Hv".
    iMod ("Hpost" with "Hv").
    iApply "Hclose"; done.
Qed.

Lemma LA_Mono {A} (α : A -> vProp Σ) β Ei Eo e Ei':
  Ei ⊆ Ei' ->
  atomic_wp α β Ei Eo e -∗ atomic_wp α β Ei' Eo e.
Proof.
  intros; iIntros "H".
  iPoseProof (LA_Frame _ _ _ _ _ (fun _ => emp%I) Ei' with "H") as "H"; first done.
  iApply LA_Csq; last (by iApply "H"); auto.
  - by iIntros "!#"; iIntros (?) "[?_]".
  - by iIntros "!#"; iIntros (??) "[?_]".
Qed.

Lemma LA_Inv' {A} β Ei Eo e N (R : A -> vProp Σ) Ei'
  {_ : ∀ x, Objective (R x)} {_ : ∀ x, Timeless (R x)}:
  ↑N ## Ei -> ↑N ∪ Ei ⊆ Ei' ->
  atomic_wp (fun x => R x)%I (fun x v => (∃y, R y) ∗ β x v)%I Ei Eo e -∗
  atomic_wp (fun _ : unit => inv N (∃x, R x))%I (fun _ v => ∃x, β x v) Ei' Eo e.
Proof.
  intros; iIntros "Ha".
  iApply LA_Mono; first done.
  unfold atomic_wp, atomic_shift.
  iIntros (Φ) "H"; iDestruct "H" as (P) "[HP #Hshift]".
  iApply "Ha"; iExists P; iFrame.
  iIntros "!# HP".
  iMod ("Hshift" with "HP") as (x) "[#Hinv Hclose]".
  iPoseProof (inv_open (↑N ∪ Ei) with "Hinv") as "HR".
  { apply union_subseteq_l. }
  iMod "HR" as "[>HR HcloseR]".
  iDestruct (monPred_objectively_elim with "HR") as (a) "HR".
  assert (Ei = (↑N ∪ Ei) ∖ ↑N) as <-.
  { by rewrite difference_union_distr_l_L difference_diag_L union_empty_l_L
               difference_disjoint_L. }
  iModIntro. iExists a; iFrame; iSplit.
  - iIntros "HR".
    iMod ("HcloseR" with "[HR]").
    { iApply objective_objectively; by iExists a. }
    iApply "Hclose"; auto.
  - iIntros (v) "[HR Hv]".
    iMod ("HcloseR" with "[HR]").
    { by iApply objective_objectively. }
    iApply "Hclose"; auto.
Qed.
End LA_rules.
