From iris.algebra Require Import gmap.

From igps Require Import weakestpre proofmode.
From igps Require Export viewpred.
From igps Require Export wp_tactics_vProp.
From igps Require Import machine lifting_vProp derived_vProp notation bigop na.
From igps.base Require Import ghosts alloc dealloc.

From iris.base_logic.lib Require Import cancelable_invariants. (* TODO: what the heck. why do I need this here? *)

Import lang.base.

Definition mdealloc : val :=
  (λ: "l" "n",
    (rec: "mdel" "i" :=
      if: "n" ≤ "i"
      then #()
      else
        Dealloc ("l" + "i") ;;
        "mdel" ("i" + #1%Z)) #0)%V.

Definition malloc : val :=
  (λ: "n",
  (rec: "malloc" <> :=
    let: "l" := Alloc in
      (rec: "mall" "i" :=
        if: "i" = "n"
        then "l"
        else
          let: "li" := Alloc in
            if: "li" = "l" + "i"
            then "mall" ("i" + #1)
            else mdealloc "l" "i" ;; Dealloc "li" ;; "malloc" #()) #1) #())%V.

Global Opaque mdealloc malloc.

Section proof.
  Context `{fG : !foundationG Σ}.
  Set Default Proof Using "Type*".

  Local Open Scope I.

  Lemma alloc (E : coPset) :
    ↑physN ⊆ E →
    {{{ ⎡PSCtx⎤ }}}
      (Alloc) @ E
    {{{ (l: loc), RET (#l); (l ↦ A) }}}%stdpp.
  Proof.
    rewrite /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    rewrite WP_Def.wp_eq /WP_Def.wp_def.
    iStartProof (uPred _).
    iIntros (???) "#kI". iIntros (V2 HV2) "Post". iIntros (V3 HV3 π) "kS".
    iApply program_logic.weakestpre.wp_mask_mono;
    last (iApply (f_alloc with "[$kI $kS]"));
    first auto.
    iNext. iIntros (l V' h' i) "(% & $ & oH & oI & % & % & %)".
    destruct H2; subst.
    iApply ("Post" $! _ _ with "[%]"); [etrans;done|].
    iExists _, _. by iFrame "oH oI".
  Qed.

  Lemma dealloc l :
    {{{ ⎡PSCtx⎤ ∗ l ↦ ? }}}
      (Dealloc #l) @ ↑physN
    {{{ RET (#()); True }}}%stdpp.
  Proof.
    rewrite /own_loc /own_loc_prim /= valloc_local_eq /valloc_local_def.
    rewrite WP_Def.wp_eq /WP_Def.wp_def.
    iStartProof (uPred _).
    iIntros (? V1) "[#kI oL]". iIntros (V2 HV2) "Post". iIntros (V3 HV3 π) "kS".
    iDestruct "oL" as (h i) "(Alloc & oH & oI)".
    rewrite HV2. rewrite {1}HV3.
    iApply (f_dealloc with "[$kI $kS $oH $oI $Alloc]").
    iNext. iIntros (?) "(% & $)". by iApply "Post".
  Qed.

  Lemma mdealloc_spec (l: loc) (n : Z) :
    ⎡PSCtx⎤ ⊢ {{{ ▷ ([∗ list] i ∈ seq 0 (Z.to_nat n), ZplusPos (Z.of_nat i) l ↦ ?) }}}
      (mdealloc #l #n) @ ↑physN
    {{{ RET (#()); True }}}.
  Proof.
    iIntros "#kI !#" (Φ) "Big Post".
    wp_bind (mdealloc (#l)%E)%E.
    wp_lam. wp_lam.
    move Eqmax : {-}n => max.
    rewrite [in X in WP X @ _ {{ _ }}]Eqmax.
    have Lemaxn: (n ≤ max)%Z by omega.
    move Eqz: {-}(0) => z.
    rewrite [in X in WP (_ X) @ _ {{ _ }}]Eqz.
    have Eqzn : (z + n = max)%Z by omega.
    have Le0z : 0 ≤ z by omega.
    rewrite (_ : 0%nat = Z.to_nat z); last by rewrite -Eqz.
    clear Eqz Eqmax.
    (iLöb as "IH" forall (n z Lemaxn Le0z Eqzn)).
    wp_rec.
    wp_op.
    unfold bool_decide.
    destruct (decide_rel _ _ _); wp_if; last by iApply "Post".
    wp_bind (Dealloc _). wp_op.
    fold (Z.gt max z) in e.
    have ?: (1 ≤ n) by omega.
    rewrite (_ : Z.to_nat n = S (Z.to_nat (n-1))); last first.
    { rewrite -Z2Nat.inj_succ; last omega. f_equal; omega. }
    rewrite [seq _ _]/=.
    iDestruct "Big" as "[B0 Big]".
    rewrite Z2Nat.id; last omega.
    iApply (dealloc with "[$kI $B0]").
    iNext. iIntros "_".
    wp_lam.
    rewrite -Z2Nat.inj_succ; last omega.
    wp_op.
    iApply ("IH" $! (n-1) (z+1) with "[%] [%] [%] Big [$Post]");
      omega.
  Qed.

  Lemma malloc_spec (n : Z) (Le1n : 1 ≤ n):
    ⎡PSCtx⎤ ⊢ {{{ True }}}
      (malloc #n) @ ↑physN
    {{{ (l : loc), RET #l;
        [∗ list] i ∈ seq 0 (Z.to_nat n), ZplusPos (Z.of_nat i) l ↦ ? }}}.
  Proof.
    iIntros "#kI !#" (Φ) "_ Post".
    wp_lam.
    set malloc'' := {-}(rec: "malloc" <> := _)%E.
    set malloc' := ltac:(let e := eval unfold malloc'' in malloc'' in tactics_vProp.reshape_val e ltac:(fun x => exact x)).
    let e := eval unfold malloc' in malloc' in
        match e with
        | context f [n] =>
          set malloc := (fun (n : Z) => ltac:(let x := context f[n] in exact x));
          unfold malloc' in malloc;
          clear malloc'' malloc'
        end.
    rewrite (eq_refl : (rec: "malloc" <> := _)%E = base.of_val (malloc n)).
    rewrite (lock (malloc n)).
    (iLöb as "IHmalloc" forall (n Le1n)).
    rewrite -lock.
    wp_rec.
    rewrite (eq_refl : (rec: "malloc" <> := _)%E = base.of_val (malloc n)) (lock (malloc n)).
    wp_bind (Alloc)%E.
    iApply (alloc with "[$kI]"); [done|].
    iNext.
    iIntros (l) "oL".
    wp_lam.
    rewrite -/(malloc _).
    move Eqz: {-}1%Z => z.
    rewrite [in X in WP (_ X) @ _ {{ _ }}]Eqz.
    have Lez1: (1 ≤ z)%Z by omega.
    move Eqmax : {-}(n) => max.
    rewrite [in X in seq _ X]Eqmax.
    have Eqzn: (z + (n-1) = max)%Z by omega.
    rewrite [in X in WP_Def.wp _ X _]Eqmax.

    iAssert (([∗ list] i ∈ seq 0 (Z.to_nat z), ZplusPos (Z.of_nat i) l ↦ ?))
      with "[oL]" as "Big".
    { rewrite -Eqz big_sepL_singleton. iDestruct "oL" as (h) "oL". by iExists _. }

    clear Eqz Eqmax.

    (iLöb as "IHmall" forall (z n Lez1 Le1n Eqzn)).
    wp_rec.
    wp_op.
    unfold bool_decide.
    destruct (decide_rel _ _ _); wp_if.
    - iClear "IHmalloc IHmall". subst max.
      iApply ("Post" with ""); by rewrite -e /=.
    - wp_bind Alloc.
      iApply (alloc with "[$kI]"); [done|].
      iNext. iIntros (l') "oL".

      wp_lam. wp_op. wp_bind (_ = _)%E.
      case: (decide (l' = ZplusPos z l)) => [EQl'|?].
      + subst. wp_op.
        rewrite bool_decide_true; last by auto. wp_if_true.
        wp_op.

        iApply ("IHmall" $! (z+1) (n-1) with "[%] [%] [%] [$Post] [Big oL]");
          [omega..|].

        rewrite Z2Nat.inj_succ; last omega.
        rewrite seq_S big_opL_app big_sepL_singleton.
        iFrame "Big".
        rewrite /= Z2Nat.id; last omega.
        iDestruct "oL" as (h) "oL".
        by iExists _.
      + subst. wp_op.
        rewrite bool_decide_false; last by auto. wp_if_false.
        wp_bind ((mdealloc _) _)%E.
        iApply (mdealloc_spec with "kI [$Big]").
        iNext. iIntros "_".
        wp_lam. wp_bind (Dealloc _)%E.
        iApply (dealloc with "[$kI oL]").
        { by iDestruct "oL" as (h) "oL"; iExists _. }
        iNext. iIntros "_".
        wp_lam. iApply ("IHmalloc" with "[%] [$Post]"); omega.
  Qed.

End proof.
