From iris.program_logic Require Import weakestpre.
From iris.base_logic Require Import lib.invariants lib.cancelable_invariants.

From igps Require Import viewpred proofmode weakestpre fractor.
From igps Require Export notation bigop.
From igps.base Require Export ghosts na_read na_write.

Program Definition valloc_local_def {Σ} h : vProp Σ :=
  MonPred (λ V, ⌜alloc_local h V⌝)%I _.
Next Obligation. intros. intros ???. iIntros "% !%". exact: alloc_local_proper. Qed.
Definition valloc_local_aux : seal (@valloc_local_def). by eexists. Qed.
Definition valloc_local {Σ} := unseal valloc_local_aux Σ.
Definition valloc_local_eq : @valloc_local = _ := seal_eq _.

Section Exclusive.
  Context {Σ : gFunctors} `{fG : foundationG Σ}.
  Set Default Proof Using "Type".

  Definition own_loc_prim l h: vProp Σ
    := (∃ i, valloc_local h ∧ ⎡Hist l h⎤ ∗ ⎡Info l 1%Qp i⎤)%I.

  Definition own_loc_na l v: vProp Σ
    := (∃ V, own_loc_prim l {[v, V]})%I.

  Definition own_loc l : vProp Σ :=
    (∃ h, own_loc_prim l h)%I.

  (* Global Instance From_own_loc_prim_exist: *)
  (*   FromExist ((own_loc_prim l h))%I *)
  (*             (λ i, (valloc_local h ∧ (☐ Hist l h) ∗ (☐ Info l 1%Qp i)))%I. *)
  (* Proof. *)
  (*   rewrite /own_loc_prim valloc_local_eq /valloc_local_def /=. *)
  (*   intros. split => ? /=.  funbi.unseal. done.  *)
  (* Qed. *)

  (* Global Instance Into_own_loc_prim_exist: *)
  (*   IntoExist ((own_loc_prim l h)) *)
  (*             (λ i, (valloc_local h ∧ (☐ Hist l h) ∗ (☐ Info l 1%Qp i)))%I. *)
  (* Proof. *)
  (*   rewrite /own_loc_prim valloc_local_eq /valloc_local_def /=. intros. split => ? /=. done. *)
  (* Qed. *)

  (* Global Instance From_own_loc_exist: *)
  (*   FromExist (own_loc l) (λ h, (own_loc_prim l h)). *)
  (* Proof. apply _. Qed. *)

  (* Global Instance Into_own_loc_exist: *)
  (*   IntoExist (own_loc l)  (λ h, (own_loc_prim l h)). *)
  (* Proof. apply _. Qed. *)

  (* Global Instance From_own_loc_na_exist: *)
  (*   ∀ V, FromExist (own_loc_na l v V) (λ V0, (own_loc_prim l {[v, V0]}) V). *)
  (* Proof. apply _. Qed. *)

  (* Global Instance Into_own_loc_na_exist: *)
  (*   ∀ V, IntoExist (own_loc_na l v V) (λ V0, (own_loc_prim l {[v, V0]}) V). *)
  (* Proof. apply _. Qed. *)

  Global Instance Frame_na_ownloc l v:
    Frame false ((own_loc_na l v)) (own_loc l) (True%I) | 1.
  Proof.
    rewrite /own_loc_na /own_loc /Frame. iStartProof (uPred _).
    iIntros (?) "[H _] /=". iDestruct "H" as (?) "?". iExists _. eauto.
  Qed.

  Global Instance valloc_local_timeless Σ h : Timeless (@valloc_local Σ h).
  Proof.
    rewrite valloc_local_eq /valloc_local_def /= /Timeless. iStartProof (uPred _).
    iIntros (?) "/=". iApply timeless.
  Qed.

  Global Instance valloc_local_Persistent Σ h : Persistent (@valloc_local Σ h).
  Proof.
    rewrite valloc_local_eq /valloc_local_def /= /Persistent. iStartProof (uPred _).
    iIntros (?) "/=". iApply persistent.
  Qed.

  (* TODO: these are probably not necessary anymore *)
  (* Global Instance own_loc_prim_timeless l h  *)
  (*   : Timeless (own_loc_prim l h). *)
  (* Proof. apply _. Qed. *)

  (* Global Instance own_loc_timeless l : Timeless (own_loc l). *)
  (* Proof. apply _. Qed. *)

  (* Global Instance own_loc_na_timeless `{foundationG Σ} l v *)
  (*   : Timeless (own_loc_na l v). *)
  (* Proof. rewrite /own_loc_na. apply _. Qed. *)

End Exclusive.

Section Fractional.
  Context {Σ : gFunctors} `{fG : foundationG Σ} `{cG : cinvG Σ}.
  Set Default Proof Using "Type".

  Notation frac_inv h := (λ l _, Hist l h)%I.
  Notation frac_local v := (λ _ _ X, ⌜X = encode v⌝)%I.

  Definition own_loc_na_frac l q v: vProp Σ
    := (∃ V, valloc_local {[v, V]}
             ∧ ⎡fractored l (frac_inv {[v, V]}) (frac_local v) q⎤)%I.

End Fractional.

Notation "l ↦ 'A'"
  := (own_loc_na l A)
  (at level 20, format "l  ↦  A") : bi_scope.
Notation "l ↦ v"
  := (own_loc_na l (VInj v))
  (at level 20, format "l  ↦  v") : bi_scope.

Notation "l ↦{ q } v"
  := (own_loc_na_frac l q (VInj v))
  (at level 20, format "l  ↦{ q }  v") : bi_scope.

Notation "l ↦ ?"
  := (own_loc l)
  (at level 20, format "l  ↦  ?") : bi_scope.


(* TODO: move to some other file *)
Lemma value_at_hd_singleton l v1 V1 v2 V2 :
  value_at_hd l {[v1, V1]} V2 (v2) -> v1 = v2 ∧ V1 = V2.
Proof.
  destruct 1 as (vx & Vx & ? & ?). abstract set_solver.
Qed.

Lemma alloc_local_singleton_na_safe l v V V':
  alloc_local {[v, V]} V' → na_safe l {[v, V]} V'.
Proof.
  move => [v1 [V1 [H1 [H2 _]]]].
  exists (v1, V1). repeat split; [auto|abstract set_solver+H1|auto].
Qed.

Lemma alloc_local_singleton_value_init v V V':
  alloc_local {[VInj v, V]} V' → init_local {[VInj v, V]} V'.
Proof.
  move => [v1 [V1 [H1 [H2 _]]]].
  exists v1, V1. repeat split; [auto|auto|].
  apply elem_of_singleton_1 in H1. by inversion H1.
Qed.


Section ExclusiveRules.
  Context `{foundationG Σ}.
  Set Default Proof Using "Type".

  Lemma na_excl l v w : (l ↦ v ∗ l ↦ w ⊢ False).
  Proof.
    (* TODO: can we do this without going to the model? *)
    iStartProof (uPred _). iIntros (?) "[H1 H2]".
    iDestruct "H1" as (? ?) "(_ & H1 & _)".
    iDestruct "H2" as (? ?) "(_ & H2 & _)".
    iApply (Hist_Exclusive with "[$H1 $H2]").
  Qed.

  Local Instance extension_proper : Proper (flip (⊑) ==> (=) ==> impl) (@sqsubseteq View _).
  Proof. do 6!intro. rewrite /impl. subst. by etrans. Qed.
  Lemma na_read (E : coPset) l v :
    ↑ physN ⊆ E →
    {{{ ⎡PSCtx⎤ ∗ l ↦ v }}}
      ([ #l]_na) @ E
    {{{ z, RET (#z); ⌜z = v⌝ ∗ (l ↦ v)  }}}.
  Proof.
    rewrite /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    rewrite WP_Def.wp_eq /WP_Def.wp_def /bi_affinely.
    iStartProof (uPred _). iIntros (???) "[#kI oNA]".
    iDestruct "oNA" as (V_0 i) "(Halloc & oH & oI)".
    iIntros (V ->) "Post". iIntros (? H1 π) "S".
    setoid_rewrite H1.
    iDestruct "Halloc" as %Halloc.
    iApply program_logic.weakestpre.wp_mask_mono;
      last (iApply (f_read_na with "[$kI $S $oH]")); first auto.
    - iSplit; iPureIntro;
      by [apply alloc_local_singleton_na_safe|
          apply alloc_local_singleton_value_init].
    - iNext. iIntros (v_r V') "(% & $ & oH & % & H)".
      iDestruct "H" as (V_1) "(% & % & %)".
      move: H5 => /value_at_hd_singleton [[->] ->].
      iApply ("Post" $! v_r).
      iSplit; first done.
      iExists _, _. iFrame "oH oI". iPureIntro.
      exists (VInj v_r), V_1. by rewrite elem_of_singleton.
  Qed.

  Lemma na_write (E : coPset) l v_w :
    ↑physN ⊆ E →
    {{{ ⎡PSCtx⎤ ∗ l ↦ ? }}}
      ([ #l]_na <- #v_w) @ E
    {{{ RET (#()); (l ↦ v_w) }}}.
  Proof.
    rewrite /own_loc /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    rewrite WP_Def.wp_eq /WP_Def.wp_def.
    iStartProof (uPred _). iIntros (???) "[#kI oL]".
    iDestruct "oL" as (h) "oL".
    iDestruct "oL" as (i) "(Halloc & oH & oI)".
    iIntros (? ->) "Post". iIntros (? -> π) "S".
    iDestruct "Halloc" as %Halloc.
    iApply program_logic.weakestpre.wp_mask_mono; last (iApply (f_write_na with "[$kI $S $oH]"));
      [auto|auto|].
    iNext. iIntros (V' h') "(% & $ & oH & % & % & % & _ & %)". subst.
    iApply "Post".
    iExists _, _. iFrame "oH oI". iPureIntro. by apply alloc_init_local.
  Qed.

End ExclusiveRules.

Section FractionalRules.
  Context {Σ : gFunctors} `{fG : foundationG Σ} `{cG : cinvG Σ}.
  Set Default Proof Using "Type".

  Global Instance na_fractional l v:
    fractional.Fractional (λ q, (l ↦{q} v)%I).
  Proof.
    rewrite /own_loc_na_frac /own_loc_prim /= valloc_local_eq /valloc_local_def.
    intros q1 q2. split => ?.
    iSplit.
    - iIntros "na".
      iDestruct "na" as (V') "(#Local & na)".
      iDestruct (fractor_splitjoin _ _ (λ _ _ X, ⌜X = encode (VInj v)⌝)%I
                                       (λ _ _ X, ⌜X = encode (VInj v)⌝)%I
        with "[na]") as "[na1 na2]".
      { iApply (fractor_mono with "[$na]"). iIntros (?) "?". by iSplit. }
      iSplitL "na1"; iExists V'; by iFrame.
    - iIntros "(na1 & na2)".
      iDestruct "na1" as (V1) "(% & na1)".
      iDestruct "na2" as (V2) "(% & na2)".
      iCombine "na1" "na2" as "na".
      rewrite fractor_join_l.
      iExists V1. iSplitR; first done.
      iApply (fractor_mono with "[$na]").
      by iIntros (?) "[? _]".
  Qed.

  Lemma na_frac_split l v q1 q2:
     l ↦{q1+q2} v ⊢ l ↦{q1} v ∗ l ↦{q2} v.
  Proof. by rewrite na_fractional. Qed.

  Lemma na_frac_split_2 l v q:
     l ↦{q} v ⊢ l ↦{q/2} v ∗ l ↦{q/2} v.
  Proof. rewrite -{1}(Qp_div_2 q). apply na_frac_split. Qed.

  Lemma na_frac_agree l v1 v2 q1 q2:
    l ↦{q1} v1 ∗ l ↦{q2} v2 ⊢ ⌜v1 = v2⌝.
  Proof.
    rewrite /own_loc_na_frac /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    constructor => V. iIntros "(na1 & na2)".
    iDestruct "na1" as (V1) "(Local1 & na1)".
    iDestruct "na2" as (V2) "(Local2 & na2)".
    iCombine "na1" "na2" as "na".
    rewrite fractor_join_l.
    iDestruct (fractor_drop with "na") as (X) "[% %]".
    iPureIntro. subst X. apply encode_inj in H0. by inversion H0.
  Qed.

  Global Instance na_exists_fractional l:
    fractional.Fractional (λ q, (∃ v, l ↦{q} v)%I).
  Proof.
    intros q1 q2.
    iSplit.
    - iIntros "na".
      iDestruct "na" as (v) "na". rewrite na_fractional.
      iDestruct "na" as "[na1 na2]".
      iSplitL "na1"; by iExists v.
    - iIntros "(na1 & na2)".
      iDestruct "na1" as (v1) "na1". iDestruct "na2" as (v2) "na2".
      pose NA := (na_frac_agree l v1 v2 q1 q2).
      iDestruct (NA with "[$na1 $na2]") as %Eq. subst v2.
      iExists v1.
      rewrite (fractional.fractional (Φ:=fun q => own_loc_na_frac l q _) q1 q2).
      iFrame "na1 na2".
      (* with "[$na1 $na2]"). *)
  Qed.

  Lemma na_frac_join l v q1 q2:
    l ↦{q1} v ∗ l ↦{q2} v ⊢ l ↦{q1+q2} v.
  Proof. by rewrite na_fractional. Qed.

  Lemma na_frac_join_1 l v1 v2 q1 q2:
    l ↦{q1} v1 ∗ l ↦{q2} v2 ⊢ l ↦{q1+q2} v1.
  Proof.
    iIntros "na".
    iDestruct (na_frac_agree with "na") as "%".
    subst v2.
    by iApply (na_frac_join with "na").
  Qed.

  Lemma na_frac_mult_splitL l v (n: positive) q:
    (l ↦{Pos2Qp n * q} v ⊢ [∗ list] _ ∈ seq 0 (Pos.to_nat n), l ↦{q} v)%I.
  Proof.
    rewrite -{1}(Pos2Nat.id n).
    change ((l ↦{_} v)%I)
      with ((λ q, (l ↦{q} v)) ((Pos2Qp (Pos.of_nat (Pos.to_nat n))
                                      * q)%Qp))%I at 1.
    rewrite big_sepL_mult_fractional; [done|lia].
  Qed.

  Lemma na_frac_mult_splitL_1 l v (n: positive):
    (l ↦{1} v ⊢ [∗ list] _ ∈ seq 0 (Pos.to_nat n), l ↦{1/n} v)%I.
  Proof. rewrite {1}(Pos2Qp_1_div n). by apply na_frac_mult_splitL. Qed.

  Lemma na_frac_mult_joinL l v (n: positive) q:
    (([∗ list] _ ∈ seq 0 (Pos.to_nat n), l ↦{q} v) ⊢ l ↦{Pos2Qp n * q} v)%I.
  Proof.
    rewrite -{2}(Pos2Nat.id n).
    change ((l ↦{_} v)%I)
      with ((λ q, (l ↦{q} v)%I) ((Pos2Qp (Pos.of_nat (Pos.to_nat n))
                                      * q)%Qp)) at 2.
    rewrite big_sepL_mult_fractional; [done|lia].
  Qed.

  Lemma na_frac_mult_exists_joinL l (n: positive) q:
    ([∗ list] _ ∈ seq 0 (Pos.to_nat n), ∃ v, l ↦{q} v)%I
    ⊢ (∃ v, l ↦{Pos2Qp n * q} v)%I.
  Proof.
    rewrite -(big_sepL_mult_fractional
                (Φ := (λ q, (∃ v, l ↦{q} v)%I))); last lia.
    by rewrite Pos2Nat.id.
  Qed.

  Lemma na_frac_mult_joinL_1 l v (n: positive):
    ([∗ list] _ ∈ seq 0 (Pos.to_nat n), l ↦{1/n} v)%I ⊢ l ↦{1} v.
  Proof. rewrite {2}(Pos2Qp_1_div n). by apply na_frac_mult_joinL. Qed.

  Lemma na_read_frac (E : coPset) l q v :
    ↑ physN ⊆ E →
    ↑fracN .@ l ⊆ E →
    {{{ ⎡PSCtx⎤ ∗ l ↦{q} v }}}
      [ #l ]_na @ E
    {{{ z, RET (#z); ⌜z = v⌝ ∗ l ↦{q} v }}}.
  Proof.
    rewrite /own_loc_na_frac /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    rewrite WP_Def.wp_eq /WP_Def.wp_def.
    iStartProof (uPred _).
    iIntros (????) "[#kI oNA]". iIntros (V H1) "Post". iIntros (V2 H2 π) "S".
    setoid_rewrite H1. setoid_rewrite H2.
    iDestruct "oNA" as (V_0) "(% & oNA)".
    iMod (fractor_open with "oNA") as (X) "(oH & HX & HClose)"; [solve_ndisj|].
    iApply program_logic.weakestpre.wp_mask_mono; last (iApply (f_read_na with "[$kI $S $oH]"));
      first solve_ndisj.
    - iSplit; iPureIntro;
      by [apply alloc_local_singleton_na_safe|
           apply alloc_local_singleton_value_init].
    - iNext. iIntros (v_r V') "(% & $ & oH & % & H)".
      iDestruct "H" as (V_1) "(% & % & %)".
      move: H6 => /value_at_hd_singleton [[->] ->].
      iApply ("Post").
      iMod ("HClose" $! (λ _ _ X, ⌜X = encode (VInj v_r)⌝)%I with "[$oH $HX]")
        as "oNA".
      iModIntro.
      iSplit; first done. iExists _. iFrame "oNA". iPureIntro.
      exists (VInj v_r), V_1. by rewrite elem_of_singleton.
  Qed.

  Lemma na_read_abs_frac (E : coPset) l q v V :
    ↑ physN ⊆ E →
    ↑fracN .@ l ⊆ E →
    {{{ ⎡PSCtx⎤ ∗ ⎡(l ↦{q} v) V⎤ ∗ monPred_in V }}}
      [ #l ]_na @ E
    {{{ z, RET (#z); ⌜z = v⌝ ∗ ⎡(l ↦{q} v) V⎤ }}}.
  Proof.
    rewrite /own_loc_na_frac /own_loc_na /own_loc_prim /=
            valloc_local_eq /valloc_local_def.
    rewrite WP_Def.wp_eq /WP_Def.wp_def.
    iStartProof (uPred _).
    iIntros (????) "[#kI (ol & %)] /=". iIntros (??) "Post". iIntros (V2 H2 π) "S".
    iDestruct "ol" as (V0) "[% ol]".
    iMod (fractor_open with "ol") as (X) "(oH & HX & HClose)"; [solve_ndisj|].
    assert (V ⊑ V2) by solve_jsl.
    iApply program_logic.weakestpre.wp_mask_mono; last (iApply (f_read_na with "[$kI $S $oH]"));
      first solve_ndisj.
    - iSplit; iPureIntro;
       [apply alloc_local_singleton_na_safe|
          apply alloc_local_singleton_value_init];
       by apply: (alloc_local_proper).
    - iNext. iIntros (v_r V') "(% & $ & oH & % & H)".
      iDestruct "H" as (V_1) "(% & % & %)".
      move: H7 => /value_at_hd_singleton [[?] ?]. subst v_r V_1.
      iApply ("Post" $! _ _ with "[%]"); [by etrans|].
      iMod ("HClose" $! (λ _ _ X, ⌜X = encode (VInj v)⌝)%I with "[$oH $HX]")
        as "ol".
      iModIntro. iSplitL ""; first done. iExists _. auto.
  Qed.

  Lemma na_frac_from_non (E: coPset) l v:
    ↑ physN ⊆ E →
    (⎡PSCtx⎤ ∗ l ↦ v ={E}=∗ l ↦{1} v)%I.
  Proof.
    rewrite /own_loc_na_frac /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    iStartProof (uPred _).
    iIntros (??) "[kI na]".
    iDestruct "na" as (V' i) "(Local & Hist & Info)".
    iExists V'. iFrame "Local".
    iApply (fractor_alloc _ _ (encode (VInj v))); [done|by iFrame].
  Qed.

  Lemma na_frac_dealloc (E: coPset) l v:
    ↑fracN.@l ⊆ E →
    (l ↦{1} v ={E}=∗ l ↦ ?)%I.
  Proof.
    intros.
    rewrite /own_loc_na_frac /own_loc /own_loc_na /own_loc_prim /= valloc_local_eq /valloc_local_def.
    iStartProof (uPred _).
    iIntros (?) "oL".
    iDestruct "oL" as (V') "[Local oL]".
    iMod (fractor_dealloc with "oL") as (? ?) "(oI & _ & >oH)"; first done.
    iExists _, _. by iFrame.
  Qed.

  Lemma na_write_frac_1 (E : coPset) l v:
    ↑ physN ⊆ E →
    {{{ ⎡PSCtx⎤ ∗ l ↦ ? }}}
      ([ #l]_na <- #v) @ E
    {{{ RET (#()); (l ↦{1} v) }}}.
  Proof.
    iIntros (??) "[#kI oL] Post".
    iApply wp_fupd.
    iApply (na_write with "[$kI $oL]"); [done|].
    iNext. iIntros "oL".
    iDestruct (na_frac_from_non with "[$kI $oL]") as ">oL"; first done.
    by iApply ("Post" with "oL").
  Qed.

  Lemma na_write_frac (E : coPset) l v v_w:
    ↑ physN ⊆ E →
    ↑fracN .@ l ⊆ E →
    {{{ ⎡PSCtx⎤ ∗ l ↦{1} v }}}
      ([ #l]_na <- #v_w) @ E
    {{{ RET (#()); (l ↦{1} v_w) }}}.
  Proof.
    iIntros (???) "[#kI oL] Post".
    iDestruct (na_frac_dealloc with "oL") as ">oL"; first done.
    by iApply (na_write_frac_1 with "[$kI $oL]").
  Qed.


End FractionalRules.


Arguments na_read [_ _] _ _ [_] _.
Arguments na_write [_ _] _ _ _ _.
