From iris.program_logic Require Import ectx_language ectxi_language language.
From stdpp Require Import fin_maps.
From igps Require Export lang.
From igps Require tactics.
Import base.


Ltac solve_closed :=
  match goal with
  | |- Closed ?X ?e =>
     let e' := tactics.W.of_expr0 e in change (Closed X (tactics.W.to_expr' e'));
     apply tactics.W.is_closed_correct0; vm_compute; exact I
  end.
Hint Extern 0 (Closed _ _) => solve_closed : typeclass_instances.

Ltac solve_to_val :=
  try match goal with
  | |- context E [language.to_val ?e] =>
     let X := context E [to_val e] in change X
  end;
  match goal with
  | |- to_val ?e = Some ?v =>
     let e' := tactics.W.of_expr0 e in change (to_val (tactics.W.to_expr0 e') = Some v);
     apply tactics.W.to_val_Some; simpl; unfold tactics.W.to_expr; reflexivity
  | |- is_Some (to_val ?e) =>
     let e' := tactics.W.of_expr0 e in change (is_Some (to_val (tactics.W.to_expr0 e')));
     apply tactics.W.to_val_is_Some, (bool_decide_unpack _); vm_compute; exact I
  end.
(* Hint Extern 10 (IntoVal _ _) => solve_to_val : typeclass_instances. *)

Ltac solve_atomic :=
  match goal with
  | |- language.atomic ?e =>
     let e' := tactics.W.of_expr0 e in change (language.atomic (tactics.W.to_expr0 e'));
     apply tactics.W.atomic_correct; vm_compute; exact I
  end.
Hint Extern 10 (language.atomic _) => solve_atomic.
(* For the side-condition of elim_upd_fupd_wp_atomic *)
Hint Extern 10 (language.atomic _) => solve_atomic : typeclass_instances.

(** Substitution *)
Ltac simpl_subst :=
  simpl;
  repeat match goal with
  | |- context [subst ?x ?er ?e] =>
      let er' := tactics.W.of_expr0 er in let e' := tactics.W.of_expr0 e in
      change (subst x er e) with (subst x (tactics.W.to_expr' er') (tactics.W.to_expr' e'));
      rewrite <-(tactics.W.to_expr_subst0 x); simpl (* ssr rewrite is slower *)
  end;
  unfold tactics.W.to_expr'.
Arguments tactics.W.to_expr : simpl never.
Arguments subst : simpl never.

(** The tactic [reshape_expr e tac] decomposes the expression [e] into an
evaluation context [K] and a subexpression [e']. It calls the tactic [tac K e']
for each possible decomposition until [tac] succeeds. *)
Ltac reshape_val e tac :=
  let rec go e :=
  lazymatch e with
  | of_val ?v => v
  | Rec ?f ?x ?e => constr:(RecV f x e)
  | Lit ?l => constr:(LitV l)
  (* | Pair ?e1 ?e2 => *)
  (*   let v1 := go e1 in let v2 := go e2 in constr:(PairV v1 v2) *)
  (* | InjL ?e => let v := go e in constr:(InjLV v) *)
  (* | InjR ?e => let v := go e in constr:(InjRV v) *)
  end in let v := go e in tac v.

Ltac reshape_expr e tac :=
  let rec go K e :=
  match e with
  | _ => tac K e
  | App ?e1 ?e2 => reshape_val e1 ltac:(fun v1 => go (AppRCtx v1 :: K) e2)
  | App ?e1 ?e2 => go (AppLCtx e2 :: K) e1
  | UnOp ?op ?e => go (UnOpCtx op :: K) e
  | BinOp ?op ?e1 ?e2 =>
     reshape_val e1 ltac:(fun v1 => go (BinOpRCtx op v1 :: K) e2)
  | BinOp ?op ?e1 ?e2 => go (BinOpLCtx op e2 :: K) e1
  | If ?e0 ?e1 ?e2 => go (IfCtx e1 e2 :: K) e0
  | Dealloc ?e => go (DeallocCtx :: K) e
  | Load ?acc ?e => go (LoadCtx acc :: K) e
  | Store ?acc ?e1 ?e2 => reshape_val e1 ltac:(fun v1 => go (StoreRCtx acc v1 :: K) e2)
  | Store ?acc ?e1 ?e2 => go (StoreLCtx acc e2 :: K) e1
  | CAS ?e0 ?e1 ?e2 => reshape_val e0 ltac:(fun v0 => first
     [ reshape_val e1 ltac:(fun v1 => go (CasRCtx v0 v1 :: K) e2)
     | go (CasMCtx v0 e2 :: K) e1 ])
  | CAS ?e0 ?e1 ?e2 => go (CasLCtx e1 e2 :: K) e0
  | FAI ?C ?e => go (FaiCtx C :: K) e
  end in go (@nil ectx_item) e.
