From stdpp Require Export coPset.
From iris.program_logic Require weakestpre.
From iris.program_logic Require Import ectxi_language language.
From igps Require Export lang viewpred.
From igps.base Require Import ghosts.
From iris.proofmode Require Import tactics.

Module WP_Def.
  Import weakestpre.
  Definition wp_def `{foundationG Σ} (E : coPset) (e : lang.base.expr) (Φ : lang.base.val → vProp Σ) : vProp Σ :=
    MonPred (
        monPred_upclosed
          (λ V, ∀ π, Seen π V -∗
              weakestpre.wp NotStuck E (e, V) (λ '(v,V), Seen π V ∗ monPred_at (Φ v) V))
      )%I _.
  Definition wp_aux : seal (@wp_def). by eexists. Qed.
  Definition wp := unseal (wp_aux).
  Definition wp_eq : wp = _ := seal_eq _.
End WP_Def.

Import WP_Def.
Arguments wp {_ _} _ _%E _.
Instance: Params (@wp) 4.

Notation "'WP' e @ E {{ Φ } }" := (wp E e%E Φ)
  (at level 20, e, Φ at level 200,
   format "'WP'  e  @  E  {{  Φ  } }") : bi_scope.
Notation "'WP' e {{ Φ } }" := (wp ⊤ e%E Φ)
  (at level 20, e, Φ at level 200,
   format "'WP'  e  {{  Φ  } }") : bi_scope.

Notation "'WP' e @ E {{ v , Q } }" := (wp E e%E (λ v, Q))
  (at level 20, e, Q at level 200,
   format "'WP'  e  @  E  {{  v ,  Q  } }") : bi_scope.
Notation "'WP' e {{ v , Q } }" := (wp ⊤ e%E (λ v, Q))
  (at level 20, e, Q at level 200,
   format "'WP'  e  {{  v ,  Q  } }") : bi_scope.

(* Texan triples *)
Notation "'{{{' P } } } e @ E {{{ x .. y , 'RET' pat ; Q } } }" :=
  (□ ∀ Φ,
      P -∗ ▷ (∀ x, .. (∀ y, Q -∗ Φ pat%V) .. ) -∗ WP e @ E {{ Φ }})%I
    (at level 20, x closed binder, y closed binder,
     format "{{{  P  } } }  e  @  E  {{{  x .. y ,  RET  pat ;  Q } } }") : bi_scope.
Notation "'{{{' P } } } e {{{ x .. y , 'RET' pat ; Q } } }" :=
  (□ ∀ Φ,
      P -∗ ▷ (∀ x, .. (∀ y, Q -∗ Φ pat%V) .. ) -∗ WP e {{ Φ }})%I
    (at level 20, x closed binder, y closed binder,
     format "{{{  P  } } }  e  {{{  x .. y ,   RET  pat ;  Q } } }") : bi_scope.
Notation "'{{{' P } } } e @ E {{{ 'RET' pat ; Q } } }" :=
  (□ ∀ Φ, P -∗ ▷ (Q -∗ Φ pat%V) -∗ WP e @ E {{ Φ }})%I
    (at level 20,
     format "{{{  P  } } }  e  @  E  {{{  RET  pat ;  Q } } }") : bi_scope.
Notation "'{{{' P } } } e {{{ 'RET' pat ; Q } } }" :=
  (□ ∀ Φ, P -∗ ▷ (Q -∗ Φ pat%V) -∗ WP e {{ Φ }})%I
    (at level 20,
     format "{{{  P  } } }  e  {{{  RET  pat ;  Q } } }") : bi_scope.

Notation "'{{{' P } } } e @ E {{{ x .. y , 'RET' pat ; Q } } }" :=
  (∀ Φ : _ → vProp _,
      P -∗ ▷ (∀ x, .. (∀ y, Q -∗ Φ pat%V) .. ) -∗ WP e @ E {{ Φ }})
    (at level 20, x closed binder, y closed binder,
     format "{{{  P  } } }  e  @  E  {{{  x .. y ,  RET  pat ;  Q } } }") : stdpp_scope.
Notation "'{{{' P } } } e {{{ x .. y , 'RET' pat ; Q } } }" :=
  (∀ Φ : _ → vProp _,
      P -∗ ▷ (∀ x, .. (∀ y, Q -∗ Φ pat%V) .. ) -∗ WP e {{ Φ }})
    (at level 20, x closed binder, y closed binder,
     format "{{{  P  } } }  e  {{{  x .. y ,  RET  pat ;  Q } } }") : stdpp_scope.
Notation "'{{{' P } } } e @ E {{{ 'RET' pat ; Q } } }" :=
  (∀ Φ : _ → vProp _, P -∗ ▷ (Q -∗ Φ pat%V) -∗ WP e @ E {{ Φ }})
    (at level 20,
     format "{{{  P  } } }  e  @  E  {{{  RET  pat ;  Q } } }") : stdpp_scope.
Notation "'{{{' P } } } e {{{ 'RET' pat ; Q } } }" :=
  (∀ Φ : _ → vProp _, P -∗ ▷ (Q -∗ Φ pat%V) -∗ WP e {{ Φ }})
    (at level 20,
     format "{{{  P  } } }  e  {{{  RET  pat ;  Q } } }") : stdpp_scope.

Section wp.
Import iprop.
Import iprop.iProp_solution.
Context `{foundationG Σ}.

Implicit Types P : vProp Σ.
Implicit Types Φ Ψ : base.val → vProp Σ.
Implicit Types φ ψ : language.val ra_lang → (iProp Σ).
Implicit Types v : base.val.
Implicit Types e : base.expr.

(* Weakest pre *)
(* TODO: figure out if wp_unfold is useful *)
(* Lemma wp_unfold E e Φ : WP e @ E {{ Φ }} ⊣⊢ wp_pre wp E e Φ. *)
(* Proof. rewrite wp_eq. apply (fixpoint_unfold wp_pre). Qed. *)

Global Instance wp_ne E e n :
  Proper (pointwise_relation _ (dist n) ==> dist n) (@wp Σ _ E e).
Proof.
  rewrite wp_eq /wp_def /=.
  intros x y z. split => t. cbn. solve_proper.
Qed.
Global Instance wp_proper E e :
  Proper (pointwise_relation _ (≡) ==> (≡)) (@wp Σ _ E e).
Proof.
  intros Φ Φ' ?; apply equiv_dist=>n; apply wp_ne=>v; apply equiv_dist.
  apply H0.
Qed.

Lemma wp_value' E Φ v : Φ v ⊢ WP base.of_val v @ E {{ Φ }}.
Proof.
  rewrite wp_eq /wp_def /=. econstructor => ?.
  apply bi.forall_intro => V.
  rewrite bi.pure_impl_forall.
  apply bi.forall_intro => ?.
  rewrite monPred_mono //.
  apply bi.forall_intro => ?. apply bi.wand_intro_l.
  (* apply bi.forall_intro => ?. apply bi.forall_intro => Eq. *)
  rewrite monPred_mono //.
  by refine (weakestpre.wp_value' (Λ:=ra_lang) _ _ _ (v,V)).
  (* by iIntros "($&?&$)". (* TODO: don't use proofmode. also figure out why not *) *)
Qed.
(* TODO: we don't seem to need this but it should be fixed and proven *)
(* Lemma wp_value_inv E Φ v : WP (language.of_val v) @ E {{ Φ }} ={E}=∗ Φ v. *)
(* Proof. *)
(*   econstructor => ?. funbi.unseal. rewrite wp_eq /wp_def /=. *)
(*   rewrite !bi.forall_elim //. *)
(*   rewrite (weakestpre.wp_value_inv (Λ:=ra_lang) _ _ (v, _)). *)
(*   (* by refine (weakestpre.wp_value_inv (Λ:=ra_lang) _ _ (v,_)). *) *)
(* Qed. *)

Import language.

Lemma wp_strong_mono_iris E1 E2 e φ ψ V0:
  E1 ⊆ E2 → (∀ v V (_ : V0 ⊑ V), φ (v,V) ={E2}=∗ ψ (v,V)) ∗
            weakestpre.wp weakestpre.NotStuck E1 (e,V0) (φ)
          ⊢ weakestpre.wp weakestpre.NotStuck E2 (e,V0) ψ.
Proof.
  iIntros (?) "[HΦ H]". iLöb as "IH" forall (e V0).
  setoid_rewrite weakestpre.wp_unfold at 3 4.
  rewrite /weakestpre.wp_pre.
  destruct (to_val ((e,V0) : ra_lang.expr)) as [v|] eqn:?.
  {
    destruct v.
    unshelve iApply ("HΦ" $! _ _ ?[H] with "[> -]").
    { by move/fmap_Some : Heqo => [? [_ [_ ->]]] //. }

    iApply (fupd_mask_mono E1 _); first done.
    iExact "H". (* TODO why does `done` not work? *)
  }
  iIntros (σ1) "Hσ". iMod (fupd_intro_mask' E2 E1) as "Hclose"; first done.
  iMod ("H" with "[$]") as "[$ H]".
  iModIntro. iNext. iIntros (e2 σ2 efs Hstep).
  iMod ("H" with "[//]") as "($ & H & $)"; auto.
  iMod "Hclose" as "_". destruct e2 as [e2 V2]. iApply ("IH" with "[HΦ] H").
  iClear "IH".
  move/prim_step_view_mono : Hstep => HV.
  iIntros (? ? ?). unshelve iApply ("HΦ" $! _ _ ?[H]). by etransitivity.
Qed.

Lemma wp_strong_mono E1 E2 e Φ Ψ :
  E1 ⊆ E2 → (∀ v, Φ v ={E2}=∗ Ψ v) ∗ WP e @ E1 {{ Φ }} ⊢ WP e @ E2 {{ Ψ }}.
Proof.
  iStartProof (uPred _). rewrite wp_eq /wp_def /=.
  iIntros (? Vo) "[VS WP]". iIntros (V HV). iSpecialize ("WP" $! V with "[//]").
  iIntros (π) "S". iSpecialize ("WP" $! π with "S").
  iApply (wp_strong_mono_iris with "[-$WP]"); first done.
  iIntros (? V' HV'). iIntros "[$ HΦ]".
  setoid_rewrite HV. setoid_rewrite HV'.
  iApply ("VS" with "HΦ").
Qed.

Lemma fupd_wp E e Φ : (|={E}=> WP e @ E {{ Φ }}) ⊢ WP e @ E {{ Φ }}.
Proof.
  rewrite wp_eq /wp_def /=. iStartProof (uPred _). iIntros (?) "/= H".
  setoid_rewrite weakestpre.wp_unfold. rewrite /weakestpre.wp_pre /=.
  iIntros (V ? ?).
  rewrite !bi.forall_elim bi.pure_impl_forall bi.forall_elim //.
  destruct (ra_lang.to_val ((e, V))) as [v|] eqn:?; [by iMod "H"|].
  iIntros "S" (σ1) "Hσ1". iMod "H". iApply ("H" with "S Hσ1").
Qed.
Lemma wp_fupd E e Φ : WP e @ E {{ v, |={E}=> Φ v }} ⊢ WP e @ E {{ Φ }}.
Proof. iIntros "H". iApply (wp_strong_mono E); try iFrame; auto. Qed.

Lemma wp_atomic E1 E2 e Φ :
  (∀ V, Atomic WeaklyAtomic ((e ,V) : ra_lang.expr)) →
  (|={E1,E2}=> WP e @ E2 {{ v, |={E2,E1}=> Φ v }}) ⊢ WP e @ E1 {{ Φ }}.
Proof.
  iStartProof (uPred _). iIntros (Hatomic ?) "H".
  rewrite wp_eq /wp_def /=. iIntros (?? π) "S".
  iApply (weakestpre.wp_atomic _ E1 E2) => //. iMod "H". iModIntro.
  iApply (weakestpre.wp_strong_mono _ _ E2 with "[-]"); [done..|by iApply "H"|].
  iIntros (?) "H". case_match. iModIntro.
  by iDestruct "H" as "[$ H]".
Qed.

Lemma wp_step_fupd E1 E2 e P Φ :
  base.to_val e = None → E2 ⊆ E1 →
  (|={E1,E2}▷=> P) -∗ WP e @ E2 {{ v, P ={E1}=∗ Φ v }} -∗ WP e @ E1 {{ Φ }}.
Proof.
  iStartProof (uPred _). iIntros (H1 H2 V) "HR". iIntros (V' ->).
  move: H1 H2. rewrite wp_eq /wp_def /=.
  do !setoid_rewrite weakestpre.wp_unfold. rewrite /weakestpre.wp_pre.
  rewrite /to_val /= /ra_lang.to_val /=.
  iIntros (-> ?) "H". iIntros (V'' HV'' π) "S".
  iSpecialize ("H" with "[//]").
  iIntros (σ1) "Hσ". iMod "HR". iMod ("H" with "S [$]") as "[$ H]".
  iModIntro; iNext; iIntros (e2 σ2 efs Hstep).
  iMod ("H" $! e2 σ2 efs with "[% //]") as "($ & H & $)".
  iMod "HR". iModIntro.
  destruct e2 as [e2 V2].
  iApply (wp_strong_mono_iris E2 with "[-$H]"); first done.
  move/prim_step_view_mono : Hstep => ?.
  iIntros (???) "[$ HP]".
  iApply ("HP" with "[HR]"). iApply ("HR" $! _ with "[%]"). repeat etrans=>//.
Qed.


(* TODO: file issue with iris bc of LanguageCtx stuff *)
(* TODO: figure out was is meant by above comment *)
Lemma wp_bind_iris K E e φ V0:
  weakestpre.wp weakestpre.NotStuck E (e,V0) (λ '(v,Vp),
      weakestpre.wp weakestpre.NotStuck E (base.fill K (base.of_val v),Vp) φ)
                ⊢ weakestpre.wp weakestpre.NotStuck E (base.fill K e,V0) φ.
Proof.
  iIntros "H".
  rewrite (_ : (base.fill _ _, _) = ectx_language.fill K (e,V0)); last first.
  { elim: K e => // ? ? IH ? /=. by rewrite IH. }
  iApply (weakestpre.wp_bind).
  (* { rewrite (_ : fill = ectx_language.fill); last done. apply _. } *)
  iApply (wp_strong_mono_iris with "[$H]"); first done.
  iIntros (???) "WP". iModIntro.
  rewrite (_ : (base.fill _ ?[e], ?[V]) = ectx_language.fill K (?e,?V)); last first.
  { elim: K (base.of_val a) => // ? ? IH ? /=. by rewrite /= IH. }
  done.
Qed.

Lemma wp_bind K E e Φ :
  WP e @ E {{ v, WP (base.fill K (base.of_val v)) @ E {{ Φ }} }} ⊢ WP (base.fill K e) @ E {{ Φ }}.
Proof.
  constructor => V_0.
  rewrite wp_eq /wp_def /=.
  iIntros "H" (V_1 ??) "S".
  iApply (wp_bind_iris).
  iSpecialize ("H" with "[%] S"); first done.
  iApply (weakestpre.wp_strong_mono with "[$H]"); [done..|].
  iIntros ([v V_2]) "[S H]".
  by iApply ("H").
Qed.

(** * Derived rules *)
Lemma wp_mono E e Φ Ψ : (∀ v, Φ v ⊢ Ψ v) → WP e @ E {{ Φ }} ⊢ WP e @ E {{ Ψ }}.
Proof.
  iIntros (HΦ) "H"; iApply (wp_strong_mono E E); auto.
  iIntros "{$H}" (v) "?". by iApply HΦ.
Qed.
Lemma wp_mask_mono E1 E2 e Φ : E1 ⊆ E2 → WP e @ E1 {{ Φ }} ⊢ WP e @ E2 {{ Φ }}.
Proof. iIntros (?) "H"; iApply (wp_strong_mono E1 E2); auto. iFrame; eauto. Qed.
Global Instance wp_mono' E e :
  Proper (pointwise_relation _ (⊢) ==> (⊢)) (@wp Σ _ E e).
Proof. by intros Φ Φ' ?; apply wp_mono. Qed.

Lemma wp_value E Φ e v : base.to_val e = Some v → Φ v ⊢ WP e @ E {{ Φ }}.
Proof. intros; rewrite -(base.of_to_val e v) //; by apply wp_value'. Qed.
Lemma wp_value_fupd' E Φ v : (|={E}=> Φ v) ⊢ WP base.of_val v @ E {{ Φ }}.
Proof. intros. by rewrite -wp_fupd -wp_value'. Qed.
Lemma wp_value_fupd E Φ e v :
  base.to_val e = Some v → (|={E}=> Φ v) ⊢ WP e @ E {{ Φ }}.
Proof. intros. rewrite -wp_fupd -wp_value //. Qed.

Lemma wp_frame_l E e Φ R : R ∗ WP e @ E {{ Φ }} ⊢ WP e @ E {{ v, R ∗ Φ v }}.
Proof. iIntros "[??]". iApply (wp_strong_mono E E _ Φ); try iFrame; eauto. Qed.
Lemma wp_frame_r E e Φ R : WP e @ E {{ Φ }} ∗ R ⊢ WP e @ E {{ v, Φ v ∗ R }}.
Proof. iIntros "[??]". iApply (wp_strong_mono E E _ Φ); try iFrame; eauto. Qed.

Lemma wp_frame_step_l E1 E2 e Φ R :
  base.to_val e = None → E2 ⊆ E1 →
  (|={E1,E2}▷=> R) ∗ WP e @ E2 {{ Φ }} ⊢ WP e @ E1 {{ v, R ∗ Φ v }}.
Proof.
  iIntros (??) "[Hu Hwp]". iApply (wp_step_fupd with "Hu"); try done.
  iApply (wp_mono with "Hwp"). by iIntros (?) "$$".
Qed.
Lemma wp_frame_step_r E1 E2 e Φ R :
  base.to_val e = None → E2 ⊆ E1 →
  WP e @ E2 {{ Φ }} ∗ (|={E1,E2}▷=> R) ⊢ WP e @ E1 {{ v, Φ v ∗ R }}.
Proof.
  rewrite [(WP _ @ _ {{ _ }} ∗ _)%I]comm; setoid_rewrite (comm _ _ R).
  apply wp_frame_step_l.
Qed.
Lemma wp_frame_step_l' E e Φ R :
  base.to_val e = None → ▷ R ∗ WP e @ E {{ Φ }} ⊢ WP e @ E {{ v, R ∗ Φ v }}.
Proof. iIntros (?) "[??]". iApply (wp_frame_step_l E E); try iFrame; eauto. Qed.
Lemma wp_frame_step_r' E e Φ R :
  base.to_val e = None → WP e @ E {{ Φ }} ∗ ▷ R ⊢ WP e @ E {{ v, Φ v ∗ R }}.
Proof. iIntros (?) "[??]". iApply (wp_frame_step_r E E); try iFrame; eauto. Qed.

Lemma wp_wand E e Φ Ψ :
  WP e @ E {{ Φ }} -∗ (∀ v, Φ v -∗ Ψ v) -∗ WP e @ E {{ Ψ }}.
Proof.
  iIntros "Hwp H". iApply (wp_strong_mono E); auto.
  iIntros "{$Hwp}" (?) "?". by iApply "H".
Qed.
Lemma wp_wand_l E e Φ Ψ :
  (∀ v, Φ v -∗ Ψ v) ∗ WP e @ E {{ Φ }} ⊢ WP e @ E {{ Ψ }}.
Proof. iIntros "[H Hwp]". iApply (wp_wand with "Hwp H"). Qed.
Lemma wp_wand_r E e Φ Ψ :
  WP e @ E {{ Φ }} ∗ (∀ v, Φ v -∗ Ψ v) ⊢ WP e @ E {{ Ψ }}.
Proof. iIntros "[Hwp H]". iApply (wp_wand with "Hwp H"). Qed.
End wp.

(** Proofmode class instances *)
Section proofmode_classes.
  Context `{foundationG Σ}.
  Implicit Types P Q : vProp Σ.
  Implicit Types Φ : base.val → vProp Σ.

  Import iprop.

  Global Instance into_forall_wp E e Φ V :
    V ⊑ V' →
    IntoForall
      (monPred_at (WP e @ E {{ Φ }})%I V)
      (λ π, Seen π V' -∗
                 weakestpre.wp weakestpre.NotStuck E (e, V') (λ '(v,V), Seen π V ∗ monPred_at (Φ v) V))%I.
  Proof.
    intros.
    rewrite wp_eq /wp_def /= /IntoForall
    bi.forall_elim
    bi.pure_impl_forall
    bi.forall_elim; try done.
  Qed.

  Global Instance from_forall_wp E e Φ V :
    FromForall
      (monPred_at (WP e @ E {{ Φ }})%I V)
      (λ V', ⌜V ⊑ V'⌝ → ∀ π, Seen π V' -∗
                 weakestpre.wp weakestpre.NotStuck E (e, V') (λ '(v,V), Seen π V ∗ monPred_at (Φ v) V))%I.
  Proof.
    rewrite wp_eq /wp_def /= /FromForall.
    apply bi.forall_intro => V0.
    rewrite bi.pure_impl_forall.
    apply bi.forall_intro => HV'.
    apply bi.forall_intro => π.
    by rewrite (bi.forall_elim V0)
    bi.pure_impl_forall
    (bi.forall_elim HV')
    (bi.forall_elim π).
  Qed.

  Global Instance into_wand_wp p q E e Φ V π:
    IntoWand p q
      (monPred_at (WP e @ E {{ Φ }})%I V)
      (Seen π V)
            (weakestpre.wp weakestpre.NotStuck E (e, V) (λ '(v,V), Seen π V ∗ monPred_at (Φ v) V))%I.
  Proof.
    rewrite wp_eq /wp_def /= /IntoWand
    bi.forall_elim
    bi.pure_impl_forall
    bi.forall_elim; [|done].
    rewrite bi.forall_elim.
    apply (_ : IntoWand _ _ _ _ _).
  Qed.

  Global Instance frame_wp p E e R Φ Ψ :
    (∀ v, Frame p R (Φ v) (Ψ v)) →
    Frame p R (WP e @ E {{ Φ }}) (WP e @ E {{ Ψ }}).
  Proof. rewrite /Frame=> HR. rewrite wp_frame_l. apply wp_mono, HR. Qed.

  Global Instance is_except_0_wp E e Φ : IsExcept0 (WP e @ E {{ Φ }}).
  Proof. by rewrite /IsExcept0 -{2}fupd_wp -except_0_fupd -fupd_intro. Qed.

  Global Instance elim_modal_bupd_wp p E e P Φ :
    ElimModal True p false (|==> P) P (WP e @ E {{ Φ }}) (WP e @ E {{ Φ }}).
  Proof.
    by rewrite /ElimModal bi.intuitionistically_if_elim (bupd_fupd E)
      fupd_frame_r bi.wand_elim_r fupd_wp.
  Qed.

  Global Instance elim_modal_fupd_wp p E e P Φ :
    ElimModal True p false (|={E}=> P) P (WP e @ E {{ Φ }}) (WP e @ E {{ Φ }}).
  Proof.
    by rewrite /ElimModal bi.intuitionistically_if_elim
      fupd_frame_r bi.wand_elim_r fupd_wp.
  Qed.

  (* lower precedence, if possible, it should always pick elim_upd_fupd_wp *)
  Global Instance elim_modal_fupd_wp_atomic p E1 E2 e P Φ :
    (∀ V, Atomic WeaklyAtomic ((e,V) : ra_lang.expr)) →
    ElimModal True p false (|={E1,E2}=> P) P
            (WP e @ E1 {{ Φ }}) (WP e @ E2 {{ v, |={E2,E1}=> Φ v }})%I | 100.
  Proof.
    intros. by rewrite /ElimModal bi.intuitionistically_if_elim
      fupd_frame_r bi.wand_elim_r wp_atomic.
  Qed.

End proofmode_classes.
