From iris.base_logic Require Import upred.
From iris.program_logic Require Import weakestpre.
From iris.program_logic Require Import ectxi_language.
From iris.proofmode Require Import coq_tactics tactics.
From igps Require Export lifting tactics_vProp ghosts.
Import uPred.

Lemma tac_wp_pure `{ghosts.foundationG Σ} K Δ Δ' E e1 e2 φ Φ :
  PureExec (Λ:=ra_lang) φ e1 e2 →
  φ →
  MaybeIntoLaterNEnvs 1 Δ Δ' →
  envs_entails Δ' (WP fill K e2 @ E {{ Φ }}) →
  envs_entails Δ (WP fill K e1 @ E {{ Φ }}).
Proof.
  rewrite envs_entails_eq=> ??? HΔ'. rewrite into_laterN_env_sound /=.
  rewrite -weakestpre.wp_bind HΔ' -lifting.wp_pure_step_later //.
  by rewrite -weakestpre.wp_bind_inv.
Qed.

Lemma tac_wp_value `{ghosts.foundationG Σ} Δ E Φ e v :
  IntoVal e v →
  envs_entails Δ (Φ v) → envs_entails Δ (WP e @ E {{ Φ }}).
Proof. rewrite envs_entails_eq=> ? ->. by apply wp_value. Qed.

Ltac wp_value_head := eapply tac_wp_value; [apply _|lazy beta].

Tactic Notation "wp_pure" open_constr(efoc) :=
  iStartProof;
  lazymatch goal with
  | |- envs_entails _ (wp ?E ?e ?Q) => reshape_expr e ltac:(fun K e' =>
    unify e' efoc;
    eapply (tac_wp_pure K);
    [simpl; apply _                 (* PureExec *)
    |try fast_done                  (* The pure condition for PureExec *)
    |apply _                        (* IntoLaters *)
    |simpl_subst; try wp_value_head (* new goal *)])
   || fail "wp_pure: cannot find" efoc "in" e "or" efoc "is not a reduct"
  | _ => fail "wp_pure: not a 'wp'"
  end.

Tactic Notation "wp_if" := wp_pure (base.If _ _ _).
Tactic Notation "wp_if_true" := wp_pure (base.If (base.Lit (base.LitInt 1)) _ _).
Tactic Notation "wp_if_false" := wp_pure (base.If (base.Lit (base.LitInt 0)) _ _).
Tactic Notation "wp_unop" := wp_pure (base.UnOp _ _).
Tactic Notation "wp_binop" := wp_pure (base.BinOp _ _ _).
Tactic Notation "wp_op" := wp_unop || wp_binop.
Tactic Notation "wp_rec" := wp_pure (base.App _ _).
Tactic Notation "wp_lam" := wp_rec.
Tactic Notation "wp_let" := wp_lam.
Tactic Notation "wp_seq" := wp_lam.


Lemma tac_wp_bind `{ghosts.foundationG Σ} K Δ E Φ e :
  envs_entails Δ (WP e @ E {{ v, WP fill K (ectxi_language.of_val v) @ E {{ Φ }} }})%I →
  envs_entails Δ (WP fill K e @ E {{ Φ }}).
Proof. rewrite envs_entails_eq=> ->. exact: wp_bind. Qed.

Ltac wp_bind_core K :=
  lazymatch eval hnf in K with
  | [] => idtac
  | _ => apply (tac_wp_bind K); simpl
  end.

Tactic Notation "wp_bind" open_constr(efoc) :=
  iStartProof;
  lazymatch goal with
  | |- envs_entails _ (wp ?E ?e ?Q) =>
    reshape_expr e ltac:(fun K e' => unify e' efoc; wp_bind_core K)
    || fail "wp_bind: cannot find" efoc "in" e
  | _ => fail "wp_bind: not a 'wp'"
  end.

(* (** wp-specific helper tactics *) *)
(* Ltac wp_bind_core K := *)
(*   lazymatch eval hnf in K with *)
(*   | [] => idtac *)
(*   | _ => etrans; [|fast_by apply (wp_bindt K)]; *)
(*          simpl; lazy beta iota zeta delta [fill_item base.fill_item]; simpl *)
(*   end. *)

(* (* Solves side-conditions generated by the wp tactics *) *)
(* Ltac wp_done := *)
(*   match goal with *)
(*   | |- base.Closed _ _ => solve_base_closed *)
(*   | |- ra_lang.Closed _ _ => solve_closed *)
(*   | |- is_Some (base.to_val _) => solve_base_to_val *)
(*   | |- is_Some (ra_lang.to_val _) => solve_to_val *)
(*   | |- base.to_val _ = Some _ => solve_base_to_val *)
(*   | |- ra_lang.to_val _ = Some _ => solve_to_val *)
(*   | |- language.to_val _ = Some _ => solve_to_val *)
(*   | _ => fast_done *)
(*   end. *)

(* Ltac wp_value_head := etrans; [|eapply wp_value; wp_done];  *)
(*          simpl; lazy beta iota zeta delta [fill_item base.fill_item]; simpl. *)

(* Ltac wp_strip_later := idtac. (* a hook to be redefined later *) *)

(* Ltac wp_seq_head := *)
(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (Seq _ _) ?Q => *)
(*     etrans; [|eapply wp_seq; wp_done]; wp_strip_later *)
(*   end. *)

(* Ltac wp_finish := intros_revert ltac:( *)
(*   rewrite /= ?to_of_val; *)
(*   try wp_strip_later; *)
(*   repeat lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (Seq _ _, ?π) ?Q => *)
(*      etrans; [|eapply wp_seq; wp_done]; wp_strip_later *)
(*   | |- _ ⊢ wp ?E _ ?Q => wp_value_head *)
(*   end). *)

(* Tactic Notation "wp_value" := *)
(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (?e, ?π) ?Q => reshape_expr e ltac:(fun K e' => *)
(*     wp_bind_core K; wp_value_head) || fail "wp_value: cannot find value in" e *)
(*   | _ => fail "wp_value: not a wp" *)
(*   end. *)

(* Tactic Notation "wp_rec" := *)

(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (?e, ?π) ?Q => reshape_expr e ltac:(fun K e' => *)
(*     match eval hnf in e' with App ?e1 _ => *)
(* (* hnf does not reduce through an of_val *) *)
(* (*      match eval hnf in e1 with Rec _ _ _ => *) *)
(*     wp_bind_core K; etrans; [|eapply wp_rec; wp_done]; simpl_subst; wp_finish *)
(* (*      end *) end) || fail "wp_rec: cannot find 'Rec' in" e *)
(*   | _ => fail "wp_rec: not a 'wp'" *)
(*   end. *)

(* Tactic Notation "wp_lam" := *)
(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (?e, ?π) ?Q => reshape_expr e ltac:(fun K e' => *)
(*     match eval hnf in e' with App ?e1 _ => *)
(* (*    match eval hnf in e1 with Rec BAnon _ _ => *) *)
(*     wp_bind_core K; etrans; [|eapply wp_lam; wp_done]; simpl_subst; wp_finish *)
(* (*    end *) end) || fail "wp_lam: cannot find 'Lam' in" e *)
(*   | _ => fail "wp_lam: not a 'wp'" *)
(*   end. *)

(* Tactic Notation "wp_let" := wp_lam. *)
(* Tactic Notation "wp_seq" := wp_let. *)

(* Tactic Notation "wp_op" := *)
(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (?e, ?π) ?Q => reshape_expr e ltac:(fun K e' => *)
(*     lazymatch eval hnf in e' with *)
(*     | BinOp LtOp _ _ => wp_bind_core K; apply wp_lt; wp_finish *)
(*     | BinOp LeOp _ _ => wp_bind_core K; apply wp_le; wp_finish *)
(*     | BinOp EqOp _ _ => *)
(*        wp_bind_core K; eapply wp_eq; wp_finish *)
(*     | BinOp _ _ _ => *)
(*        wp_bind_core K; etrans; *)
(*          [|eapply wp_bin_op; try fast_done]; wp_finish *)
(*     | UnOp _ _ => *)
(*        wp_bind_core K; etrans; *)
(*          [|eapply wp_un_op; try fast_done]; wp_finish *)
(*     end) || fail "wp_op: cannot find 'BinOp' or 'UnOp' in" e *)
(*   | _ => fail "wp_op: not a 'wp'" *)
(*   end. *)

(* (* Tactic Notation "wp_proj" := *) *)
(* (*   lazymatch goal with *) *)
(* (*   | |- _ ⊢ wp ?E ?e ?Q => reshape_expr e ltac:(fun K e' => *) *)
(* (*     match eval hnf in e' with *) *)
(* (*     | Fst _ => wp_bind_core K; etrans; [|eapply wp_fst; wp_done]; wp_finish *) *)
(* (*     | Snd _ => wp_bind_core K; etrans; [|eapply wp_snd; wp_done]; wp_finish *) *)
(* (*     end) || fail "wp_proj: cannot find 'Fst' or 'Snd' in" e *) *)
(* (*   | _ => fail "wp_proj: not a 'wp'" *) *)
(* (*   end. *) *)

(* Tactic Notation "wp_if" := *)
(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (?e, ?π) ?Q => reshape_expr e ltac:(fun K e' => *)
(*     match eval hnf in e' with *)
(*     | If _ _ _ => *)
(*       wp_bind_core K; *)
(*       etrans; [|eapply wp_if_true || eapply wp_if_false]; wp_finish *)
(*     end) || fail "wp_if: cannot find 'If' in" e *)
(*   | _ => fail "wp_if: not a 'wp'" *)
(*   end. *)

(* (* Tactic Notation "wp_match" := *) *)
(* (*   lazymatch goal with *) *)
(* (*   | |- _ ⊢ wp ?E ?e ?Q => reshape_expr e ltac:(fun K e' => *) *)
(* (*     match eval hnf in e' with *) *)
(* (*     | Case _ _ _ => *) *)
(* (*       wp_bind_core K; *) *)
(* (*       etrans; [|first[eapply wp_match_inl; wp_done|eapply wp_match_inr; wp_done]]; *) *)
(* (*       simpl_subst; wp_finish *) *)
(* (*     end) || fail "wp_match: cannot find 'Match' in" e *) *)
(* (*   | _ => fail "wp_match: not a 'wp'" *) *)
(* (*   end. *) *)

(* Tactic Notation "wp_bind" open_constr(efoc) := *)
(*   lazymatch goal with *)
(*   | |- _ ⊢ wp ?E (?e, ?π) ?Q => reshape_expr e ltac:(fun K e' => *)
(*     match e' with *)
(*     | efoc => unify e' efoc; wp_bind_core K *)
(*     end) || fail "wp_bind: cannot find" efoc "in" e *)
(*   | _ => fail "wp_bind: not a 'wp'" *)
(*   end. *)
