From iris.program_logic Require Import language ectxi_language.
From iris.proofmode Require Import coq_tactics tactics.
From iris.bi Require Import bi.
From igps Require Import tactics_vProp viewpred weakestpre ghosts lang.surface.
Import WP_Def.
From igps Require Import lifting_vProp. (* Import after WP_Def because of wp_eq name clash *)

Lemma tac_wp_expr_eval `{foundationG Σ} Δ E Φ e e' :
  e = e' →
  envs_entails Δ (WP e' @ E {{ Φ }}) → envs_entails Δ (WP e @ E {{ Φ }}).
Proof. by intros ->. Qed.

Ltac wp_expr_eval t :=
  try (iStartProof; eapply tac_wp_expr_eval; [t; reflexivity|]).

Ltac wp_expr_simpl := wp_expr_eval simpl.
Ltac wp_expr_simpl_subst := wp_expr_eval simpl_subst.

Lemma tac_wp_pure `{foundationG Σ} Δ Δ' E e1 e2 φ Φ :
  (PureExec (Λ:=surface_lang) φ (e1) (e2)) →
  φ →
  MaybeIntoLaterNEnvs 1 Δ Δ' →
  envs_entails Δ' (WP e2 @ E {{ Φ }}) →
  envs_entails Δ (WP e1 @ E {{ Φ }}).
Proof.
  rewrite envs_entails_eq=> ??? HΔ'. rewrite into_laterN_env_sound /=.
  rewrite HΔ' -wp_pure_step_later //.
Qed.

Lemma tac_wp_value `{foundationG Σ} Δ E Φ e v :
  (IntoVal (e) (v)) →
  envs_entails Δ (Φ v) → envs_entails Δ (WP e @ E {{ Φ }}).
Proof. rewrite envs_entails_eq=> IV ->. now apply: wp_value. Qed.

Ltac wp_value_head := eapply tac_wp_value; [apply _|lazy beta].

Tactic Notation "wp_pure" open_constr(efoc) :=
  iStartProof;
  lazymatch goal with
  | |- envs_entails _ (WP_Def.wp ?E ?e ?Q) => tactics_vProp.reshape_expr e ltac:(fun K e' =>
    unify e' efoc;
    eapply tac_wp_pure;
    [
      simple apply (language.pure_exec_ctx _ (H:=ectx_language.ectx_lang_ctx (Λ:=surface_ectx_lang) K) e' _ _ _) (* FIXME *)
                                    (* PureExec *)
    |try fast_done                  (* The pure condition for PureExec *)
    |apply _                        (* IntoLaters *)
    |wp_expr_simpl_subst; try wp_value_head (* new goal *)
    ])
   || fail "wp_pure: cannot find" efoc "in" e "or" efoc "is not a reduct"
  | _ => fail "wp_pure: not a 'wp'"
  end.


Import lang.base.
Tactic Notation "wp_if" := wp_pure (If _ _ _).
Tactic Notation "wp_if_true" := wp_pure (If (Lit (LitInt 1)) _ _).
Tactic Notation "wp_if_false" := wp_pure (If (Lit (LitInt 0)) _ _).
Tactic Notation "wp_unop" := wp_pure (UnOp _ _).
Tactic Notation "wp_binop" := wp_pure (BinOp _ _ _).
Tactic Notation "wp_op" := wp_unop || wp_binop.
Tactic Notation "wp_rec" := wp_pure (App _ _).
Tactic Notation "wp_lam" := wp_rec.
Tactic Notation "wp_let" := wp_lam.
Tactic Notation "wp_seq" := wp_lam.

Lemma tac_wp_bind `{foundationG Σ} K Δ E Φ e :
  envs_entails Δ (WP e @ E {{ v, WP fill K (of_val v) @ E {{ Φ }} }})%I →
  envs_entails Δ (WP fill K e @ E {{ Φ }}).
Proof. rewrite envs_entails_eq=> ->. by apply wp_bind. Qed.

Ltac wp_bind_core K :=
  lazymatch eval hnf in K with
  | [] => idtac
  | _ => apply (tac_wp_bind K); simpl
  end.

Tactic Notation "wp_bind" open_constr(efoc) :=
  iStartProof;
  lazymatch goal with
  | |- envs_entails _ (wp ?E ?e ?Q) =>
    reshape_expr e ltac:(fun K e' => unify e' efoc; wp_bind_core K)
    || fail "wp_bind: cannot find" efoc "in" e
  | _ => fail "wp_bind: not a 'wp'"
  end.

Tactic Notation "wp_apply" open_constr(lem) :=
  iPoseProofCore lem as false true (fun H =>
    lazymatch goal with
    | |- envs_entails _ (wp ?E ?e ?Q) =>
      reshape_expr e ltac:(fun K e' =>
        wp_bind_core K; iApplyHyp H; try iNext; simpl) ||
      lazymatch iTypeOf H with
      | Some (_,?P) => fail "wp_apply: cannot apply" P
      end
    | _ => fail "wp_apply: not a 'wp'"
    end).
